/*
 * Copyright (c) 1997, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Shared code for gluing COM interfaces into MOM servers.
 */

#ifndef _FLUKE_NEST_SERVER_COM_H
#define _FLUKE_NEST_SERVER_COM_H 1

#include <fluke/nest/server/server.h>
#include <oskit/com.h>
#include <oskit/error.h>
#include <fluke/mutex.h>


/*
 * This structure describes a lock variable and how to lock and unlock it.
 * If the `lock' function pointer is null, use fluke_mutex_lock;
 * if the `unlock' function pointer is null, use fluke_mutex_unlock.
 *
 * The functions below provide a simple lock/unlock interface.
 */
struct nest_com_lock_data {
	void (*lock)(void *);
	void (*unlock)(void *);
	void *data;
};

static inline void
nest_com_lock(const struct nest_com_lock_data *ld)
{
	if (ld->lock)
		(*ld->lock)(ld->data);
	else if (ld->data)
		fluke_mutex_lock(ld->data);
}

static inline void
nest_com_unlock(const struct nest_com_lock_data *ld)
{
	if (ld->unlock)
		(*ld->unlock)(ld->data);
	else if (ld->data)
		fluke_mutex_unlock(ld->data);
}


/*
 * This is all we know in this glue layer about the MOM objects we implement.
 *
 */
#define COMSTRUCT(mom_iface, com_iface)					      \
struct nest_com_##mom_iface##_struct {					      \
	mom_obj_header_t header;					      \
	oskit_##com_iface##_t *com;					      \
	struct nest_com_lock_data lock_data;			      	      \
}

#define LOCK	nest_com_lock(&_obj->lock_data)
#define UNLOCK	nest_com_unlock(&_obj->lock_data)

#define COMCALL(call, args...) ({					      \
	oskit_error_t err;						      \
	LOCK;								      \
	err = oskit_##call(_obj->com ,##args);			      	      \
	UNLOCK;								      \
	if (err) {							      \
		err2exc(err, _ev);					      \
	}								      \
	err; })

#define COMCALL_V(call, args...) ({					      \
	LOCK;								      \
	oskit_##call(_obj->com ,##args);				      \
	UNLOCK;								      \
	})

static inline void
err2exc(oskit_error_t err, mom_exc_t *exc)
{
	switch (err) {
	case OSKIT_E_UNEXPECTED:
	case OSKIT_E_NOTIMPL:
	case OSKIT_E_NOINTERFACE:
		mom_set_exc(exc, OSKIT_ENOSYS);
		break;
	default:
		mom_set_exc(exc, err);
		break;
	}

}

COMSTRUCT(obj, iunknown);

static inline void *
query(struct nest_com_obj_struct *obj, const oskit_iid_t *iid, mom_exc_t *_ev)
{
	void *new;
	oskit_error_t rc = obj->com->ops->query(obj->com, iid, &new);
	if (rc) {
		err2exc(rc, _ev);
		return NULL;
	}
	return new;
}

#define COMQUERY(iface) query((void *) _obj, &oskit_##iface##_iid, _ev)

#define COMQUERYCALL(iface, call, args...) ({				      \
	oskit_error_t __rc;						      \
	oskit_##iface##_t *__iface;					      \
	LOCK;								      \
	__iface = COMQUERY(iface);					      \
	if (__iface) {							      \
 		__rc = oskit_##iface##_##call(__iface ,##args);		      \
		oskit_##iface##_release(__iface);			      \
		UNLOCK;							      \
		if (__rc)						      \
			err2exc(__rc, _ev);				      \
	}								      \
	else {								      \
		UNLOCK;							      \
		mom_set_exc(_ev, ENOTSUP);				      \
		__rc = ENOTSUP;						      \
	} __rc; })

/*
 * Create a MOM class for COM objects to be exported with
 * `nest_com_instantiate_ref'.
 */
int nest_com_class_create(mom_dispatch_t dispatcher, mom_class_t *class);

#if FLASK
#define COMCLASSINIT(mom_iface, com_iface)				      \
	if (! nest_com_##com_iface##_class &&				      \
	    nest_com_class_create(nest_com_##mom_iface##_server_secure,       \
				  &nest_com_##com_iface##_class))	      \
		return ENOMEM;
#else
#define COMCLASSINIT(mom_iface, com_iface)				      \
	if (! nest_com_##com_iface##_class &&				      \
	    nest_com_class_create(nest_com_##mom_iface##_server,	      \
				  &nest_com_##com_iface##_class))	      \
		return ENOMEM;
#endif


/*
 * Instantiate a COM object as a MOM object in the given class.
 * If non-null, `lock_data' indicates a lock to be held while
 * calling COM operations on the object; the lock data structure is copied,
 * so its storage can and should be reclaimed by the caller.
 * The MOM object will have `extra_bytes' more space allocated
 * after the `nest_com_obj_struct' at the beginning of the object.
 * If `*newobj' is zero, this returns null if no MOM object already exists.
 * On return `*newobj' is nonzero iff it created a new object, indicating
 * the extra data should be initialized.  A null return with `*newobj' nonzero
 * means it could not allocate memory for the MOM object.
 *
 * If *newobj is nonzero, this call may need to take the lock in `lock_data',
 * so it must not be held by the caller.
 */
void *nest_com_instantiate_object(mom_class_t class, void *comobj,
				  const struct nest_com_lock_data *lock_data,
				  size_t extra_bytes,
				  oskit_bool_t *newobj);

/*
 * Instantiate a COM object as a MOM object in the given class, as above
 * (with zero extra bytes of data).  Returns a reference to the MOM object,
 * consumes the `comobj' COM reference no matter what, and sets `*_ev' to
 * indicate errors, so this function is convenient to call in a return
 * statement from a server work function.
 *
 * The `lock_data' lock must not be held by the caller, as this will use it.
 */
mom_ref_t nest_com_instantiate_ref(mom_class_t class, void *comobj,
				   const struct nest_com_lock_data *lock_data,
				   mom_exc_t *_ev);

/*
 * If REF refers to a MOM object in this nester that wraps a COM
 * object, add a ref to that COM object and return it;
 * otherwise return null.
 * This call does its own locking; locks should not be held by the caller.
 */
void *nest_com_obj_lookup(mom_ref_t ref);

/*
 * Same as above, but queries the object for IID and returns that interface,
 * or null if the object doesn't answer to it.
 * This call does its own locking; locks should not be held by the caller.
 */
void *nest_com_obj_lookup_iid(const oskit_iid_t *iid, mom_ref_t ref);


/*
 * Type of COM->MOM glue objects.
 * The operations do make CP RPCs to the contained MOM reference.
 */
struct com_cp_obj {
	struct oskit_iunknown_ops *ops;
	mom_ref_t ref;
	oskit_u32_t count;
};

/*
 * Macro for writing COM operations that are wrappers around CP calls.
 */
#define CPCALL(result, call, args...)					      \
	mom_exc_t exc;							      \
	oskit_error_t rc;						      \
	result cp_##call(((struct com_cp_obj *) obj)->ref ,##args, &exc);     \
	rc = com_cp_error(&exc); if (rc) return rc;
#define CPCALL_V(result, call, args...)					      \
	mom_exc_t exc;							      \
	cp_##call(((struct com_cp_obj *) obj)->ref ,##args, &exc);	      \
	com_cp_error(&exc);

/*
 * These are the add_ref and release operations for glue objects.
 */
oskit_u32_t com_cp_add_ref(struct com_cp_obj *);
oskit_u32_t com_cp_release(struct com_cp_obj *);

/*
 * Convert a CP exception into a COM error code.
 */
oskit_error_t com_cp_error(mom_exc_t *_ev);

/*
 * Consume the given MOM reference and create a wrapper COM object
 * to contain it.  The given `ops' pointer should use com_cp_add_ref
 * and com_cp_release, and provide operations written using CPCALL.
 */
oskit_error_t com_cp_instantiate_ref(const void *ops, mom_ref_t ref,
				    void **comobj);

/*
 * Classes for libnest-provided MOM->COM adapters, and functions
 * to initialize them.  Once these are initialized,
 * nest_com_instantiate_ref can be used to wrap COM objects
 * as MOM objects and export references to them.
 */
extern mom_class_t nest_com_file_class, nest_com_openfile_class;
extern mom_class_t nest_com_dir_class;
int nest_com_fs_init(void);

#include <oskit/fs/file.h>
/*
 * Instantiate the given COM oskit_file object as a MOM object
 * supporting File and also other protocols corresponding to
 * the COM i/o interfaces the underlying object supports.
 * XXX this function knows about a few things it might support.
 */
mom_ref_t
nest_com_instantiate_file(oskit_file_t *file,
			  const struct nest_com_lock_data *lock_data,
			  mom_exc_t *_ev);

extern mom_class_t nest_com_stream_class;
int nest_com_io_init(void);

extern mom_class_t nest_com_ttystream_class, nest_com_opentty_class;
int nest_com_termios_init(void);

extern mom_class_t nest_com_socket_class;
int nest_com_socket_init();

/*
 * COM operation tables for libnest-provided COM->CP client adapters.
 * Pass the appropriate one of these to com_cp_instantiate_ref to
 * wrap MOM references for CP objects in COM objects whose operations
 * make CP calls on the contained references.
 */
extern struct oskit_filesystem_ops com_cp_filesystem_ops;


#if FLASK
/*
 * Create a call context for the given SID.
 *
 * The returned context may be used in a call to
 * oskit_set_call_context, and may be queried for
 * either a oskit_principal interface or a oskit_comsid
 * interface.
 */
oskit_error_t oskit_create_call_context(security_id_t sid,
					oskit_iunknown_t **out_context);


/*
 * Set the client call context based on the specified SID _csid.
 *
 * On failure, set _ev and return _ret.
 */
#include <assert.h>
#define CCONSET(_csid,_ev,_ret) {					      \
	oskit_iunknown_t *unki;						      \
        oskit_error_t rc; 						      \
	rc = oskit_create_call_context(_csid, &unki);	                      \
	if (rc) { mom_set_exc(_ev, rc); return _ret; }		              \
        rc = oskit_set_call_context(unki);				      \
	unki->ops->release(unki);					      \
	if (rc) { mom_set_exc(_ev, rc); return _ret; }			      \
}

/*
 * Perform a permission check with a target SID obtained
 * from the specified COM object _comobj.  If _comobj
 * does not support the oskit_comsid interface, then use
 * __mysid.
 *
 * On failure, set _ev.
 */
#define MOMCHECKCOM(_sclass,_ssid,_tclass,_comobj,_perm,_ev) {		      \
        oskit_comsid_t *sidi;						      \
        security_id_t objsid;						      \
        security_class_t objclass;					      \
        oskit_error_t rc;						      \
	LOCK;								      \
        rc = _comobj->ops->query(_comobj, &oskit_comsid_iid, (void**)&sidi);  \
        if (!rc) {							      \
            rc = oskit_comsid_get(sidi, &objclass, &objsid);		      \
            oskit_comsid_release(sidi);					      \
        }								      \
	UNLOCK;								      \
        if (rc)	{							      \
            objsid = __mysid;						      \
        }								      \
        MOMCHECK(_sclass,_ssid,_tclass,objsid,_perm,_ev);		      \
}

#endif

#endif
