/*
 * Copyright (c) 1995 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#ifndef _FLUKE_OB_H_
#define _FLUKE_OB_H_

#include <oskit/machine/types.h>

/* Include machine-dependent definitions first.  */
#include <fluke/machine/ob.h>

typedef unsigned char fluke_type_t;

/* Fluke object types */
#define FLUKE_NULL		0	/* Invalid object type */
#define FLUKE_TASK		1
#define FLUKE_THREAD		2
#define FLUKE_MUTEX		3
#define FLUKE_COND		4
#define FLUKE_PORT		5
#define FLUKE_PSET		6
#define FLUKE_CATALOG		7
#define FLUKE_REF		8
#define FLUKE_MAPPING		9
#define FLUKE_REGION		10
#define FLUKE_MI_MAX		10

#define FLUKE_ARCH_DEP		128

/* x86-specific object types */
#define FLUKE_LDT		128	/* Local descriptor table */
#define FLUKE_FPU_STATE		129	/* FPU state save area */
#define FLUKE_IOPB		130	/* I/O permission bitmap */
#define FLUKE_MD_MAX		130

/* User-memory sizes of all the Fluke object types.  */
#define FLUKE_CATALOG_SIZE	(1 << FLUKE_CATALOG_SHIFT)
#define FLUKE_COND_SIZE		(1 << FLUKE_COND_SHIFT)
#define FLUKE_MUTEX_SIZE	(1 << FLUKE_MUTEX_SHIFT)
#define FLUKE_PORT_SIZE		(1 << FLUKE_PORT_SHIFT)
#define FLUKE_PSET_SIZE		(1 << FLUKE_PSET_SHIFT)
#define FLUKE_REF_SIZE		(1 << FLUKE_REF_SHIFT)
#define FLUKE_TASK_SIZE		(1 << FLUKE_TASK_SHIFT)
#define FLUKE_MAPPING_SIZE	(1 << FLUKE_MAPPING_SHIFT)
#define FLUKE_REGION_SIZE	(1 << FLUKE_REGION_SHIFT)
#define FLUKE_THREAD_SIZE	(1 << FLUKE_THREAD_SHIFT)

#define FLUKE_MINOB_SIZE	(1 << FLUKE_MINOB_SHIFT)
#define FLUKE_MAXOB_SIZE	(1 << FLUKE_MAXOB_SHIFT)

/*
 * Most objects have a hash code
 * that can be set when the object is created
 * but is constant thereafter.
 */
typedef oskit_addr_t fluke_hash_t;

/*
 * Macro used to define structure tags and type names
 * representing active Fluke objects.
 * Since all such objects are opaque, no real structure elements are needed;
 * just a size and corresponding natural alignment.
 */
#define FLUKE_OB_DEFINE(name, size)				\
	typedef struct name {					\
		char x[size] __attribute((aligned(size)));	\
	} name##_t;

/*
 * Error codes
 * XXX move to error.h?
 * XXX rename to fluke_status_t?
 */
typedef unsigned fluke_error_t;

#define FLUKE_SUCCESS			0x00	/* everything went OK */
#define FLUKE_NO_MEMORY			0x01	/* no kernel memory */

/*
 * Return code for mutex_trylock() call.
 */
#define FLUKE_MUTEX_HELD		0x01

typedef unsigned fluke_ipc_status_t;

/*
 * These bits in the status code indicate what operation was in progress
 * when the error or "exceptional situation" occurred.
 * This is necessary for combination IPC operations
 * to allow the client to know in what stage the problem occurred,
 * and how the results of the operation should be interpreted.
 */
#define FLUKE_OP_MASK			0xf0
#define FLUKE_OP_IPC_RECEIVE		0x00
#define FLUKE_OP_IPC_OVER		0x10
#define FLUKE_OP_IPC_SEND		0x20
#define FLUKE_OP_IPC_ACK		0x30
#define FLUKE_OP_IPC_CONNECT		0x40
#define FLUKE_OP_IPC_CALL		0x50
#define FLUKE_OP_IPC_WAIT		0x60
#define FLUKE_OP_IPC_REPLY		0x70

/* 
 * Specific return codes used in various situations.
 */
#define FLUKE_IPC_CONNECT_INVALID_DEST	0x41	/* invalid or stale port ref */
#define FLUKE_IPC_ACK_DISCONNECTED	0x31	/* other side disconnected */
#define FLUKE_IPC_SEND_DISCONNECTED	0x21	/* other side disconnected */
#define FLUKE_IPC_OVER_DISCONNECTED	0x11	/* other side disconnected */
#define FLUKE_IPC_RECV_REVERSED		0x00	/* connection was reversed */
#define FLUKE_IPC_RECV_DISCONNECTED	0x01	/* other side disconnected */
#define FLUKE_IPC_RECV_MORE_DATA	0x02	/* need more data buffers */
#define FLUKE_IPC_RECV_MORE_REFS	0x03	/* need more receive refs */
#define FLUKE_IPC_WAIT_ONEWAY		0x61	/* one-way msg received */
#define FLUKE_IPC_WAIT_IDEMPOTENT	0x62	/* idempotent req received */
#define FLUKE_IPC_REPLY_DISCONNECTED	0x71	/* caller didn't hang around */

/*
 * Flask additions
 */
#define FLUKE_IPC_CONNECT_INVALID_SERVER    0x42 /* server SID mismatch */
#define FLUKE_IPC_CONNECT_PERMISSION_DENIED 0x43 /* permission denied */
#define FLUKE_IPC_WAIT_PERMISSION_DENIED    0x63 /* permission denied */

/* Object array entry used in fluke_region_search call.  */
struct fluke_obspec
{
	void		*ob;
	fluke_type_t	type;
};
typedef struct fluke_obspec fluke_obspec_t;

void fluke_nop(void);

#endif /* _FLUKE_OB_H_ */
