/*
 * Copyright (c) 1995, 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#ifndef _FLUKE_X86_THREAD_H_
#define _FLUKE_X86_THREAD_H_

/* To set the default sched_state for a thread */

struct fluke_thread_state
{
	fluke_hash_t		hash;

	/* Virtual address of the thread object itself
	   in the task to which the thread is assigned.  */
	oskit_addr_t		thread_va;

	/* Scheduler-assigned thread identifier value.  */
	int			sched_val;

	/* General thread flags; see definitions below.  */
	unsigned		flags;

	/* Saved processor state.  */
	unsigned		eax;
	unsigned		ebx;
	unsigned		ecx;
	unsigned		edx;
	unsigned		esi;
	unsigned		edi;
	unsigned		ebp;
	unsigned		esp;
	unsigned		eip;
	unsigned		eflags;

	/*
	 * These locations act as "extra registers"
	 * which are used to save the minimum message words
	 * passed to the receiver during IPC operations.
	 * Yep, this is yet another horrible kludge
	 * resulting from the x86's scarcity of registers.
	 */
	unsigned		min_msg[2];

	/*
	 * Exception handlers for each of the main types of thread exceptions.
	 * These only need to be valid if THREAD_EXCEPTION_ENABLE flag is set.
	 */
	unsigned		trap_handler;
	unsigned		interrupt_handler;
	unsigned		client_alert_handler;
	unsigned		server_alert_handler;

	/*
	 * Save locations for thread exceptions.
	 * When an exception is dispatched to the thread,
	 * the original values of eax, ecx, edx, and eip are saved here,
	 * and the registers are set as follows:
	 *	eax = pointer to current thread,
	 *		so the exception handler has "a place to stand".
	 *	ecx = exception code
	 *	edx = exception subcode
	 *	eip = exception handler entrypoint
	 *		(one of the _handler values above).
	 *	eflags = original eflags with TF bit clear
	 */
	unsigned		saved_eax;
	unsigned		saved_ecx;
	unsigned		saved_edx;
	unsigned		saved_eip;
	unsigned		saved_eflags;

	/*
	 * These are only relevant for threads acting as servers.
	 */
	struct fluke_pset	*server_pset;
	unsigned		server_esp;
	unsigned		server_oneway_eip;
	unsigned		server_idempotent_eip;
	unsigned		server_reliable_eip;

	/*
	 * Proportional-share scheduler state.
	 * Inherited tickets and tickets for the thread's own use
	 * (hierarchical scheduling).
	 */
	unsigned		tickets;
	unsigned		own_tickets;

	/* Leave at least 5 pad words for:
		- task
		- scheduler
		- waiting_for
		- server
		- client
	*/
};
typedef struct fluke_thread_state fluke_thread_state; /*XXX obsolete */
typedef struct fluke_thread_state fluke_thread_state_t;

#if FLASK
struct fluke_thread_state_secure
{
    /*
     * For each of the reliable IPC references (client_ref, server_ref)
     * in a thread's state, there are two related effective SIDs:
     * the effective SID of the thread in its role as a server or client
     * to its peer in the reliable IPC connection, and the effective
     * SID of the peer.  The effective SID of the peer in the thread's
     * state must equal the corresponding effective SID in the peer
     * thread's state in order for the connection to be active.
     * This ensures that neither side of a connection can change its
     * effective SID for the connection without the awareness of the 
     * other side.
     */

    /* the effective SID of this thread in its role as a server */
    security_id_t			effective_server_sid;

    /* the effective SID of this thread's client */
    security_id_t			client_effective_client_sid;

    /* the effective SID of this thread in its role as a client */
    security_id_t			effective_client_sid;

    /* the effective SID of this thread's server */
    security_id_t			server_effective_server_sid;


    /* the ordinary Fluke thread state */
    struct fluke_thread_state	state;
};
typedef struct fluke_thread_state_secure fluke_thread_state_secure_t;
#endif


/*
 * Flag definitions for fluke_thread_state.flags
 */
#define FLUKE_THREAD_EXCEPTION_ENABLE		0x00000001
#define FLUKE_THREAD_INTERRUPT_PENDING		0x00000002
#define FLUKE_THREAD_CLIENT_ALERT_PENDING	0x00000004
#define FLUKE_THREAD_SERVER_ALERT_PENDING	0x00000008
#define FLUKE_THREAD_ASYNC_PENDING		0x0000000e /* mask of above */
#define FLUKE_THREAD_STOPPED			0x00000010
#define FLUKE_THREAD_CLIENT_SENDER		0x00000040
#define FLUKE_THREAD_SERVER_SENDER		0x00000080
#define FLUKE_THREAD_USES_FPU			0x00010000

/*
 * Default ticket values for proportional-share based scheduling. 
 */
#define FLUKE_THREAD_DEFAULT_TICKETS	8

/*
 * Saved exception state structure used by
 * fluke_thread_get_saved_state and fluke_thread_set_saved_state.
 */
struct fluke_thread_saved_state
{
	unsigned		saved_eax;
	unsigned		saved_ecx;
	unsigned		saved_edx;
	unsigned		saved_eip;
	unsigned		saved_eflags;
};
typedef struct fluke_thread_saved_state fluke_thread_saved_state_t;


/*
 * Set the server entry information for the current thread.
 * This information will be used the next time this thread
 * waits for and successfully receives a message or connection.
 */
void fluke_thread_set_server_info(
	struct fluke_pset	*server_pset,
	unsigned		server_esp,
	unsigned		server_oneway_eip,
	unsigned		server_idempotent_eip,
	unsigned		server_reliable_eip);


#if defined(__GNUC__) && defined(FLUKE_INLINING)

#include <oskit/compiler.h>
#ifndef FLUKE_ENTRY_INLINE
#define FLUKE_ENTRY_INLINE OSKIT_INLINE
#endif
#ifdef GPROF
#define e(name) extern char _fluke_##name##_[];
#include <fluke/x86/syscall_list.h>
#undef e
#include <oskit/x86/proc_reg.h>
#include <fluke/profile.h>
#endif

FLUKE_ENTRY_INLINE fluke_error_t fluke_thread_create(
	fluke_thread_t		*thread)
{
	fluke_error_t rc;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_thread_create_hash_);
#endif
	asm volatile("call	_fluke_thread_create_hash"
		: "=a" (rc) : "a" (thread), "d" (0) : "ecx", "edx");

	return rc;
}

FLUKE_ENTRY_INLINE fluke_error_t fluke_thread_create_hash(
	fluke_thread_t	*thread,
	fluke_hash_t	*hashp)
{
	fluke_error_t rc;
	fluke_hash_t hash;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_thread_create_hash_);
#endif
	hash = hashp ? *hashp : 0;
	asm volatile("call _fluke_thread_create_hash"
		: "=a" (rc), "=d" (hash)
		: "a" (thread), "d" (hash)
		: "ecx");
	if (hashp) *hashp = hash;

	return rc;
}

FLUKE_ENTRY_INLINE void fluke_thread_destroy(
	fluke_thread_t		*thread)
{
#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_thread_destroy_);
#endif
	asm volatile("call	_fluke_thread_destroy"
		: : "a" (thread) : "eax", "ecx", "edx");
}

FLUKE_ENTRY_INLINE void fluke_thread_reference(
	fluke_thread_t		*thread,
	struct fluke_ref	*out_thread_ref)
{
#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_thread_reference_);
#endif
	asm volatile("call	_fluke_thread_reference"
		: : "a" (thread), "d" (out_thread_ref) : "eax", "ecx", "edx");
}

FLUKE_ENTRY_INLINE void fluke_thread_move(
	fluke_thread_t		*old_va,
	fluke_thread_t		*new_va)
{
#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_thread_move_);
#endif
	asm volatile("call	_fluke_thread_move"
		: : "a" (old_va), "d" (new_va) : "eax", "ecx", "edx");
}

#ifdef GPROF
/*
 * XXX: inline all system calls for GPROF counting.
 */

#else

/*
 * No inline versions of these are provided on the x86
 * because there are too many parameters to pass in registers,
 * and it's easier just to implement them as normal functions
 * that pass their parameter stack frames directly on to the kernel.
 */
#if 0
void fluke_thread_get_state(...);
void fluke_thread_set_state(...);

void fluke_thread_enable_exceptions(fluke_thread_t *cur_thread);
void fluke_thread_disable_exceptions(fluke_thread_t *cur_thread);
#endif 0

#endif GPROF


FLUKE_ENTRY_INLINE void fluke_thread_get_client(
	fluke_thread_t		*cur_thread,
	struct fluke_ref	*client_ref,
	unsigned		*flags)
{
#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_thread_get_client_);
#endif
	asm volatile("call	_fluke_thread_get_client"
		: "=c" (*flags)
		: "a" (cur_thread), "d" (client_ref)
		: "eax", "ecx", "edx");
}

#if FLASK
FLUKE_ENTRY_INLINE void fluke_thread_get_client_secure(
	fluke_thread_t			*cur_thread,
	struct fluke_ref		*client_ref,
        security_id_t			*effective_server_sid, 
        security_id_t			*client_effective_client_sid, 
	unsigned			*flags)
{
#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_thread_get_client_secure_);
#endif
	asm volatile("call	_fluke_thread_get_client_secure"
		: "=c" (*flags),
		  "=S" (*effective_server_sid),
                  "=D" (*client_effective_client_sid)
	        : "a" (cur_thread), "d" (client_ref)
		: "eax", "ecx", "edx", "esi", "edi");
}
#endif

FLUKE_ENTRY_INLINE void fluke_thread_set_client(
	fluke_thread_t		*cur_thread,
	struct fluke_ref	*client_ref,
	unsigned		flags)
{
#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_thread_set_client_);
#endif
	asm volatile("call	_fluke_thread_set_client"
		: : "a" (cur_thread), "d" (client_ref), "c" (flags)
		: "eax", "ecx", "edx");
}

#if FLASK
FLUKE_ENTRY_INLINE void fluke_thread_set_client_secure(
	fluke_thread_t		*cur_thread,
	struct fluke_ref	*client_ref,
        security_id_t		effective_server_sid,
        security_id_t		client_effective_client_sid,
	unsigned		flags)
{
#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_thread_set_client_secure_);
#endif
	asm volatile("call	_fluke_thread_set_client_secure"
	        : : "a" (cur_thread), "d" (client_ref), "c" (flags), "S" (effective_server_sid), "D" (client_effective_client_sid)
		: "eax", "ecx", "edx", "esi", "edi");
}
#endif

FLUKE_ENTRY_INLINE void fluke_thread_get_server(
	fluke_thread_t		*cur_thread,
	struct fluke_ref	*server_ref,
	unsigned		*flags)
{
#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_thread_get_server_);
#endif
	asm volatile("call	_fluke_thread_get_server"
		: "=c" (*flags)
		: "a" (cur_thread), "d" (server_ref)
		: "eax", "ecx", "edx");
}

#if FLASK
FLUKE_ENTRY_INLINE void fluke_thread_get_server_secure(
	fluke_thread_t		*cur_thread,
	struct fluke_ref	*server_ref,
        security_id_t		*effective_client_sid,
        security_id_t		*server_effective_server_sid,
	unsigned		*flags)
{
#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_thread_get_server_secure_);
#endif
	asm volatile("call	_fluke_thread_get_server_secure"
		: "=c" (*flags),
		  "=S" (*effective_client_sid),
                  "=D" (*server_effective_server_sid)
	        : "a" (cur_thread), "d" (server_ref)
		: "eax", "ecx", "edx", "esi", "edi");
}
#endif

FLUKE_ENTRY_INLINE void fluke_thread_set_server(
	fluke_thread_t		*cur_thread,
	struct fluke_ref	*server_ref,
	unsigned		flags)
{
#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_thread_set_server_);
#endif
	asm volatile("call	_fluke_thread_set_server"
		: : "a" (cur_thread), "d" (server_ref), "c" (flags)
		: "eax", "ecx", "edx");
}

#if FLASK
FLUKE_ENTRY_INLINE void fluke_thread_set_server_secure(
	fluke_thread_t		*cur_thread,
	struct fluke_ref	*server_ref,
        security_id_t		effective_client_sid, 
        security_id_t		server_effective_server_sid,
	unsigned		flags)
{
#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_thread_set_server_secure_);
#endif
	asm volatile("call	_fluke_thread_set_server_secure"
	     : : "a" (cur_thread), "d" (server_ref), "c" (flags), "S" (effective_client_sid), "D" (server_effective_server_sid)
		: "eax", "ecx", "edx", "esi", "edi");
}
#endif

FLUKE_ENTRY_INLINE void fluke_thread_interrupt(
	fluke_thread_t		*thread)
{
#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_thread_interrupt_);
#endif
	asm volatile("call	_fluke_thread_interrupt"
		: : "a" (thread) : "eax", "ecx", "edx");
}

FLUKE_ENTRY_INLINE fluke_thread_t *fluke_thread_self(void)
{
	fluke_thread_t *thread;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_thread_self_);
#endif
	asm volatile("call	_fluke_thread_self"
		: "=a" (thread) : : "ecx", "edx");

	return thread;
}

FLUKE_ENTRY_INLINE void fluke_thread_get_saved_state(
	fluke_thread_t			*cur_thread,
	struct fluke_thread_saved_state	*out_saved_state)
{
#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_thread_get_saved_state_);
#endif
	asm volatile("call	_fluke_thread_get_saved_state"
		: "=S" (out_saved_state->saved_eax),
		  "=c" (out_saved_state->saved_ecx),
		  "=d" (out_saved_state->saved_edx),
		  "=D" (out_saved_state->saved_eip),
		  "=b" (out_saved_state->saved_eflags)
		: "a" (cur_thread));
	/* Note: eax is _not_ trashed */
}

FLUKE_ENTRY_INLINE void fluke_thread_set_saved_state(
	fluke_thread_t			*cur_thread,
	struct fluke_thread_saved_state	*saved_state)
{
#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_thread_set_saved_state_);
#endif
	asm volatile("
		movl	12(%%ebx),%%edi
		movl	16(%%ebx),%%ebx
		call	_fluke_thread_set_saved_state
	" :
	  : "a" (cur_thread),
	    "S" (saved_state->saved_eax),
	    "c" (saved_state->saved_ecx),
	    "d" (saved_state->saved_edx),
	    "b" (saved_state)
	  : "ebx", "ecx", "edx", "esi", "edi");
	/* Note: eax is _not_ trashed */
}

FLUKE_ENTRY_INLINE void fluke_thread_get_handlers(
	struct fluke_thread		*cur_thread,
	oskit_addr_t			*trap_handler,
	oskit_addr_t			*interrupt_handler,
	oskit_addr_t			*client_alert_handler,
	oskit_addr_t			*server_alert_handler)
{
	oskit_addr_t th, ih, cah, sah;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_thread_get_handlers_);
#endif
	asm volatile("call	_fluke_thread_get_handlers"
		: "=c" (trap_handler ? *trap_handler : th),
		  "=d" (interrupt_handler ? *interrupt_handler : ih),
		  "=S" (client_alert_handler ? *client_alert_handler : cah),
		  "=D" (server_alert_handler ? *server_alert_handler : sah)
		: "a" (cur_thread)
		: "eax");
}

FLUKE_ENTRY_INLINE void fluke_thread_set_handlers(
	struct fluke_thread		*cur_thread,
	oskit_addr_t			trap_handler,
	oskit_addr_t			interrupt_handler,
	oskit_addr_t			client_alert_handler,
	oskit_addr_t			server_alert_handler)
{
#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_thread_set_handlers_);
#endif
	asm volatile("call	_fluke_thread_set_handlers"
		:
		: "a" (cur_thread),
		  "c" (trap_handler),
		  "d" (interrupt_handler),
		  "S" (client_alert_handler),
		  "D" (server_alert_handler)
		: "eax", "ecx", "edx", "esi", "edi");
}

FLUKE_ENTRY_INLINE void fluke_thread_yield(
	fluke_thread_t *cur_thread,
	fluke_thread_t *target)
{
#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_thread_yield_);
#endif
	asm volatile("call      _fluke_thread_yield"
		: : "a" (cur_thread), "d" (target) : "eax", "ecx", "edx");
}

#endif /* __GNUC__ */

#endif /* _FLUKE_X86_THREAD_H_ */
