/*
 * Copyright (c) 1997-1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * Contributed by the Computer Security Research division,
 * INFOSEC Research and Technology Office, NSA.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#if FLASK

#include "avc.h"
#include "debug.h"
#include "kalloc.h"
#include "mutex.h"
#include "private_interfaces.h"

#include <oskit/c/stdio.h>	/* NULL, printf */

#include <oskit/c/sys/time.h>
extern volatile struct timeval time;

#ifdef DEBUG
/*
 * Allow all operations, even if permission would be denied.
 */
static oskit_bool_t avc_always_allow = TRUE;

/*
 * Generate diagnostic output on permission failures, even if
 * avc_always_allow is TRUE.
 */
static oskit_bool_t avc_be_verbose = TRUE;

#endif				/* DEBUG */

#if AVC_KEEP_STATS
/* cache statistics */
unsigned        avc_cache_stats[AVC_NSTATS];
#endif

link_t          security_port_link;


/*
 * The access vector cache must initially contain sufficient information to
 * permit the system to bootstrap to the point where the security policy
 * server becomes available.
 */

typedef struct {
    security_class_t sclass;
    security_id_t   ssid;
    security_class_t tclass;
    security_id_t   tsid;
    access_vector_t allowed;
    oskit_bool_t    wired;
}               avc_initial_state_t;

static avc_initial_state_t avc_initial_state[] = {
#include "avc_initial_state.h"
};

#define AVC_INITIAL_STATE_SIZE (sizeof(avc_initial_state)/sizeof(avc_initial_state_t))


typedef struct avc_node {
    struct avc_entry ae;
    oskit_bool_t    wired;
    struct avc_node *next;

#ifdef DEBUG
    /* debugging information */
#define CONTEXTSZ 200
    char            smcontext[CONTEXTSZ];
    char            sacontext[CONTEXTSZ];
    char            secontext[CONTEXTSZ];
    char            tmcontext[CONTEXTSZ];
    char            tacontext[CONTEXTSZ];
    char            tecontext[CONTEXTSZ];
#endif				/* DEBUG */
}               avc_node_t;

static struct avc_node *avc_node_freelist = NULL;


#define AVC_CACHE_SLOTS 128
#define AVC_CACHE_MAXNODES 102

typedef struct {
    struct mutex    lock;
    avc_node_t     *slots[AVC_CACHE_SLOTS];
    unsigned int    lru_hint;	/* LRU hint for reclaim scan */
    unsigned int    activeNodes;
    oskit_u32_t     latest_notif;
}               avc_cache_t;


static avc_cache_t avc_cache;

#define AVC_HASH(sclass,ssid,tclass,tsid) \
    ((tsid ^ (tclass<<2) ^ (ssid<<1) ^ (sclass<<4)) & (AVC_CACHE_SLOTS - 1))


static avc_node_t *
avc_reclaim_node(void)
{
    avc_node_t     *prev, *current;
    int             hvalue, try;


    hvalue = avc_cache.lru_hint;
    for (try = 0; try < 2; try++) {
        do {
	    prev = NULL;
	    current = avc_cache.slots[hvalue];
	    while (current) {
	        if (!current->ae.used)
		    /* found a node that has not been "recently" used */
		    goto found;

	        if (!current->wired)
		    /*
		     * As we scan for a node which has not been "recently" used,
		     * mark all unwired nodes as if they had not been recently
		     * used.
		     */
		    current->ae.used = FALSE;

	        prev = current;
	        current = current->next;
	    }
	    hvalue = (hvalue + 1) & (AVC_CACHE_SLOTS - 1);
        } while (hvalue != avc_cache.lru_hint);
    }

    /* all nodes in the cache are wired; nothing to reclaim */
    return NULL;

found:
    avc_cache.lru_hint = hvalue;

    if (prev == NULL)
	avc_cache.slots[hvalue] = current->next;
    else
	prev->next = current->next;

    /* update the entry's version to invalidate any pointers to it */
    current->ae.version++;

    return current;
}


static avc_node_t *
avc_claim_node(security_class_t sclass,
	       security_id_t ssid,
	       security_class_t tclass,
	       security_id_t tsid,
	       oskit_bool_t wired)
{
    avc_node_t     *new;
    int             hvalue;


    /* compute the hash value for the key */
    hvalue = AVC_HASH(sclass, ssid, tclass, tsid);

    /* find a node to store the new (key, datum) pair */
    if (avc_cache.activeNodes == AVC_CACHE_MAXNODES) {
	/* the cache is full -- so reclaim a node */
	new = avc_reclaim_node();
	if (!new)
	    return NULL;
    } else {
	/* the cache can still grow -- try to allocate a new node */
	if (avc_node_freelist) {
	    new = avc_node_freelist;
	    avc_node_freelist = avc_node_freelist->next;
	} else {
	    new = (avc_node_t *) kalloc(sizeof(avc_node_t));
	    new->ae.version = 0;
	}
	if (!new) {
	    /* out of memory -- try reclaiming a node */
	    new = avc_reclaim_node();
	    if (!new)
		return NULL;
	} else {
	    /* an active node has been allocated */
	    avc_cache.activeNodes++;
	}
    }

    new->ae.lock = &avc_cache.lock;
    new->ae.sclass = sclass;
    new->ae.ssid = ssid;
    new->ae.tclass = tclass;
    new->ae.tsid = tsid;
    new->ae.used = TRUE;
    new->wired = wired;
    new->next = avc_cache.slots[hvalue];
    avc_cache.slots[hvalue] = new;
#ifdef DEBUG
    new->smcontext[0] = 0;
    new->sacontext[0] = 0;
    new->secontext[0] = 0;
    new->tmcontext[0] = 0;
    new->tacontext[0] = 0;
    new->tecontext[0] = 0;
#endif				/* DEBUG */

    return new;
}


OSKIT_INLINE avc_node_t *
avc_search_node(security_class_t sclass,
		security_id_t ssid,
		security_class_t tclass,
		security_id_t tsid
#if AVC_KEEP_STATS
		,int *probes
#endif
)
{
    avc_node_t     *current;
    int             hvalue;
#if AVC_KEEP_STATS
    int             tprobes = 1;
#endif


    /* compute the hash value for the key */
    hvalue = AVC_HASH(sclass, ssid, tclass, tsid);

    /* scan the appropriate chain in the hash table */
    current = avc_cache.slots[hvalue];
    while (current != NULL &&
	   (sclass != current->ae.sclass ||
	    ssid != current->ae.ssid ||
	    tclass != current->ae.tclass ||
	    tsid != current->ae.tsid)) {
#if AVC_KEEP_STATS
	tprobes++;
#endif
	current = current->next;
    }

    if (current == NULL) {
	/* cache miss */
	return NULL;
    }
    /* cache hit */

#if AVC_KEEP_STATS
    if (probes)
	*probes = tprobes;
#endif

    current->ae.used = TRUE;

    return current;
}


void
avc_init(void)
{
    avc_initial_state_t *init_node;
    avc_node_t     *node;
    int             index;

    link_init(&security_port_link);

    /*
     * Clear the access vector cache
     */
    for (index = 0; index < AVC_CACHE_SLOTS; index++)
	avc_cache.slots[index] = 0;
    avc_cache.lru_hint = 0;
    avc_cache.activeNodes = 0;
    avc_cache.latest_notif = 0;
    mutex_init(&avc_cache.lock);

#if AVC_KEEP_STATS
    for (index = 0; index < AVC_NSTATS; index++)
	avc_cache_stats[index] = 0;
#endif

    /*
     * Load the initial state into the access vector cache
     */
    mutex_lock(&avc_cache.lock);
    for (index = 0; index < AVC_INITIAL_STATE_SIZE; index++) {
	init_node = &avc_initial_state[index];

	node = avc_search_node(init_node->sclass,
			       init_node->ssid,
			       init_node->tclass,
			       init_node->tsid
#if AVC_KEEP_STATS
			       ,0
#endif
	    );

	if (node)
	  {
	    node->ae.allowed |= init_node->allowed;
	    node->ae.decided |= init_node->allowed;
	  }
	else
	  {
	    node = avc_claim_node(init_node->sclass,
				  init_node->ssid,
				  init_node->tclass,
				  init_node->tsid,
				  init_node->wired);
	    if (!node)
	      panic("avc: unable to add initial state entry");

	    node->ae.allowed = init_node->allowed;
	    node->ae.decided = init_node->allowed;
	    node->ae.expiration = 0;
	  }
    }

    mutex_unlock(&avc_cache.lock);
}


#ifdef DEBUG

#include <flask/class_to_string.h>
#include <flask/common_perm_to_string.h>
#include <flask/av_inherit.h>
#include <flask/av_perm_to_string.h>

void
avc_dump_av(security_class_t sclass,
	    security_class_t tclass,
	    access_vector_t av)
{
    char          **common_pts = 0;
    access_vector_t common_base = 0;
    int             index, index2, perm;


    if (av == 0) {
	printf(" null");
	return;
    }

    for (index = 0; index < AV_INHERIT_SIZE; index++) {
	if ((av_inherit[index].sclass == sclass) &&
	    (av_inherit[index].tclass == tclass)) {
	    common_pts = av_inherit[index].common_pts;
	    common_base = av_inherit[index].common_base;
	    break;
	}
    }

    if (index == AV_INHERIT_SIZE) 
    {
	for (index = 0; index < FLUKE_AV_INHERIT_SIZE; index++) {
	    if ((fluke_av_inherit[index].sclass == sclass) &&
		(fluke_av_inherit[index].tclass == tclass)) {
		common_pts = fluke_av_inherit[index].common_pts;
		common_base = fluke_av_inherit[index].common_base;
		break;
	    }
	}
    }

    printf(" {");
    index = 0;
    perm = 1;
    while (perm < common_base) {
	if (perm & av)
	    printf(" %s", common_pts[index]);
	index++;
	perm <<= 1;
    }

    while (index < sizeof(access_vector_t) * 8) {
	if (perm & av) {
	    for (index2 = 0; index2 < AV_PERM_TO_STRING_SIZE; index2++) {
		if ((av_perm_to_string[index2].sclass == sclass) &&
		    (av_perm_to_string[index2].tclass == tclass) &&
		    (av_perm_to_string[index2].value == perm))
		    break;
	    }
	    if (index2 < AV_PERM_TO_STRING_SIZE)
		printf(" %s", av_perm_to_string[index2].name);
	    else
	    {
		for (index2 = 0; index2 < FLUKE_AV_PERM_TO_STRING_SIZE; index2++) {
		    if ((fluke_av_perm_to_string[index2].sclass == sclass) &&
			(fluke_av_perm_to_string[index2].tclass == tclass) &&
			(fluke_av_perm_to_string[index2].value == perm))
			    break;
		}
		if (index2 < FLUKE_AV_PERM_TO_STRING_SIZE)
			printf(" %s", fluke_av_perm_to_string[index2].name);
	    }
	}
	index++;
	perm <<= 1;
    }

    printf(" }");
}


void
avc_sid_equiv_debug(fluke_type_t stype,
		    security_id_t ssid,
		    fluke_type_t ttype,
		    security_id_t tsid)
{
    printf("avc:  (%d, %d) %ld != %ld\n",
	   stype, ttype, ssid, tsid);
}
#endif				/* DEBUG */

#include <flask/initial_sid_to_string.h>

#define DISPLAY_SID(sid) \
if (sid <= SECINITSID_NUM) \
    printf(" %s)", initial_sid_to_string[sid]); \
else \
    printf(" %ld)", (sid)); 


int
avc_compute_access_vector(security_class_t sclass,
			  security_id_t ssid,
			  security_class_t tclass,
			  security_id_t tsid,
			  access_vector_t requested,
			  access_vector_t * allowed,
			  struct avc_entry ** out_ae,
			  unsigned *out_aever)
{
    struct s_thread *th;
    avc_node_t     *node;
#if AVC_KEEP_STATS
    int             probes;
#endif

    /*
     * Check the access vector cache
     */
    mutex_lock(&avc_cache.lock);

#if AVC_KEEP_STATS
    avc_cache_stats[AVC_CAV_LOOKUPS]++;
#endif
    node = avc_search_node(sclass, ssid, tclass, tsid
#if AVC_KEEP_STATS
			   ,&probes
#endif
	);
    if (node) {
	/*
	 * Check the validity of a cache entry.
	 * 
	 * If a decision on any of the requested permissions is not provided in
	 * the entry, then the entry is invalid.
	 */
	if ((node->ae.decided & requested) == requested) {
	    /* found a valid entry */

#if AVC_KEEP_STATS
	    avc_cache_stats[AVC_CAV_HITS]++;
	    avc_cache_stats[AVC_CAV_PROBES] += probes;
#endif
#ifdef DEBUG
	    if ((requested & node->ae.allowed) != requested) {
		/* permission denied */

		if (avc_be_verbose) {
		    printf("avc:  source (%s,", class_to_string[sclass]);
		    if (node->smcontext[0] && node->sacontext[0]) {
			printf(" (%s %s", node->smcontext, node->sacontext);
			if (node->secontext[0])
			    printf(" %s))", node->secontext);
			else
			    printf("))");
		    } else {
			DISPLAY_SID(node->ae.ssid);
		    }

		    printf(" target (%s,", class_to_string[tclass]);
		    if (node->tmcontext[0] && node->tacontext[0]) {
			printf(" (%s %s", node->tmcontext, node->tacontext);
			if (node->tecontext[0])
			    printf(" %s))", node->tecontext);
			else
			    printf("))");
		    } else {
			DISPLAY_SID(node->ae.tsid);
		    }

		    printf(" requested");
		    avc_dump_av(sclass, tclass, requested);
		    printf("\n");
		}
		if (avc_always_allow)
		    node->ae.allowed |= requested;
	    }
#endif				/* DEBUG */

#if AVC_ALWAYS_MISS
	    /*
	     * Invalidate the requested permission in the cache entry, so
	     * that each check of the permission will cause a security fault.
	     * This provides a sense of the worst-case performance impact of
	     * Flask permission checks.
	     */
	    if (!node->wired)
		node->ae.decided &= ~requested;
#endif

	    *allowed = node->ae.allowed;
#if AVC_ENTRY_ALWAYS_MISS
	    /*
	     * Don't validate second-level cache entries in IPC references,
	     * again ensuring that we take a security fault.
	     */
	    *out_ae = 0;
#else
	    *out_ae = &node->ae;
	    *out_aever = node->ae.version;
#endif
	    mutex_unlock(&avc_cache.lock);
	    return 0;
	} else {
	    /* reuse this node to store the result of the fault */

#if AVC_KEEP_STATS
	    avc_cache_stats[AVC_CAV_MISSES]++;
#endif

	    mutex_unlock(&avc_cache.lock);

	    th = current_thread();
	    assert(th);
	    th->sclass = sclass;
	    th->ssid = ssid;
	    th->tclass = tclass;
	    th->tsid = tsid;
	    th->requested = requested;

	    return KR_SECURITY_FAULT;
	}
    }
#if AVC_KEEP_STATS
    avc_cache_stats[AVC_CAV_MISSES]++;
#endif

    /* claim a node to store the results of the fault */
    node = avc_claim_node(sclass, ssid, tclass, tsid, FALSE);
    if (!node) {
	mutex_unlock(&avc_cache.lock);
	return KR_RESTART;
    }
    node->ae.allowed = 0;
    node->ae.decided = 0;
    mutex_unlock(&avc_cache.lock);

    th = current_thread();
    assert(th);
    th->sclass = sclass;
    th->ssid = ssid;
    th->tclass = tclass;
    th->tsid = tsid;
    th->requested = requested;

    return KR_SECURITY_FAULT;
}


int
do_private_avc_has_permission(void *args)
{
    struct avc_has_permission_args *a = args;
    int            *exc_code = a->exc_code;
    access_vector_t allowed;
    int             rc;

    *exc_code = 1;		/* permission denied */

    rc = avc_compute_access_vector(a->sclass, a->ssid,
				   a->tclass, a->tsid,
				   a->requested, &allowed,
				   0, 0);
    if (rc)
	return rc;

    if ((a->requested & allowed) != a->requested)
	*exc_code = 1;		/* permission denied */
    else
	*exc_code = 0;		/* permission granted */

    return 0;
}


#ifdef DEBUG
static int
decode_longstring(char **buf_currentp,
		  char *endp,
		  char *outcp)
{
    int             index, clen, bp;
    char           *cp;


    /* flick_fluke_decode_new_chunk_align(4, 2, 0, 0); */
    bp = (unsigned int) *buf_currentp;
    bp = (bp + 3) & ~3;
    *buf_currentp = (char *) bp;

    /* flick_fluke_decode_unsigned32(0, _temp_len); */
    clen = *((unsigned int *) *buf_currentp);
    assert(clen <= CONTEXTSZ);
    if (clen > CONTEXTSZ)
	return -1;

    /* flick_fluke_decode_end_chunk(4); */
    *buf_currentp += 4;

    cp = *buf_currentp;
    assert((cp + clen) <= endp);
    if ((cp + clen) > endp)
	return -1;

    for (index = 0; index < clen; index++)
	outcp[index] = cp[index];

    /* flick_fluke_decode_end_chunk(_temp_len); */
    *buf_currentp += clen;
    return 0;
}
#endif				/* DEBUG */


int
avc_update_cache(security_class_t sclass,
		 security_id_t ssid,
		 security_class_t tclass,
		 security_id_t tsid,
		 fluke_ipc_word_t * buf,
		 oskit_size_t words)
{
    avc_node_t     *node;
    oskit_u32_t     seqno;
#ifdef DEBUG
    char           *bp, *endp;
#endif				/* DEBUG */


    mutex_lock(&avc_cache.lock);

    assert(words >= 4);
    if (words < 4)
	goto bad;
    seqno = buf[3];

    if (seqno < avc_cache.latest_notif) {
	mutex_unlock(&avc_cache.lock);
	return 0;
    }
    node = avc_search_node(sclass, ssid, tclass, tsid
#if AVC_KEEP_STATS
			   ,0
#endif
	);
    if (!node) {
	mutex_unlock(&avc_cache.lock);
	return 0;
    }
    node->ae.allowed = buf[0];
    node->ae.decided = buf[1];
    node->ae.expiration = buf[2];

    /* limit allowed by decided */
    node->ae.allowed &= node->ae.decided;

#ifdef DEBUG
    bp = (char *) &buf[4];
    endp = (char *) &buf[words];

    if (decode_longstring(&bp, endp, node->smcontext) ||
	decode_longstring(&bp, endp, node->sacontext) ||
	decode_longstring(&bp, endp, node->secontext) ||
	decode_longstring(&bp, endp, node->tmcontext) ||
	decode_longstring(&bp, endp, node->tacontext) ||
	decode_longstring(&bp, endp, node->tecontext))
	goto bad;
#endif				/* DEBUG */

    mutex_unlock(&avc_cache.lock);

    return 0;

bad:
    mutex_unlock(&avc_cache.lock);

    return -1;
}

#if AVC_KEEP_STATS
void
avc_dump_stats(char *tag)
{
    printf("%s avc:  entry:  %d lookups == %d hits + %d misses\n",
	   tag,
	   avc_cache_stats[AVC_ENTRY_LOOKUPS],
	   avc_cache_stats[AVC_ENTRY_HITS],
	   avc_cache_stats[AVC_ENTRY_MISSES]);

    printf("%s avc:  cav:  %d lookups == %d hits + %d misses\n",
	   tag,
	   avc_cache_stats[AVC_CAV_LOOKUPS],
	   avc_cache_stats[AVC_CAV_HITS],
	   avc_cache_stats[AVC_CAV_MISSES]);

    printf("%s avc:  cav:  %d/%d probe/hit ratio\n",
	   tag,
	   avc_cache_stats[AVC_CAV_PROBES],
	   avc_cache_stats[AVC_CAV_HITS]);
}
#endif

#ifdef DEBUG
void
avc_debug_handle_fault(security_class_t sclass,
		       security_id_t ssid,
		       security_class_t tclass,
		       security_id_t tsid,
		       access_vector_t requested)
{
    avc_node_t     *node;


    if (avc_be_verbose) {
	printf("avc:  source (%s, %ld)", class_to_string[sclass], ssid);
	printf(" target (%s, %ld)", class_to_string[tclass], tsid);
	printf(" requested");
	avc_dump_av(sclass, tclass, requested);
	printf("\n");
    }
    if (!avc_always_allow)
	panic("avc:  no security policy server");

    mutex_lock(&avc_cache.lock);
    node = avc_search_node(sclass, ssid, tclass, tsid
#if AVC_KEEP_STATS
			   ,0
#endif
			   );
    if (!node)
	panic("avc:  no security policy server");

    node->ae.allowed |= requested;
    node->ae.decided |= requested;
    node->ae.expiration = 0;

    mutex_unlock(&avc_cache.lock);

    return;
}


void
avc_dump(char *tag)
{
    int             index, chain_len, max_chain_len, slots_used;
    avc_node_t     *node;


    mutex_lock(&avc_cache.lock);

#if AVC_KEEP_STATS
    avc_dump_stats(tag);
#endif

    slots_used = 0;
    max_chain_len = 0;
    for (index = 0; index < AVC_CACHE_SLOTS; index++) {
	node = avc_cache.slots[index];
	if (node) {
	    printf("\n%s avc:  slot %d:\n", tag, index);
	    slots_used++;
	    chain_len = 0;
	    while (node) {
		printf("\tsource (%s,", class_to_string[node->ae.sclass]);
		if (node->smcontext[0] && node->sacontext[0]) {
		    printf(" (%s %s", node->smcontext, node->sacontext);
		    if (node->secontext[0])
			printf(" %s))", node->secontext);
		    else
			printf("))");
		} else {
		    DISPLAY_SID(node->ae.ssid);
		}

		printf(" target (%s,", class_to_string[node->ae.tclass]);
		if (node->tmcontext[0] && node->tacontext[0]) {
		    printf(" (%s %s", node->tmcontext, node->tacontext);
		    if (node->tecontext[0])
			printf(" %s))", node->tecontext);
		    else
			printf("))");
		} else {
		    DISPLAY_SID(node->ae.tsid);
		}
		printf(" allowed");
		avc_dump_av(node->ae.sclass, node->ae.tclass, node->ae.allowed);
		printf("\n");

		chain_len++;
		node = node->next;
	    }

	    if (chain_len > max_chain_len)
		max_chain_len = chain_len;
	}
    }

    printf("\n%s avc:  %d nodes and %d slots used, longest chain length %d\n",
	   tag, avc_cache.activeNodes, slots_used, max_chain_len);

    printf("%s avc:  latest_notif=%d\n", tag, avc_cache.latest_notif);

    mutex_unlock(&avc_cache.lock);
}
#endif


int
do_private_avc_add(void *args)
{
    struct avc_add_args *a = args;
    avc_node_t     *node;
    oskit_u32_t     seqno = a->seqno;
    int            *exc_code = a->exc_code;


    *exc_code = 0;

    mutex_lock(&avc_cache.lock);

    if (seqno < avc_cache.latest_notif) {
	mutex_unlock(&avc_cache.lock);
	return 0;
    }
    node = avc_search_node(a->sclass, a->ssid, a->tclass, a->tsid
#if AVC_KEEP_STATS
			   ,0
#endif
	);
    if (!node) {
	node = avc_claim_node(a->sclass, a->ssid, a->tclass, a->tsid,
			      a->wired);
	if (!node) {
	    mutex_unlock(&avc_cache.lock);
	    return KR_RESTART;
	}
	node->ae.allowed = a->allowed;
	node->ae.decided = a->decided;
	node->ae.expiration = a->expiration;
	mutex_unlock(&avc_cache.lock);

	return 0;
    }
    if (!a->replace) {
	mutex_unlock(&avc_cache.lock);
	*exc_code = EEXIST;
	return 0;
    }
    node->wired = a->wired;
    if (node->wired)
	node->ae.used = TRUE;

    node->ae.allowed = a->allowed;
    node->ae.decided = a->decided;
    node->ae.expiration = a->expiration;

    /* limit allowed by decided */
    node->ae.allowed &= node->ae.decided;

    mutex_unlock(&avc_cache.lock);

    return 0;
}


static int
avc_remove_node(avc_node_t * prev, avc_node_t * current, int index)
{
    /* update the node's predecessor in the hash table chain */
    if (prev == NULL) {
	/* the node was the first element of the chain */
	avc_cache.slots[index] = current->next;
    } else {
	/* the node had a predecessor in the chain */
	prev->next = current->next;
    }

    /* update the entry's version to invalidate any pointers to it */
    current->ae.version++;

    /* free the node */
    current->next = avc_node_freelist;
    avc_node_freelist = current;
    avc_cache.activeNodes--;

    return 0;
}


static int
avc_remove_one(security_class_t sclass,
	       security_id_t ssid,
	       security_class_t tclass,
	       security_id_t tsid,
	       oskit_bool_t removewired,
	       int *exc_code)
{
    avc_node_t     *prev, *current;
    int             hvalue;


    *exc_code = 0;

    /* compute the hash value for the key */
    hvalue = AVC_HASH(sclass, ssid, tclass, tsid);

    /* scan the appropriate chain in the hash table */
    prev = NULL;
    current = avc_cache.slots[hvalue];
    while (current != NULL &&
	   (sclass != current->ae.sclass ||
	    ssid != current->ae.ssid ||
	    tclass != current->ae.tclass ||
	    tsid != current->ae.tsid)) {
	prev = current;
	current = current->next;
    }

    if (current == NULL) {
	/* cache miss */
	*exc_code = ENOENT;
	return 0;
    }
    /* cache hit */

    if (current->wired && !removewired) {
	/* only remove wired entries if explicitly requested */
	*exc_code = ENOENT;
	return 0;
    }
    return avc_remove_node(prev, current, hvalue);
}


/*
 * Note that argument order is important; only the input parameters to
 * avc_remove may be wildcarded; an avc node never has wildcarded fields.
 */
#define AVC_CLASSCMP(x,y) ((x) == (y) || (x) == SECCLASS_WILD)
#define AVC_SIDCMP(x,y) ((x) == (y) || (x) == SECSID_WILD)

static int
avc_remove_any(security_class_t sclass,
	       security_id_t ssid,
	       security_class_t tclass,
	       security_id_t tsid,
	       oskit_bool_t removewired,
	       int *exc_code)
{
    avc_node_t     *prev, *current, *temp;
    int             index, rc;


    *exc_code = 0;

    for (index = 0; index < AVC_CACHE_SLOTS; index++) {
	prev = NULL;
	current = avc_cache.slots[index];
	while (current != NULL) {
	    if (AVC_CLASSCMP(sclass, current->ae.sclass) &&
		AVC_SIDCMP(ssid, current->ae.ssid) &&
		AVC_CLASSCMP(tclass, current->ae.tclass) &&
		AVC_SIDCMP(tsid, current->ae.tsid) &&
		(removewired || !current->wired)) {
		temp = current;
		current = current->next;
		rc = avc_remove_node(prev, temp, index);
		if (rc)
		    return rc;
		continue;
	    }
	    prev = current;
	    current = current->next;
	}
    }

    return 0;
}


int
do_private_avc_reset(void *args)
{
    struct avc_reset_args *a = args;
    int            *exc_code = a->exc_code;
    avc_initial_state_t *init_node;
    avc_node_t     *node;
    int             index, rc;


    mutex_lock(&avc_cache.lock);

    /* Flush the cache */
    rc = avc_remove_any(SECCLASS_WILD, SECSID_WILD,
			SECCLASS_WILD, SECSID_WILD,
			TRUE,
			exc_code);
    if (rc) {
	mutex_unlock(&avc_cache.lock);
	return rc;
    }
    /* Reload the initial state */
    for (index = 0; index < AVC_INITIAL_STATE_SIZE; index++) {
	init_node = &avc_initial_state[index];
	node = avc_search_node(init_node->sclass,
			       init_node->ssid,
			       init_node->tclass,
			       init_node->tsid
#if AVC_KEEP_STATS
			       ,0
#endif
	    ); 

	if (node)
	  {
	    node->ae.allowed |= init_node->allowed;
	    node->ae.decided |= init_node->allowed;
	  }
	else
	  {
	    node = avc_claim_node(init_node->sclass,
				  init_node->ssid,
				  init_node->tclass,
				  init_node->tsid,
				  init_node->wired);
	    if (!node)
	      panic("avc: unable to reset initial state entry");
	    node->ae.allowed = init_node->allowed;
	    node->ae.decided = init_node->allowed;
	    node->ae.expiration = 0;
	  }
    }

    mutex_unlock(&avc_cache.lock);

    *exc_code = 0;
    return 0;
}


typedef struct reset_callback_node {
    int             (*callback) (void);
    struct reset_callback_node *next;
}               reset_callback_node_t;

static reset_callback_node_t *reset_callbacks = NULL;

int
avc_register_reset_callback(int (*callback) (void))
{
    reset_callback_node_t *c;

    c = (reset_callback_node_t *) kalloc(sizeof(reset_callback_node_t));
    if (!c)
	return ENOMEM;

    c->callback = callback;

    mutex_lock(&avc_cache.lock);
    c->next = reset_callbacks;
    reset_callbacks = c;
    mutex_unlock(&avc_cache.lock);

    return 0;
}


/* separated since it must repeatedly restart due to security faults */
int
do_private_avc_reset_migrated(void *args)
{
    struct avc_reset_args *a = args;
    reset_callback_node_t *c;
    int            *exc_code = a->exc_code;
    int             seqno = a->seqno;
    int             rc;

    /* recheck migrated permissions, restarting as necessary */
    c = reset_callbacks;
    while (c) {
	rc = c->callback();
	if (rc)
	    return rc;
	c = c->next;
    }

    mutex_lock(&avc_cache.lock);

    if (seqno > avc_cache.latest_notif)
	avc_cache.latest_notif = seqno;

    mutex_unlock(&avc_cache.lock);

    *exc_code = 0;
    return 0;
}


int
do_private_avc_remove(void *args)
{
    struct avc_remove_args *a = args;
    security_class_t sclass = a->sclass;
    security_id_t   ssid = a->ssid;
    security_class_t tclass = a->tclass;
    security_id_t   tsid = a->tsid;
    oskit_bool_t    removewired = a->wired;
    int             rc, *exc_code = a->exc_code;


    *exc_code = 0;

    mutex_lock(&avc_cache.lock);

    if (sclass != SECCLASS_WILD && ssid != SECSID_WILD &&
	tclass != SECCLASS_WILD && tsid != SECSID_WILD) {
	rc = avc_remove_one(sclass, ssid, tclass, tsid, removewired, exc_code);
    } else {
	rc = avc_remove_any(sclass, ssid, tclass, tsid, removewired, exc_code);
    }

    mutex_unlock(&avc_cache.lock);

    return rc;
}


int
do_private_avc_revoke(void *args)
{
    struct avc_revoke_args *a = args;
    security_class_t sclass = a->sclass;
    security_id_t   ssid = a->ssid;
    security_class_t tclass = a->tclass;
    security_id_t   tsid = a->tsid;
    access_vector_t revoke = a->revoke;
    oskit_bool_t    revokewired = a->wired;
    avc_node_t     *node;
    int            *exc_code = a->exc_code;

    *exc_code = 0;

    mutex_lock(&avc_cache.lock);

    node = avc_search_node(sclass, ssid, tclass, tsid
#if AVC_KEEP_STATS
			   ,0
#endif
	);

    if (node && node->wired && !revokewired) {
	mutex_unlock(&avc_cache.lock);
	*exc_code = EBUSY;	/* XXX */
	return 0;
    }
    if (node)
	node->ae.allowed &= ~revoke;

    mutex_unlock(&avc_cache.lock);

    return 0;
}


typedef struct revoke_callback_node {
    security_class_t sclass;
    security_id_t   ssid;
    security_class_t tclass;
    security_id_t   tsid;
    access_vector_t revoke;

    int             (*callback) (security_class_t sclass,
				                 security_id_t ssid,
				                 security_class_t tclass,
				                 security_id_t tsid,
				                 access_vector_t revoke);

    struct revoke_callback_node *next;
}               revoke_callback_node_t;

static revoke_callback_node_t *revoke_callbacks = NULL;


int
avc_register_revoke_callback(security_class_t sclass,
			     security_id_t ssid,
			     security_class_t tclass,
			     security_id_t tsid,
			     access_vector_t revoke,
			     int (*callback) (security_class_t sclass,
					      security_id_t ssid,
					      security_class_t tclass,
					      security_id_t tsid,
					      access_vector_t revoke))
{
    revoke_callback_node_t *c;

    c = (revoke_callback_node_t *) kalloc(sizeof(revoke_callback_node_t));
    if (!c)
	return ENOMEM;

    c->sclass = sclass;
    c->ssid = ssid;
    c->tclass = tclass;
    c->tsid = tsid;
    c->revoke = revoke;

    c->callback = callback;

    mutex_lock(&avc_cache.lock);
    c->next = revoke_callbacks;
    revoke_callbacks = c;
    mutex_unlock(&avc_cache.lock);

    return 0;
}


int
do_private_avc_stop_all_threads(void *args)
{
    return s_ob_stop_all_threads();
}


int
do_private_avc_release_all_threads(void *args)
{
    return s_ob_release_all_threads();
}


int
do_private_avc_revoke_migrated(void *args)
{
    struct avc_revoke_args *a = args;
    security_class_t sclass = a->sclass;
    security_id_t   ssid = a->ssid;
    security_class_t tclass = a->tclass;
    security_id_t   tsid = a->tsid;
    access_vector_t revoke = a->revoke;
    oskit_u32_t     seqno = a->seqno;
    revoke_callback_node_t *c;
    int             rc, *exc_code = a->exc_code;

    *exc_code = 0;

    c = revoke_callbacks;
    while (c) {
	if (AVC_CLASSCMP(c->sclass, sclass) && AVC_CLASSCMP(c->tclass, tclass) &&
	    AVC_SIDCMP(c->ssid, ssid) && AVC_SIDCMP(c->tsid, tsid) &&
	    (c->revoke & revoke)) {
	    rc = c->callback(sclass, ssid, tclass, tsid, (c->revoke & revoke));
	    if (rc)
		return rc;
	}
	c = c->next;
    }

    mutex_lock(&avc_cache.lock);

    if (seqno > avc_cache.latest_notif)
	avc_cache.latest_notif = seqno;

    mutex_unlock(&avc_cache.lock);

    return 0;
}


typedef struct try_remove_callback_node {
    security_class_t sclass;
    security_id_t   ssid;
    security_class_t tclass;
    security_id_t   tsid;
    access_vector_t permissions;

    int             (*callback) (security_class_t sclass,
				                 security_id_t ssid,
				                 security_class_t tclass,
				                 security_id_t tsid,
				                 access_vector_t permissions,
			                    access_vector_t * out_retained);

    struct try_remove_callback_node *next;
}               try_remove_callback_node_t;

static try_remove_callback_node_t *try_remove_callbacks = NULL;

int
avc_register_try_remove_callback(security_class_t sclass,
				 security_id_t ssid,
				 security_class_t tclass,
				 security_id_t tsid,
				 access_vector_t permissions,
				 int (*callback) (security_class_t sclass,
						  security_id_t ssid,
						  security_class_t tclass,
						  security_id_t tsid,
						access_vector_t permissions,
					    access_vector_t * out_retained))
{
    try_remove_callback_node_t *c;

    c = (try_remove_callback_node_t *) kalloc(sizeof(try_remove_callback_node_t));
    if (!c)
	return ENOMEM;

    c->sclass = sclass;
    c->ssid = ssid;
    c->tclass = tclass;
    c->tsid = tsid;
    c->permissions = permissions;

    c->callback = callback;

    mutex_lock(&avc_cache.lock);
    c->next = try_remove_callbacks;
    try_remove_callbacks = c;
    mutex_unlock(&avc_cache.lock);

    return 0;
}


int
do_private_avc_try_remove(void *args)
{
    struct avc_try_remove_args *a = args;
    security_class_t sclass = a->sclass;
    security_id_t   ssid = a->ssid;
    security_class_t tclass = a->tclass;
    security_id_t   tsid = a->tsid;
    access_vector_t permissions = a->permissions;
    oskit_bool_t    removewired = a->wired;
    oskit_u32_t     seqno = a->seqno;
    access_vector_t *out_retained = a->retained;
    try_remove_callback_node_t *c;
    int             rc, *exc_code = a->exc_code;
    avc_node_t     *node;
    access_vector_t tretained, cretained;

    *exc_code = 0;
    tretained = 0;

    c = try_remove_callbacks;
    while (c) {
	if (AVC_CLASSCMP(c->sclass, sclass) &&
	    AVC_CLASSCMP(c->tclass, tclass) &&
	    AVC_SIDCMP(c->ssid, ssid) &&
	    AVC_SIDCMP(c->tsid, tsid) &&
	    (c->permissions & permissions)) {
	    cretained = 0;
	    rc = c->callback(sclass, ssid, tclass, tsid,
			     (c->permissions & permissions),
			     &cretained);
	    if (rc)
		return rc;
	    cretained &= permissions;	/* just in case callback is verbose */
	    tretained |= cretained;
	    if (tretained == permissions)
		/* no need to continue searching */
		break;
	}
	c = c->next;
    }

    mutex_lock(&avc_cache.lock);
    node = avc_search_node(sclass, ssid, tclass, tsid
#if AVC_KEEP_STATS
			   ,0
#endif
	);

    if (node && node->wired && !removewired)
	/* all permissions granted in a wired node are also retained */
	tretained |= (permissions & node->ae.allowed);

    if (tretained != permissions) {
	/* not all specified permissions were retained */

	if (node) {
	    /*
	     * Remove any of the specified permissions which are allowed in
	     * the avc but not retained in the state.
	     */
	    node->ae.decided &= ~(permissions & node->ae.allowed & ~tretained);
	}
	if (seqno > avc_cache.latest_notif)
	    avc_cache.latest_notif = seqno;
    }
    mutex_unlock(&avc_cache.lock);

    *out_retained = tretained;

    return 0;
}


int
do_private_security_set(void *args)
{
    struct security_set_args *a = args;
    int             rc;


    rc = kentry_ref_extract_type(current_thread(),
				 a->ref_uva,
				 FLUKE_PORT,
				 &security_port_link);
    assert(rc == 0);
    return rc;
}
#endif
