/*
 * Copyright (c) 1997-1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * Contributed by the Computer Security Research division,
 * INFOSEC Research and Technology Office, NSA.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#if FLASK

#ifndef _FLUKE_KERNEL_AVC_H_
#define _FLUKE_KERNEL_AVC_H_

/*
 * Indicate whether or not all subject-subject permissions should be granted
 * when the SIDs are equal.
 */
#define AVC_OPTIMIZE_SID_EQUALITY 0

/*
 * Invalidate the requested permission in the cache entry, so
 * that each check of the permission will cause a security fault.
 * This provides a sense of the worst-case performance impact of
 * Flask permission checks.
 */
#define AVC_ALWAYS_MISS		0

/*
 * Don't validate second-level cache entries in IPC references,
 * again ensuring that we take a security fault.
 */
#define AVC_ENTRY_ALWAYS_MISS	0

/*
 * Record/print AVC stats
 */
#ifdef DEBUG
#define AVC_KEEP_STATS	 	1
#else
#define AVC_KEEP_STATS	  	0
#endif

#include <fluke/ob.h>
#include <fluke/ipc.h>
#include <fluke/exception.h>
#include <flask/security.h>
#include <flask/av_permissions.h>
#include "exception.h"
#include "s_thread.h"


struct avc_entry {
    /* lock prior to examining or modifying any other field */
    struct mutex   *lock;

    /* the access query (with the exception of requested) */
    security_class_t sclass;
    security_id_t   ssid;
    security_class_t tclass;
    security_id_t   tsid;

    /* the access ruling */
    access_vector_t allowed;
    access_vector_t decided;
    unsigned int    expiration;

    /* version number; incremented if reclaimed or removed */
    unsigned        version;

    /* used "recently" */
    oskit_bool_t    used;
};


#if AVC_KEEP_STATS
#define AVC_ENTRY_LOOKUPS        0
#define AVC_ENTRY_HITS	         1
#define AVC_ENTRY_MISSES         2
#define AVC_CAV_LOOKUPS          3
#define AVC_CAV_HITS             4
#define AVC_CAV_PROBES           5
#define AVC_CAV_MISSES           6
#define AVC_NSTATS               7
extern unsigned avc_cache_stats[AVC_NSTATS];
#endif


extern void     avc_init(void);

extern int
avc_compute_access_vector(security_class_t sclass,
			  security_id_t ssid,
			  security_class_t tclass,
			  security_id_t tsid,
			  access_vector_t requested,
			  access_vector_t * allowed,
			  struct avc_entry ** out_ae,
			  unsigned *out_aever);

extern int
avc_security_fault(security_class_t sclass,
		   security_id_t ssid,
		   security_class_t tclass,
		   security_id_t tsid,
		   access_vector_t requested);

extern int
avc_update_cache(security_class_t sclass,
		 security_id_t ssid,
		 security_class_t tclass,
		 security_id_t tsid,
		 fluke_ipc_word_t * buf,
		 oskit_size_t words);

#ifdef DEBUG
extern void
avc_debug_handle_fault(security_class_t sclass,
		       security_id_t ssid,
		       security_class_t tclass,
		       security_id_t tsid,
		       access_vector_t requested);
#endif				/* DEBUG */

extern void     avc_dump(char *tag);


OSKIT_INLINE int
avc_has_permission(security_class_t sclass,
		   security_id_t ssid,
		   security_class_t tclass,
		   security_id_t tsid,
		   access_vector_t requested)
{
    struct avc_entry *ae;
    unsigned        aever;
    access_vector_t allowed;
    int             rc;

#if AVC_OPTIMIZE_SID_EQUALITY
    if (ssid == tsid)
	return 0;
#endif

    rc = avc_compute_access_vector(sclass, ssid, tclass, tsid, requested,
				   &allowed, &ae, &aever);
    if (rc)
	return (rc);

    if ((requested & allowed) != requested)
	EXC_RETURN_INSANITY(PERMISSION_DENIED);

    return 0;
}


OSKIT_INLINE int
avc_has_permission_reset(security_class_t sclass,
			 security_id_t ssid,
			 security_class_t tclass,
			 security_id_t tsid,
			 access_vector_t requested,
			 oskit_bool_t * revoke)
{
    struct avc_entry *ae;
    unsigned        aever;
    access_vector_t allowed;
    int             rc;

#if AVC_OPTIMIZE_SID_EQUALITY
    if (ssid == tsid)
	return 0;
#endif

    rc = avc_compute_access_vector(sclass, ssid, tclass, tsid, requested,
				   &allowed, &ae, &aever);
    if (rc)
	return (rc);

    *revoke = ((requested & allowed) != requested);

    return 0;
}


#ifdef DEBUG
void
avc_sid_equiv_debug(fluke_type_t stype,
		    security_id_t ssid,
		    fluke_type_t ttype,
		    security_id_t tsid);
#endif				/* DEBUG */

OSKIT_INLINE int
avc_sid_equiv(fluke_type_t stype,
	      security_id_t ssid,
	      fluke_type_t ttype,
	      security_id_t tsid)
{
    if (ssid != tsid) {
#ifdef DEBUG
	avc_sid_equiv_debug(stype, ssid, ttype, tsid);
#endif				/* DEBUG */
	EXC_RETURN_INSANITY(PERMISSION_DENIED);
    }
    return 0;
}


/*
 * Register the specified callback for reset requests.
 * 
 * All registered callbacks will be invoked on a reset request after the cache
 * has been cleared.
 */
int             avc_register_reset_callback(int (*callback) (void));


/*
 * Register the specified callback for revocation requests which match the
 * specified sclass, ssid, tclass, tsid and revoke parameters.  The class and
 * SID parameters to this function may be wildcarded, in which case they will
 * match any value in a revocation request. The revoke value for the callback
 * matches the revoke value in a revocation request if they have a nonnull
 * intersection.
 * 
 * All matching callbacks will be invoked on a revocation request, and will be
 * provided with the actual class and SID parameters to the revocation
 * request.  The revoke parameter to the callback will be the intersection of
 * the actual revoke parameter to the request and the registered revoke
 * vector for the callback. These callbacks are expected to deal with
 * migrated permissions.
 */
extern int
avc_register_revoke_callback(security_class_t sclass,
			     security_id_t ssid,
			     security_class_t tclass,
			     security_id_t tsid,
			     access_vector_t revoke,
			     int (*callback) (security_class_t sclass,
					      security_id_t ssid,
					      security_class_t tclass,
					      security_id_t tsid,
					      access_vector_t revoke));


/*
 * Register the specified callback for try_remove requests.
 */
    extern int
                    avc_register_try_remove_callback(security_class_t sclass,
					                 security_id_t ssid,
				                    security_class_t tclass,
					                 security_id_t tsid,
				                access_vector_t permissions,
				   int (*callback) (security_class_t sclass,
						    security_id_t ssid,
						    security_class_t tclass,
						    security_id_t tsid,
						access_vector_t permissions,
					   access_vector_t * out_retained));

#endif				/* _FLUKE_KERNEL_AVC_H_ */

#endif
