/*
 * Copyright (c) 1997 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include "config_hier_sched.h"
#ifdef CONFIG_HIER_SCHED

#include "config_hier_sched_set_parent.h"

#define ROOT_OWN_TICKETS 8


#include "currency.h"
#include "list.h"
#include <stdio.h>
#include <stdlib.h>
#include "debug.h"
#ifndef _TESTING_
#include "s_thread.h"
#endif

#define RESCALE_PRECISION 100
#define RESCALE_NEW_MINIMUM 4

#define CURRS(t) ((t)->sched_state.curr_state)

#define BASE_PARENT(t) ((CURRS(t).total_tickets) ?\
	CURRS(t).base_tickets*CURRS(t).own_tickets/CURRS(t).total_tickets \
	:0)
#define PARENT_RESCALE_FACTOR(t) \
	((RESCALE_NEW_MINIMUM*CURRS(t).total_tickets * RESCALE_PRECISION)/ \
	CURRS(t).own_tickets / CURRS(t).base_tickets)

#define CHILD_RESCALE_FACTOR(p,c) \
	((RESCALE_NEW_MINIMUM*CURRS(p).total_tickets * RESCALE_PRECISION)/ \
	CURRS(p).base_tickets / CURRS(c).inherit_tickets)

#define BASE_CHILD(p,c) (CURRS(p).total_tickets ?  \
	CURRS(p).base_tickets*CURRS(c).inherit_tickets/ \
	CURRS(p).total_tickets : 0)

/*#define BASE_DEBUG*/

#ifdef BASE_DEBUG
#undef SOURCE
#define SOURCE "currency"
#define BD_BEGIN() DEBUG_BASE_BEGIN()
#define BD_END() DEBUG_BASE_END()
#define BD_BEGIN_ARGS(fmt, args...) DEBUG_BASE_BEGIN_ARGS(fmt,args)
#define BD_END_ARGS(fmt, args...) DEBUG_BASE_END_ARGS(fmt,args)
#define BD() DEBUG_BASE()
#define BD_ARGS(fmt, args...) DEBUG_BASE_ARGS(fmt,args)
#else
#define BD_BEGIN()
#define BD_END()
#define BD_BEGIN_ARGS(args...)
#define BD_END_ARGS(args...)
#define BD()
#define BD_ARGS(args...)
#endif

static void recalculate_tickets(currency_t *c, s_thread_t *t);

#ifdef _TESTING_
void
stride_client_modify(struct s_thread *t, int new_tickets)
{
	t->sched_state.tickets = new_tickets;
}
#endif

#ifdef DEBUG
static int
currency_find_thread(s_thread_t *root, s_thread_t *t)
{
	if (root == t)
		return 1;
	else {
		s_thread_t *child;
		list_t *l = &(CURRS(root).child_list);
		if (!list_empty(l)) {
			list_go_first(l);
			while (!list_end(l)) {
				child = list_get(l);
				list_next(l);
				if (currency_find_thread(child, t))
					return 1;
			}
		}
	}
	return 0;
}

#endif

static void
rescale(currency_t *c, unsigned factor)
{
	CURRS(c->root).base_tickets = (CURRS(c->root).base_tickets*factor)
		/RESCALE_PRECISION;
	recalculate_tickets(c, c->root);
}


static void
recalculate_tickets(currency_t *c, s_thread_t *t)
{
	/* Here we could optize by making all the changes and
	   reordering all of those that changed at a time */
	s_thread_t *child;
	unsigned num_tickets;
	list_t *l = &(CURRS(t).child_list);
	BD_BEGIN_ARGS("t=%p", t);

#ifdef DEBUG
	if (!CURRS(t).base_tickets) {
		printf("thread=%p\n", t);
		currency_dump_state(t);
		currency_dump_tree(c);
	}
#endif
	assert(CURRS(t).base_tickets); /* For debugging only */
	assert(CURRS(t).own_tickets); /* For debugging only */
	num_tickets = BASE_PARENT(t);
	if (!num_tickets && CURRS(t).base_tickets && CURRS(t).own_tickets) {
		rescale(c, PARENT_RESCALE_FACTOR(t));
	} else {
		stride_client_modify(t, num_tickets);
		if (!list_empty(l)) {
			list_go_first(l);
			while (!list_end(l)) {
				child = list_get(l);
				list_next(l);
				assert(CURRS(child).inherit_tickets);
				num_tickets = BASE_CHILD(t, child);
				if (!num_tickets && CURRS(t).base_tickets &&
				    CURRS(child).inherit_tickets) {
					rescale(c,
						CHILD_RESCALE_FACTOR(t, child));
					break;
				} else {
					CURRS(child).base_tickets = num_tickets;
					recalculate_tickets(c, child);
				}
			}
		}
	}
	BD_END();
}


void
currency_init(currency_t *c, s_thread_t *t, unsigned base_tickets)
{
	BD_BEGIN_ARGS("c=%p, t=%p, tickets=%u", c, t, base_tickets);
	c->root = t;
	CURRS(t).base_tickets = base_tickets;
	CURRS(t).total_tickets = CURRS(t).own_tickets = ROOT_OWN_TICKETS;
	CURRS(t).inherit_tickets = 0;
	BD_END();
}


/*
 * Currency modify
 */
static void
remove_from_current_parent(currency_t *c, s_thread_t *t)
{
	BD_BEGIN_ARGS("t=%p", t);
	/* First I remove children from his old parent */
	list_find_remove(&(CURRS(CURRS(t).parent).child_list), t);
	/* Reflect the change in tickets to his old brothers */
	CURRS(CURRS(t).parent).total_tickets -= CURRS(t).own_tickets;
	assert (CURRS(CURRS(t).parent).total_tickets >= 0);
	recalculate_tickets(c, CURRS(t).parent);
	BD_END();
}

static void
change_ticket_values(currency_t *c, s_thread_t *t, s_thread_t *parent,
		     unsigned inherit_tickets, unsigned own_tickets)
{
	BD_BEGIN();
	/* Reflect the change in the number of
	   tickets to its brothers */
	CURRS(CURRS(t).parent).total_tickets =
		CURRS(CURRS(t).parent).total_tickets-
		CURRS(t).inherit_tickets+inherit_tickets;
	CURRS(t).inherit_tickets = inherit_tickets;
	CURRS(t).total_tickets =
		CURRS(t).total_tickets - CURRS(t).own_tickets + own_tickets;
	CURRS(t).own_tickets = own_tickets;
	recalculate_tickets(c, CURRS(t).parent);
	BD_END();
}


/* I to do it all at once to make it more efficient */
void
currency_modify(currency_t *c, s_thread_t *t, s_thread_t *parent,
		unsigned inherit_tickets, unsigned own_tickets)
{
	BD_BEGIN_ARGS("c=%p, t=%p, parent=%p, inh_tick=%u, own_tick=%u",
		      c, t, parent, inherit_tickets, own_tickets);
	if (t != c->root) {
		assert(parent != NULL);
		if (CURRS(t).parent != parent) { /* Parent changes */
			if (CURRS(t).parent != NULL)
				remove_from_current_parent(c, t);
			/* I insert children in the new parent */
			list_insert(&(CURRS(parent).child_list), t);
			CURRS(t).parent = parent;
			/* Update added tickets to new parent */
			CURRS(parent).total_tickets += CURRS(t).inherit_tickets;
		}
		change_ticket_values(c, t, parent, inherit_tickets,
				     own_tickets);
	}
	BD_END();
}


void
currency_state_init(s_thread_t *t)
{
	BD_BEGIN_ARGS("t=%p", t);
	CURRS(t).own_tickets = 4;
	CURRS(t).total_tickets = 4;
	CURRS(t).inherit_tickets = 4;
	CURRS(t).base_tickets = 0;
	CURRS(t).parent = NULL;
	list_create(&(CURRS(t).child_list));
	BD_END();
}


#ifdef DEBUG
void
currency_dump_state(s_thread_t *t)
{
	DEBUG_INDENT();
	printf("Thread %p, own_t=%u, tot_t=%u, inh_t=%u, b_t=%u,"
	       " p=%p, t->tick=%u\n", t, CURRS(t).own_tickets,
	       CURRS(t).total_tickets, CURRS(t).inherit_tickets,
	       CURRS(t).base_tickets, CURRS(t).parent, t->sched_state.tickets);
}


void
currency_dump_tree(currency_t *c)
{
	BD_BEGIN();
	DEBUG_INDENT();
	printf("Currency tree dump\n");
	DEBUG_INDENT();
	printf("------------------\n");
	if (c->root != NULL)
		currency_dump(c->root);
	BD_END();
}

/* PRECOND: t must be <> NULL */
void
currency_dump(s_thread_t *t)
{
	s_thread_t *child;
	list_t *l = &(CURRS(t).child_list);

	currency_dump_state(t);
	if (!list_empty(l)) {
		list_go_first(l);
		while (!list_end(l)) {
			child = list_get(l);
			list_next(l);
			currency_dump(child);
		}
	}
}

int
currency_check(s_thread_t *t)
{
	s_thread_t *child;
	list_t *l = &(CURRS(t).child_list);

	if (t->sched_state.tickets > 1024*8 || t->sched_state.tickets <= 0) {
		return 1;
	}
	if (!list_empty(l)) {
		list_go_first(l);
		while (!list_end(l)) {
			child = list_get(l);
			list_next(l);
			if (currency_check(child))
				return 1;
		}
	}
	return 0;
}

int
currency_check_tree(currency_t *c)
{
	/* We can't check the tree if we are using rescaling as we look for
	   maximum number of tickets */
#ifdef NO_RESCALING
	if (c->root != NULL)
		return currency_check(c->root);
	else
#endif
		return 0;
}
#endif


#ifndef _TESTING_
void
currency_set_state(currency_t *c, struct s_thread *target,
		   fluke_thread_state_t *state)
{
	s_thread_t *t;

	/*BD_BEGIN_ARGS("thread=%p, sched_state.kserver=%d",
		      target, target->sched_state.kserver);*/
#ifdef CONFIG_HIER_SCHED_SET_PARENT
	if (!target->sched_state.kserver) {
		t = (s_thread_t *)link_follow(&target->sched_state.sched_link);
		currency_modify(c, target, t, state->tickets,
				state->own_tickets);
		if (t)
			s_ob_unlock(&t->sob);
	} else {
		/* If is the kserver make it a child of the root thread */
		currency_modify(c, target, c->root, state->tickets,
				state->own_tickets);
	}
#else
	t = current_thread();
	currency_modify(c, target, t, state->tickets, state->own_tickets);
#endif
	BD_END();
}

/* Get the currency state */
void
currency_get_state(currency_t *c, struct s_thread *target,
		   fluke_thread_state_t *state)
{
	BD_BEGIN_ARGS("thread=%p", target);
	state->own_tickets = CURRS(target).own_tickets;
	state->tickets = CURRS(target).inherit_tickets;
	/* The parent field is still saved from last time it was set */
	BD_END();
}

#endif
static void
reparent(s_thread_t *new_p, s_thread_t *old_p, list_t *l)
{
	s_thread_t *child;
	unsigned size, total_tickets, child_tickets;

	size = list_size(l);
	list_go_first(l);

	total_tickets = CURRS(old_p).total_tickets - CURRS(old_p).own_tickets;
	child_tickets = 0;
	while (!list_empty(l)) {
		child = list_get(l);
		/* Update tickets of the child */
		CURRS(child).inherit_tickets = CURRS(old_p).inherit_tickets*
			CURRS(child).inherit_tickets/total_tickets;
		/*Just for debbuging. We don't whant to have 0 inh_tickets*/
		assert(CURRS(child).inherit_tickets);
		/* Until I implememt ticket factor or other mechanism */
		assert(CURRS(child).inherit_tickets);
		/* Change the parent */
		CURRS(child).parent = new_p;
		list_insert(&(CURRS(new_p).child_list), child);
		list_remove(l);
		child_tickets += CURRS(child).inherit_tickets;
	}
	/* We have to do this because we might loose
	   tickets in the integer divisions */
	CURRS(new_p).total_tickets +=
		child_tickets - CURRS(old_p).inherit_tickets;
}


/* A thread is being destroyed. Clean the currency state */
void
currency_thread_destroy(currency_t *c, s_thread_t *th)
{

	BD_BEGIN();
#ifdef DEBUG
	/*
	 * Check that thread to destroy is in the tree.
	 * XXX note that we do occasionally create threads which we never
	 * set_state on (and never run).  They will have parent == 0.
	 */
	if (!currency_find_thread(c->root, th) && CURRS(th).parent) {
		printf("currency_thread_destroy called on non existent "
		       "thread %p\n", th);
		currency_dump_tree(c);
	} else
#endif
	if (CURRS(th).parent) {
		/* Check we are not destroying the parent thread */
		assert(c->root != th);

		/* First I remove children from his old parent */
		list_find_remove(&(CURRS(CURRS(th).parent).child_list), th);

		/* Check if he has any children */
		if (list_empty(&(CURRS(th).child_list))) {
			/* Reflect the change in tickets to his old brothers */
			CURRS(CURRS(th).parent).total_tickets -=
				CURRS(th).inherit_tickets;
			assert(CURRS(CURRS(th).parent).total_tickets >= 0);
		} else {/* Deal with the orphans */
			reparent(CURRS(th).parent, th, &(CURRS(th).child_list));
			list_destroy(&(CURRS(th).child_list));
		}
		recalculate_tickets(c, CURRS(th).parent);
		BD_END();
	}
}


#endif
