/*
 * Copyright (c) 1996, 1998, 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include <setjmp.h>

#include "csw.h"
#include "spin_lock.h"
#include "s_thread.h"
#include "dispatch.h"
#include "preempt.h"
#include "debug.h"
#include "config_smp.h"
#include "config_instrument.h"
#include "config_nopreempt.h"
#include "config_fdev.h"
#include "config_prio_inherit.h"
#include "s_instrument.h"
#include <stdio.h>
#include "sched.h"
#include "per_cpu.h"

#include <fluke/thread.h>

/*#define BASE_DEBUG*/
#ifdef BASE_DEBUG
#define BD_BEGIN() DEBUG_BASE_BEGIN()
#define BD_END() DEBUG_BASE_END()
#define BD() DEBUG_BASE()
#define BD_ARGS(fmt, args...) DEBUG_BASE_ARGS(fmt,args)
#else
#define BD_BEGIN()
#define BD_END()
#define BD()
#define BD_ARGS(args...)
#endif

#ifdef PREEMPTSTATS
#define STAMPTIME(th) { \
	asm volatile("rdtsc" : "=a" ((th)->disp_state.stamp) :: "eax", "edx"); \
	(th)->disp_state.valid = 1; \
}

int preempts, userpreempts, expreempts;
#else
#define STAMPTIME(th)
#endif

#ifndef CONFIG_NOPREEMPT
struct preempter disp_preempter;
#endif


#if DEBUG
oskit_u32_t disp_run_tick;
#endif

void thread_ready(struct s_thread *thread)
{
	int kpreempt, resched;

	assert(thread != current_thread());

	save_preemption_enable(kpreempt);
	disable_preemption();

	spin_lock(&sched_lock);

	INSTRUMENT_EVENT(thread, INSTRUMENT_EVENT_READY, 0);
	STAMPTIME(thread);
	resched = sched_thread_ready(thread);

	spin_unlock(&sched_lock);

#ifndef CONFIG_NOPREEMPT
	if (resched)
		register_preempter(&disp_preempter, disp_yield);
#endif

	restore_preemption_enable(kpreempt);
}


/*
 * This is the dispatcher, which saves the cur_thread state and
 * runs the next thread on the run queue.
 */

s_thread_t *pd_thread_dispatch(
	struct s_thread		*cur_thread,
	spin_lock_t		*unlock_lock,
	struct s_thread		*inheritor,
	struct link		*inheritor_link)
{
	struct s_thread *next_thread;

	assert_preemption_disabled();
	assert(cur_thread == current_thread());
	BD_ARGS("cur_thread=%p",cur_thread);

	spin_lock(&cur_thread->disp_state.lock);

	/*
	 * It's now safe to unlock the specified spin lock;
	 * anyone who locks it and tries to wake us up
	 * will encounter our thread's disp_state.lock instead.
	 */
	spin_unlock(unlock_lock);

#ifndef CONFIG_PRIO_INHERIT
	if (inheritor_link) {
		assert(!inheritor);
		inheritor = (s_thread_t *) link_consume(inheritor_link);
		if (inheritor)
			s_thread_unlock(inheritor);
	}
#else
	if (inheritor) {
		assert(!inheritor_link);
		assert(inheritor != cur_thread);
		spin_lock(&inheritor->disp_state.lock);
		next_thread = inheritor;
	}
	else if (inheritor_link) {
		assert(!inheritor);
		inheritor = (s_thread_t *) link_consume(inheritor_link);
		if (!inheritor)
			goto yield;
		assert(inheritor != cur_thread);
		spin_lock(&inheritor->disp_state.lock);
		s_thread_unlock(inheritor);
		next_thread = inheritor;
	}
	else
	yield:
#endif
	{
		/*
		 * Find another ready thread and run it.
		 * There are at least NCPUs idle threads which never block;
		 * (hah!)  this guarantees that we will never find the
		 * ready queue empty.  (This is a temporary no-brainer
		 * scheduler; it will be replaced with a CPU inheritance
		 * scheduler.)
		 */
		spin_lock(&sched_lock);
		next_thread = sched_thread_next();
		spin_unlock(&sched_lock);
		BD_ARGS("cur_thread=%p next_thread=%p",
			cur_thread, next_thread);

#ifdef CONFIG_SMP
		if (next_thread == cur_thread) {
			/*
			 * Another processor readied the current thread
			 * before it could finish yielding, and now it's
			 * the most eligible thread to be scheduled again!
			 */
			spin_unlock(&cur_thread->disp_state.lock);
			return cur_thread;
		}
#else
		assert(next_thread != cur_thread);
#endif
		spin_lock(&next_thread->disp_state.lock);
	}

	/*
	 * The next thread will release the lock on our own thread.
	 * We can't release it ourselves because we'll still be using our stack
	 * clear up until the time when the switch is actually completed.
	 */
	next_thread->disp_state.last_thread = cur_thread;

#ifdef CONFIG_SMP
	/*
	 * Assign the CPU to the new thread.  We must do this before we
	 * switch into it so that `my_cpu' is defined for interrupt
	 * handlers.  For the same reason, we must leave the old thread's
	 * pointer intact until after we have switched.
	 */
	next_thread->disp_state.cpu = cur_thread->disp_state.cpu;
#endif

	INSTRUMENT_EVENT(cur_thread,
			 INSTRUMENT_EVENT_SWITCH_OUT, 0);

	/* Switch to the selected next thread. */
	cur_thread = csw_switch(cur_thread, next_thread);

	INSTRUMENT_EVENT(cur_thread,
			 INSTRUMENT_EVENT_SWITCH_IN, 0);

#ifdef CONFIG_SMP
	assert(cur_thread->disp_state.cpu);
	assert(cur_thread->disp_state.cpu->number == smp_find_cur_cpu());
	cur_thread->disp_state.last_thread->disp_state.cpu = 0;
#endif

	/*
	 * Change the user-visible thread identity.
	 * This used to be done in the kentry layer when returning
	 * KR_RESTART to user mode, a heuristic indicating that a
	 * context switch occured.  However, it didn't always work
	 * because we don't KR_RESTART after blocking on a kernel
	 * mutex or when preempted; i.e., anything but a cond_wait.
	 */
	csw_set_uthread_va(cur_thread);

#if DEBUG			/* XXX not smp safe */
	cur_thread->disp_state.run_tick = ++disp_run_tick;
#endif

	/*
	 * We were woken up by switching from another thread.
	 * Both threads are still locked.
	 * Just unlock both threads and return.
	 */
	spin_unlock(&cur_thread->disp_state.last_thread->disp_state.lock);
	spin_unlock(&cur_thread->disp_state.lock);

	return cur_thread;
}

/*
 * Supervisor-mode yield() function.
 * If there is another higher priority runnable thread run it.
 * Returns a pointer to the (possibly new) current thread.
 */

s_thread_t *pd_disp_yield(s_thread_t *cur_thread)
{
#ifndef CONFIG_NOPREEMPT
	unregister_preempter(&disp_preempter);
#endif

	spin_lock(&sched_lock);
	if (sched_thread_yield()) {
#ifdef PREEMPTSTATS
		extern int preemptfromuser;

		preempts++;
		if (preemptfromuser) {
			userpreempts++;
			preemptfromuser = 0;
		}
#endif
		STAMPTIME(cur_thread);
		cur_thread = pd_thread_dispatch(cur_thread, &sched_lock, 0, 0);
	} else
		spin_unlock(&sched_lock);

	return cur_thread;
}



#ifndef CONFIG_NOPREEMPT
/*
 * Preempt if there is a runnable thread with priority
 * greater than or equal to that of the current thread.
 */
void
disp_yield(struct preempter *p)
{
	assert_preemption_disabled();
	BD();

	/*
	 * Call the dispatcher to yield the CPU to another thread.
	 * Note that we're blocking rather than waiting -
	 * e.g., other threads won't be able to manhandle us.
	 */
	(void)pd_disp_yield(current_thread());
}
#endif



void thread_disp_state_init(
	struct s_thread		*thread)
{
	BD_ARGS("thread=%p",thread);
	spin_lock_init(&thread->disp_state.lock);
#ifdef CONFIG_SMP
	thread->disp_state.cpu = 0;
#endif
#ifdef PREEMPTSTATS
	thread->disp_state.scheds = 0;
	thread->disp_state.valid = 0;
	thread->disp_state.lattotal = 0;
	thread->disp_state.latmin = ~0;
	thread->disp_state.latmax = 0;
#endif
	do_debug(thread->disp_state.last_thread = 0);
	do_debug(thread->disp_state.run_tick = 0);
}


#ifndef CONFIG_NONBLOCKING
static int thread_disp_init_func(void)
{
	s_thread_t *cur_thread = current_thread();

	/*
	 * Unlock the current and previous threads
	 * before returning to the actual startup function.
	 */
	BD();
#ifdef CONFIG_SMP
	assert(cur_thread->disp_state.cpu);
	if (smp_message_pass_enable[smp_find_cur_cpu()])
		assert(cur_thread->disp_state.cpu->number
		       == smp_find_cur_cpu());
	cur_thread->disp_state.last_thread->disp_state.cpu = 0;
#endif

	/*
	 * Change the user-visible thread identity.
	 */
	csw_set_uthread_va(cur_thread);

	do_debug(cur_thread->disp_state.run_tick = ++disp_run_tick);
	spin_unlock(&cur_thread->disp_state.last_thread->disp_state.lock);
	spin_unlock(&cur_thread->disp_state.lock);
	enable_preemption();

	return 0;
}

void thread_disp_state_setup_stack(
	struct s_thread		*thread,
	void			(*start_func)(void),
	void			*start_sp)
{
	/*
	 * Set up the thread's context switch state
	 * so that the first time we context switch to it,
	 * it'll run thread_disp_init_func()
	 * and then jump to the specified start_func.
	 */
	BD_ARGS("thread=%p",thread);
	thread_csw_state_setup_stack(thread, thread_disp_init_func,
				     start_func, start_sp);
}
#endif /* CONFIG_NONBLOCKING */


void thread_disp_state_destroy(
	struct s_thread		*thread)
{
	/*
	 * We don't actually have to do anything here,
	 * because the thread is being held captive
	 * and is therefore not on any wait queue.
	 */
#ifdef PREEMPTSTATS
	thread_disp_dump_preemptstats(thread);
#endif
}

void disp_boot(void)
{
	BD();
}

/*
 * Start the boot processor running real threads.
 * Never returns.
 */
void
disp_start(struct s_thread *cur_th)
{
	BD();
	assert(cur_th == current_thread());
	assert(cur_th == &startup_thread);

	disable_preemption();

	sched_first_thread(cur_th);

	enable_preemption();

	/*
	 * From here to user mode.
	 */
	csw_start(cur_th);

	/* Should never reach here... */
	otsan();
}

#ifdef PREEMPTSTATS
void disp_dump_preemptstats(void)
{
	extern int ticks, pdticks;
#ifdef HEARTBEAT_THREAD
	extern unsigned int hbruns, hbcycles, hbnotready, hbrunning, hbdcycles;
	extern struct preempter heartbeat_preempter;
#endif
#ifdef CONFIG_FDEV
	extern struct preempter fdev_timer_preempter;
#endif
	struct preempter *p;

	printf("%d ticks (%d preemption disabled)\n", ticks, pdticks);
	printf("%d preempts (%d user mode, %d explicit)\n",
	       preempts, userpreempts-expreempts, expreempts);

	p = &disp_preempter;
	printf("disp_yield: %d immediate, %d already queued, %d delayed, %d cycles/delay\n",
	       p->immediate, p->queued, p->delayed, p->dcycles/p->delayed);
#ifdef HEARTBEAT_THREAD
	p = &heartbeat_preempter;
	printf("heartbeat_func: %d immediate, %d already queued, %d delayed, %d cycles/delay\n",
	       p->immediate, p->queued, p->delayed, p->dcycles/p->delayed);
	printf("%d heartbeat runs (%d notready, %d running), %d cycles/delay, %d cycles/run\n",
	       hbruns, hbnotready, hbrunning,
	       hbdcycles/hbruns, hbcycles/hbruns);
#endif
#ifdef CONFIG_FDEV
	p = &fdev_timer_preempter;
	printf("fdev_timer: %d immediate, %d already queued, %d delayed, %d cycles/delay\n",
	       p->immediate, p->queued, p->delayed, p->dcycles/p->delayed);

#endif
}

void thread_disp_dump_preemptstats(
	struct s_thread		*thread)
{
	printf("T%p", thread);
#ifdef DEBUG
	printf("(%16.16s)", thread->sob.label);
#endif
	printf(": pri=%d", thread->sched_state.pri);
	if (thread->disp_state.scheds)
#if 1
		printf(", scheds=%d, latency(min/ave/max)=%u/%u/%u",
		       thread->disp_state.scheds, 
		       thread->disp_state.latmin,
		       thread->disp_state.lattotal / thread->disp_state.scheds,
		       thread->disp_state.latmax);
#else
		printf(", scheds=%d lattot=%u, latmin=%u, latmax=%u",
		       thread->disp_state.scheds, thread->disp_state.lattotal,
		       thread->disp_state.latmin, thread->disp_state.latmax);
#endif
	printf("\n");
}
#endif

#ifdef CONFIG_SMP
#include <pthread.h>

/*
 * This is run in the root thread of the second and later CPUs.
 * Preemption is disabled.
 */
void
smp_disp_boot(struct s_thread *cur_th)
{
	/*
	 * Lock against the master CPU still running on this thread.
	 * We know that it held this lock before we were called,
	 * and won't unlock it until it's no longer running cur_th.
	 */
	spin_lock(&cur_th->disp_state.lock);
	assert(cur_th->disp_state.cpu->number == smp_find_cur_cpu());
	spin_unlock(&cur_th->disp_state.lock);

	idle_boot();

	/*
	 * From here to user mode.
	 */
	csw_start(cur_th);
}
#endif CONFIG_SMP
