/*
 * Copyright (c) 1996, 1998, 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#ifndef _FLUKE_KERNEL_DISPATCH_H_
#define _FLUKE_KERNEL_DISPATCH_H_

#include <oskit/boolean.h>

#include "spin_lock.h"
#include "wait_queue.h"
#include "preempt.h"
#include <fluke/ipc.h>

#include "config_user_sched.h"
#include "config_smp.h"
#include "config_nonblocking.h"
#include "config_nopreempt.h"

typedef struct disp_state
{
	spin_lock_t		lock;

	/* Link on wait or ready queue.  */
	struct s_thread		*next;

	/* Wakeup information.  */
	struct s_thread		*last_thread;

#if DEBUG
	/* `disp_run_tick' value when this thread was last switched into */
	oskit_u32_t		run_tick;
#endif

#ifdef CONFIG_SMP
	/*
	 * Per-CPU structure for the CPU currently assigned to this thread.
	 * This is null when this thread not running.
	 */
	struct fluke_cpu_info	*cpu;
#endif

#ifdef CONFIG_USER_SCHED
	/*
	 * Thread we are donating our CPU to; his provider points back to us.
	 * If we are readied, we cancel this thread.
	 */
	struct s_thread		*donating_to;

	/*
	 * Thread providing CPU to us; his donating_to points back to us.
	 * If we block, we dispatch back to him.  This is nonzero only if
	 * we have control of the CPU; the chain of provider links goes back
	 * to a root thread representing one real CPU.
	 */
	struct s_thread		*provider;

	/*
	 * This is used with disp_yield to request that this thread
	 * yield its CPU back to its provider.
	 */
	struct preempter	preempter;

	/*
	 * Scheduler port link.
	 */
	struct link		sched_link;

	/*
	 * This member is only set while the thread is in
	 * blocked in the WAIT_SCHEDULER state.
	 * Port alias of the scheduler port that led
	 * to the scheduler pset wait, so the IPC can be completed
	 * without chasing the port and pset links again.
	 */
	fluke_alias_t		alias;
#endif

#ifdef PREEMPTSTATS
	int		scheds;
	int		stamp, valid;
	unsigned	lattotal;
	unsigned	latmin, latmax;
#endif
} disp_state;

struct link;

/*
 * Ready a thread (make it runnable).
 * Assumes that the thread has already been removed
 * from any other queue it might already have been on,
 * and that the caller has exclusive control over the thread.
 */
void thread_ready(struct s_thread *thread);

/*
 * Put the current thread asleep and select another thread to run.
 * Assumes interrupts are disabled and the specified spin lock
 * is locked and will prevent the current thread from being woken
 * until it is unlocked (atomically) by this routine.
 * Returns a pointer to the new current thread.
 */
struct s_thread *pd_thread_dispatch(
	struct s_thread		*cur_thread,
	spin_lock_t		*unlock_lock,
	struct s_thread		*inheritor,
	struct link		*inheritor_link);

/* Initialize the per-thread dispatcher state in a new thread. */
void thread_disp_state_init(
	struct s_thread		*thread);

/*
 * In blocking kernels, set up a new thread's stack
 * so that when it is dispatched for the first time,
 * it calls 'start_func' with 'start_stack_pointer' as the stack pointer.
 */
#ifndef CONFIG_NONBLOCKING
void thread_disp_state_setup_stack(
	struct s_thread		*thread,
	void			(*start_func)(void),
	void			*start_stack_pointer);
#endif

/*
 * Place the current thread into a permanent coma,
 * waiting to be destroyed by another thread.
 * Interrupts must be disabled on entry.
 * Releases the lock on the specified wait queue
 * atomically with respect to the block operation.
 */
void thread_disp_stop(
	struct s_thread		*cur_thread,
	spin_lock_t		*release_lock);

/*
 * Destroy the dispatcher state for a thread.
 * Assumes the thread has stopped itself using thread_disp_stop().
 */
void thread_disp_state_destroy(
	struct s_thread		*thread);

/*
 * Yield the processor to another runnable thread if possible.
 * Takes the current thread as a parameter,
 * and returns the (new) current thread as its result.
 */
struct s_thread *pd_disp_yield(
	struct s_thread		*cur_thread);

#ifndef CONFIG_NOPREEMPT
/*
 * Give the processor to a higher priority thread (if there is one).
 */
void disp_yield(
	struct preempter	*p);
#endif

/*
 * Initialize the dispatcher layer at boot time.
 */
void disp_boot(void);

/*
 * This is called after all *_boot() routines have been called;
 * it starts the system running and switches to the first thread,
 * which must be captured (not yet known to anyone).
 * This function never returns.
 */
void disp_start(
	struct s_thread		*first_thread);

#ifdef PREEMPTSTATS
void thread_disp_dump_preemptstats(
	struct s_thread		*thread);
#endif

/*
 * Start a new processor running the given first thread.
 */
void smp_disp_boot(
	struct s_thread		*first_thread);


#endif /* _FLUKE_KERNEL_DISPATCH_H_ */
