/*
 * Copyright (c) 1996 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Fluke page fault handling.
 * XXX this stuff is a mess.
 */

#include <fluke/mapping.h>
#include <fluke/exception.h>
#include <oskit/page.h>

#include "fault.h"
#include "exception.h"
#include "s_region.h"
#include "s_mapping.h"
#include "frag.h"
#include "s_task.h"
#include "s_thread.h"
#include "s_port.h"
#include "debug.h"

#include "config_instrument.h"
#include "s_instrument.h"

#if FLASK
#include "avc.h"
#endif /* FLASK */

/* XXX move these elsewhere */
#define min(a, b) ((a) <= (b) ? (a) : (b))
#define max(a, b) ((a) < (b) ? (b) : (a))

/*
 * Handle an address translation fault for task ``ftask'' at address ``fva''
 * with access ``fprot'' in the context of thread ``fthread.''
 *
 * We locate the ``lowest'' task in the hierarchy that:
 *
 *	1. if the faulting task, has no mapping for the address.
 *	2. if the faulting task, has a mapping that disallows the access.
 *	3. if the faulting task, has a valid translation
 *	4. has a translation for the address that allows the indicated access.
 *	5. has a region covering the address that disallows the access.
 *
 * #1,#2 return KR_USER_EXCEPTION
 *
 * #3 returns KR_SUCCESS with a pointer to the appropriate fragment.
 * The mapping unmap lock for that fragment is held on return.
 *
 * #4 the translation found is used to create a translation in the faulting
 * task and returns KR_SUCCESS with a pointer to the appropriate fragment.
 * The mapping unmap lock for that fragment is held on return.
 *
 * #5 returns KR_PAGE_FAULT with a pointer to the region.
 * The region object lock is held on return.
 */
int task_fault(s_thread_t *fthread, s_task_t *ftask, oskit_addr_t fva,
	       fluke_prot_t fprot, frag_t **fragp, s_region_t **regionp,
	       oskit_addr_t *regoffp)
{
	s_task_t *task, *ntask;
	s_mapping_t *mapping, *fmapping;
	frag_t *frag, *ffrag;
	s_region_t *region, *nregion;
	oskit_addr_t va, rva, off;
	oskit_addr_t msva, meva;
	fluke_prot_t mprot;
#if FLASK
	security_id_t segment_sid;
        struct avc_entry *ae;
        unsigned aever;
	fluke_prot_t aprot;
	int rc;
#endif /* FLASK */

	/* XXX lint */
	msva = meva = mprot = 0;

	assert(fprot & (FLUKE_PROT_READ|FLUKE_PROT_WRITE|FLUKE_PROT_EXEC));
	assert(regionp == 0 || regoffp != 0);

	/*
	 * Thread exception state for non-zero return.
	 * It should be ignored on zero returns so we just always set it.
	 */
	exc_set_pagefault(fthread, fva, fprot);

	INSTRUMENT_EVENT(fthread, INSTRUMENT_EVENT_MIGHT_FAULT, fva);
	
	/*
	 * Lookup and lock task as necessary.
	 * If not explicitly specified, use the task of the faulting thread.
	 */
	if (ftask == 0) {
		ftask = fthread->task_state.task;
		if (ftask == 0) {
			thread_task_state_start(fthread);
			if (fthread->task_state.task == 0) {
				INSTRUMENT_EVENT(fthread, KR_USER_EXCEPTION, fva);
				return (KR_USER_EXCEPTION);
			}
			
			/* Now that we have a task, retry the faulting access */
			INSTRUMENT_EVENT(fthread, KR_RESTART, fva);
			return (KR_RESTART);
		}
	}
	s_task_lock(ftask);

	fmapping = 0;
	ffrag = 0;
	va = fva;
	task = ftask;
	region = 0;
#if FLASK
	segment_sid = SECSID_NULL;
#endif 

	while (1) {
		assert(task && task->sob.active);

#if FLASK
		if (task->segment_sid)
		{
		    assert(!segment_sid);	
		    segment_sid = task->segment_sid;
		}
#endif 

		/*
		 * Scan the task mapping list for a mapping covering
		 * the faulting address.  If the hint is valid, start
		 * searching from there, either forward or backward as
		 * appropriate.  If no hint, scan forward from the beginning.
		 *
		 * Note we are accessing the mapping start and size fields
		 * without holding any mapping lock.  This is ok since only
		 * mapping_destroy and mapping_set_state can change these
		 * fields and those routines always acquire the task lock
		 * (which we do hold) first.
		 */
		if (!queue_empty(&task->mappings)) {
			mapping = task->mhint;
			if (mapping == 0)
				mapping = (s_mapping_t *)
					queue_first(&task->mappings);
			if (va < mapping->start_va) {
				mapping = (s_mapping_t *)
					queue_prev(&mapping->tlink);
				while (!queue_end(&task->mappings,
						  (queue_entry_t)mapping)) {
					if (va >= mapping->start_va &&
					    va < mapping->start_va + mapping->size)
						break;

					mapping = (s_mapping_t *)
						queue_prev(&mapping->tlink);
				}
			} else {
				while (!queue_end(&task->mappings,
						  (queue_entry_t)mapping)) {
					if (va >= mapping->start_va &&
					    va < mapping->start_va + mapping->size)
						break;

					mapping = (s_mapping_t *)
						queue_next(&mapping->tlink);
				}
			}
		} else
			mapping = (s_mapping_t *)queue_first(&task->mappings);

		/*
		 * No mapping was found.
		 * If this mapping was derived from a region (i.e. is not the
		 * initial task) send an exception to that region's keeper if
		 * possible.  Otherwise we return a failure. 
		 */
		if (queue_end(&task->mappings, (queue_entry_t)mapping)) {
			task->mhint = 0;
			s_task_unlock(task);
			if (fmapping)
				s_mapping_f_unlock(fmapping);
			if (region) {
				if (regionp) {
					*regionp = region;
					*regoffp = va - region->start_va;
				} else
					s_region_unlock(region);
				INSTRUMENT_EVENT(fthread, KR_PAGE_FAULT, fva);
				return (KR_PAGE_FAULT);
			}
			INSTRUMENT_EVENT(fthread, KR_USER_EXCEPTION, fva);
			return (KR_USER_EXCEPTION);
		}

		task->mhint = mapping;
		s_mapping_f_lock(mapping);
		s_task_unlock(task);
		
		/*
		 * Mapping was found.
		 * If this is the initial mapping, we take special actions.
		 */
		if (fmapping == 0) {
			/*
			 * Check mapping protection if mapping is for the
			 * faulting task; i.e. protections in a mapping
			 * only apply for the one task for which the mapping
			 * is made.  If this is the faulting task and the
			 * indicated access is not allowed, exception goes
			 * to the thread.
			 */
			if ((mapping->prot & fprot) != fprot) {
				s_mapping_f_unlock(mapping);
				INSTRUMENT_EVENT(fthread, KR_USER_EXCEPTION, fva);
				return (KR_USER_EXCEPTION);
			}
			fmapping = mapping;

			/*
			 * Record attributes of the source ``range.''
			 * The range is the largest piece of the source frag
			 * that maps to the faulting mapping.  The source frag
			 * is the one from which we ultimately obtain the
			 * information for a kernel-handled fault.  This
			 * information allows us to create large HW pages if
			 * we desire.  This information is not useful (except
			 * for the protection) if the underlying HW is using a
			 * fixed page size.
			 */
			msva = mapping->start_va;
			meva = msva + mapping->size;
			mprot = mapping->prot;
		} else {
			msva = max(msva, mapping->start_va);
			meva = min(meva, mapping->start_va + mapping->size);
		}

		/*
		 * Check for a VtoP translation associated with this mapping.
		 * If one exists, we may be able to resolve the fault ourselves.
		 */
		s_mapping_u_lock(mapping);
		frag = mapping_frag_locate(mapping,
					   va-mapping->start_va, PAGE_SIZE);
		if (frag) {
			assert(frag->mapping == mapping);

			/*
			 * Translation exists, check its protection.
			 * If it allows access, we can resolve the fault
			 * by creating a new translation based on this one.
			 */
			if ((frag_prot(frag) & fprot) == fprot) {
				s_mapping_f_unlock(mapping);

				/*
				 * Requested address is valid.
				 * Just return frag if desired.
				 */
				if (mapping == fmapping) {
					if (fragp)
						*fragp = frag;
					else
						s_mapping_u_unlock(mapping);
					return (0);
				}
				
				if (region)
					s_region_unlock(region);
				break;
			}
			/*
			 * Translation exists but doesn't allow protection.
			 * If this is the initial mapping, remember this
			 * fragment.
			 */
			else if (mapping == fmapping)
				ffrag = frag;
		}
		s_mapping_u_unlock(mapping);

		/*
		 * Next check the region from which the mapping is derived.
		 * If there is no region, it is as though the mapping
		 * disallowed all access; i.e., a fault is sent to the
		 * previous region keeper or the thread.
		 */
		nregion = (s_region_t *)link_follow(&mapping->region);
		if (nregion == 0) {
			if (mapping != fmapping)
				s_mapping_f_unlock(mapping);
			s_mapping_f_unlock(fmapping);
			if (region) {
				if (regionp) {
					*regionp = region;
					*regoffp = va - region->start_va;
				} else
					s_region_unlock(region);
				INSTRUMENT_EVENT(fthread, KR_PAGE_FAULT, fva);
				return (KR_PAGE_FAULT);
			}
			INSTRUMENT_EVENT(fthread, KR_USER_EXCEPTION, fva);
			return (KR_USER_EXCEPTION);
		}
		if (region)
			s_region_unlock(region);
		region = nregion;

		/*
		 * If the offset of the fault in the mapping is beyond the
		 * end of its associated region, exception goes to the
		 * region's keeper.
		 */
		off = mapping->region_off + (va - mapping->start_va);
		if (off >= region->size) {
			if (mapping != fmapping)
				s_mapping_f_unlock(mapping);
			s_mapping_f_unlock(fmapping);
			if (regionp) {
				*regionp = region;
				*regoffp = off;
			} else
				s_region_unlock(region);
			INSTRUMENT_EVENT(fthread, KR_PAGE_FAULT, fva);
			return (KR_PAGE_FAULT);
		}

		/*
		 * We compute the fault address within the region and that
		 * becomes the virtual address for region faults or for the
		 * next round of checks.
		 */
		rva = region->start_va + off;

		/*
		 * If new task == faulting task or last task we are in a cycle.
		 * Note that these are not the only possible cycles, just ones
		 * we can pick up easily.
		 */
		ntask = (s_task_t *)link_follow(&region->task);
#if 0
/* XXX task may legitimately map its own memory more than once */
		assert(ntask == 0 || (ntask != ftask && ntask != task));
#endif

		/*
		 * Region has no task (treated as if region had zero access
		 * permission) or disallows access, send a fault to region
		 * keeper.
		 */
		if (ntask == 0 || (region->prot & fprot) != fprot) {
			if (ntask)
				s_task_unlock(ntask);
			if (mapping != fmapping)
				s_mapping_f_unlock(mapping);
			s_mapping_f_unlock(fmapping);
			if (regionp) {
				*regionp = region;
				*regoffp = off;
			} else
				s_region_unlock(region);
			INSTRUMENT_EVENT(fthread, KR_PAGE_FAULT, fva);
			return (KR_PAGE_FAULT);
		}
		
		/*
		 * If we get here, then we have trace through a mapping and
		 * its region and not found the source of the fault.  Now we
		 * back up a level and try again.
		 */
		task = ntask;
		va = rva;
		off = meva - msva;
		msva = region->start_va + mapping->region_off +
			(msva - mapping->start_va);
		meva = min(msva + off, region->start_va + region->size);
		mprot &= region->prot;

		if (mapping != fmapping)
			s_mapping_f_unlock(mapping);
	}

	/*
	 * If we have broken out of the loop it is because we have found
	 * no one restricting the access and we have discovered a physical
	 * translation with sufficient protection to allow the access (which
	 * indicates that no higher level region is preventing the access).
	 *
	 * We know that we will break out of the loop since we will eventually
	 * get back to the kernel task and the kernel frag covers all of
	 * physical memory.
	 *
	 * Locks: fault mapping fault lock should be held (preventing other
	 * faults for the same mapping) as well as source mapping unmap lock
	 * (preventing changes to fragment we are cloning).  We must acquire
	 * the fault mapping unmap lock to update its fragment chain.
	 */
	assert(mapping != fmapping);
	assert(frag_mapping(frag) == mapping);
	assert((mprot & fprot) == fprot);

	assert_mutex_locked(&fmapping->flock);
	assert_mutex_locked(&mapping->ulock);

#if FLASK
	/*
	 * Compute the mandatory memory permissions based on the segment SID.
	 */

	if (!segment_sid)
		segment_sid = frag_segment_sid(frag);

	assert(SUBJECT_SUBJECT__READ == FLUKE_PROT_READ);
	assert(SUBJECT_SUBJECT__WRITE == FLUKE_PROT_WRITE);
	assert(SUBJECT_SUBJECT__EXECUTE == FLUKE_PROT_EXEC);
	assert(SUBJECT_SUBJECT__CREATE_OBJECT == FLUKE_PROT_CREATE_OB);
	assert(sizeof(fluke_prot_t) == sizeof(access_vector_t));
	
	rc = avc_compute_access_vector(SECCLASS_SUBJECT,
				       ftask->sob.sid,
				       SECCLASS_SUBJECT,
				       segment_sid,
				       fprot,	
				       (access_vector_t *) &aprot,
				       &ae, &aever);
	if (rc)
	{
	    s_mapping_u_unlock(mapping);
	    s_mapping_f_unlock(fmapping);
	    return rc;
	}

	if ((fprot & aprot) != fprot)
	{
	    s_mapping_u_unlock(mapping);
	    s_mapping_f_unlock(fmapping);
	    EXC_RETURN_INSANITY(PERMISSION_DENIED);
	}

	/*
	 * Act as though the mapping protection were really the
	 * intersection of the "discretionary" mapping protection 
	 * and the mandatory protection.
	 */
	mprot &= aprot;
#endif /* FLASK */

	s_mapping_u_lock(fmapping);
	if (ffrag) {
		assert(frag_start_pa(ffrag) >= frag_start_pa(frag) &&
		       frag_start_pa(ffrag) + frag_size(ffrag) <=
		       frag_start_pa(frag) + frag_size(frag));
		frag_protect(ffrag, frag_prot(frag) & mprot);
	} else {
		/*
		 * We need to further reduce the [msva-meva] range to
		 * accomodate the size of the fragment we are inheriting from.
		 */
		msva = max(msva, mapping->start_va + frag_offset(frag));
		meva = min(meva, mapping->start_va + frag_offset(frag) +
			   frag_size(frag));
		assert(msva <= meva);

		off = va - (mapping->start_va + frag_offset(frag));
		assert(off >= 0 && off < frag_size(frag));
		ffrag = mapping_frag_validate(fmapping, fva,
					      frag_start_pa(frag) + off,
					      frag_prot(frag) & mprot,
#if FLASK
					      fva - (va - msva), meva - msva, 
					      segment_sid);
#else
		                              fva - (va - msva), meva - msva);
#endif 
	}
	assert(ffrag->mapping == fmapping);
	s_mapping_u_unlock(mapping);
	s_mapping_f_unlock(fmapping);
	if (fragp)
		*fragp = ffrag;
	else
		s_mapping_u_unlock(fmapping);

	return (0);
}

/*
 * XXX could be a macro somewhere
 */
int task_lookup(s_thread_t *thread, s_task_t *task, oskit_addr_t va,
		fluke_prot_t prot, frag_t **fragp)
{
	return (task_fault(thread, task, va, prot, fragp, 0, 0));
}

