/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Internal routines to handle fragments.
 */
#include <malloc.h>
#include <oskit/lmm.h>
#include <oskit/page.h>
#include <fluke/ob.h>
#include <fluke/vm.h>

#include "frag.h"
#include "s_task.h"
#include "s_mapping.h"
#include "pmap.h"
#include "kalloc.h"
#include "debug.h"

#ifdef DEBUG
/* turn this on if you suspect ob caching is not working correctly */
int checkallfrags = 0;
#else
#define checkallfrags 0
#endif

#ifdef MEMSTATS
int mstats_frags, mstats_maxfrags;
unsigned long mstats_fragpagescovered;
#ifdef USEAMM
int mstats_freeentries, mstats_maxfreeentries;
#endif
#endif

#ifdef USEAMM
/*
 * Support routines
 */
amm_entry_t *
frag_entry_alloc(struct amm *amm, oskit_addr_t addr, oskit_size_t size, int flags)
{
	amm_entry_t *entry;
	unsigned shift;

	if (flags == AMM_FREE) {
		if ((sizeof *entry & (sizeof *entry - 1)) == 0)
			for (shift = 0; (1 << shift) < sizeof *entry; shift++);
		else
			shift = 0;
		entry = kalloc_aligned(sizeof *entry, shift);
#ifdef MEMSTATS
		if (++mstats_freeentries > mstats_maxfreeentries)
			mstats_maxfreeentries = mstats_freeentries;
#endif
	} else {
		frag_t *frag;

		if ((sizeof *frag & (sizeof *frag - 1)) == 0)
			for (shift = 0; (1 << shift) < sizeof *frag; shift++);
		else
			shift = 0;
		frag = kalloc_aligned(sizeof *frag, shift);
#ifdef MEMSTATS
		if (++mstats_frags > mstats_maxfrags)
			mstats_maxfrags = mstats_frags;
#endif
		entry = &frag->entry;
	}
	assert(entry);

	return entry;
}

void
frag_entry_free(amm_t *amm, amm_entry_t *entry)
{
	s_mapping_t *mapping;
	frag_t *frag;
	oskit_addr_t frag_va;

	if (amm_entry_flags(entry) == AMM_FREE) {
		kfree(entry, sizeof *entry);
#ifdef MEMSTATS
		mstats_freeentries--;
#endif
		return;
	}

	assert(amm);
	mapping = AMMTOMAP(amm);
	frag = AMMETOFRAG(entry);
	assert_mutex_locked(&mapping->ulock);

	/*
	 * XXX hack, hack
	 * If frag->mapping == 0 then we are here following a join
	 * and we shouldn't invalidate the obtab or TLB as the VtoP
	 * mappings are still valid (just covered by a different frag).
	 */
	if (frag->mapping) {
		/*
		 * If any objects have been looked up using this fragment,
		 * then the task may have object links referring to them
		 * in its obtab, which need to be invalidated.
		 */
		frag_va = mapping->start_va + frag_offset(frag);
		if (checkallfrags || frag->obs_cached) {
			s_task_t *task = mapping->taskp;
			int i;

			assert(task);
			s_task_u_lock(task);
			for (i = 0; i < OBTAB_SIZE; i++) {
				unsigned uva = task->obtab[i].uva;

				if (uva >= frag_va &&
				    uva < frag_va + frag_size(frag)) {
					task->obtab[i].uva = -1;
					link_destroy(&task->obtab[i].link);
					assert(frag->obs_cached);
				}
			}
			s_task_u_unlock(task);
		}
		pmap_remove(mapping->taskp, frag_va, frag_size(frag));
#ifdef MEMSTATS
		mstats_fragpagescovered -= atop(frag_size(frag));
#endif
	}

	kfree(frag, sizeof *frag);
#ifdef MEMSTATS
	mstats_frags--;
#endif
}

int
frag_entry_split(amm_t *amm, amm_entry_t *entry, oskit_addr_t addr,
		 amm_entry_t **headp, amm_entry_t **tailp)
{
	amm_entry_t *nentry;
	oskit_size_t size;

	size = addr - amm_entry_start(entry);
	nentry = frag_entry_alloc(amm, addr, amm_entry_size(entry) - size,
				  amm_entry_flags(entry));
	assert(nentry);

	*headp = entry;
	*tailp = nentry;

	if (amm_entry_flags(entry) != AMM_FREE) {
		frag_t *frag, *nfrag;

		frag = AMMETOFRAG(entry);
		assert(frag->mapping != 0);
		assert_mutex_locked(&frag->mapping->ulock);
		nfrag = AMMETOFRAG(nentry);
		nfrag->start_pa = frag->start_pa + size;
		nfrag->mapping = frag->mapping;
		nfrag->obs_cached = frag->obs_cached;	/* XXX conservative */
#if FLASK
		nfrag->segment_sid = frag->segment_sid;
#endif
	}

	return 0;
}

int
frag_entry_join(amm_t *amm, amm_entry_t *head, amm_entry_t *tail,
		amm_entry_t **new)
{
	*new = head;
	if (amm_entry_flags(head) != AMM_FREE) {
		frag_t *hfrag, *tfrag;

		hfrag = AMMETOFRAG(head);
		tfrag = AMMETOFRAG(tail);

		assert(hfrag->mapping != 0);
		assert(tfrag->mapping != 0);
		assert(hfrag->mapping == tfrag->mapping);
		assert_mutex_locked(&hfrag->mapping->ulock);

		/*
		 * Frags must be physcially contiguous and part of the
		 * same mapping in order to join.
		 */
		if (hfrag->start_pa + amm_entry_size(head) != tfrag->start_pa
#if FLASK
		    || (hfrag->segment_sid != tfrag->segment_sid)
#endif 
		    )
			return 1;

		if (!hfrag->obs_cached)
			hfrag->obs_cached = tfrag->obs_cached;

		/*
		 * XXX zap the mapping pointer so that the free function
		 * doesn't invalidate the translations.
		 */
		tfrag->mapping = 0;
	}

	return 0;
}
#else
/*
 * Allocate a frag.
 * Does not associate the fragment with any mapping, hence nothing need be
 * locked.
 */
frag_t *frag_create(void)
{
	frag_t *frag;

	frag = kalloc_aligned(sizeof *frag, 0);
	if (!frag)
		panic("frag_create: no memory!");

	frag->mapping = 0;

#if FLASK
	frag->segment_sid = SECSID_NULL;
#endif 
#ifdef MEMSTATS
	if (++mstats_frags > mstats_maxfrags)
		mstats_maxfrags = mstats_frags;
#endif

	return frag;
}

/*
 * Destroy a frag.
 * Should not be part of a mapping so nothing need be locked.
 */
int frag_destroy(frag_t *frag)
{
	assert(frag->mapping == 0);

	kfree(frag, sizeof *frag);
#ifdef MEMSTATS
	mstats_frags--;
#endif
	return (0);
}

/*
 * Insert a fragment into a mapping's fragment list.
 * Fragments are kept sorted by starting offset.
 * Mapping unmap lock should be held.
 */
void frag_insert(s_mapping_t *mapping, frag_t *frag)
{
	assert_mutex_locked(&mapping->ulock);
	assert(mapping->taskp != 0);
	assert(frag->mapping == 0);
	assert(frag_offset(frag) + frag_size(frag) <= mapping->size);

	if (queue_empty(&mapping->fragments)) {
		queue_enter(&mapping->fragments, frag, frag_t *, mlink);
	} else {
		frag_t *nfrag;

		queue_iterate(&mapping->fragments, nfrag, frag_t *, mlink)
			if (nfrag->offset >= frag->offset)
				break;
		queue_enter_before(&mapping->fragments, nfrag,
				   frag, frag_t *, mlink)
	}
	frag->mapping = mapping;
	frag->obs_cached = 0;

	pmap_enter(mapping->taskp, mapping->start_va + frag_offset(frag),
		   frag->start_pa, frag_size(frag), frag_prot(frag));
	/* XXX needs rc */
#ifdef MEMSTATS
	mstats_fragpagescovered += atop(frag_size(frag));
#endif
}

/*
 * Remove a fragment from its mapping list and invalidate any translations.
 * Mapping unmap lock should be held.
 */
void frag_delete(s_mapping_t *mapping, frag_t *frag)
{
	oskit_addr_t frag_start_va;

	assert_mutex_locked(&mapping->ulock);
	assert(frag->mapping == mapping);

	frag_start_va = mapping->start_va + frag->offset;

	/*
	 * If any objects have been looked up using this fragment,
	 * then the task may have object links referring to them
	 * in its obtab, which need to be invalidated.
	 */
	if (checkallfrags || frag->obs_cached) {
		s_task_t *task = mapping->taskp;
		int i;

		assert(task);
		s_task_u_lock(task);
		for (i = 0; i < OBTAB_SIZE; i++) {
			unsigned uva = task->obtab[i].uva;

			if ((uva >= frag_start_va) &&
			    (uva < frag_start_va + frag->size)) {
				task->obtab[i].uva = -1;
				link_destroy(&task->obtab[i].link);
				assert(frag->obs_cached);
			}
		}
		s_task_u_unlock(task);
	}

	/*
	 * Invalidate any PMAP-layer mappings for this fragment in the task.
	 */
	pmap_remove(mapping->taskp, frag_start_va, frag->size);
#ifdef MEMSTATS
	mstats_fragpagescovered -= atop(frag_size(frag));
#endif

	/*
	 * Finally, remove the fragment from the task's fragment list.
	 */
	queue_remove(&mapping->fragments, frag, frag_t *, mlink);
	frag->mapping = 0;
	if (mapping->fhint == frag)
		mapping->fhint = 0;
}

/*
 * Potentially split a frag into two or three pieces, one which includes
 * any addresses in the range [offset - offset+size] and one or two with
 * the remainder.  Returns a pointer to the new entry associated with the
 * address range:
 *
 * 1. If the frag and the range are disjoint:
 *	*nfragp == 0
 * 2. If the frag is totally contained in the range:
 *	*nfragp == frag
 * 3. If the frag is split:
 *	*nfragp == adjusted frag, new fragments created for the rest
 *
 * The new entry is not inserted in the mapping, the other fragments are.
 * Mapping should be locked.
 *
 * XXX this could be cleaned up considerably.
 */
void frag_split(s_mapping_t *mapping, frag_t *frag,
		oskit_addr_t offset, oskit_size_t size, frag_t **nfragp)
{
	oskit_addr_t eoffset = offset + size;
	frag_t *nfrag;

	assert_mutex_locked(&mapping->ulock);
	assert(frag->mapping == mapping);

	/*
	 * 1. Disjoint
	 */
	if (offset > frag->offset + frag->size || eoffset < frag->offset) {
		*nfragp = 0;
		return;
	}

	/*
	 * All other cases will result in a frag being returned.
	 * Remove that frag from the mapping now.
	 */
	frag_delete(mapping, frag);

	if (offset < frag->offset)
		offset = frag->offset;
	if (eoffset > frag->offset + frag->size)
		eoffset = frag->offset + frag->size;

	/*
	 * 2. Total containment
	 */
	if (offset == frag->offset && eoffset == frag->offset + frag->size) {
		*nfragp = frag;
		return;
	}

	/*
	 * 3. Frag split
	 *
	 * Desired frag clipped from the front.  Initialize and insert the
	 * new frag for the remaining region, adjust the old and return.
	 */
	if (offset == frag->offset && eoffset < frag->offset + frag->size) {
		nfrag = frag_create();
		nfrag->offset = eoffset;
		nfrag->start_pa = frag->start_pa + eoffset - offset;
		nfrag->size = frag->offset + frag->size - nfrag->offset;
		nfrag->prot = frag->prot;
#if FLASK
		nfrag->segment_sid = frag->segment_sid;
#endif 
		frag_insert(mapping, nfrag);
		frag->offset = offset;
		frag->size = eoffset - frag->offset;
		*nfragp = frag;
		return;
	}
	/*
	 * Desired frag clipped from the end.  Initialize and insert the
	 * new frag for the remaining region, adjust the old and return.
	 */
	if (offset > frag->offset && eoffset == frag->offset + frag->size) {
		nfrag = frag_create();
		nfrag->offset = frag->offset;
		nfrag->start_pa = frag->start_pa;
		nfrag->size = offset - nfrag->offset;
		nfrag->prot = frag->prot;
#if FLASK
		nfrag->segment_sid = frag->segment_sid;
#endif 
		frag_insert(mapping, nfrag);
		frag->offset = offset;
		frag->start_pa = nfrag->start_pa + nfrag->size;
		frag->size = eoffset - frag->offset;
		*nfragp = frag;
		return;
	}
	/*
	 * Desired frag extracted from the middle.
	 * Create new frags at start and end and insert them.
	 */
	nfrag = frag_create();
	nfrag->offset = frag->offset;
	nfrag->start_pa = frag->start_pa;
	nfrag->size = offset - nfrag->offset;
	nfrag->prot = frag->prot;
#if FLASK
	nfrag->segment_sid = frag->segment_sid;
#endif 
	frag_insert(mapping, nfrag);
	nfrag = frag_create();
	nfrag->offset = eoffset;
	nfrag->start_pa = frag->start_pa + eoffset - frag->offset;
	nfrag->size = frag->offset + frag->size - nfrag->offset;
	nfrag->prot = frag->prot;
#if FLASK
	nfrag->segment_sid = frag->segment_sid;
#endif 
	frag_insert(mapping, nfrag);
	frag->start_pa = frag->start_pa + offset - frag->offset;
	frag->offset = offset;
	frag->size = eoffset - frag->offset;
	*nfragp = frag;
	return;
}
#endif

/*
 * XXX Simplified version of mapping_frag_validate.
 * Primary difference is not entering the mapping in the pmap
 * (which has already been done).
 */
#if FLASK
void
mapping_frag_boot(s_mapping_t *mapping, oskit_addr_t va, oskit_addr_t pa,
		  oskit_size_t size, fluke_prot_t prot, 
		  security_id_t segment_sid)
#else
void
mapping_frag_boot(s_mapping_t *mapping, oskit_addr_t va, oskit_addr_t pa,
		  oskit_size_t size, fluke_prot_t prot)
#endif /* FLASK */
{
	frag_t *nfrag;
	oskit_addr_t off;
#ifdef USEAMM
	amm_entry_t *entry;
#endif

	off = va - mapping->start_va;

	/*
	 * Create the fragment and enter it in the list.
	 */
#ifdef USEAMM
	entry = frag_entry_alloc(0, 0, 0, AMM_ALLOCATED);
	nfrag = AMMETOFRAG(entry);
	nfrag->start_pa = pa + off;
	nfrag->mapping = mapping;
	nfrag->obs_cached = 0;
#if FLASK
	nfrag->segment_sid = segment_sid;
#endif 
	if (amm_modify(&mapping->fragments, off, size,
		       AMM_ALLOCATED|prot, entry))
		panic("mapping_frag_validate: no memory");
#else
	nfrag = frag_create();
	nfrag->offset = off;
	nfrag->size = size;
	nfrag->prot = prot;
	nfrag->start_pa = pa + off;
	nfrag->mapping = mapping;
	nfrag->obs_cached = 0;
#if FLASK
	nfrag->segment_sid = segment_sid;
#endif 
	queue_enter(&mapping->fragments, nfrag, frag_t *, mlink);
#endif
}

void
mapping_frag_init(s_mapping_t *mapping)
{
#ifdef USEAMM
	amm_init_gen(&mapping->fragments, AMM_FREE, 0,
		     frag_entry_alloc, frag_entry_free,
		     frag_entry_split, frag_entry_join);
#else
	mapping->fhint = 0;
	queue_init(&mapping->fragments);
#endif
}

/*
 * Return pointer to the first frag containing any part of the range specified.
 * May return 0 if no part of the range belongs to a frag.
 * Mapping unmap lock should be held.
 */
frag_t *
mapping_frag_locate(s_mapping_t *mapping, oskit_addr_t offset, oskit_size_t size)
{
#ifdef USEAMM
	amm_entry_t *entry;
#else
	frag_t *frag;
#endif
	oskit_addr_t eoffset;

	assert_mutex_locked(&mapping->ulock);

	if (size == 0)
		return 0;

	eoffset = offset + size;
#ifdef USEAMM
	do {
		entry = amm_find_addr(&mapping->fragments, offset);
		if (amm_entry_flags(entry) & AMM_ALLOCATED)
			return AMMETOFRAG(entry);
		offset = amm_entry_end(entry);
	} while (offset < eoffset);
#else
	frag = mapping->fhint;
	if (frag == 0 || offset < frag->offset)
		frag = (frag_t *)queue_first(&mapping->fragments);
	while (!queue_end(&mapping->fragments, (queue_entry_t)frag)) {
		if (offset < frag->offset + frag->size) {
			if (eoffset > frag->offset) {
				mapping->fhint = frag;
				return frag;
			}
			break;
		}

		frag = (frag_t *)queue_next(&frag->mlink);
	}
	mapping->fhint = 0;
#endif
	return 0;
}

/*
 * Modify or remove all fragments in the indicated range of a mapping.
 */
void
mapping_frag_protect(s_mapping_t *mapping, oskit_addr_t offset,
		     oskit_size_t size, fluke_prot_t prot)
{
#ifdef USEAMM
	oskit_addr_t eoffset;
	amm_entry_t *entry;
	int rc;

	/*
	 * If not freeing, we cannot just amm_modify the whole range as that
	 * would change the attributes of free entries as well as allocated
	 * ones.  Hence we iterate and choose only the allocated entries.
	 *
	 * XXX we cannot use amm_iterate as:
	 *	- we might have to clip the first and last entries since
	 *	  amm_iterate passes in entries that overlap to any extent
	 *	  with the desired range
	 *	- we are changing the attributes which may cause entries to
	 *	  split or merge, amm_iterate doesn't handle this right now
	 */
	if (prot == FLUKE_PROT_NONE) {
		rc = amm_modify(&mapping->fragments, offset, size, AMM_FREE, 0);
		assert(rc == 0);
		return;
	}

	eoffset = offset + size;
	while (offset < eoffset) {
		entry = amm_find_addr(&mapping->fragments, offset);

		/*
		 * Entry is not allocated or already has the correct protection
		 */
		if (amm_entry_flags(entry) == AMM_FREE ||
		    (amm_entry_flags(entry) & FLUKE_PROT_ALL) == prot) {
			offset = amm_entry_end(entry);
			continue;
		}

		/*
		 * Otherwise, see if we need to clip the start/end of the
		 * entry.  If so, do it and look up again (since our entry
		 * may no longer be valid).
		 */
		if (offset > amm_entry_start(entry) ||
		    eoffset < amm_entry_end(entry)) {
			oskit_addr_t off, eoff;

			off = amm_entry_start(entry);
			if (offset > off)
				off = offset;
			eoff = amm_entry_end(entry);
			if (eoffset < eoff)
				eoff = eoffset;
			rc = amm_modify(&mapping->fragments, off, eoff - off,
					AMM_ALLOCATED|prot, 0);
			assert(rc == 0);
			entry = amm_find_addr(&mapping->fragments, offset);
		}

		/*
		 * Now adjust the frag protection.
		 */
		assert(offset == amm_entry_start(entry));
		assert(eoffset >= amm_entry_end(entry));
		frag_protect(AMMETOFRAG(entry), prot);

		offset = amm_entry_end(entry);
	}
#else
	frag_t *frag;

	/*
	 * XXX this implementation has the bad effect that it will remove and
	 * then reinsert a mapping even if the protection is already correct.
	 */
	while ((frag = mapping_frag_locate(mapping, offset, size)) != 0) {
		frag_t *nfrag;

		frag_split(mapping, frag, offset, size, &nfrag);
		assert(nfrag != 0);
		offset = nfrag->offset + nfrag->size;
		assert(nfrag->size <= size);
		size -= nfrag->size;
		if (prot == FLUKE_PROT_NONE)
			frag_destroy(nfrag);
		else {
			nfrag->prot &= prot;
			frag_insert(mapping, nfrag);
		}
	}
#endif
}

/*
 * Create a frag describing the translation of a VA to a PA in a mapping
 * with a given protection.
 *
 * ``rsva'' and ``rsize'' describe a range including the VA which is the
 * maximum sized frag we could create with the PA and protection parameters.
 *
 * XXX belongs in unmap layer
 */
#if FLASK
frag_t *
mapping_frag_validate(s_mapping_t *mapping, oskit_addr_t va, oskit_addr_t pa,
		      fluke_prot_t prot, oskit_addr_t rsva, oskit_size_t rsize,
		      security_id_t segment_sid)
#else
frag_t *
mapping_frag_validate(s_mapping_t *mapping, oskit_addr_t va, oskit_addr_t pa,
		      fluke_prot_t prot, oskit_addr_t rsva, oskit_size_t rsize)
#endif 
{
	frag_t *nfrag;
	oskit_addr_t off, toff, noff;
	oskit_size_t size;
#ifdef USEAMM
	amm_entry_t *entry;
#endif

	assert_mutex_locked(&mapping->ulock);
	assert(rsva <= va && va <= rsva + rsize);

	toff = va - mapping->start_va;
	off = rsva - mapping->start_va;
	size = rsize;

	/*
	 * Lookup the maximum sized frag in the target mapping.
	 * If there is some overlap with existing fragments,
	 * we will have to reduce the size of the new frag.
	 */
#ifdef USEAMM
	entry = amm_find_addr(&mapping->fragments, toff);
	assert(amm_entry_flags(entry) == AMM_FREE);

	noff = amm_entry_start(entry);
	if (noff > off) {
		size -= (noff - off);
		off = noff;
	}
	if (off + size > amm_entry_end(entry))
		size = amm_entry_end(entry) - off;

	entry = frag_entry_alloc(0, 0, 0, AMM_ALLOCATED);
	nfrag = AMMETOFRAG(entry);
	nfrag->start_pa = pa - toff + off;
	nfrag->mapping = mapping;
	nfrag->obs_cached = 0;
#if FLASK
	nfrag->segment_sid = segment_sid;
#endif 
	if (amm_modify(&mapping->fragments, off, size,
		       AMM_ALLOCATED|prot, entry))
		panic("mapping_frag_validate: out of memory");

	/*
	 * XXX re-lookup frag as it may not be the one we just allocated
	 * (if amm_modify merged entries).
	 */
	entry = amm_find_addr(&mapping->fragments, off);
	nfrag = AMMETOFRAG(entry);
	assert(nfrag->mapping == mapping);

	pmap_enter(mapping->taskp, mapping->start_va + off,
		   pa - toff + off, size, prot);
#ifdef MEMSTATS
	mstats_fragpagescovered += atop(size);
#endif
#else
	while ((nfrag = mapping_frag_locate(mapping, off, size)) != 0) {

		/* single target page should not be covered */
		assert(off != toff || size != PAGE_SIZE);

		/*
		 * If overlap is before the target address, truncate the start.
		 */
		noff = nfrag->offset + nfrag->size;
		if (noff <= toff) {
			size = off + size - noff;
			off = noff;
		}

		/*
		 * If overlap is after the target address, truncate the end.
		 */
		else {
			noff = nfrag->offset;
			assert(noff > toff);
			size = noff - off;
		}
	}

	/*
	 * Create, initialize and insert a new fragment
	 */
	nfrag = frag_create();
	nfrag->offset = off;
	nfrag->start_pa = pa - toff + off;
	nfrag->size = size;
	nfrag->prot = prot;
#if FLASK
	nfrag->segment_sid = segment_sid;
#endif 
	frag_insert(mapping, nfrag);
#endif

	return nfrag;
}

/*
 * Relax the protection of a fragment and modify the actual HW translations.
 * Effected mapping unmap lock should be held.
 *
 * This is a short-cut version of mapping_frag_protect used by the fault path.
 * Here we have already identified the target fragment so there is no need to
 * look it up.
 */
void
frag_protect(frag_t *frag, fluke_prot_t prot)
{
	oskit_addr_t frag_va;

	assert(frag->mapping != 0);
	assert_mutex_locked(&frag->mapping->ulock);

	frag_va = frag->mapping->start_va + frag_offset(frag);

	/*
	 * If any objects have been looked up using this fragment,
	 * then the task may have object links referring to them
	 * in its obtab, which need to be updated.
	 */
	if (checkallfrags || frag->obs_cached) {
		s_task_t *task = frag->mapping->taskp;
		int i;

		assert(task);
		s_task_u_lock(task);
		for (i = 0; i < OBTAB_SIZE; i++) {
			unsigned uva = task->obtab[i].uva;

			if (uva >= frag_va && uva < frag_va + frag_size(frag)) {
				task->obtab[i].prot = prot;
				assert(frag->obs_cached);
			}
		}
		s_task_u_unlock(task);
	}

	pmap_protect(frag->mapping->taskp, frag_va, frag_size(frag),
		     frag_prot(frag), prot);

#ifdef USEAMM
	/*
	 * We perform the amm_modify after everything else since it may
	 * merge the fragment with surrounding fragments.
	 */
	(void)amm_modify(&frag->mapping->fragments,
			 frag_offset(frag), frag_size(frag),
			 AMM_ALLOCATED|prot, &frag->entry);
#else
	frag->prot = prot;
#endif
}

#ifdef MEMSTATS
void
frag_dump_mstats(void)
{
	printf("  Frag memory:       %10d bytes (%10d bytes max)\n",
	       mstats_frags*sizeof(frag_t), mstats_maxfrags*sizeof(frag_t));
#ifdef USEAMM
	printf("  Free entry memory: %10d bytes (%10d bytes max)\n",
	       mstats_freeentries*sizeof(amm_entry_t),
	       mstats_maxfreeentries*sizeof(amm_entry_t));
#endif
	printf("    %d frags covering %10ld pages of VM\n",
	       mstats_frags, mstats_fragpagescovered);
}
#endif

#ifdef DEBUG
void
frag_dump(frag_t *frag, int indent)
{
	int i;

	for (i = 0; i < indent; i++) putchar(' ');
	printf("offset=0x%08x, start_pa=0x%08x, size=0x%08x, prot=%c%c%c%c\n",
	       frag_offset(frag),
	       frag->start_pa,
	       frag_size(frag),
	       (frag_prot(frag)&8)?'C':'-',
	       (frag_prot(frag)&4)?'r':'-',
	       (frag_prot(frag)&2)?'w':'-',
	       (frag_prot(frag)&1)?'x':'-');
}

void
mapping_frag_dump(s_mapping_t *mapping, int indent)
{
	int i;

	for (i = 0; i < indent; i++) putchar(' ');
	printf("frags:\n");
#ifdef USEAMM
	amm_dump(&mapping->fragments);
#else
	{
		frag_t *frag;

		queue_iterate(&mapping->fragments, frag, frag_t *, mlink)
			frag_dump(frag, indent+1);
	}
#endif
}
#endif
