/*
 * Copyright (c) 1996 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#ifndef _FLUKE_KERNEL_FRAG_H_
#define _FLUKE_KERNEL_FRAG_H_

#define USEAMM	/* XXX */

#include <oskit/machine/types.h>
#include <oskit/queue.h>
#include <fluke/vm.h>
#ifdef USEAMM
#include <oskit/amm.h>
#endif

#if FLASK
#include <flask/security.h>
#endif 

struct s_mapping;

/*
 * Physical mapping information.
 *
 * A fragment describes a contiguous range of physical memory (a ``page'').
 * mapped into a particular task mapping.
 */
typedef struct fragment {
#ifdef USEAMM
	amm_entry_t		entry;		/* offset, size, prot */
#else
	oskit_addr_t		offset;		/* offset in mapping */
	oskit_size_t		size;		/* size of range */
	fluke_prot_t		prot;		/* protection attribute */
	queue_chain_t		mlink;		/* other frags in mapping */
#endif
	oskit_addr_t		start_pa;	/* physical address */
	struct s_mapping	*mapping;	/* mapping we are a part of */
	oskit_bool_t		obs_cached;	/* task has cached objects */
#if FLASK
	security_id_t		segment_sid;	/* SID (Flask only) */
#else
	int			pad;		/* Keep Flask/Fluke same size */
#endif
} frag_t;

#ifdef USEAMM
#define AMMETOFRAG(entry)	((frag_t *)(entry))
#define frag_offset(frag)	amm_entry_start(&(frag)->entry)
#define frag_size(frag)		amm_entry_size(&(frag)->entry)
#define frag_prot(frag)		(amm_entry_flags(&(frag)->entry)&FLUKE_PROT_ALL)
#else
#define frag_offset(frag)	(frag)->offset
#define frag_size(frag)		(frag)->size
#define frag_prot(frag)		(frag)->prot
#endif
#define frag_start_pa(frag)	(frag)->start_pa
#define frag_mapping(frag)	(frag)->mapping
#define frag_obs_cached(frag)	(frag)->obs_cached
#if FLASK
#define frag_segment_sid(frag)	(frag)->segment_sid
#endif 

#define frag_vtop(frag, va) \
	(frag_start_pa(frag) + \
		((va) - frag_mapping(frag)->start_va) - frag_offset(frag))
#define frag_ptov(frag, pa) \
	(frag_mapping(frag)->start_va + \
		((pa) - frag_start_pa(frag)) + frag_offset(frag))

#define frag_set_obs_cached(frag)	(frag)->obs_cached = 1
#define frag_clear_obs_cached(frag)	(frag)->obs_cached = 0

frag_t *frag_create(void);
int frag_destroy(frag_t *frag);
void frag_protect(frag_t *frag, fluke_prot_t prot);

void mapping_frag_boot(struct s_mapping *mapping, oskit_addr_t va,
		       oskit_addr_t pa, oskit_size_t size,
#if FLASK
		       fluke_prot_t prot, security_id_t segment_sid);
#else
		       fluke_prot_t prot);
#endif /* FLASK */
void mapping_frag_init(struct s_mapping *mapping);
frag_t *mapping_frag_locate(struct s_mapping *mapping,
			    oskit_addr_t offset, oskit_size_t size);
void mapping_frag_protect(struct s_mapping *mapping, oskit_addr_t offset,
			  oskit_size_t size, fluke_prot_t prot);
frag_t *mapping_frag_validate(struct s_mapping *mapping, oskit_addr_t va,
			      oskit_addr_t pa, fluke_prot_t prot,
#if FLASK
			      oskit_addr_t rsva, oskit_size_t rsize,
			      security_id_t segment_sid);
#else
			      oskit_addr_t rsva, oskit_size_t rsize);
#endif /* FLASK */

#if FLASK
void mapping_frag_revoke(struct s_mapping *mp, security_id_t tsid,
                         access_vector_t revoke, oskit_bool_t try_remove,
			 access_vector_t *out_retained);
int mapping_frag_reset(struct s_mapping *mp, security_id_t task_sid);
#endif

#endif /* _FLUKE_KERNEL_FRAG_H_ */
