/*
 * Copyright (c) 1996-1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include "spin_lock.h"
#include "s_thread.h"
#include "dispatch.h"
#include "debug.h"
#include "preempt.h"

#include "config_user_sched.h"

#ifndef CONFIG_USER_SCHED

#include <fluke/thread.h>
#include <fluke/debug.h>
#include "pthread.h"
#include "prog.h"		/* XXX for do_in_supervisor_mode */
#include "round_robin.h"

void *run_idle_thread(void *arg);


/*
 * This creates a user-mode thread which makes calls in supervisor-mode
 * to the thread_yield function, which does a context switch if there is
 * a runable thread.
 *
 * If this function is called multiple times, it will generate an
 * idle thread for each call.
 */
void
idle_boot()
{
	static int num = 0;
	pthread_t thread;
	pthread_attr_t attr;

	pthread_attr_init(&attr);
	attr.prio = SCHED_PRIIDLE;

	printf("Creating idle thread %d\n", num);
	pthread_create(&thread, &attr, run_idle_thread,
		(void *)num++);
#ifdef DEBUG
	/*
	 * XXX thread used to label itself, but most of the time it would
	 * appear unlabeled since it would never get to run.
	 */
	{
		char label[16];
		snprintf(label, sizeof(label), "IDLE %d", num);
		fluke_debug_label_flob((fluke_thread_t *)thread, label);
	}
#endif
}


/*
 * Supervisor-mode half of the idle thread.
 */
static int
do_yield(void *arg)
{
	int rc;
	s_thread_t *cur_thread;

	cur_thread = current_thread();
	/* If another thread can run, schedule it instead of me. */
	rc = thread_yield(cur_thread);

#ifdef CONFIG_NONBLOCKING
	/* Only block if we are still in the idle thread */
	if (cur_thread == current_thread())
#endif
	if (!preempt_pending()) {
		/* Block until the next interrupt.  Keep the CPU cool. */
		__asm__("hlt\n");
		/* Back from low-level interrupt handler. */
	}

	/*
	 * Any preempters pending will now run; if a new thread is
	 * added to the run queue, we will yield to it when we come
	 * back in the top.
	 */
	return rc;
}


/*
 * This is the user-mode `half' of the idle thread.
 * Now that we have fluke_thread_yield(), this could be re-written
 * to use that, but then we wouldn't be HLTing the CPU.
 */
static void *
run_idle_thread(void *arg)
{
	while (1) {
		do_in_supervisor_mode(do_yield, NULL);
	}

	return 0;
}

#endif	/* ! CONFIG_USER_SCHED */
