/*
 * Copyright (c) 1996-1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include <fluke/ipc.h>
#include <fluke/exception.h>
#include <fluke/thread.h>/*XXX*/
#include <fluke/machine/syscall.h>

#include "xfer.h"
#include "link.h"
#include "s_ref.h"
#include "s_port.h"
#include "ipc.h"
#include "s_thread.h"
#include "debug.h"

#include "config_instrument.h"
#include "s_instrument.h"

#include "config_epreempt.h"
#ifdef CONFIG_EPREEMPT
#include <oskit/page.h>
#define EPREEMPT_IPC_MAX_XFER_WORDS	(2 * PAGE_SIZE)	/* XXX */
#endif

#if FLASK
#include "avc.h"
#include "avc_ipc.h"
#endif 

/*
 * - Disconnection can only be done by one of the threads involved,
 *   never by any "third party" threads.
 *   Either thread can disconnect directly if the other is in a wait state;
 *   otherwise, the thread wanting to disconnect must cancel the other
 *   and then go into the WAIT_IPC_???_DSCN state itself.
 *
 * - Similarly, pickling and unpickling connections can only be done
 *   by one of the threads involved, and only when the other thread is waiting.
 *
 * - The current orientation (flow direction) of an active connection
 *   is always indicated by the server_sender flag in the server's ipc_state.
 *   It can only be changed by one of the threads
 *   while it is holding the other captive.
 */

void thread_ipc_state_init(s_thread_t *th)
{
	th->ipc_state.server = 0;
	th->ipc_state.client = 0;
	th->ipc_state.flags = 0;
	link_init(&th->ipc_state.client_link);
	link_init(&th->ipc_state.server_link);
	th->ipc_state.idemp_server = 0;
	th->ipc_state.idemp_client = 0;
#if FLASK
	th->ipc_state.effective_server_sid = th->sob.sid;
	th->ipc_state.client_effective_client_sid = SECSID_NULL;
	th->ipc_state.effective_client_sid = th->sob.sid;
	th->ipc_state.server_effective_server_sid = SECSID_NULL;
	th->ipc_state.server_wants_client_sid = FALSE;
#endif 
}

void thread_ipc_state_destroy(s_thread_t *th)
{
	assert(th->ipc_state.client == 0);
	assert(th->ipc_state.server == 0);
	link_destroy(&th->ipc_state.client_link);
	link_destroy(&th->ipc_state.server_link);
	assert(th->ipc_state.idemp_client == 0);
	assert(th->ipc_state.idemp_server == 0);
}

static void ipc_pickle(s_thread_t *client, s_thread_t *server)
{
	assert(client->ipc_state.server == server);
	assert(server->ipc_state.client == client);

	link_destroy(&server->ipc_state.client_link);
	link_produce(&server->ipc_state.client_link, &client->sob);
	server->ipc_state.client = 0;

	link_destroy(&client->ipc_state.server_link);
	link_produce(&client->ipc_state.server_link, &server->sob);
	client->ipc_state.server = 0;
}

static void ipc_unpickle(s_thread_t *client, s_thread_t *server)
{
	/*
	 * Null out the cross-links so they don't get used again.
	 */
	link_destroy(&client->ipc_state.server_link);
	link_init(&client->ipc_state.server_link);
	link_destroy(&server->ipc_state.client_link);
	link_init(&server->ipc_state.client_link);

	/*
	 * Check the direction of the reestablished connection.
	 * If the two parties don't agree, then disconnect.
	 */
	if (((client->ipc_state.flags & FLUKE_THREAD_CLIENT_SENDER) &&
	     (server->ipc_state.flags & FLUKE_THREAD_SERVER_SENDER)) ||
	    (!(client->ipc_state.flags & FLUKE_THREAD_CLIENT_SENDER) &&
	     !(server->ipc_state.flags & FLUKE_THREAD_SERVER_SENDER)))
		return;

	client->ipc_state.server = server;
	server->ipc_state.client = client;

#if FLASK
	client->ipc_state.server_effective_server_sid = 
	  server->ipc_state.effective_server_sid;
	server->ipc_state.client_effective_client_sid = 
	  client->ipc_state.effective_client_sid;
#endif
}

/*
 * Pickle any connections that ``client'' has with a server.
 * XXX must be called with current thread as the client.
 */
int
ipc_client_pickle(s_thread_t *client, s_thread_t *pickler,
		  link_t *slink, unsigned *flags)
{
	s_thread_t *server;

	assert(client);
	assert(client == current_thread());

	server = client->ipc_state.server;
	if (server != 0) {
		assert(server->ipc_state.client == client);
		if (pickler == server) {
			ipc_pickle(client, server);
		} else if (thread_wakeup(server, WAIT_ANYTHING)) {
			ipc_pickle(client, server);
			thread_ready(server);
		} else {
			/* XXX layering violation */
			if (pickler != client)
				cond_wait_other(&client->stop_cond, pickler);
			return thread_wait(client, WAIT_IPC_PICKLE,
					   server, 0 /* cancel it */);
		}
	}

	if (slink)
		link_create_copy(&client->ipc_state.server_link, slink);
	if (flags)
		*flags = client->ipc_state.flags & FLUKE_THREAD_CLIENT_SENDER;

	return 0;
}

/*
 * Revive any connection that ``client'' has with a server.
 * XXX must be called with current thread as the client.
 */
int
ipc_client_unpickle(s_thread_t *client, int no_wait)
{
	s_thread_t *server;

	assert(client);
	assert(client == current_thread());
	assert(client->ipc_state.server == 0);

	/* Try to revive a half-connection */
	server = (s_thread_t*)link_follow(&client->ipc_state.server_link);
	if (!server)
		return 0;
	if (server->sob.type != FLUKE_THREAD) /* XXX */
	{
		s_ob_unlock(&server->sob);
		return 0;
	}
	if (!thread_wakeup(server, WAIT_IPC_UNPICKLE)) /* XXX UNPICKLABLE? */
	{
		if (no_wait) {
			s_thread_unlock(server);
			return 0;
		}
		thread_cancel(server);
		return thread_wait(client, WAIT_IPC_UNPICKLE,
				   server, -1 /* unlock it */);
	}
	s_thread_unlock(server);

	/*
	 * The server is now locked and captured.
	 * Make sure it's actually us it wants to talk to.
	 */
#if FLASK
	if (!link_compare_ptr(&server->ipc_state.client_link, &client->sob) ||
	    (client->ipc_state.server_effective_server_sid &&
	     (client->ipc_state.server_effective_server_sid !=
	      server->ipc_state.effective_server_sid)) ||
	    (server->ipc_state.client_effective_client_sid &&
	     (server->ipc_state.client_effective_client_sid !=
	      client->ipc_state.effective_client_sid))) {
#else
	if (!link_compare_ptr(&server->ipc_state.client_link, &client->sob)) {
#endif 
		if (no_wait) {
			thread_ready(server);
			return 0;
		}
		return thread_handoff(client, WAIT_IPC_UNPICKLE, server);
	}

	ipc_unpickle(client, server);

	thread_ready(server);
	return 0;
}

/*
 * Pickle any connections that ``server'' has with a client.
 * XXX must be called with current thread as the server.
 */
int
ipc_server_pickle(s_thread_t *server, s_thread_t *pickler,
		  link_t *clink, unsigned *flags)
{
	s_thread_t *client;

	assert(server);
	assert(server == current_thread());

	client = server->ipc_state.client;
	if (client != 0) {
		assert(client->ipc_state.server == server);
		if (pickler == client) {
			ipc_pickle(client, server);
		} else if (thread_wakeup(client, WAIT_ANYTHING)) {
			ipc_pickle(client, server);
			thread_ready(client);
		} else {
			if (pickler != server)
				thread_ready(pickler);
			return thread_wait(server, WAIT_IPC_PICKLE,
					   client, 0 /* cancel it */);
		}
	}

	if (clink)
		link_create_copy(&server->ipc_state.client_link, clink);
	if (flags)
		*flags = server->ipc_state.flags & FLUKE_THREAD_SERVER_SENDER;

	return 0;
}

/*
 * Revive any connection that ``server'' has with a client.
 * XXX must be called with current thread as the server.
 */
int
ipc_server_unpickle(s_thread_t *server, int no_wait)
{
	s_thread_t *client;

	assert(server == current_thread());
	assert(server->ipc_state.client == 0);

	/* Try to revive a half-connection */
	client = (s_thread_t*)link_follow(&server->ipc_state.client_link);
	if (!client)
		return 0;
	if (client->sob.type != FLUKE_THREAD) /* XXX */
	{
		s_ob_unlock(&client->sob);
		return 0;
	}
	if (!thread_wakeup(client, WAIT_IPC_UNPICKLE)) /* XXX UNPICKLABLE? */
	{
		if (no_wait) {
			s_thread_unlock(client);
			return 0;
		}
		thread_cancel(client);
		return thread_wait(server, WAIT_IPC_UNPICKLE,
				   client, -1 /* unlock it */);
	}
	s_thread_unlock(client);

	/*
	 * The client is now locked and captured.
	 * Make sure it's actually us it wants to talk to.
	 */
#if FLASK
	if (!link_compare_ptr(&client->ipc_state.server_link, &server->sob) ||
	    (client->ipc_state.server_effective_server_sid &&
	     (client->ipc_state.server_effective_server_sid !=
	      server->ipc_state.effective_server_sid)) ||
	    (server->ipc_state.client_effective_client_sid &&
	     (server->ipc_state.client_effective_client_sid !=
	      client->ipc_state.effective_client_sid))) {
#else
	if (!link_compare_ptr(&client->ipc_state.server_link, &server->sob)) {
#endif 
		if (no_wait) {
			thread_ready(client);
			return 0;
		}
		return thread_handoff(server, WAIT_IPC_UNPICKLE, client);
	}

	ipc_unpickle(client, server);

	thread_ready(client);
	return 0;
}

int ipc_client_idemp_disconnect(s_thread_t *client)
{
	s_thread_t *server = client->ipc_state.idemp_server;
	if (server != 0)
	{
		assert(server->ipc_state.idemp_client == client);
		if (thread_wakeup(server, WAIT_ANYTHING)) {
			server->ipc_state.idemp_client = 0;
			client->ipc_state.idemp_server = 0;
			thread_ready(server);
		} else {
			return thread_wait(client, WAIT_IPC_CLI_DSCN,
					   server, 0 /* cancel it */);
		}
	}
	return 0;
}

int ipc_server_idemp_disconnect(s_thread_t *server)
{
	s_thread_t *client = server->ipc_state.idemp_client;
	if (client != 0)
	{
		assert(client->ipc_state.idemp_server == server);
		if (thread_wakeup(client, WAIT_ANYTHING)) {
			server->ipc_state.idemp_client = 0;
			client->ipc_state.idemp_server = 0;
			thread_ready(client);
		} else {
			return thread_wait(server, WAIT_IPC_SRV_DSCN,
					   client, 0 /* cancel it */);
		}
	}
	return 0;
}


int ipc_client_disconnect(s_thread_t *client)
{
	s_thread_t *server = client->ipc_state.server;
	int rc;

	/*
	 * Make sure any connection we're involved in is actually active.
	 * Since unpickling a connection may involve waiting,
	 * this means that the standard disconnect is a "soft" disconnect,
	 * which the other side can potentially delay arbitrarily.
	 * For a "hard" disconnect, null out the appropriate link,
	 * or better yet, destroy the connected thread and re-create it.
	 */
	if (server == 0) {
		rc = ipc_client_unpickle(client, 0);
		if (rc)
			return rc;
		server = client->ipc_state.server;
	}

	if (server != 0)
	{
		assert(server->ipc_state.client == client);
		if (thread_wakeup(server, WAIT_ANYTHING)) {
			server->ipc_state.client = 0;
			client->ipc_state.server = 0;
#if FLASK
			FLASK_IPC_STATE_DISCONNECT(client,server);
#endif 
			thread_ready(server);
		} else {
			return thread_wait(client, WAIT_IPC_CLI_DSCN,
					   server, 0 /* cancel it */);
		}
	}

	return 0;
}

int ipc_server_disconnect(s_thread_t *server)
{
	s_thread_t *client = server->ipc_state.client;
	int rc;

	/* Unpickle (if necessary) and break any reliable connection. */
	if (client == 0) {
		rc = ipc_server_unpickle(server, 0);
		if (rc)
			return rc;
		client = server->ipc_state.client;
	}
	if (client != 0)
	{
		assert(client->ipc_state.server == server);
		if (thread_wakeup(client, WAIT_ANYTHING)) {
			server->ipc_state.client = 0;
			client->ipc_state.server = 0;
#if FLASK
			FLASK_IPC_STATE_DISCONNECT(client,server);
#endif 
			thread_ready(client);
		} else {
			return thread_wait(server, WAIT_IPC_SRV_DSCN,
					   client, 0 /* cancel it */);
		}
	}

	/* Break any idempotent connection as well. */
	return ipc_server_idemp_disconnect(server);
}

int thread_ipc_state_stop(struct s_thread *cur_th, struct s_thread *stopper)
{
	s_thread_t *server;
	int rc;

	assert(cur_th == current_thread());

	/* Break out of any idempotent call we're making */
	server = cur_th->ipc_state.idemp_server;
	if (stopper == server) {
		server->ipc_state.idemp_client = 0;
		cur_th->ipc_state.idemp_server = 0;
	} else {
		rc = ipc_client_idemp_disconnect(cur_th);
		if (rc)
			return rc;
	}

	/* Break out of any idempotent call we're servicing */
	rc = ipc_server_idemp_disconnect(cur_th);
	if (rc)
		return rc;

	/* Pickle any reliable connection we are a server of (to a client). */
	rc = ipc_server_pickle(cur_th, stopper, 0, 0);
	if (rc)
		return rc;

	/* Pickle any reliable connection we are a client of (to a server). */
	rc = ipc_client_pickle(cur_th, stopper, 0, 0);
	if (rc)
		return rc;

	return 0;
}

void thread_ipc_canceled(s_thread_t *cur_th)
{
	s_thread_t *client, *server;

	assert(cur_th);
	assert(cur_th == current_thread());

	/* Break any idempotent call we may be involved in if necessary. */
	server = cur_th->ipc_state.idemp_server;
	if ((server != 0) && thread_wakeup(server, WAIT_IPC_SRV_DSCN)) {
		assert(server->ipc_state.idemp_client == cur_th);
		cur_th->ipc_state.idemp_server = 0;
		server->ipc_state.idemp_client = 0;
		thread_ready(server);
	}
	client = cur_th->ipc_state.idemp_client;
	if ((client != 0) && thread_wakeup(client, WAIT_IPC_CLI_DSCN)) {
		assert(client->ipc_state.idemp_server == cur_th);
		cur_th->ipc_state.idemp_client = 0;
		client->ipc_state.idemp_server = 0;
		thread_ready(client);
	}

	/* Perform any requested operation on our reliable server link. */
	server = cur_th->ipc_state.server;
	if (server != 0) {
		if (thread_wakeup(server, WAIT_IPC_SRV_DSCN))
		{
			assert(server->ipc_state.client == cur_th);
			cur_th->ipc_state.server = 0;
			server->ipc_state.client = 0;
#if FLASK
			FLASK_IPC_STATE_DISCONNECT(cur_th,server);
#endif 
			thread_ready(server);
		}
		else if (thread_wakeup(server, WAIT_IPC_PICKLE))
		{
			assert(server->ipc_state.client == cur_th);
			ipc_pickle(cur_th, server);
			thread_ready(server);
		}
	} else {
		int rc = ipc_client_unpickle(cur_th, 1);
		assert(rc == 0);
	}

	/* Perform any requested operation on our reliable client link. */
	client = cur_th->ipc_state.client;
	if (client != 0) {
		if (thread_wakeup(client, WAIT_IPC_CLI_DSCN))
		{
			assert(client->ipc_state.server == cur_th);
			cur_th->ipc_state.client = 0;
			client->ipc_state.server = 0;
#if FLASK
			FLASK_IPC_STATE_DISCONNECT(client,cur_th);
#endif 
			thread_ready(client);
		}
		if ((client != 0) && thread_wakeup(client, WAIT_IPC_PICKLE))
		{
			assert(client->ipc_state.server == cur_th);
			ipc_pickle(client, cur_th);
			thread_ready(client);
		}
	} else {
		int rc = ipc_server_unpickle(cur_th, 1);
		assert(rc == 0);
	}
}

void thread_ipc_get_state(s_thread_t *th, fluke_thread_state_t *state)
{
	assert((th->ipc_state.flags & ~THREAD_IPCFLAGS) == 0);
	state->flags |= th->ipc_state.flags;

	state->server_pset = (struct fluke_pset *)
				th->ipc_state.server_pset_uva;
	state->server_esp  = th->ipc_state.server_esp;
	state->server_oneway_eip  = th->ipc_state.server_oneway_eip;
	state->server_idempotent_eip  = th->ipc_state.server_idempotent_eip;
	state->server_reliable_eip  = th->ipc_state.server_reliable_eip;

#ifdef IPC_STATE_MD
	thread_ipc_get_state_md(th, state);
#endif
}

#if FLASK
void thread_ipc_get_state_secure(s_thread_t *th, 
				 fluke_thread_state_secure_t *state)
{
    state->effective_server_sid = th->ipc_state.effective_server_sid;
    state->client_effective_client_sid = 
	    th->ipc_state.client_effective_client_sid;
    state->effective_client_sid = th->ipc_state.effective_client_sid;
    state->server_effective_server_sid = 
	    th->ipc_state.server_effective_server_sid;
}
#endif

void thread_ipc_set_state(s_thread_t *th, fluke_thread_state_t *state)
{
	th->ipc_state.flags = state->flags & THREAD_IPCFLAGS;

	th->ipc_state.server_pset_uva = (oskit_addr_t) state->server_pset;
	th->ipc_state.server_esp = state->server_esp;
	th->ipc_state.server_oneway_eip = state->server_oneway_eip;
	th->ipc_state.server_idempotent_eip = state->server_idempotent_eip;
	th->ipc_state.server_reliable_eip = state->server_reliable_eip;

#ifdef IPC_STATE_MD
	thread_ipc_set_state_md(th, state);
#endif
}

/*
 * Push data and/or refs over to the waiting receiver.
 * Assumes we have already synchronized with the receiver
 * and are holding it captive (we've woken but not readied it).
 * On return, the receiver is still captive.
 */
int ipc_reliable_transfer(s_thread_t *sender, s_thread_t *receiver)
{
	int rc;

	assert(sender == current_thread());

	/*
	 * The receiver is ready to receive our message.
	 * First transfer any plain data buffers we have.
	 */
	while (1)
	{
		oskit_size_t words;
		int remote_rc;

		/*
		 * Make sure we have a send buffer.
		 */
		while (IPC_SEND_BUF_WORDS(sender) == 0)
		{
			fluke_ipc_buf_t buf;

			if (IPC_SEND_BUF_COUNT(sender) == 0)
			{
				/*
				 * Sender has no more buffers or words,
				 * so we're all done transferring plain data.
				 * (We need Java's labeled break stmt here...)
				 */
				goto send_data_done;
			}

			/* Copyin the next send buffer descriptor.  */
			rc = copyin_local(
				IPC_SEND_BUF_TAB(sender), &buf,
				sizeof(fluke_ipc_buf_t) / sizeof(fluke_ipc_word_t));
			if (rc) {
				thread_ready(receiver);
				return rc;
			}

			IPC_SEND_BUF_BUF(sender) = (fluke_ipc_word_t)buf.buf;
			IPC_SEND_BUF_WORDS(sender) = buf.words;
			IPC_SEND_BUF_TAB(sender) += sizeof(buf);
			IPC_SEND_BUF_COUNT(sender)--;
		}

		/*
		 * Make sure the receiver has a receive buffer.
		 */
		while (IPC_RECV_BUF_WORDS(receiver) == 0)
		{
			fluke_ipc_buf_t buf;

			if (IPC_RECV_BUF_COUNT(receiver) == 0)
			{
				/*
				 * Receiver's run out of receive space:
				 * stop and wait for it to get some more.
				 */
				IPC_FINISH_RECEIVE(receiver,
						   FLUKE_IPC_RECV_MORE_DATA);
				thread_ready(receiver);
				return KR_RESTART;
			}

			/* Copyin the next receive buffer descriptor.  */
			rc = copyin_remote(
				receiver,
				IPC_RECV_BUF_TAB(receiver), &buf,
				sizeof(fluke_ipc_buf_t) / sizeof(fluke_ipc_word_t));
			if (rc) {
				receiver->wait_state.resume_rc = rc;
				thread_ready(receiver);
				INSTRUMENT_EVENT(sender, KR_RESTART, 0);
				return KR_RESTART;
			}

			IPC_RECV_BUF_BUF(receiver) = (fluke_ipc_word_t)buf.buf;
			IPC_RECV_BUF_WORDS(receiver) = buf.words;
			IPC_RECV_BUF_TAB(receiver) += sizeof(buf);
			IPC_RECV_BUF_COUNT(receiver)--;
		}

		/*
		 * Transfer as much data as possible in one chunk.
		 */
		words = IPC_SEND_BUF_WORDS(sender);
		if (IPC_RECV_BUF_WORDS(receiver) < words)
			words = IPC_RECV_BUF_WORDS(receiver);
#ifdef CONFIG_EPREEMPT
		/*
		 * Restrict the size to limit the delay between preemption
		 * checks.
		 */
		if (words > EPREEMPT_IPC_MAX_XFER_WORDS)
			words = EPREEMPT_IPC_MAX_XFER_WORDS;
#endif
		remote_rc = 0;
		rc = copyto_remote(receiver,
				   IPC_SEND_BUF_BUF(sender),
				   IPC_RECV_BUF_BUF(receiver),
				   &words,
				   &remote_rc);

		/*
		 * Keep track of the number of words successfully transferred
		 * even if an error occurred in the source or destination task.
		 */
		IPC_SEND_BUF_WORDS(sender) -= words;
		IPC_SEND_BUF_BUF(sender) += words * sizeof(fluke_ipc_word_t);
		IPC_RECV_BUF_WORDS(receiver) -= words;
		IPC_RECV_BUF_BUF(receiver) += words * sizeof(fluke_ipc_word_t);

#ifdef CONFIG_EPREEMPT
		if (preempt_pending()) {
#ifdef PREEMPTSTATS
			extern int expreempts;
			expreempts++;
#endif
			thread_ready(receiver);
			INSTRUMENT_EVENT(sender, KR_RESTART, 0);
			return KR_RESTART;
		}
#endif
		if (rc) {
			thread_ready(receiver);
			return rc;
		}
		if (remote_rc) {
			receiver->wait_state.resume_rc = remote_rc;
			thread_ready(receiver);
			INSTRUMENT_EVENT(sender, KR_RESTART, 0);
			return KR_RESTART;
		}
	}
	send_data_done:

	/*
	 * We're done with all plain data;
	 * now transfer any references.
	 */
	while (IPC_SEND_REF_COUNT(sender) > 0)
	{
		oskit_addr_t send_ref_uva, recv_ref_uva;
		s_ref_t *ref;
		struct link l;

		assert(sizeof(oskit_addr_t) == sizeof(fluke_ipc_word_t));

		if (IPC_RECV_REF_COUNT(receiver) == 0)
		{
			/*
			 * Receiver's run out of receive references:
			 * wait for it to get some more.
			 */
			IPC_FINISH_RECEIVE(receiver, FLUKE_IPC_RECV_MORE_REFS);
			thread_ready(receiver);
			return KR_RESTART;
		}

		/* Find the reference object addresses on both sides.  */
		rc = copyin_local(
			IPC_SEND_REF_TAB(sender),
			&send_ref_uva, 1);
		if (rc) {
			thread_ready(receiver);
			return rc;
		}
		rc = copyin_remote(
			receiver,
			IPC_RECV_REF_TAB(receiver),
			&recv_ref_uva, 1);
		if (rc) {
			receiver->wait_state.resume_rc = rc;
			thread_ready(receiver);
			INSTRUMENT_EVENT(sender, KR_RESTART, 0);
			return KR_RESTART;
		}

		/* Look up the link from the source ref.  */
		rc = task_ob_lookup_type(sender, send_ref_uva,
					 FLUKE_PROT_READ, FLUKE_REF,
					 (s_ob_t**)&ref);
		if (rc) {
			thread_ready(receiver);
			INSTRUMENT_EVENT(sender, rc, 0);
			return rc;
		}
		s_ref_create_link(ref, &l);
		s_ref_unlock(ref);

		/* Transfer it into the target ref.  */
		rc = task_ob_lookup_type(receiver, recv_ref_uva,
					 FLUKE_PROT_WRITE, FLUKE_REF,
					 (s_ob_t**)&ref);
		if (rc)
		{
			link_destroy(&l);
			receiver->wait_state.resume_rc = rc;
			thread_ready(receiver);
			INSTRUMENT_EVENT(sender, KR_RESTART, 0);
			return KR_RESTART;
		}
		s_ref_link_transfer(&l, ref);
		s_ref_unlock(ref);

		link_destroy(&l);

		IPC_SEND_REF_COUNT(sender)--;
		IPC_SEND_REF_TAB(sender) += sizeof(oskit_addr_t);
		IPC_RECV_REF_COUNT(receiver)--;
		IPC_RECV_REF_TAB(receiver) += sizeof(oskit_addr_t);
	}

	/*
	 * Set the success code in the sender's register state.
	 * If sending was the last phase of its operation,
	 * we are the last to touch its state before it returns to the user.
	 */
	IPC_STATUS(sender) = 0;

	return 0;
}

int ipc_speculative_transfer(s_thread_t *sender, s_thread_t *receiver,
			     struct fluke_ipc_params *send_params,
			     struct fluke_ipc_params *recv_params)
{
	int rc;

	assert(sender == current_thread());

	/*
	 * The receiver is ready to receive our message.
	 * First transfer any plain data buffers we have.
	 */
	while (1)
	{
		oskit_size_t words;
		int remote_rc;

		/*
		 * Make sure we have a send buffer.
		 */
		while (send_params->send_buf.words == 0)
		{
			if (send_params->send_buf_count == 0)
			{
				/*
				 * Sender has no more buffers or words,
				 * so we're all done transferring plain data.
				 * (We need Java's labeled break stmt here...)
				 */
				goto send_data_done;
			}

			/* Copyin the next send buffer descriptor.  */
			rc = copyin_local(
				(oskit_addr_t)send_params->send_buf_tab,
				&send_params->send_buf,
				sizeof(fluke_ipc_buf_t) / sizeof(fluke_ipc_word_t));
			if (rc) {
				thread_ready(receiver);
				return rc;
			}

			send_params->send_buf_tab++;
			send_params->send_buf_count--;
		}

		/*
		 * Make sure the receiver has a receive buffer.
		 */
		while (recv_params->recv_buf.words == 0)
		{
			if (recv_params->recv_buf_count == 0)
			{
				/*
				 * Receiver's run out of receive space;
				 * just drop the rest of the sender's data.
				 */
				goto send_data_done;
			}

			/* Copyin the next receive buffer descriptor.  */
			rc = copyin_remote(
				receiver,
				(oskit_addr_t)recv_params->recv_buf_tab,
				&recv_params->recv_buf,
				sizeof(fluke_ipc_buf_t) / sizeof(fluke_ipc_word_t));
			if (rc) {
				receiver->wait_state.resume_rc = rc;
				thread_ready(receiver);
				INSTRUMENT_EVENT(sender, KR_RESTART, 0);
				return KR_RESTART;
			}

			recv_params->recv_buf_tab++;
			recv_params->recv_buf_count--;
		}

		/*
		 * Transfer as much data as possible in one chunk.
		 */
		words = send_params->send_buf.words;
		if (recv_params->recv_buf.words < words)
			words = recv_params->recv_buf.words;
		remote_rc = 0;
		rc = copyto_remote(receiver,
				   (oskit_addr_t)send_params->send_buf.buf,
				   (oskit_addr_t)recv_params->recv_buf.buf,
				   &words,
				   &remote_rc);

		/*
		 * Keep track of the number of words successfully transferred
		 * even if an error occurred in the source or destination task.
		 */
		send_params->send_buf.words -= words;
		send_params->send_buf.buf += words * sizeof(fluke_ipc_word_t);
		recv_params->recv_buf.words -= words;
		recv_params->recv_buf.buf += words * sizeof(fluke_ipc_word_t);

		if (rc) {
			thread_ready(receiver);
			return rc;
		}
		if (remote_rc) {
			receiver->wait_state.resume_rc = remote_rc;
			thread_ready(receiver);
			INSTRUMENT_EVENT(sender, KR_RESTART, 0);
			return KR_RESTART;
		}
	}
	send_data_done:

	/*
	 * We're done with all plain data;
	 * now transfer any references.
	 */
	while ((send_params->send_ref_count > 0) &&
	       (recv_params->recv_ref_count > 0))
	{
		oskit_addr_t send_ref_uva, recv_ref_uva;
		s_ref_t *ref;
		struct link l;

		assert(sizeof(oskit_addr_t) == sizeof(fluke_ipc_word_t));

		/* Find the reference object addresses on both sides.  */
		rc = copyin_local(
			(oskit_addr_t)send_params->send_ref_tab,
			&send_ref_uva, 1);
		if (rc) {
			thread_ready(receiver);
			return rc;
		}
		rc = copyin_remote(
			receiver,
			(oskit_addr_t)recv_params->recv_ref_tab,
			&recv_ref_uva, 1);
		if (rc) {
			receiver->wait_state.resume_rc = rc;
			thread_ready(receiver);
			INSTRUMENT_EVENT(sender, KR_RESTART, 0);
			return KR_RESTART;
		}

		/* Look up the link from the source ref.  */
		rc = task_ob_lookup_type(sender, send_ref_uva,
					 FLUKE_PROT_READ, FLUKE_REF,
					 (s_ob_t**)&ref);
		if (rc) {
			thread_ready(receiver);
			return rc;
		}
		s_ref_create_link(ref, &l);
		s_ref_unlock(ref);

		/* Transfer it into the target ref.  */
		rc = task_ob_lookup_type(receiver, recv_ref_uva,
					 FLUKE_PROT_WRITE, FLUKE_REF,
					 (s_ob_t**)&ref);
		if (rc)
		{
			link_destroy(&l);
			receiver->wait_state.resume_rc = rc;
			thread_ready(receiver);
			INSTRUMENT_EVENT(sender, KR_RESTART, 0);
			return KR_RESTART;
		}
		s_ref_create_link(ref, &l);
		s_ref_unlock(ref);

		link_destroy(&l);

		send_params->send_ref_count--;
		send_params->send_ref_tab++;
		recv_params->recv_ref_count--;
		recv_params->recv_ref_tab++;
	}

	return 0;
}

int ipc_client_find_server(struct s_thread *const client, int disconnect_code)
{
	struct s_thread *server = client->ipc_state.server;
	int rc;

	assert(client == current_thread());

	if (server == 0) {
		rc = ipc_client_unpickle(client, 0);
		if (rc)
			return rc;
		server = client->ipc_state.server;
	}
	if (server == 0) {
		if (disconnect_code == FLUKE_IPC_RECV_DISCONNECTED) {
			IPC_FINISH_RECEIVE(client, FLUKE_IPC_RECV_DISCONNECTED);
		} else {
			IPC_STATUS(client) = disconnect_code;
			EXC_SET_IP(client, fluke_nop__entry);
		}
		INSTRUMENT_EVENT(client, KR_RESTART, 0);
		return KR_RESTART;
	}

	return 0;
}

int ipc_server_find_client(struct s_thread *const server, int disconnect_code)
{
	struct s_thread *client = server->ipc_state.client;
	int rc;

	assert(server == current_thread());

	if (client == 0) {
		rc = ipc_server_unpickle(server, 0);
		if (rc)
			return rc;
		client = server->ipc_state.client;
	}
	if (client == 0) {
		if (disconnect_code == FLUKE_IPC_RECV_DISCONNECTED) {
			IPC_FINISH_RECEIVE(server, FLUKE_IPC_RECV_DISCONNECTED);
		} else {
			IPC_STATUS(server) = disconnect_code;
			EXC_SET_IP(server, fluke_nop__entry);
		}
		INSTRUMENT_EVENT(server, KR_RESTART, 0);
		return KR_RESTART;
	}

	return 0;
}

/*
 * FLASK:  The Flask modifications are not included in the
 * FLASK:  machine-independent IPC path components, since
 * FLASK:  they are not currently used.  See x86/ipc_md.c. 
 */

/*** One-way IPC path components ***/

#ifndef MD_IPC_SEND
int ipc_send(struct s_thread *const client)
{
	s_thread_t *server;
	fluke_ipc_params_t params;
	int rc;

	rc = s_port_capture(client, &server);
	if (rc)
		return rc;

	/*
	 * Extract the send and receive parameters
	 * from the sender's and receiver's register state.
	 */
	params.send_buf.words =	IPC_SEND_BUF_WORDS(client);
	params.send_buf.buf =	(fluke_ipc_word_t*)IPC_SEND_BUF(client);
	params.send_buf_count =	IPC_SEND_BUF_COUNT(client);
	params.send_buf_tab =	(fluke_ipc_buf_t*)IPC_SEND_BUF_TAB(client);
	params.send_ref_count =	IPC_SEND_REF_COUNT(client);
	params.send_ref_tab =	(struct fluke_ref**)IPC_SEND_REF_TAB(client);
	params.recv_buf.words =	IPC_RECV_BUF_WORDS(server);
	params.recv_buf.buf =	(fluke_ipc_word_t*)IPC_RECV_BUF(server);
	params.recv_buf_count =	IPC_RECV_BUF_COUNT(server);
	params.recv_buf_tab =	(fluke_ipc_buf_t*)IPC_RECV_BUF_TAB(server);
	params.recv_ref_count =	IPC_RECV_REF_COUNT(server);
	params.recv_ref_tab =	(struct fluke_ref**)IPC_RECV_REF_TAB(server);

	/* Speculatively transfer our message to it. */
	ipc_transfer_min_msg(client, server);
	rc = ipc_speculative_transfer(client, server, &params, &params);
	if (rc)
		return rc;

	/* The transfer succeeded, so release both parties and return. */
	IPC_STATUS(client) = 0;
	IPC_RECV_BUF_WORDS(server) = params.recv_buf.words;
	IPC_RECV_BUF_COUNT(server) = params.recv_buf_count;
	IPC_RECV_REF_COUNT(server) = params.recv_ref_count;
	IPC_FINISH_RECEIVE(server, FLUKE_IPC_WAIT_ONEWAY);
	thread_ready(server);
	return 0;
}
#endif /* MD_IPC_SEND */

/*** Idempotent IPC path components ***/

#ifndef MD_IPC_CALL
int ipc_call(struct s_thread *const client, oskit_bool_t secure)
{
	struct s_thread *server;
	fluke_ipc_params_t params;
	int rc;

	/* Break any existing idempotent connection. */
	rc = ipc_client_idemp_disconnect(client);
	if (rc)
		return rc;
	assert(client->ipc_state.idemp_server == 0);

	/* Capture a server thread. */
	rc = s_port_capture(client, &server);
	if (rc)
		return rc;

	/*
	 * Extract the send and receive parameters
	 * from the sender's and receiver's register state.
	 */
	params.send_buf.words =	IPC_SEND_BUF_WORDS(client);
	params.send_buf.buf =	(fluke_ipc_word_t*)IPC_SEND_BUF(client);
	params.send_buf_count =	IPC_SEND_BUF_COUNT(client);
	params.send_buf_tab =	(fluke_ipc_buf_t*)IPC_SEND_BUF_TAB(client);
	params.send_ref_count =	IPC_SEND_REF_COUNT(client);
	params.send_ref_tab =	(struct fluke_ref**)IPC_SEND_REF_TAB(client);
	params.recv_buf.words =	IPC_RECV_BUF_WORDS(server);
	params.recv_buf.buf =	(fluke_ipc_word_t*)IPC_RECV_BUF(server);
	params.recv_buf_count =	IPC_RECV_BUF_COUNT(server);
	params.recv_buf_tab =	(fluke_ipc_buf_t*)IPC_RECV_BUF_TAB(server);
	params.recv_ref_count =	IPC_RECV_REF_COUNT(server);
	params.recv_ref_tab =	(struct fluke_ref**)IPC_RECV_REF_TAB(server);

	/* Speculatively transfer our request message to the server. */
	ipc_transfer_min_msg(client, server);
	rc = ipc_speculative_transfer(client, server, &params, &params);
	if (rc)
		return rc;

	/* The transfer succeeded, so update the receiver's register state. */
	client->ipc_state.idemp_server = server;
	server->ipc_state.idemp_client = client;
	IPC_RECV_BUF_WORDS(server) = params.recv_buf.words;
	IPC_RECV_BUF_COUNT(server) = params.recv_buf_count;
	IPC_RECV_REF_COUNT(server) = params.recv_ref_count;
	IPC_FINISH_RECEIVE(server, FLUKE_IPC_WAIT_IDEMPOTENT);

	return thread_handoff(client, WAIT_IPC_CALL, server);
}
#endif /* MD_IPC_CALL */

#ifndef MD_IPC_REPLY
int ipc_reply(struct s_thread *const server)
{
	struct s_thread *client = server->ipc_state.idemp_client;
	wait_val_t wval;
	fluke_ipc_params_t params;
	int rc;

	assert(server == current_thread());
	assert(client == 0 || client->ipc_state.idemp_server == server);

	if (client == 0) {
		/*
		 * The idempotent IPC was probably interrupted for some reason,
		 * so just throw away the reply and return success.
		 */
		return 0;
	}

	/* Capture the client thread. */
	wval = thread_wakeup(client, WAIT_IPC_IDEMPOTENT);
	if (wval == 0) {
		/*
		 * Client thread wasn't still waiting on the call,
		 * so it was probably interrupted somehow.
		 * Again, just discard the reply;
		 * we will disconnect at the next wait_receive().
		 */
		return 0;
	}
	if (wval == WAIT_IPC_EXCEPTION) {
		/*
		 * This call was caused by an exception;
		 * Need to read the minimum message to see if the exception
		 * handler succeeded.  If not we arrange to return a user
		 * exception.  In either case we restart the client thread.
		 */
#if 0
/*
 * XXX figure out how to do this in an MI way.
 * Maybe ipc_transfer_min_msg returns an rc value.
 */
		rc = copyin_local(IPC_SEND_BUF(server), min_msg,
				  FLUKE_MIN_MSG_WORDS);
		if (rc) {
			thread_ready(client);
			return rc;
		}
		if (min_msg[0])
			client->wait_state.resume_rc = KR_USER_EXCEPTION;
#endif
		client->ipc_state.idemp_server = 0;
		server->ipc_state.idemp_client = 0;
		thread_ready(client);
		return 0;
	}
	assert(wval == WAIT_IPC_CALL);

	/*
	 * Extract the send and receive parameters
	 * from the sender's and receiver's register state.
	 */
	params.send_buf.words =	IPC_SEND_BUF_WORDS(server);
	params.send_buf.buf =	(fluke_ipc_word_t*)IPC_SEND_BUF(server);
	params.send_buf_count =	IPC_SEND_BUF_COUNT(server);
	params.send_buf_tab =	(fluke_ipc_buf_t*)IPC_SEND_BUF_TAB(server);
	params.send_ref_count =	IPC_SEND_REF_COUNT(server);
	params.send_ref_tab =	(struct fluke_ref**)IPC_SEND_REF_TAB(server);
	params.recv_buf.words =	IPC_RECV_BUF_WORDS(client);
	params.recv_buf.buf =	(fluke_ipc_word_t*)IPC_RECV_BUF(client);
	params.recv_buf_count =	IPC_RECV_BUF_COUNT(client);
	params.recv_buf_tab =	(fluke_ipc_buf_t*)IPC_RECV_BUF_TAB(client);
	params.recv_ref_count =	IPC_RECV_REF_COUNT(client);
	params.recv_ref_tab =	(struct fluke_ref**)IPC_RECV_REF_TAB(client);

	/*
	 * Speculatively transfer our reply message to the client.
	 * If this gets interrupted, the client will restart from scratch.
	 */
	ipc_transfer_min_msg(server, client);
	rc = ipc_speculative_transfer(server, client, &params, &params);
	if (rc)
		return rc;

	/* The transfer succeeded, so disconnect and release the client. */
	client->ipc_state.idemp_server = 0;
	server->ipc_state.idemp_client = 0;
	IPC_RECV_BUF_WORDS(client) = params.recv_buf.words;
	IPC_RECV_BUF_COUNT(client) = params.recv_buf_count;
	IPC_RECV_REF_COUNT(client) = params.recv_ref_count;
	IPC_FINISH_RECEIVE(client, 0);
	thread_ready(client);
	return 0;
}
#endif /* MD_IPC_REPLY */

/*** Client-side reliable IPC path components ***/

#ifndef MD_IPC_CLIENT_1_CONNECT_SEND
int ipc_client_1_connect_send(struct s_thread *const client,
			      oskit_addr_t connected_ip)
{
	struct s_thread *server;
	int rc;

	rc = ipc_client_disconnect(client);
	if (rc)
		return rc;
	assert(client->ipc_state.server == 0);

	/* Establish a connection to a new server thread. */
	rc = s_port_capture(client, &server);
	if (rc)
		return rc;
	client->ipc_state.server = server;
	server->ipc_state.client = client;
	client->ipc_state.flags |= FLUKE_THREAD_CLIENT_SENDER;
	server->ipc_state.flags &= ~FLUKE_THREAD_SERVER_SENDER;

	ipc_transfer_min_msg(client, server);

	EXC_SET_IP(client, connected_ip);
	EXC_SET_IP(server, fluke_ipc_server_receive__entry);

	/* Perform the send operation. */
	return ipc_reliable_transfer(client, server);
}
#endif /* MD_IPC_CLIENT_1_CONNECT_SEND */

#ifndef MD_IPC_CLIENT_1_ACK_SEND
int ipc_client_1_ack_send(struct s_thread *const client, oskit_addr_t acked_ip)
{
	struct s_thread *server;
	wait_val_t wval;
	int rc;

	rc = ipc_client_find_server(client, FLUKE_IPC_ACK_DISCONNECTED);
	if (rc)
		return rc;

	server = client->ipc_state.server;
	assert(server != 0);

	/* Capture the server if it's waiting to send. */
	if ((wval = thread_wakeup(server, WAIT_IPC_SRV_SENDER)) == 0) {

		/*
		 * Server's not in the proper wait state.
		 * Check the current connection direction.
		 */
		if (client->ipc_state.flags & FLUKE_THREAD_CLIENT_SENDER)
			EXC_RETURN_INSANITY(NOT_RECEIVER);

		/*
		 * It's going the right direction,
		 * but the server just isn't ready - wait for it.
		 */
		return thread_wait(client, WAIT_IPC_CLI_ASEND,
				   server, WAIT_IPC_SRV_SENDER);
	}
	if (wval != WAIT_IPC_SRV_ORECV) {
		assert(wval == WAIT_IPC_SRV_SEND);

		/*
		 * The server is still trying to send data to us.
		 * Just handoff to it and let it drain its message.
		 */
		return thread_handoff(client, WAIT_IPC_CLI_ASEND, server);
	}

	/* Both the client and the server are ready to reverse, so do so. */
	client->ipc_state.flags |= FLUKE_THREAD_CLIENT_SENDER;
	server->ipc_state.flags &= ~FLUKE_THREAD_SERVER_SENDER;
	ipc_transfer_min_msg(client, server);

	EXC_SET_IP(client, acked_ip);
	EXC_SET_IP(server, fluke_ipc_server_receive__entry);

	return ipc_reliable_transfer(client, server);
}
#endif /* MD_IPC_CLIENT_1_ACK_SEND */

int ipc_client_1_send(struct s_thread *const client, wait_val_t *out_wval)
{
	struct s_thread *server;
	wait_val_t wval;
	int rc;

	rc = ipc_client_find_server(client, FLUKE_IPC_SEND_DISCONNECTED);
	if (rc)
		return rc;

	server = client->ipc_state.server;
	assert(server != 0);

	/* Capture the server if it's trying to receive. */
	wval = thread_wakeup(server, WAIT_IPC_SRV_RECEIVER);
	if (wval == 0) {
		if (!(client->ipc_state.flags & FLUKE_THREAD_CLIENT_SENDER))
			EXC_RETURN_INSANITY(NOT_SENDER);
		return thread_wait(client, WAIT_IPC_CLI_SEND,
				   server, WAIT_IPC_SRV_RECEIVER);
	}
	if (wval == WAIT_IPC_SRV_RECV) {
		rc = ipc_reliable_transfer(client, server);
		if (rc)
			return rc;
	} else {
		assert(wval == WAIT_IPC_SRV_ASEND);
		/*
		 * The impatient server is acking us before we're done.
		 * Therefore, just throw away our send data.
		 */
	}

	*out_wval = wval;
	return 0;
}

#ifndef MD_IPC_CLIENT_2_OVER_RECEIVE
int ipc_client_2_over_receive(struct s_thread *const client, wait_val_t wval)
{
	struct s_thread * const server = client->ipc_state.server;

	assert(server != 0);

	/*
	 * Since the server still thinks it's receiving data,
	 * notify it that the message has ended.
	 */
	if (wval == WAIT_IPC_SRV_RECV)
		IPC_FINISH_RECEIVE(server, 0);
	else
		assert(wval == WAIT_IPC_SRV_ASEND);

	/* Hand control to the server. */
	return thread_handoff(client, WAIT_IPC_CLI_ORECV, server);
}
#endif /* MD_IPC_CLIENT_2_OVER_RECEIVE */

int ipc_client_12_receive(struct s_thread *const client)
{
	struct s_thread *server;
	wait_val_t wval;
	int rc;

	rc = ipc_client_find_server(client, FLUKE_IPC_RECV_DISCONNECTED);
	if (rc)
		return rc;

	server = client->ipc_state.server;
	assert(server != 0);

	/* Capture the server thread as the sender. */
	if ((wval = thread_wakeup(server, WAIT_IPC_SRV_SENDER)) == 0) {
		if (client->ipc_state.flags & FLUKE_THREAD_CLIENT_SENDER)
			EXC_RETURN_INSANITY(NOT_RECEIVER);
		return thread_wait(client, WAIT_IPC_CLI_RECV,
				   server, WAIT_IPC_SRV_SENDER);
	}
	if (wval != WAIT_IPC_SRV_SEND) {
		assert(wval == WAIT_IPC_SRV_ORECV);

		/* The server says it's done sending, so just return. */
		thread_ready(server);
		IPC_FINISH_RECEIVE(client, 0);
		return KR_RESTART;
	}

	return thread_handoff(client, WAIT_IPC_CLI_RECV, server);
}

/*** Server-side reliable IPC path components ***/

#ifndef MD_IPC_SERVER_2_ACK_SEND
int ipc_server_1_ack_send(struct s_thread *const server, oskit_addr_t acked_ip)
{
	struct s_thread *client;
	wait_val_t wval;
	int rc;

	rc = ipc_server_find_client(server, FLUKE_IPC_ACK_DISCONNECTED);
	if (rc)
		return rc;

	client = server->ipc_state.client;
	assert(client != 0);

	/* Capture the client if it's waiting to send. */
	if ((wval = thread_wakeup(client, WAIT_IPC_CLI_SENDER)) == 0) {

		/*
		 * Client's not in the proper wait state.
		 * Check the current connection direction.
		 */
		if (server->ipc_state.flags & FLUKE_THREAD_SERVER_SENDER)
			EXC_RETURN_INSANITY(NOT_RECEIVER);

		/*
		 * It's going the right direction,
		 * but the client just isn't ready - wait for it.
		 */
		return thread_wait(server, WAIT_IPC_SRV_ASEND,
				   client, WAIT_IPC_CLI_SENDER);
	}
	if (wval != WAIT_IPC_CLI_ORECV) {
		assert(wval == WAIT_IPC_CLI_SEND);

		/*
		 * The client is still trying to send data to us.
		 * Just handoff to it and let it drain its message.
		 */
		return thread_handoff(server, WAIT_IPC_SRV_ASEND, client);
	}

	/* Both the server and the client are ready to reverse, so do so. */
	server->ipc_state.flags |= FLUKE_THREAD_SERVER_SENDER;
	client->ipc_state.flags &= ~FLUKE_THREAD_CLIENT_SENDER;
	ipc_transfer_min_msg(server, client);

	EXC_SET_IP(server, acked_ip);
	EXC_SET_IP(client, fluke_ipc_client_receive__entry);

	return ipc_reliable_transfer(server, client);
}
#endif /* MD_IPC_SERVER_2_ACK_SEND */

int ipc_server_1_send(struct s_thread *const server, wait_val_t *out_wval)
{
	struct s_thread *client;
	wait_val_t wval;
	int rc;

	rc = ipc_server_find_client(server, FLUKE_IPC_SEND_DISCONNECTED);
	if (rc)
		return rc;

	client = server->ipc_state.client;
	assert(client != 0);

	/* Capture the client if it's trying to receive. */
	wval = thread_wakeup(client, WAIT_IPC_CLI_RECEIVER);
	if (wval == 0) {
		if (!(server->ipc_state.flags & FLUKE_THREAD_SERVER_SENDER))
			EXC_RETURN_INSANITY(NOT_SENDER);
		return thread_wait(server, WAIT_IPC_SRV_SEND,
				   client, WAIT_IPC_CLI_RECEIVER);
	}
	if (wval == WAIT_IPC_CLI_RECV) {
		rc = ipc_reliable_transfer(server, client);
		if (rc)
			return rc;
	} else {
		assert(wval == WAIT_IPC_CLI_ASEND);
		/*
		 * The impatient client is acking us before we're done.
		 * Therefore, just throw away our send data.
		 */
	}

	*out_wval = wval;
	return 0;
}

#ifndef MD_IPC_SERVER_2_WAIT_RECEIVE
int ipc_server_2_wait_receive(struct s_thread *const server, wait_val_t wval,
			      oskit_addr_t wait_receive_ip)
{
	struct s_thread * const client = server->ipc_state.client;

	/*
	 * Break the connection with the client thread and release it.
	 * It's much more efficient to do this manually here
	 * than to call ipc_server_disconnect() after releasing the client,
	 * because right now we know the client is passive (captured),
	 * whereas ipc_server_disconnect() would see it as "not waiting"
	 * and cause us to sleep ourselves, producing extra context switches.
	 */
	server->ipc_state.client = 0;
	client->ipc_state.server = 0;
	IPC_FINISH_RECEIVE(client, FLUKE_IPC_RECV_DISCONNECTED);
	thread_ready(client);

	/* Update the server's IPC state. */
	EXC_SET_IP(server, wait_receive_ip);

	return s_pset_wait_receive(server);
}
#endif /* MD_IPC_SERVER_2_WAIT_RECEIVE */

#ifndef MD_IPC_SERVER_2_OVER_RECEIVE
int ipc_server_2_over_receive(struct s_thread *const server, wait_val_t wval)
{
	struct s_thread * const client = server->ipc_state.client;

	assert(client != 0);

	/*
	 * Since the client still thinks it's receiving data,
	 * notify it that the message has ended.
	 */
	if (wval == WAIT_IPC_CLI_RECV)
		IPC_FINISH_RECEIVE(client, 0);
	else
		assert(wval == WAIT_IPC_CLI_ASEND);

	/* Hand control to the client. */
	return thread_handoff(server, WAIT_IPC_SRV_ORECV, client);
}
#endif /* MD_IPC_SERVER_2_OVER_RECEIVE */

#ifndef MD_IPC_SERVER_3_WAIT_RECEIVE
int ipc_server_3_wait_receive(struct s_thread *const server,
			      oskit_addr_t wait_receive_ip)
{
	/* Update the server's IPC state. */
	EXC_SET_IP(server, wait_receive_ip);

	return s_pset_wait_receive(server);
}
#endif /* MD_IPC_SERVER_3_WAIT_RECEIVE */

int ipc_server_12_receive(struct s_thread *const server)
{
	struct s_thread *client;
	wait_val_t wval;
	int rc;

	rc = ipc_server_find_client(server, FLUKE_IPC_RECV_DISCONNECTED);
	if (rc)
		return rc;

	client = server->ipc_state.client;
	assert(client != 0);

	/* Capture the client thread as the sender. */
	if ((wval = thread_wakeup(client, WAIT_IPC_CLI_SENDER)) == 0) {
		if (server->ipc_state.flags & FLUKE_THREAD_SERVER_SENDER)
			EXC_RETURN_INSANITY(NOT_RECEIVER);
		return thread_wait(server, WAIT_IPC_SRV_RECV,
				   client, WAIT_IPC_CLI_SENDER);
	}
	if (wval != WAIT_IPC_CLI_SEND) {
		assert(wval == WAIT_IPC_CLI_ORECV);

		/* The client says it's done sending, so just return. */
		thread_ready(client);
		IPC_FINISH_RECEIVE(server, 0);
		return KR_RESTART;
	}

	return thread_handoff(server, WAIT_IPC_SRV_RECV, client);
}
