/*
 * Copyright (c) 1996 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#ifndef _FLUKE_KERNEL_IPC_H_
#define _FLUKE_KERNEL_IPC_H_

#include "link.h"
#include "mutex.h"
#include "wait.h"
#include "cond.h"

#if FLASK
#include <flask/security.h>
#endif

/* Per-thread state relating to IPC.  */
struct ipc_state
{
	/*
	 * Flags defining the direction of the current reliable connection:
	 * contains FLUKE_THREAD_CLIENT_SENDER and FLUKE_THREAD_SERVER_SENDER.
	 * (This pair is defined below as THREAD_IPCFLAGS.)
	 */
	unsigned		flags;

	/*
	 * Pointer to the server thread to which this client is attached,
	 * or NULL if no active connection exists.
	 * Invariant: server->client == this.
	 * Invariant: server_link == NULL.
	 */
	struct s_thread		*server;

	/*
	 * Reference to the client thread to which this server is attached.
	 * Invariant: client->server == this.
	 * Invariant: client_link == NULL.
	 */
	struct s_thread		*client;

	/*
	 * Client and server links for this thread,
	 * used when only a half-connection is present
	 * or when the thread is stopped.
	 * When the thread is live, these fields are thread-private;
	 * when stopped, the fields are locked by the thread's s_ob lock.
	 */
	link_t			client_link;
	link_t			server_link;

	/*
	 * Idempotent client/server connection pointers.
	 * There are no pickled versions of these,
	 * since breaking an idempotent connection just cancels it.
	 */
	struct s_thread		*idemp_server;
	struct s_thread		*idemp_client;

	/*
	 * Server-side IPC entrypoint information,
	 * set using fluke_thread_set_server_info().
	 */
	oskit_addr_t		server_pset_uva;
	unsigned		server_esp;
	unsigned		server_oneway_eip;
	unsigned		server_idempotent_eip;
	unsigned		server_reliable_eip;

#if FLASK
        /* 
	 * The effective SID of this thread in its role as
	 * a server waiting on a port set.
	 */
	security_id_t		effective_server_sid;

	/*
	 * The effective client SID of this thread's client.
	 * The connection is only complete if this SID is identical
	 * to the effective_client_sid field in the client's state.
	 */
	security_id_t		client_effective_client_sid;

        /* 
	 * The effective SID of this thread in its role as
	 * a client in a reliable IPC connection.
	 *
	 * The effective client SID for an idempotent or
	 * one-way IPC has no relationship to this effective 
	 * client SID.
	 */
	security_id_t		effective_client_sid;

	/*
	 * The effective server SID of this thread's server.
	 * The connection is only complete if this SID is identical
	 * to the effective_server_sid field in the server's state.
	 */
	security_id_t		server_effective_server_sid;

	/*
	 * A flag indicating whether or not this thread
	 * wants to be supplied with the effective SID of
	 * its client when it is awakened from wait_receive.
	 * This flag is set to true when this thread
	 * uses a wait_receive_secure call, and causes
	 * a min_msg register to be used for the
	 * effective client SID rather than for the
	 * corresponding min_msg word.
	 */
	oskit_bool_t		server_wants_client_sid;
#endif 

	/* Include machine-dependent state, if any. */
#ifdef IPC_STATE_MD
	struct ipc_state_md	md;
#endif
};

#define THREAD_IPCFLAGS		(FLUKE_THREAD_CLIENT_SENDER |	\
				 FLUKE_THREAD_SERVER_SENDER)


struct link;
struct s_thread;
struct fluke_ipc_params;
struct fluke_thread_state;
#if FLASK
struct fluke_thread_state_secure;
#endif

void thread_ipc_state_init(struct s_thread *th);

void thread_ipc_state_destroy(struct s_thread *th);

void thread_ipc_state_start(struct s_thread *th);

int thread_ipc_state_stop(struct s_thread *th, struct s_thread *stopper);

void thread_ipc_state_get(struct s_thread *th, struct link *client_link,
			  struct link *server_link);

void thread_ipc_state_set(struct s_thread *th, struct link *client_link,
			  struct link *server_link);

/*
 * Pickle any active connection this thread is involved in.
 * Pickler is the thread requesting the pickle operation.
 */
int ipc_client_pickle(struct s_thread *client, struct s_thread *pickler,
		      struct link *slink, unsigned *flags);
int ipc_server_pickle(struct s_thread *server, struct s_thread *pickler,
		      struct link *clink, unsigned *flags);

/*
 * Unpickle any pickled connection this thread is involved in.
 * Must only be called if there is no active connection (yet).
 * On successful return, if there is no active connection,
 * then there is no connection at all.
 * These functions may wait, causing a KR_RESTART return.
 */
int ipc_client_unpickle(struct s_thread *client, int no_wait);
int ipc_server_unpickle(struct s_thread *server, int no_wait);

/*
 * Ensure that the appropriate reliable connection exists and is active.
 */
int ipc_client_find_server(struct s_thread *const client, int disconnect_code);
int ipc_server_find_client(struct s_thread *const server, int disconnect_code);

/* Disconnect any reliable connection with me as the client. */
int ipc_client_disconnect(struct s_thread *client);

/* Disconnect any idempotent connection with me as the client. */
int ipc_client_idemp_disconnect(struct s_thread *client);

/* Disconnect any reliable or idempotent connection with me as the server. */
int ipc_server_disconnect(struct s_thread *server);

/*
 * This is called from the kernel entry layer
 * after a thread is cancelled and rolled back to its user mode entrypoint.
 * It checks to see if the cancel was related to IPC, and if so, handles it.
 * In particular, it checks to see if the server wants to disconnect.
 */
void thread_ipc_canceled(struct s_thread *cur_thread);

/*
 * Called by the thread layer to get and set the IPC parts of a thread's state.
 */
void thread_ipc_get_state(struct s_thread *th,
			  struct fluke_thread_state *state);
void thread_ipc_set_state(struct s_thread *th,
			  struct fluke_thread_state *state);
#if FLASK
void thread_ipc_get_state_secure(struct s_thread *th,
				 struct fluke_thread_state_secure *state);
#endif

#ifdef IPC_STATE_MD
/*
 * These are called by the above routines
 * to get and set the MD part of a thread's IPC state.
 */
void thread_ipc_get_state_md(struct s_thread *th,
			     struct fluke_thread_state *state);
void thread_ipc_set_state_md(struct s_thread *th,
			     struct fluke_thread_state *state);
#endif

/*
 * This routine copies the minimum message from sender to receiver,
 * immediately after a connection establishment or reversal.
 * It is always machine-dependent,
 * and generally just copies over a few registers.
 * It may not fail, because the establishment or reversal is already done
 * and cannot easily be undone.
 */
void ipc_transfer_min_msg(struct s_thread *sender, struct s_thread *receiver);

/*
 * This routine handles all "bulk" data transfer for reliable IPC
 * (i.e., transfer of all data beyond minimal message data).
 * It expects 'sender' to be the current thread,
 * and 'receiver' to be currently captured by the sender.
 * On successful (zero) return, the receiver is still captured;
 * on unsuccessful (nonzero) return, the receiver has been released.
 * In either case, the states of both the sender and the receiver
 * have been updated to reflect the amount of data transferred so far.
 */
int ipc_reliable_transfer(struct s_thread *sender, struct s_thread *receiver);

/*
 * This routine handles all "bulk" data transfer for one-way and idempotent IPC.
 * It expects 'sender' to be the current thread,
 * and 'receiver' to be currently captured by the sender.
 * On successful (zero) return, the receiver is still captured;
 * on unsuccessful (nonzero) return, the receiver has been released.
 * The receiver running out of space counts as "success";
 * any extra send data is simply dropped.
 * The specified IPC params structures (which may be the same)
 * are used instead of the sender's and receiver's register state.
 */
int ipc_speculative_transfer(struct s_thread *sender,
			     struct s_thread *receiver,
			     struct fluke_ipc_params *send_params,
			     struct fluke_ipc_params *recv_params);

/*
 * The following routines implement the unreliable IPC paths.
 */
#if FLASK
int ipc_send(struct s_thread *const client, oskit_bool_t secure);
int ipc_call(struct s_thread *const client, oskit_bool_t secure);
#else
int ipc_send(struct s_thread *const client);
int ipc_call(struct s_thread *const client);
#endif
int ipc_reply(struct s_thread *const client);

/*
 * The following routines implement components of reliable IPC paths.
 * Most paths are built from two phases: 1 (send) and 2 (receive).
 * All phase 1 routines are called with the current thread in an unknown state,
 * and on successful return, the other thread is captured as the receiver.
 * All phase 2 routines are called with the other captured as receiver,
 * and return with everything all done, ready to return from the syscall.
 * Routines that implement both phases 1 and 2 are marked _12_.
 */
#if FLASK
int ipc_client_1_connect_send(struct s_thread *const client,
			      oskit_bool_t secure,
			      oskit_addr_t connected_ip);
#else
int ipc_client_1_connect_send(struct s_thread *const client,
			      oskit_addr_t connected_ip);
#endif
int ipc_client_1_ack_send(struct s_thread *const client, oskit_addr_t acked_ip);
int ipc_client_1_send(struct s_thread *const client, wait_val_t *out_wval);
int ipc_client_2_over_receive(struct s_thread *const client, wait_val_t wval);
int ipc_client_12_receive(struct s_thread *const client);

int ipc_server_1_ack_send(struct s_thread *const server, oskit_addr_t acked_ip);
int ipc_server_1_send(struct s_thread *const server, wait_val_t *out_wval);
int ipc_server_2_wait_receive(struct s_thread *const server, wait_val_t wval,
			      oskit_addr_t wait_receive_ip);
int ipc_server_2_over_receive(struct s_thread *const server, wait_val_t wval);
int ipc_server_3_wait_receive(struct s_thread *const server, 
			      oskit_addr_t wait_receive_ip);
int ipc_server_12_receive(struct s_thread *const server);

#if FLASK
void thread_ipc_revoke(struct s_thread *thread, 
		      security_id_t ssid,
		      security_id_t tsid,
		      access_vector_t revoke,
		      oskit_bool_t try_remove,
		      access_vector_t *out_retained);

int thread_ipc_reset(struct s_thread *thread, oskit_bool_t *out_reset);

#define FLASK_IPC_STATE_DISCONNECT(client,server) \
    server->ipc_state.client_effective_client_sid = SECSID_NULL; \
    server->ipc_state.effective_server_sid = server->sob.sid;    \
    client->ipc_state.server_effective_server_sid = SECSID_NULL; \
    client->ipc_state.effective_client_sid = client->sob.sid; 
#endif

#endif /* _FLUKE_KERNEL_IPC_H_ */
