/*
 * Copyright (c) 1995, 1997, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Basic kernel memory allocator module.
 * This module uses the List Memory Manager (LMM) provided by the OS toolkit;
 * the only thing it really does is add a global lock to provide thread safety.
 */

#include <malloc.h>
#include <oskit/lmm.h>

#include "kalloc.h"
#include "mutex.h"
#include "debug.h"
#include "preempt.h"

#include "config_kalloc_debug.h"

#ifdef MEMSTATS
unsigned long mstats_physmem;
unsigned long mstats_bytes, mstats_maxbytes;
#endif

mutex_t kalloc_lock;

#ifdef CONFIG_KALLOC_DEBUG
#include <oskit/memdebug.h>
void* memdebug_alloc(size_t alignment, size_t bytes, unsigned flags,
		     unsigned char initval, char caller, char *file, int line);
int memdebug_free(void* mem, int wipeval, char caller, int blocksize,
		  char *file, int line);
void mem_lock_init(void) { }
void mem_lock(void) { mutex_lock(&kalloc_lock); }
void mem_unlock(void) { mutex_unlock(&kalloc_lock); }
int morecore(size_t need) { return 0; }
#endif

void *kalloc_gen(oskit_size_t size, unsigned align_bits, lmm_flags_t flags)
{
	void *mem;

#ifdef CONFIG_KALLOC_DEBUG
	mem = memdebug_alloc(1 << align_bits, size, flags,
			     0xaa, 0, 0, 0);
	mutex_lock(&kalloc_lock);
#else
	mutex_lock(&kalloc_lock);
	if (align_bits == 0)
		mem = lmm_alloc(&malloc_lmm, size, flags);
	else
		mem = lmm_alloc_aligned(&malloc_lmm, size, flags, align_bits,
					0);
#endif
#ifdef MEMSTATS
	if (mem) {
		mstats_bytes += size;
		if (mstats_bytes > mstats_maxbytes)
			mstats_maxbytes = mstats_bytes;
	}
#endif
	mutex_unlock(&kalloc_lock);
	return mem;
}


void kfree(void *mem, oskit_size_t size)
{
#ifdef CONFIG_KALLOC_DEBUG
	memdebug_free(mem, 0xee, 0, size, 0, 0);
	mutex_lock(&kalloc_lock);
#else
	mutex_lock(&kalloc_lock);
	lmm_free(&malloc_lmm, mem, size);
#endif

#ifdef MEMSTATS
	mstats_bytes -= size;
#endif
	mutex_unlock(&kalloc_lock);
}

void kalloc_boot(void)
{
	mutex_init(&kalloc_lock);
#ifdef MEMSTATS
	mstats_physmem = lmm_avail(&malloc_lmm, 0);
#endif
}

#ifdef MEMSTATS
void kalloc_dump_mstats()
{
	printf("Approx phys mem:     %10lu bytes\n",
	       mstats_physmem);
	printf("Kernel memory:       %10lu bytes (%10lu bytes max)\n",
	       mstats_bytes, mstats_maxbytes);
}
#endif
