/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Utility routines related to user-space references,
 * used by the other kentry-layer syscall routines.
 */

#include <fluke/exception.h>

#include "s_thread.h"
#include "s_ref.h"

/*
 * Change the specified user-space reference to contain the provided link.
 * Consumes the link on either success or failure.
 */
int kentry_ref_associate(
	s_thread_t		*cur_th,
	oskit_addr_t		ref_uva,
	struct link		*link)
{
	s_ref_t *ref;
	int rc;

	rc = task_ob_lookup_type(cur_th, ref_uva,
				 FLUKE_PROT_WRITE, FLUKE_REF,
				 (s_ob_t **)&ref);
	if (rc)
	{
		link_destroy(link);
		return rc;
	}

	s_ref_link_transfer(link, ref);
	link_destroy(link);
	s_ref_unlock(ref);

	return 0;
}

/*
 * Lookup a reference object at a given VA and ensure it is of the correct type
 * (note that NULL is always correct).
 * Nothing should be locked on call, nothing is returned locked.
 */
int kentry_ref_extract_type(
	s_thread_t		*cur_th,
	oskit_addr_t		ref_uva,
	fluke_type_t		type,
	struct link		*out_link)
{
	s_ref_t *ref;
	s_ob_t *ob;
	int rc;

	rc = task_ob_lookup_type(cur_th, ref_uva,
				 FLUKE_PROT_READ, FLUKE_REF,
				 (s_ob_t **)&ref);
	if (rc)
		return rc;
	s_ref_create_link(ref, out_link);
	s_ref_unlock(ref);

	ob = link_follow(out_link);
	if (ob && ob->type != type)
	{
		s_ob_unlock(ob);
		link_destroy(out_link);
		exc_set_insanity(cur_th, FLUKE_INSANITY_WRONG_REF_TYPE | type);
		return KR_USER_EXCEPTION;
	}
	if (ob)
		s_ob_unlock(ob);

	return 0;
}

int kentry_ob_reference(
	s_thread_t		*thread,
	oskit_addr_t		ob_uva,
	oskit_addr_t		ref_uva,
	fluke_type_t		type)
{
	struct link ob_link;
	s_ob_t *ob;
	int rc;

	/*
	 * Look up the object and create a link to it.
	 */
	rc = task_ob_lookup_type(thread, ob_uva, FLUKE_PROT_WRITE, type, &ob);
	if (rc)
		return rc;
	link_produce(&ob_link, ob);
	s_ob_unlock(ob);

	/*
	 * Look up the reference and insert the object link into it.
	 * This function always consumes the reference we produced above.
	 */
	rc = kentry_ref_associate(thread, ref_uva, &ob_link);
	return rc;
}
