/*
 * Copyright (c) 1997, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include "prog.h"
#include "debug.h"
#include <errno.h>
#include <oskit/boolean.h>
#include <fluke/nest/server/com.h>
#include <fluke/nest/server/fdev.h>
#include <oskit/page.h>

/* for my task; could reference the kserver task directly */
#include <fluke/cp/startup.h>
extern struct cp_startup *__startup;


/*
 * This is the FDev device interface for the kserver.
 * It uses MOM directly, rather than a MOM->COM wrapper.
 */

mom_class_t fdev_class;
nest_FDev_t *interface;

struct nest_FDev_struct {
	mom_obj_header_t header;

	/* and per-object data here... */
	/* XXX: state is stored in the per-task structure */
};


static void
dead_fdev(void *object)
{
	printf("In %s\n", __FUNCTION__);
}

/*
 * Init routine
 * This creates a single class and object for use by everyone.
 */

char master_bitmap[65536/8 + 1];

#include "per_cpu.h"

void
ks_initialize_fdev(void)
{
	static int init = 0;

	if (!init) {
		int rc;
		init = 1;
		rc = mom_class_create(nest_module, dead_fdev,
			nest_FDev_server, &fdev_class);
		assert(rc == 0);

		/* Allocate a new MOM object */
		rc = mom_obj_create(nest_module, fdev_class,
			__alignof(*interface), sizeof(*interface),
			&interface);
		assert(!rc);
		assert(interface);

		/* init bitmap */
		memset(master_bitmap, 0xff, sizeof(master_bitmap));

		/* XXX: port 0x80 is used for timing delays. */


		/* disallow driver access to these ranges */
#if 0	/* ISA DMA ports */
		io_alloc(master_bitmap, 0x00, 0x10); /* dma1 */
		io_alloc(master_bitmap, 0x80, 0x10); /* dma page reg */
		io_alloc(master_bitmap, 0xc0, 0x20); /* dma2 */
#endif
		io_alloc(master_bitmap, 0x40, 0x20); /* timer */
		io_alloc(master_bitmap, 0x70, 0x10); /* rtc */
		io_alloc(master_bitmap, 0xf0, 0x02); /* fpu */
		io_alloc(master_bitmap, 0xf8, 0x08); /* fpu */

		io_alloc(master_bitmap, 0x20, 0x02); /* master pic */
		io_alloc(master_bitmap, 0xa0, 0x02); /* slave pic */
	}
}


/*
 * We replace part of the parent interface.
 * This overrides the implementation in libnest.
 */
#include <fluke/nest/server/parent.h>

mom_ref_t
nest_parent_get_fdev(nest_Parent_1_0_t *obj, mom_exc_t *env)
{
	mom_ref_t ret;

	mom_obj_reference(interface, &ret);
	return ret;
}


/*
 * Server work funcitons.
 */

void
nest_fdev_registerhandler(nest_FDev_t *obj, oskit_u16_t irq,
	oskit_u32_t flags, mom_exc_t *env)
{
	fluke_ref_t *client_ref;
	unsigned client_flags;
	unsigned rc;

	/* get the thread reference for the thread talking to me */

	/* create an empty reference */
	client_ref = (fluke_ref_t *)smemalign(FLUKE_REF_SIZE, FLUKE_REF_SIZE);
	rc = fluke_ref_create(client_ref); assert(rc == 0);

	/* get the client thread reference */
	fluke_thread_get_client(fluke_thread_self(), client_ref,
		&client_flags);

	/* call the kernel's register-handler code */
	rc = my_fdev_irq_register(client_ref, irq, flags);

	if (rc) {
		printf("rc = %d\n", rc);
		mom_set_exc(env, rc);
	}

	fluke_ref_destroy(client_ref);
}


void
nest_fdev_unregisterhandler(nest_FDev_t *obj, oskit_u16_t irq,
	mom_exc_t *env)
{
	fluke_ref_t *client_ref;
	unsigned client_flags;
	unsigned rc;

	/* get the thread reference for the thread talking to me */

	/* create an empty reference */
	client_ref = (fluke_ref_t *)smemalign(FLUKE_REF_SIZE, FLUKE_REF_SIZE);
	rc = fluke_ref_create(client_ref); assert(rc == 0);

	/* get the client thread reference */
	fluke_thread_get_client(fluke_thread_self(), client_ref,
		&client_flags);

	/* call the kernel's register-handler code */
	rc = my_fdev_irq_unregister(client_ref, irq);

	if (rc) {
		printf("rc = %d\n", rc);
		mom_set_exc(env, rc);
	}

	fluke_ref_destroy(client_ref);
}



/*
 * I/O Space management routines.
 */

/* Supervisor-mode half of ioalloc */
struct my_alloc_args {
	unsigned start;
	unsigned size;
	fluke_ref_t *client_ref;
};

#include "s_ref.h"
#include "s_thread.h"

#ifdef IOPL0
int
do_nest_fdev_ioalloc(void *a)
{
	struct my_alloc_args *args = a;
	s_ref_t *sref;
	struct s_thread *client_thread;
	int rc;

	/* find the s_ref for the fluke ref to find the s_thread */
	rc = task_ob_lookup_type(current_thread(),
			(oskit_addr_t)args->client_ref,
			FLUKE_PROT_READ|FLUKE_PROT_WRITE,
			FLUKE_REF, (struct s_ob **)&sref);
	assert(rc == 0);

	client_thread = s_ref_follow(sref);
	assert(client_thread);

	/*
	 * Okay, now I have the s_thread thread object.
	 * Follow pointer in that to get its task.
	 */
#if 0
	printf("%d bytes iospace (task %x) at %x\n", args->size,
		client_thread->task_state.task, args->start);
#endif

	/* add to the task's I/O permissions bitmap */
	rc = io_alloc(client_thread->task_state.task->bitmask,
		args->start, args->size);
	if (rc) {
		printf("alloc at %x failed, %d bytes\n", args->start,
			args->size);
	}
	assert(!rc);

	/* XXX: ensure updated bitmap takes affect immediatly */

	return 0;
}
#endif IOPL0

/* nest_fdev interfaces */
oskit_bool_t
nest_fdev_ioalloc(nest_FDev_t *obj, fluke_size_t start, fluke_size_t size,
	mom_exc_t *env)
{
	int rc;

	rc = io_alloc(master_bitmap, start, size);
	if (!rc) {
#ifdef IOPL0
		struct my_alloc_args args;
		unsigned client_flags;
		fluke_ref_t *client_ref;

		/* create an empty reference */
		client_ref = (fluke_ref_t *)smemalign(FLUKE_REF_SIZE,
			FLUKE_REF_SIZE);
		rc = fluke_ref_create(client_ref); assert(rc == 0);

		/* get the client thread reference */
		fluke_thread_get_client(fluke_thread_self(), client_ref,
			&client_flags);

		args.start = start;
		args.size = size;
		args.client_ref = client_ref;
		do_in_supervisor_mode(do_nest_fdev_ioalloc, &args);

		fluke_ref_destroy(client_ref);
#endif IOPL0
		return TRUE;
	} else
		return FALSE;
}


void
nest_fdev_iofree(nest_FDev_t *obj, fluke_size_t start, fluke_size_t size,
	mom_exc_t *env)
{
	io_free(master_bitmap, start, size);

	/* XXX take back from app! */
}


/*
 * Okay, when a driver requests physical memory, what do we do?
 * Well, we allocate memory with the specified flags, create a
 * region containing that memory, and return a reference to that
 * region.  This region will be cleaned up ``somehow'' using
 * the mythical reference counting.
 * Note that multiples of pages are always returned for simplicity here.
 */
mom_ref_t
nest_fdev_physalloc(nest_FDev_t *obj,
	fluke_size_t size, oskit_u32_t flags, fluke_size_t *phys,
	fluke_size_t *regionoff, mom_exc_t *env)
{
	/* XXX Plagerized my mmap support */

	mom_ref_t ret;
	int rc;
	fluke_region_t *region;
	fluke_region_state_t rstate;

	*phys = smemalignf(PAGE_SIZE, size, flags);

	*regionoff = 0;
	assert(size % PAGE_SIZE == 0);

	/* XXX: use lmm_alloc_gen to get memory < 128MB
	 * so it can be mapped at virt == phys. */
	region = (fluke_region_t *)
		smemalign(FLUKE_REGION_SIZE, FLUKE_REGION_SIZE);
	assert(region);

	/* XXX: This region is never reclaimed! */
	rc = fluke_region_create(region);
	assert(!rc);

	rstate.start_va = *phys;
	rstate.size = size;
	rstate.prot = FLUKE_PROT_READ | FLUKE_PROT_WRITE;

	fluke_region_set_state(region, &rstate, &__startup->task_ref, 0);

	/*
	 * I munge a mom_ref_t manually, as I want to
	 * to return a region reference instead of a port reference
	 */
	rc = mom_ref_allocate(&ret);
	assert(!rc);
	fluke_region_reference(region,
			       mom_ref_receive_fluke_ref(ret, FLUKE_REGION));

	return ret;
}

#if 0

void
nest_fdev_physfree(nest_FDev_t *obj, mom_ref_t mem, mom_exc_t *env)
{
}

#endif

/*
 * PCI Configuration access is done in supervisor mode, since kserver
 * doesn't have any more permission than normal nesters.
 */
struct config_read {
	int bus, device, function, reg;
	oskit_u32_t *data;
};
struct config_write {
	int bus, device, function, reg;
	oskit_u32_t data;
};

int pci_config_read(char bus, char device, char function, char port,
	unsigned *data);

int pci_config_write(char bus, char device, char function, char port,
	unsigned data);

static int
do_config_read(void *arg)
{
	struct config_read *tmp = arg;
	pci_config_read(tmp->bus, tmp->device, tmp->function,
		tmp->reg, tmp->data);
	return 0;
}

static int
do_config_write(void *arg)
{
	struct config_write *tmp = arg;
	pci_config_write(tmp->bus, tmp->device, tmp->function,
		tmp->reg, tmp->data);
	return 0;
}

/*
 * PCI configuration space is always 32-bits wide.
 * Other parameters are all 8 bits or less.
 */
void
nest_fdev_pci_config_read(nest_FDev_t *obj, unsigned char pci_bus,
		unsigned char pci_device, unsigned char pci_function,
		unsigned char pci_register, oskit_u32_t *data, mom_exc_t *env)
{
	struct config_read tmp = { pci_bus, pci_device, pci_function,
		pci_register, data };

	/* XXX: do checking */

	do_in_supervisor_mode(do_config_read, &tmp);
}

void
nest_fdev_pci_config_write(nest_FDev_t *obj, unsigned char pci_bus,
		unsigned char pci_device, unsigned char pci_function,
		unsigned char pci_register, oskit_u32_t data, mom_exc_t *env)
{
	struct config_read tmp = { pci_bus, pci_device, pci_function,
		pci_register, data };

	/* XXX: do checking */

	do_in_supervisor_mode(do_config_write, &tmp);
}

void
nest_fdev_pci_config_check(nest_FDev_t *obj, mom_exc_t *env)
{
	/* pci_config_init returns 0 on suceess, otherwise send exception */
	if (pci_config_init())
		mom_set_exc(env, ENOTSUP);
}


void
nest_fdev_pci_config_alloc(nest_FDev_t *obj, unsigned char pci_bus,
		unsigned char pci_device, unsigned char pci_function,
		mom_exc_t *env)
{
	printf("%s\n",__FUNCTION__);
	mom_set_exc(env, ENOTSUP);
}

void
nest_fdev_pci_config_free(nest_FDev_t *obj, unsigned char pci_bus,
		unsigned char pci_device, unsigned char pci_function,
		mom_exc_t *env)
{
	printf("%s\n",__FUNCTION__);
	mom_set_exc(env, ENOTSUP);
}


#include <oskit/x86/pc/pic.h>

int
my_pic_enable_irq(void *param)
{
	pic_enable_irq((char)param);
	return 0;
}

int
my_pic_disable_irq(void *param)
{
	pic_disable_irq((char)param);
	return 0;
}


/*
 * XXX: These will go away when disabling interrupts is handled
 * in user-mode by blocking the interrupt-handler thread.
 */
void
nest_fdev_irq_enable(nest_FDev_t *obj, oskit_u16_t irq, mom_exc_t *env)
{
	do_in_supervisor_mode(my_pic_enable_irq, irq);
}

void
nest_fdev_irq_disable(nest_FDev_t *obj, oskit_u16_t irq, mom_exc_t *env)
{
	do_in_supervisor_mode(my_pic_disable_irq, irq);
}
