/*
 * Copyright (c) 1997-1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Initialization of the kserver layer.
 * Creates the kserver port, port set, and service thread.
 * Called in user mode by prog_start()
 * before loading and starting the first external user-level program.
 */

#include "kserver.h"
#include "preempt.h"
#include <fluke/nest/spawn.h>
#include <fluke/nest/ref_pool.h>
#include <fluke/nest/server/com.h>
#include <oskit/fs/dir.h>
#include <oskit/fs/bmodfs.h>
#include <fluke/nest/server/fs.h>
#include "config_fdev.h"
#include "config_fdev_console.h"
#include "config_flob_check.h"
#include "config_profiling.h"
#include <oskit/c/sys/time.h>
#include <oskit/dev/dev.h>
#include <fluke/debug.h>
#include <fcntl.h>

static fluke_task_t first_child_task;
#ifdef HEARTBEAT_THREAD
#include "round_robin.h"	/* XXX SCHED_PRIINTR */
#endif

/*
 * These symbols are defined in libsac/fs.c, along with an
 * initializer for them.  We replace them with our own and initialize
 * them specially for use with the bmod filesystem.
 */
mom_ref_t __root_dir, __cwd_dir;
fluke_mutex_t __fs_mutex;
mode_t __not_cmask;

extern struct timeval time;	/* Current real time clock.  */

#if FLASK
/*
 * These symbols are defined in libsac/sid.c, along with an initializer.
 * We replace it with our own and initialize them specially.
 */
mom_ref_t     __security_ref;
security_id_t __mysid;
#endif

oskit_dir_t *bmod_root;


void
kserver(char **argv, char **envp)
{
	int rc;
	nest_nester_t *pserver;
	mom_ref_t pref;
	fluke_ref_t *ktask_ref, *kth_ref;
	struct cp_startup *app_startup;
#if FLASK
	oskit_iunknown_t *unki;
	oskit_avc_t *avc;
	oskit_dir_t *bmod_rroot, *rdevdir;
#endif
	mom_ref_t bmod_root_ref;
	static fluke_mutex_t bmod_mutex;
	struct nest_com_lock_data bmod_lock_data = { NULL, NULL, &bmod_mutex };
	mom_exc_t exc;
	oskit_dir_t *devdir;
	oskit_file_t *timefile;
	oskit_file_t *kmemfile;
	int appfd;

	printf("Kernel server starting up...command line `%s", argv[0]);
	for (rc = 1; argv[rc]; ++rc)
		printf(" %s", argv[rc]);
	printf("'\n");

	/*
	 * XXX MM library expects to extract a task reference from the
	 * startup structure.
	 */
	{
		static struct cp_startup kserver_startup;
		extern fluke_task_t kernel_task;
		extern struct cp_startup *__startup;

		__startup = &kserver_startup;
		(void)fluke_ref_create(&__startup->task_ref);
		fluke_task_reference(&kernel_task, &__startup->task_ref);
	}

	refholder_init();	/* XXX */

	if (!nest_module) {
		rc = mom_module_create(&nest_module);
		assert(!rc);
	}
	DMARK;

	ref_pool_init();

#if FLASK
	rc = mom_ref_allocate(&__security_ref);
	assert(rc == 0);
	__mysid = SECINITSID_KERNEL;
#endif

#ifdef PREEMPTSTATS
	{
		extern void disp_dump_preemptstats(void);
		extern void s_ob_dump_thread_preemptstats(void);

		atexit(disp_dump_preemptstats);
		atexit(s_ob_dump_thread_preemptstats);
	}
#endif

	/*
	 * Initialize memory allocation support.
	 */
	kserver_mempool_init();
	DMARK;

#if FLASK
	/*
	 * Initialize the interface to the kernel avc.
	 */
	kserver_avc_init(&avc);
	DMARK;
#endif

	/*
	 * Initialize the bmod filesystem and install it as the root fs.
	 */
	rc = fluke_mutex_create(&bmod_mutex); assert(rc == 0);
	fluke_debug_label_flob(&bmod_mutex, "bmod_mutex");

	rc = nest_com_fs_init ();
	assert(rc == 0);

#if FLASK
	bmod_rroot = oskit_bmod_init();
	rc = kserver_oskit_filepsid_init(bmod_rroot);
	assert(rc == 0);
	rc = oskit_sfs_wrap(bmod_rroot,
			    SECINITSID_KSFS,
			    kserver_oskit_filepsid,
			    avc,
			    &bmod_root);
	assert(rc == 0);
	rc = oskit_create_call_context(__mysid, &unki);
	assert(rc == 0);
	rc = oskit_set_call_context(unki);
	assert(rc == 0);
	unki->ops->release(unki);
#else
	bmod_root = oskit_bmod_init();
#endif
	assert(bmod_root);
	/*
	 * Instantiate bmod_root as a MOM dir object; this eats a ref.
	 * We need one of our own to give fdev below, so we add a ref.
	 */
	oskit_dir_addref(bmod_root); /* so we have one of our own left */
	exc._type = 0;
	bmod_root_ref = nest_com_instantiate_file((oskit_file_t *) bmod_root,
						  &bmod_lock_data,
						  &exc);
	assert(exc._type == exc_mom_no_exception &&
	       ! mom_ref_null(bmod_root_ref));


	/*
	 * Make the bmod root directory our own root and current dir.
	 */
	__root_dir = bmod_root_ref;
	__cwd_dir = bmod_root_ref;
	mom_ref_acquire(bmod_root_ref);

	/*
	 * Initialize the mutex used by libsac fs code.
	 */
	fluke_mutex_create(&__fs_mutex);
	fluke_debug_label_flob(&__fs_mutex, "__fs_mutex");

	/*
	 * Init the fdev server support code.
	 */
	ks_initialize_fdev();

	/*
	 * Create a /dev directory where we will link in devices.
	 */
	rc = oskit_dir_mkdir(bmod_root, "dev", 0777);
	assert(rc == 0);
	rc = oskit_dir_lookup(bmod_root, "dev", (oskit_file_t **)&devdir);
	assert(rc == 0);

	/*
	 * Create a /dev/time file whose contents will point directly
	 * at the variable `time', which is updated with the time of day.
	 */
#if FLASK
	rc = oskit_dir_lookup(bmod_rroot, "dev", (oskit_file_t **) &rdevdir);
	assert(rc == 0);
	rc = oskit_dir_create(rdevdir, "time", 1, 0644, &timefile);
	assert(rc == 0);
#else
	rc = oskit_dir_create(devdir, "time", 1, 0644, &timefile);
	assert(rc == 0);
#endif

	rc = oskit_bmod_file_set_contents(timefile, &time, sizeof time,
					  sizeof time, 0, 1);
	assert(rc == 0);
	oskit_file_release(timefile);


	/*
	 * Create a /dev/kmem file
	 */
#if FLASK
	rc = oskit_dir_create(rdevdir, "kmem", 1, 0600, &kmemfile);
	assert(rc == 0);
#else
	rc = oskit_dir_create(devdir, "kmem", 1, 0777, &kmemfile);
	assert(rc == 0);
#endif

	rc = oskit_bmod_file_set_contents(kmemfile, 0, osenv_mem_phys_max(),
				osenv_mem_phys_max(), 0, 1);
	assert(rc == 0);
	oskit_file_release(kmemfile);


#ifdef CONFIG_FDEV
	/*
	 * Initialize fdev devices.
	 * Do this after the root filesystem has been initialized,
	 * so that the devices may be mounted into the filesystem.
	 */
	kserver_oskit_dev_init();
#endif
#ifndef CONFIG_FDEV_CONSOLE
	/*
	 * Initialize the non-fdev console.
	 */
	kserver_console_init();
#endif

	DMARK;

	/*
	 * Now we are all done with bmod setup.
	 * The only refs keeping the bmod dir alive are those to
	 * the MOM object help in __root_dir, __cwd_dir, and children.
	 */
	oskit_dir_release(devdir);
#ifndef CONFIG_PROFILING
	oskit_dir_release(bmod_root);
#endif
#if FLASK
	oskit_dir_release(rdevdir);
	oskit_dir_release(bmod_rroot);
#endif

#ifdef HEARTBEAT_THREAD
	/*
	 * Start the heartbeat thread.
	 */
	{
		pthread_t thread;
		pthread_attr_t attr;

		pthread_attr_init(&attr);
		attr.prio = SCHED_PRIINTR;

		pthread_create(&thread, &attr, /* (void *(*)()) */
			heartbeat_thread, NULL);
	}
#endif
#ifdef PREEMPTSTATS
	/*
	 * Now that device initialization is over, enable stat keeping
	 */
	{
		extern struct preempter disp_preempter;
		preempt_enable_stats(&disp_preempter);
	}
#ifdef CONFIG_FDEV
	{
		extern struct preempter fdev_timer_preempter;
		preempt_enable_stats(&fdev_timer_preempter);
	}
#endif
#endif

#if FLASK
	/* XXX:  why not use spawn_task_create_secure? */
	rc = private_task_create_secure(&first_child_task,
					SECINITSID_SECURITY);
	assert(rc == 0);
#else
	rc = fluke_task_create(&first_child_task);
	assert(rc == 0);
#endif
	fluke_debug_label_flob(&first_child_task, "first_child_task");

	pserver = cp_default_pserver(&first_child_task, argv[0]);

	/*
	 * Make a reference to the parent object.
	 * This will be used for mempool calls and become the task keeper ref.
	 * Since the keeper ref does not hold a connection, we are responsible
	 * for keeping this object alive until the child dies.  (In fact
	 * we never release this reference, since when the child dies,
	 * we reboot.)
	 */
	rc = mom_obj_reference(pserver, &pref);
	assert(rc == 0);

	ktask_ref = alloc_scratch_ref();
	assert(ktask_ref);
	fluke_task_reference(&kernel_task, ktask_ref);
	kth_ref = alloc_scratch_ref();
	assert(kth_ref);
	fluke_thread_reference((fluke_thread_t *)fluke_thread_self(), kth_ref);

	appfd = open(argv[0], O_RDONLY, 0);
	if (appfd < 0)
		panic("cannot open first program `%s'", argv[0]);

	/*
	 * Load the program file and set it off.
	 */
#if FLASK
	rc = spawn_secure_bootstrap(
#else
	rc = spawn(
#endif
		   argv[0], appfd, argv, envp,
		   ktask_ref,
		   kth_ref,
		   pref,
		   0,
		   FLUKE_PROT_ALL,
		   &pserver->backing_list,
		   &app_startup,
#if FLASK
		   &first_child_task,
		   SECINITSID_SECURITY,
		   mom_ref_fluke_ref(__security_ref)
#else
		   &first_child_task
#endif
		   );

	close(appfd);
#if FLASK
	rc = private_security_set(mom_ref_fluke_ref(__security_ref));
        assert(rc == 0);
#endif

	mom_ref_release(pref);
	free_scratch_ref(kth_ref);
	free_scratch_ref(ktask_ref);

	if (rc)
		panic("cannot start first program `%s': %x\n", argv[0], rc);

	DPRINTF("leaving %s\n", __FUNCTION__);

	/*
	 * This loader thread's job is done.
	 * The implicitly created kserver threads will handle requests
	 * from the child task.
	 */
}
