/*
 * Copyright (c) 1997-1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include "kserver.h"
#include <fluke/nest/server/server.h>
#include <fluke/mm/mm.h>
#include <fluke/cp/startup.h>
#include <fluke/debug.h>

#include <oskit/lmm.h>		/* lmm_avail */

#ifdef MEMSTATS
unsigned long mstats_ks_bytes, mstats_ks_maxbytes;
unsigned long mstats_ks_ubytes, mstats_ks_maxubytes;
static void kserver_premm_exit(void);
static void kserver_postmm_exit(void);
extern int sys_debug_dump_memuse(void);
#endif

/*
 * Initialize everything needed for mempools.
 */
void
kserver_mempool_init()
{
#ifdef MEMSTATS
	printf("MEMSTATS: kserver startup:\n");
	(void)sys_debug_dump_memuse();
#endif

	mminfo.mm_ident = "KMM";
	mminfo.mm_membase = FLUKE_VM_MIN;
	mminfo.mm_memsize = FLUKE_VM_MAX - FLUKE_VM_MIN;
	mminfo.mm_vmsize = mminfo.mm_memsize;
	mminfo.mm_vminuse = mminfo.mm_vmsize - lmm_avail(&malloc_lmm, 0);
	/*
	 * XXX must be eager allocation for now to avoid deadlock
	 * condition in the kpreempt kernel (didn't track it down,
	 * it was attempting to retake a mapping lock in region_restrict
	 * when disabling a mapping in the libmm/mm_seg frag-join code).
	 * This was with fdev enabled, so maybe we were violating some
	 * assumption of the fdev code concerning memory being inherently
	 * wired...
	 */
	mminfo.mm_flags = MM_EAGERALLOC;
#if FLASK
	/*
	 * XXX right now we aggressively destroy flobs since we cannot
	 * move them.  Moving flobs in Flask requires that the target memory
	 * have the same SID as the source and the MM library isn't that
	 * smart yet!
	 *
	 * XXX update.  Aggressive destruction almost always results in
	 * a panic on exit.  So for now we just don't cleanup on final exit.
	 */
	mminfo.mm_flags |= MM_NOCLEANUPONEXIT;
#else
	mminfo.mm_flags |= MM_DELAYFLOBDESTROY;
#endif
	mminfo.mm_policy = &mm_null_policy;

#ifdef MEMSTATS
	atexit(kserver_postmm_exit);
#endif
	mm_setup();
#ifdef MEMSTATS
	atexit(kserver_premm_exit);
#endif

}

#if FLASK
int
fluke_segment_task_create(fluke_task_t *task, security_id_t sid)
{
    return private_segment_task_create(task, sid);
}
#endif

#ifdef MEMSTATS
void *
#if FLASK
mm_malloc(size_t size, void *daddr, 
	  security_id_t memory_sid,
	  security_id_t *out_replacement_sid)
#else
mm_malloc(size_t size, void *daddr)
#endif
{
	void *mem;

	mem = smemalign(PAGE_SIZE, size);
	if (mem) {
		mstats_ks_ubytes += size;
		if (mstats_ks_ubytes > mstats_ks_maxubytes)
			mstats_ks_maxubytes = mstats_ks_ubytes;
	}

#if FLASK
	*out_replacement_sid = 0;
#endif
	return mem;
}

void
#if FLASK
mm_mfree(void *mem, size_t size, security_id_t sid)
#else
mm_mfree(void *mem, size_t size)
#endif
{
	sfree(mem, size);
	mstats_ks_ubytes -= size;
}

void
kserver_dump_mstats(void)
{
	printf("  Kserver internal:  %10lu bytes (%10lu bytes max)\n",
		mstats_ks_bytes-mstats_ks_ubytes,
		mstats_ks_maxbytes-mstats_ks_maxubytes);
	printf("  Kserver exported:  %10lu bytes (%10lu bytes max)\n",
		mstats_ks_ubytes, mstats_ks_maxubytes);
}

void
kserver_premm_exit(void)
{
	printf("MEMSTATS: before kernel mempool cleanup:\n");
	(void)sys_debug_dump_memuse();
}

void
kserver_postmm_exit(void)
{
	printf("MEMSTATS: after kernel mempool cleanup:\n");
	(void)sys_debug_dump_memuse();
}
#endif
