/*
 * Copyright (c) 1997, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * Contributed by the Computer Security Research division,
 * INFOSEC Research and Technology Office, NSA.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#if FLASK

#include <oskit/boolean.h>
#include <oskit/fs/filepsid.h>
#include <oskit/io/absio.h>
#include "debug.h"
#include <stddef.h>
#include <ctype.h>
#include "kserver/kserver.h"
#include "kserver/ks_hashtab.h"
#include <fluke/sac/security.h>
#include <flask/initial_sid_to_string.h>

extern mom_ref_t __security_ref;
extern oskit_bool_t ss_up;


#define INOTAB_SIZE 23

static unsigned int inohash(hashtab_key_t key)
{
    unsigned int h;

    h = (unsigned int) key;
    return h % INOTAB_SIZE;
}

static int inocmp(hashtab_key_t key1, hashtab_key_t key2)
{
    return !(key1 == key2);
}


struct filepsid
{
    oskit_filepsid_t psidi;	/* COM file persistent SID interface */
    unsigned count;		/* reference count */
    hashtab_t ino2sid; 		/* ino -> sid */
};


oskit_filepsid_t *kserver_oskit_filepsid;

static struct oskit_filepsid_ops filepsid_ops;


static struct filepsid *filepsid_create(void)
{
    struct filepsid *psid;

    psid = malloc(sizeof(struct filepsid));
    if (!psid)
	    return NULL;

    psid->psidi.ops = &filepsid_ops;
    psid->count = 1;
    psid->ino2sid = hashtab_create(inohash, inocmp, INOTAB_SIZE);

    if (!psid->ino2sid)
    {
	free(psid);
	return NULL;
    }

    return psid;
}


/*
 * Skip preceeding whitespace, 
 * record the beginning of the field and 
 * terminate the field with a null character.
 */
#define next_field(p, end, field) \
      while (p < end && isspace(*p)) \
	p++; \
      field = p; \
      while (p < end && !isspace(*p)) \
	p++; \
      if (p < end) \
          *p++ = '\0'; 

typedef struct file_context {
  ino_t ino;
  mandatory_context_t mcontext;
  mandatory_context_t acontext;
  struct file_context *next;
} file_context_t;

static struct file_context *file_contexts = NULL;

oskit_error_t kserver_oskit_filepsid_init(oskit_dir_t *rootdir)
{
  char *buf, *bufp, *name, *nameb, *namep;
  char *bufend, *pathname, *mcontext, *acontext;
  oskit_error_t rc;
  oskit_stat_t sb;
  oskit_size_t actual;
  oskit_file_t *file;
  oskit_dir_t *dir;
  oskit_absio_t *absio;
  struct file_context *fc;
  security_id_t sid;
  struct filepsid *psid;


  psid = filepsid_create();
  if (!psid)
    return OSKIT_ENOMEM;

  kserver_oskit_filepsid = &psid->psidi;

  /* load the entire file context mapping file into memory */
  rc = oskit_dir_lookup(rootdir, "initial_file_contexts", &file);
  if (rc)
    {
#ifdef DEBUG
      printf("kserver:  initial_file_contexts doesn't exist:  default labels only");
#endif
      return 0;
    }
  rc = oskit_file_stat(file, &sb);
  assert(rc == 0);
  buf = malloc(sb.size+1); /* one extra for final terminator */
  assert(buf);
  rc = oskit_file_query(file, &oskit_absio_iid, (void**)&absio);
  assert(rc == 0);
  oskit_file_release(file);
  rc = oskit_absio_read(absio, buf, 0, sb.size, &actual);
  assert(rc == 0);
  assert(actual == sb.size);
  oskit_absio_release(absio);
  bufp = buf;
  bufend = buf + sb.size;
  *bufend++ = '\n'; /* terminate with whitespace */

  /* process the entries in the mapping file */
  while (bufp < bufend)
    {
      /* extract the pathname */
      next_field(bufp,bufend,pathname);
      
      if (bufp == bufend)
	break;

      /* lookup the basename */
      dir = rootdir;
      oskit_dir_addref(dir);
      nameb = strdup(pathname);
      name = nameb;
      while (1) {
	namep = strchr(name, '/');
	if (!namep) 
	  break;
	*namep = '\0';
	
	rc = oskit_dir_lookup(dir, name, &file);
	oskit_dir_release(dir);
	if (rc)
	  break;

	do ++namep; while (*namep == '/'); /* Skip multiple slashes.  */
	name = namep;
	
	rc = oskit_file_query(file, &oskit_dir_iid, (void **) &dir);
	oskit_file_release(file); 
	if (rc)
	  break;
      }

      if (!rc)
	{
	  /* lookup the final component */
	  rc = oskit_dir_lookup(dir, name, &file);
	  oskit_dir_release(dir);

	  if (!rc)
	    {
	      /* acquire the corresponding inode number */
	      rc = oskit_file_stat(file, &sb);
	      oskit_file_release(file);
	    }
	}

      free(nameb);

      if (rc)
	{
	  /* ignore the entry if the file doesn't exist */
#ifdef DEBUG
	  printf("kserver:  %s doesn't exist, skipping...\n", pathname);
#endif
	  sb.ino = 0;
	  rc = 0;
	}

      /* extract the mandatory context or initial SID name */
      next_field(bufp,bufend,mcontext);

      /* check to see if it is an initial SID name */
      for (sid = 1; sid <= SECINITSID_NUM; sid++)
	{
	  if (!strcmp(mcontext, initial_sid_to_string[sid]))
	    {
	      if (sb.ino) {
		/* yes - we can immediately bind the inode to the SID */
#ifdef DEBUG
		printf("kserver:  labeling %s (ino 0x%x) with SID %ld\n",
		       pathname, sb.ino, sid);
#endif
		rc = hashtab_insert(psid->ino2sid,
				    (hashtab_key_t) sb.ino,
				    (hashtab_datum_t) sid);
		assert(rc == 0);
	      }
	      break;
	    }
	}

      if (sid > SECINITSID_NUM) {
	/* no - it is not an initial SID name, so it must be a mcontext */

	/* obtain the auth context */
	next_field(bufp,bufend,acontext);	  
	if (sb.ino) {
	  /*
	   * We can only bind the inode to the (mcontext,acontext) pair now.
	   * The translation of the pair to a SID and the binding of the
	   * inode to that SID must be deferred until the ss is up.
	   * See psid_get.
	   */ 
#ifdef DEBUG
	  printf("kserver:  labeling %s (ino 0x%x) with context (%s, %s)\n",
		 pathname, sb.ino, mcontext, acontext);
#endif
	  fc = malloc(sizeof(struct file_context));
	  assert(fc);
	  fc->ino = sb.ino;
	  fc->mcontext = strdup(mcontext);
	  fc->acontext = strdup(acontext);
	  fc->next = file_contexts;
	  file_contexts = fc;
	}
      }
    }

  /* free the copy of the file contents */
  free(buf);

  return 0;
}




/*
 * oskit_filepsid methods
 */

static OSKIT_COMDECL psid_query(oskit_filepsid_t *p,
				const struct oskit_guid *iid,
				void **out_ihandle)
{
    struct filepsid *psid;

    psid = (struct filepsid*)p;
    assert(psid);
    assert(psid->count);

    if (memcmp(iid, &oskit_iunknown_iid, sizeof(*iid)) == 0 ||
	memcmp(iid, &oskit_filepsid_iid, sizeof(*iid)) == 0) {
	*out_ihandle = &psid->psidi;
	++psid->count;
	return 0;
    }

    *out_ihandle = NULL;
    return OSKIT_E_NOINTERFACE;
}


static OSKIT_COMDECL_U psid_addref(oskit_filepsid_t *p)
{
    struct filepsid *psid;

    psid = (struct filepsid*)p;
    assert(psid);
    assert(psid->count);

    return ++psid->count;
}


static OSKIT_COMDECL_U psid_release(oskit_filepsid_t *p)
{
    struct filepsid *psid;
    unsigned newcount;

    psid = (struct filepsid*)p;
    assert(psid);
    assert(psid->count);

    newcount = --psid->count;
    if (newcount == 0)
    {
	hashtab_destroy(psid->ino2sid);
	free(psid);
    }

    return newcount;
}


static oskit_bool_t file_context_translated = FALSE;

static OSKIT_COMDECL psid_get(oskit_filepsid_t *p,
			      oskit_file_t *file,
			      security_id_t *outsid)
{
    struct filepsid *psid;
    oskit_stat_t stats;
    oskit_error_t rc;
    struct file_context *fc, *temp;
    security_id_t sid;
    mom_exc_t ev;


    psid = (struct filepsid*)p;
    assert(psid);
    assert(psid->count);

    if (ss_up && !file_context_translated)
      {
	file_context_translated = TRUE;
	fc = file_contexts;
	while (fc) {
	    sid = cp_security_context_to_sid(__security_ref, 
					     fc->mcontext, 
					     fc->acontext,
					     "",
					     &ev);
	    if (ev._type)
	      {
		printf("kserver:  unable to translate (%s, %s)\n", 
		       fc->mcontext, fc->acontext);
	      }
	    else
	      {
#ifdef DEBUG
		printf("kserver:  obtained SID %ld for context (%s, %s) [ino 0x%x]\n",
		       sid, fc->mcontext, fc->acontext, fc->ino);
#endif
		if (hashtab_insert(psid->ino2sid,
				    (hashtab_key_t) fc->ino,
				    (hashtab_datum_t) sid))
		  printf("kserver:  unable to store ino 0x%x -> sid %ld\n", 
			 fc->ino, sid);
	      }
	    free(fc->mcontext);
	    free(fc->acontext);
	    temp = fc;
	    fc = fc->next;
	    free(temp);
	  }
      }

    rc = oskit_file_stat(file, &stats);
    if (rc)
	    return rc;

    if (!ss_up)
      {
	fc = file_contexts;
	while (fc) {
	    if (fc->ino == stats.ino)
	      {
		printf("kserver:  attempt to access ino 0x%x with security context (%s, %s) prior to translation\n", fc->ino, fc->mcontext, fc->acontext);
		return OSKIT_EACCES;
	      }
	    fc = fc->next;
	}
      }

    sid = (security_id_t) hashtab_search(psid->ino2sid,
					 (hashtab_key_t) stats.ino);
    if (sid)
      {
	*outsid = sid;
#if VERBOSE
	if (sid <= SECINITSID_NUM)
	  printf("psid_get:  ino 0x%x -> sid %s\n", stats.ino, initial_sid_to_string[sid]);
	else
	  printf("psid_get:  ino 0x%x -> sid %ld\n", stats.ino, sid);
#endif
	return 0;
      }

#if VERBOSE
    printf("psid_get:  ino 0x%x -> sid ksfile\n", stats.ino);
#endif
    *outsid = SECINITSID_KSFILE;
    return 0;
}


static OSKIT_COMDECL psid_set(oskit_filepsid_t *p,
			     oskit_file_t *file,
			     security_id_t newsid)
{
    struct filepsid *psid;
    oskit_stat_t stats;
    oskit_error_t rc;

    psid = (struct filepsid*)p;
    assert(psid);
    assert(psid->count);

    rc = oskit_file_stat(file, &stats);
    if (rc)
	    return rc;

    return hashtab_replace(psid->ino2sid,
			   (hashtab_key_t) stats.ino,
			   (hashtab_datum_t) newsid,
			   NULL, NULL);
}


static OSKIT_COMDECL psid_unset(oskit_filepsid_t *p,
				oskit_ino_t ino)
{
    struct filepsid *psid;

    psid = (struct filepsid*)p;
    assert(psid);
    assert(psid->count);

    return hashtab_remove(psid->ino2sid,
			  (hashtab_key_t) ino,
			  NULL,
			  NULL);
}


static OSKIT_COMDECL psid_create(oskit_filepsid_t *p,
				security_id_t csid,
				security_id_t dsid,
				security_class_t fclass,
				security_id_t *outsid)
{
    struct filepsid *psid;
    security_id_t sid;
    mom_exc_t ev;

    psid = (struct filepsid*)p;
    assert(psid);
    assert(psid->count);

    if (ss_up)
      {
	sid = cp_security_transition_sid(__security_ref, csid, dsid, fclass, &ev);
	if (ev._type)
	  {
#ifdef DEBUG
	    printf("kserver:  transition_sid(%ld,%ld,%d) returned 0x%x\n",
		   csid, dsid, fclass, ev._type);
#endif
	    return OSKIT_EACCES;
	  }
	*outsid = sid;
      }
    else
      {
	*outsid = dsid;
      }

    return 0;
}


static OSKIT_COMDECL psid_chown(oskit_filepsid_t *p,
			       security_id_t fsid,
			       oskit_u32_t mask,
			       const oskit_stat_t *stats,
			       security_id_t *outsid)
{
    struct filepsid *psid;

    psid = (struct filepsid*)p;
    assert(psid);
    assert(psid->count);

    /* kserver:  no Unix identity attributes; chown has no effect */
    *outsid = fsid;
    return 0;
}


static struct oskit_filepsid_ops filepsid_ops = {
    psid_query,
    psid_addref,
    psid_release,
    psid_get,
    psid_set,
    psid_unset,
    psid_create,
    psid_chown,
};

#endif
