/*
 * Copyright (c) 1995 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#ifndef _FLUKE_KERNEL_LINK_H_
#define _FLUKE_KERNEL_LINK_H_

#include <oskit/machine/types.h>
#include <oskit/compiler.h>
#include <fluke/ob.h>

typedef struct link
{
	/*
	 * Pointer to the actual object this link points to,
	 * or NULL if this link is invalid.
	 */
	struct s_ob		*target;

} link_t;

#define LINK_NULL	((struct link*)0)


/*
 * Initialize a link to null (pointing to no object).
 */
OSKIT_INLINE void link_init(
	struct link		*link)
{
	link->target = 0;
}

#include "config_inlining.h"

#ifndef CONFIG_INLINING
/*
 * Destroy a link and reinitialize it to null (pointing to no object).
 */
void link_reinit(
	struct link		*link);

/*
 * Initialize a link to point to an object.
 * The target object must be known to exist, but it need not be locked.
 */
void link_produce(
	struct link		*link,
	struct s_ob		*target);

/*
 * Create a copy of a link (either an object link or a number link).
 * The memory pointed to by 'dest' is assumed to be uninitialized.
 */
void link_create_copy(
	struct link		*src,
	struct link		*dest);

/*
 * Destroy a link, leaving its contents undefined.
 */
void link_destroy(
	struct link		*link);

/*
 * Given two already-initialized links,
 * change the 'dest' link to point to whatever 'src' points to.
 */
void link_transfer(
	struct link		*src,
	struct link		*dest);

/*
 * Try to follow a link and lock the object it points to.
 * If successful, returns a pointer to the now-locked object.
 * If the link is stale (or becomes stale before the object can be locked),
 * this function returns null.
 */
struct s_ob *link_follow(
	struct link		*link);

/*
 * Try to follow a link and obtain the hash of the referenced object.
 * If successful, returns the hash.
 * If the link is stale this function returns 0.
 * The object is not locked at any point during this operation.
 */
fluke_hash_t link_hash(
	struct link		*link);

/*
 * Find the target of a link and consume the link itself.
 * If successful, returns a pointer to the now-locked object.
 * If the link was stale, returns null.
 * In either case, the contents of the link is left uninitialized.
 */
struct s_ob *link_consume(
	struct link		*link);

/*
 * Compare two links to see if they point to the same thing.
 * If one of the links is NULL (or stale) but the other is invalid,
 * this routine always returns false, as you would expect.
 * However, if both links are NULL (or stale),
 * then this routine may return either true or false.
 * e.g. depending on whether they _used_ to point to the same object.
 */
int link_compare(
	struct link		*link1,
	struct link		*link2);

/*
 * Like link_compare, but compare a link against a pointer to an object.
 */
int link_compare_ptr(
	struct link		*link,
	struct s_ob		*ptr);

#endif CONFIG_INLINING

#endif /* _FLUKE_KERNEL_LINK_H_ */
