/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Simple kernel mutex implementation
 * for use higher-level kernel code.
 */
#ifndef _FLUKE_KERNEL_MUTEX_H_
#define _FLUKE_KERNEL_MUTEX_H_

#include "dispatch.h"
#include "config_kpreempt.h"
#include "config_smp.h"

/*
 * Three configurations
 *
 * Preemptible kernel
 *	Require full blocking mutex support.
 *	
 * Non-preemptible kernel, multiprocessor
 *	Blocking mutexes reduce to spinlocks.
 *
 * Non-preemptible kernel, uniprocessor
 *	Requires no mutexes.
 */

#ifdef CONFIG_KPREEMPT

struct mutex {
	struct wait_queue	queue;
	struct s_thread		*owner;
};
typedef struct mutex mutex_t;

void mutex_init(
	struct mutex		*m);
void mutex_destroy(
	struct mutex		*m);
void mutex_lock(
	struct mutex		*m);
int mutex_lock_cancelable(
	struct mutex		*m);
int mutex_trylock(
	struct mutex		*m);
void mutex_unlock(
	struct mutex		*m);

/*
 * Same as the normal lock/unlock routines above,
 * but these expect preemption to be disabled on entry
 * and leave preemption disabled on exit.
 * Also, they expect to receive the current thread as a parameter.
 */
void pd_mutex_lock(
	struct s_thread		*cur_thread,
	struct mutex		*m);
void pd_mutex_unlock(
	struct s_thread		*cur_thread,
	struct mutex		*m);

/*
 * Atomically release the mutex and enter a permanent uncancelable wait.
 * Called by a thread when it's being destroyed by another thread.
 * (The other thread is presumably waiting to grab this mutex,
 * and will destroy us as soon as it successfully does so.)
 */
void mutex_unlock_stop(
	struct mutex		*m);

#define assert_mutex_locked(m)		assert((m)->owner != 0)
#define assert_mutex_unlocked(m)	assert((m)->owner == 0)

#else	/* !CONFIG_KPREEMPT */

#ifdef CONFIG_SMP

struct mutex {
	spin_lock_t		lock;
};
typedef struct mutex mutex_t;

#define mutex_init(m)		spin_lock_init(&(m)->lock)
#define mutex_destroy(m)	/* do nothing */
#define mutex_lock(m)		spin_lock(&(m)->lock)
#define mutex_trylock(m)        spin_try_lock(&(m)->lock)
#define mutex_unlock(m)		spin_unlock(&(m)->lock)
#define mutex_unlock_ints_disabled(th, m) spin_unlock(&(m)->lock)
#define mutex_unlock_stop(m)	thread_disp_stop(current_thread(), &(m)->lock)
#define pd_mutex_lock(th,m)	spin_lock(&(m)->lock)
#define pd_mutex_unlock(th,m)	spin_unlock(&(m)->lock)

#define assert_mutex_locked(m)		assert_spin_lock_locked(&(m)->lock)
#define assert_mutex_unlocked(m)	assert_spin_lock_unlocked(&(m)->lock)

#else	/* !CONFIG_SMP */

struct mutex {
	/* nothing */
};
typedef struct mutex mutex_t;

#define mutex_init(m)		/* do nothing */
#define mutex_destroy(m)	/* do nothing */
#define mutex_lock(m)		/* do nothing */
#define mutex_trylock(m)        1
#define mutex_unlock(m)		/* do nothing */
#define mutex_unlock_ints_disabled(th, m) 	/* do nothing */
#define mutex_unlock_stop(m)	thread_disp_stop(current_thread(), ???)
#define pd_mutex_lock(th,m)	/* do nothing */
#define pd_mutex_unlock(th,m)	/* do nothing */

#define assert_mutex_locked(m)		/* do nothing */
#define assert_mutex_unlocked(m)	/* do nothing */

#endif	/* CONFIG_SMP */
#endif	/* CONFIG_KPREEMPT */

#endif /* _FLUKE_KERNEL_MUTEX_H_ */
