/*
 * Copyright (c) 1997, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include <stddef.h>

#include "preempt.h"
#include "per_cpu.h"		/* preempters list is in my_cpu */
#include "intr.h" /* XXX layering */
#include "debug.h"

#ifdef CONFIG_KPREEMPT

/*
 * A number of routines with preemption assertions are called before
 * the scheduler is initiailized (and hence, preemption enabled).
 * For example, kalloc indirectly via malloc from the multiboot startup.
 * This is a hack to avoid those assertion failures.
 */
#ifndef NDEBUG
int preempt_initialized;
#endif

void preempt_boot(void)
{
#ifndef NDEBUG
	preempt_initialized = 1;
#endif

	preempt_enable_flag = 1; /* this processor only */
}

#ifdef DEBUG
#include "s_thread.h"
void checkpreempt(void)
{
	if (!preempt_enable_flag) {
		exc_state_dump(&current_thread()->exc_state);
		dump_stack_trace();
		panic("trap: preemption off on return to user mode");
	}
}
#endif
#endif

void
id_register_preempter(struct preempter *p, void (*func)(struct preempter *p))
{
	assert_interrupts_disabled();
	assert_preemption_disabled();

	if (p->func == NULL) {
		p->func = func;
		p->next = my_cpu.preempters;
		my_cpu.preempters = p;
#ifdef PREEMPTSTATS
		if (p->enabled) {
			p->delayed++;
			asm volatile("rdtsc"
				     : "=a" (p->dstart) :: "eax", "edx");
		}
#endif
	}
#ifdef PREEMPTSTATS
	else
		if (p->enabled) p->queued++;
#endif
}

void
id_unregister_preempter(struct preempter *p)
{
	assert_interrupts_disabled();
	assert_preemption_disabled();

	if (p->func != NULL) {
		struct preempter **pp;
		assert(my_cpu.preempters != NULL);
		for (pp = &my_cpu.preempters; *pp != p; pp = &(*pp)->next)
			assert(*pp != NULL);
		*pp = p->next;
		p->func = NULL;
	}
}

void
register_preempter(struct preempter *p, void (*func)(struct preempter *p))
{
	assert_interrupts_enabled();
	assert_preemption_disabled();

	disable_interrupts();
	id_register_preempter(p, func);
	enable_interrupts();
}

void
intr_preempter(struct preempter *p, void (*func)(struct preempter *p))
{
	assert_interrupts_disabled();

	if (preempt_enable_flag) {

		/* If preemption is enabled now, run the handler now. */
		disable_preemption();
#ifdef PREEMPTSTATS
		if (p->enabled) p->immediate++;
#endif
		enable_interrupts();
		func(p);
		disable_interrupts();
		enable_preemption();

	} else {

		/* Otherwise register it to be run later. */
		id_register_preempter(p, func);
	}

	assert_interrupts_disabled();
}

void
unregister_preempter(struct preempter *p)
{
	assert_interrupts_enabled();

	disable_interrupts();
	id_unregister_preempter(p);
	enable_interrupts();
}

#ifdef PREEMPTSTATS
int preemptfromuser;

void
u_run_preempters(void)
{
	assert_interrupts_enabled();
	assert_preemption_enabled();

	disable_preemption();
	disable_interrupts();
	while (my_cpu.preempters != NULL) {
		extern void disp_yield(struct preempter *);
		struct preempter *p = my_cpu.preempters;
		void (*func)(struct preempter *p) = p->func;

		my_cpu.preempters = p->next;
		p->func = NULL;
		if (func == disp_yield)
			preemptfromuser = 1;
		{
			unsigned int tmp;

			asm volatile("rdtsc" : "=a" (tmp) :: "eax", "edx");
			p->dcycles += (tmp - p->dstart);
		}
		enable_interrupts();
		func(p);
		disable_interrupts();
		preemptfromuser = 0;

		assert_preemption_disabled();
	}
	enable_interrupts();
	enable_preemption();
}
#endif

void
run_preempters(void)
{
	int iflag;

	assert_preemption_enabled();

	/*
	 * Note that interrupts are not necessarily enabled.
	 * In intr_preempter, if the called function schedules
	 * another preempter, run_preemper will be called as a result of
	 * reenabling preemption but interrupts will be off.
	 */
	save_interrupt_enable(iflag);

	disable_preemption();
	disable_interrupts();
	while (my_cpu.preempters != NULL) {
		struct preempter *p = my_cpu.preempters;
		void (*func)(struct preempter *p) = p->func;

		my_cpu.preempters = p->next;
		p->func = NULL;

#ifdef PREEMPTSTATS
		if (p->enabled) {
			unsigned int tmp;

			asm volatile("rdtsc" : "=a" (tmp) :: "eax", "edx");
			p->dcycles += (tmp - p->dstart);
		}
#endif
		enable_interrupts();
		func(p);
		disable_interrupts();

		assert_preemption_disabled();
	}
	restore_interrupt_enable(iflag);
	enable_preemption();
}

#if !defined(CONFIG_NOPREEMPT) && !defined(CONFIG_KPREEMPT)
int
preempt_pending(void)
{
	return (my_cpu.preempters != NULL);
}
#endif
