/*
 * Copyright (c) 1997, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#ifndef _FLUKE_KERNEL_PREEMPT_H_
#define _FLUKE_KERNEL_PREEMPT_H_

#include "config_nopreempt.h"
#include "config_kpreempt.h"
#include "config_nonblocking.h"

/*
 * This structure describes the simple linked list of
 * registered preempters to be called.  The head of the
 * chain is stored in the per_cpu data structure (my_cpu.preempters).
 */
struct preempter {
	void (*func)(struct preempter *p);
	struct preempter *next;
#ifdef PREEMPTSTATS
	unsigned int enabled;
	unsigned int immediate;
	unsigned int queued;
	unsigned int delayed;
	unsigned int dcycles;
	unsigned int dstart;
#endif
};


/*
 * Register a preemption function to be called later
 * when it is safe to wake up and switch to other threads and such.
 * It's safe to re-register a preempter that's already registered.
 * This is one of the very few functions
 * that short interrupt handlers can safely call.
 * Assumes preemption is disabled;
 * id_register_preempter also assumes interrupts are disabled.
 * XXX can this be called when preemption is enabled?
 */
void id_register_preempter(struct preempter *p,
			   void (*func)(struct preempter *p));
void register_preempter(struct preempter *p,
			void (*func)(struct preempter *p));

/*
 * Activate a preempter from an interrupt handler.
 * Interrupts are disabled but preemption may be enabled or disabled.
 * If preemption is enabled, runs the preempter function immediately;
 * otherwise, registers it to be called when preemption is re-enabled.
 */
void intr_preempter(struct preempter *p, void (*func)(struct preempter *p));

/*
 * Unregister a preempter if it's currently registered.
 * Interrupts disabled throughout id_unregister_preempter.
 */
void id_unregister_preempter(struct preempter *p);
void unregister_preempter(struct preempter *p);

/*
 * Run all the currently registered preemption handlers,
 * unregistering them in the process.
 * Interrupts and preemption enabled on entry and exit,
 * but preempter functions are called with preemption disabled.
 */
void run_preempters(void);

/*
 * Returns non-zero if there is a preemption event scheduled for this CPU.
 * Note: there is no need to explicitly check for this in the fully-preemptible
 * kernel and we don't want to check for it in the non-preemptible case.
 */
#if defined(CONFIG_NOPREEMPT) || defined(CONFIG_KPREEMPT)
#define preempt_pending()	(0)
#else
int preempt_pending(void);
#endif

/*
 * Initialize the kernel preemption layer, must be called before
 * anything else here.
 */
void preempt_boot(void);

#ifdef PREEMPTSTATS
#define preempt_enable_stats(p)	(p)->enabled = 1
#endif

/*** Kernel Preemption Support ***/

/*
 * Full kernel preemption is optional, and only works in the blocking model,
 * because the preempted thread's stack has to be retained
 * in order for it to be restarted transparently later.
 */
#ifdef CONFIG_KPREEMPT
#ifdef CONFIG_NONBLOCKING
#error kernel preemption is incompatible with the nonblocking model!
#endif
#ifdef CONFIG_NOPREEMPT
#error kernel preemption is incompatible with no enable_nopreempt option!
#endif

extern int preempt_initialized;

#include "per_cpu.h"
#define preempt_enable_flag	my_cpu.preemption_enabled

/*
 * When full kernel preemptibility is enabled (blocking model only),
 * these macros are used on entry into and exit from preemption-critical code,
 * such as the scheduling and thread waiting/blocking code itself.
 */

#define disable_preemption()						\
	(preempt_enable_flag = 0)
#define enable_preemption()						\
	(preempt_enable_flag = 1,					\
	 my_cpu.preempters ? run_preempters() : (void)0)
#define save_preemption_enable(flag)					\
	((flag) = preempt_enable_flag)
#define restore_preemption_enable(flag)					\
	((flag) ? enable_preemption() : (void)0)

#define assert_preemption_disabled() \
	assert(!preempt_initialized || !preempt_enable_flag)
#define assert_preemption_enabled() \
	assert(!preempt_initialized || preempt_enable_flag)

#else /* not CONFIG_KPREEMPT */

/*
 * When full kernel preemptibility is disabled,
 * the whole kernel is non-preemptible by default anyway,
 * so there's no distinguished preemption-sensitive to protect.
 */
#define preempt_enable_flag		(0)	/* always disabled */
#define disable_preemption()		((void)0)
#define enable_preemption()		((void)0)
#define save_preemption_enable(flag)	((flag) = 0) /* silence warnings */
#define restore_preemption_enable(flag)	((void)0)
#define assert_preemption_disabled()	((void)0)
#define assert_preemption_enabled()	((void)0)

#endif /* not CONFIG_KPREEMPT */


#endif /* _FLUKE_KERNEL_PREEMPT_H_ */
