/*
 * Copyright (c) 1996 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * Contributed by the Computer Security Research division,
 * INFOSEC Research and Technology Office, NSA.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#ifndef _FLUKE_PRIVATE_INTERFACES_H_
#define _FLUKE_PRIVATE_INTERFACES_H_

/* FLASK */

#include <flask/security.h>
#include <fluke/task.h>
#include <fluke/thread.h>
#include <fluke/region.h>
#include <fluke/mapping.h>
#include <fluke/port.h>
#include <fluke/ref.h>
#include <fluke/ipc.h>

/*
 * These interfaces are used by the Flask version of the
 * kserver and are implemented via do_in_supervisor_mode.
 */


struct security_set_args 
{
    oskit_addr_t ref_uva;
};

extern int do_private_security_set(void *args);
extern int private_security_set(fluke_ref_t *ref);


/*
 * The kserver uses the private_task_create_secure
 * interface to create the task for its child with
 * the security SID.  If the child task is created
 * in the child's memory at some point, then this
 * will go away.
 */

struct task_create_secure_args 
{
    oskit_addr_t uva;
    security_id_t sid;
};

extern int do_private_task_create_secure(void *args);
extern int private_task_create_secure(fluke_task_t *task,
				      security_id_t sid);


/*
 * The kserver uses the private_segment_task_create
 * interface to supply the kernel with the SID
 * of the segment which is associated with the newly
 * created segment task.
 */

struct segment_task_create_args 
{
    oskit_addr_t uva;
    security_id_t sid;
};

extern int do_private_segment_task_create(void *args);
extern int private_segment_task_create(fluke_task_t *task,
				       security_id_t sid);


/*
 * The kserver uses the private_avc_reset, private_avc_remove,
 * private_avc_revoke and private_avc_add interfaces to service 
 * Avc requests from the security policy server.  
 * See fluke/cp/Avc.idl.
 */

extern int do_private_avc_stop_all_threads(void *args);
extern int do_private_avc_release_all_threads(void *args);

struct avc_reset_args
{
  oskit_u32_t seqno;
  int *exc_code;
};

extern int do_private_avc_reset(void *args);
extern int do_private_avc_reset_migrated(void *args);
extern int private_avc_reset(oskit_u32_t seqno);


struct avc_remove_args
{
    security_class_t sclass;
    security_id_t ssid;
    security_class_t tclass;
    security_id_t tsid;
    oskit_bool_t wired;
    int	*exc_code;	
};

extern int do_private_avc_remove(void *args);
extern int private_avc_remove(security_class_t sclass,
			      security_id_t ssid,
			      security_class_t tclass,
			      security_id_t tsid,
			      oskit_bool_t wired);


struct avc_try_remove_args
{
    security_class_t sclass;
    security_id_t ssid;
    security_class_t tclass;
    security_id_t tsid;
    access_vector_t permissions;
    oskit_bool_t wired;
    oskit_u32_t seqno;
    access_vector_t *retained;
    int *exc_code;
};

extern int do_private_avc_try_remove(void *args);
extern int private_avc_try_remove(security_class_t sclass,
                                  security_id_t ssid,
                                  security_class_t tclass,
                                  security_id_t tsid,
                                  access_vector_t permissions,
                                  oskit_bool_t removewired,
                                  oskit_u32_t seqno,
                                  access_vector_t *out_retained);


struct avc_revoke_args
{
    security_class_t sclass;
    security_id_t ssid;
    security_class_t tclass;
    security_id_t tsid;
    access_vector_t revoke;
    oskit_bool_t wired;
    oskit_u32_t seqno;
    int	*exc_code;	
};

extern int do_private_avc_revoke(void *args);
extern int do_private_avc_revoke_migrated(void *args);
extern int private_avc_revoke(security_class_t sclass,
			      security_id_t ssid,
			      security_class_t tclass,
			      security_id_t tsid,
			      access_vector_t revoke,
			      oskit_bool_t wired,
			      oskit_u32_t seqno);


struct avc_add_args
{
    security_class_t sclass;
    security_id_t ssid;
    security_class_t tclass;
    security_id_t tsid;
    access_vector_t allowed;
    access_vector_t decided;
    unsigned int expiration;
    oskit_bool_t wired;
    oskit_bool_t replace;
    oskit_u32_t seqno;
    int	      *exc_code;		
};

extern int do_private_avc_add(void *args);
extern int private_avc_add(security_class_t sclass,
			   security_id_t ssid,
			   security_class_t tclass,
			   security_id_t tsid,
			   access_vector_t allowed,
			   access_vector_t decided,
			   unsigned int expiration,
			   oskit_bool_t wired,	
			   oskit_bool_t replace,
			   oskit_u32_t seqno);


/*
 * The kserver uses the private_avc_has_permission interface to
 * perform permission checks on operations on kserver objects
 * (eg segment, mempool, file, filesystem, avc).  This may not be
 * optimal; perhaps it would be more efficient for the kserver
 * to maintain its own access vector cache and directly perform
 * idempotent calls to the security policy server on cache misses.
 */
struct avc_has_permission_args
{
    security_class_t sclass;
    security_id_t   ssid;
    security_class_t tclass;
    security_id_t   tsid;
    access_vector_t requested;
    int		  *exc_code;
};

extern int do_private_avc_has_permission(void *args);
extern int private_avc_has_permission(security_class_t sclass,
				      security_id_t ssid,
				      security_class_t tclass,
				      security_id_t tsid,
				      access_vector_t requested);

/* FLASK */

#endif _FLUKE_PRIVATE_INTERFACES_H_
