/*
 * Copyright (c) 1996 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Routines to initialize and control kernel profiling.
 */

#include "config_profiling.h"
#ifdef CONFIG_PROFILING

#include <stdio.h>
#include <string.h>

#include "s_thread.h"
#include "kalloc.h"
#include "kserver.h"

#include "debug.h"

#include <fluke/machine/syscall.h>
#include <fluke/profile.h>
#include <sys/types.h>
#include <sys/gmon.h>
#include <fcntl.h>

#include <oskit/io/bufio.h>
#include <oskit/fs/openfile.h>

struct gmonparam _gmonparam = { GMON_PROF_OFF };

extern char _start[];
extern char etext[];

void
profiling_boot(void) 
{
	char *cp;
	struct gmonparam *p = &_gmonparam;

	p->lowpc = ROUNDDOWN((unsigned long)_start, 
			     HISTFRACTION * sizeof(HISTCOUNTER));
	p->highpc = ROUNDUP((unsigned long)etext, 
			    HISTFRACTION * sizeof(HISTCOUNTER));
	p->textsize = p->highpc - p->lowpc;

	printf("Profiling kernel, textsize=%ld [%lx..%lx]\n",
	       p->textsize, p->lowpc, p->highpc);

	p->kcountsize = p->textsize / HISTFRACTION;
	p->hashfraction = HASHFRACTION;
	p->fromssize = p->textsize / HASHFRACTION;
	p->tolimit = p->textsize * ARCDENSITY / 100;

	if (p->tolimit < MINARCS)
		p->tolimit = MINARCS;
	else if (p->tolimit > MAXARCS)
		p->tolimit = MAXARCS;
	p->tossize = p->tolimit * sizeof(struct tostruct);

#if 0
	printf("allocating buffer: kcountsize = %d, fromssize = %d, "
		"tolimit = %d\n", p->kcountsize, p->fromssize, p->tolimit);
#endif

	cp = (char *)kalloc(p->kcountsize + p->fromssize + p->tossize);
	if (cp == 0) {
		printf("prof_init: couldn't allocate profiling buffer!\n");
		return;
	}

	memset(cp, 0, p->kcountsize + p->fromssize + p->tossize);

	p->tos = (struct tostruct *)cp;
	cp += p->tossize;
	p->kcount = (HISTCOUNTER *)cp;
	cp += p->kcountsize;
	p->froms = (unsigned short *)cp;

#if 0
	printf("p->froms = %lx\n", (u_long)p->froms);
	printf("p->fromssize = %lx\n", (u_long)p->fromssize);
	printf("p->tos = %lx\n", (u_long)p->tos);
	printf("p->kcount = %lx\n", (u_long)p->kcount);
#endif

}


int
sys_doprof(void)
{
	struct s_thread *cur_th = current_thread();
	int foo;
	register struct gmonparam *p;

	p = &_gmonparam;

	foo = EXC_SYSPARAM(cur_th, doprof, foo);

	switch (foo) {
	case FLUKE_PROF_ON:
		p->state = GMON_PROF_ON; 
		printf("Turning on kernel profiling; enabling irq 8\n");
		prof_clock_init(OSKIT_PROFHZ);
		break;
	case FLUKE_PROF_OFF:
		p->state = GMON_PROF_OFF;
		/* FALLTHRU */
	case FLUKE_PROF_DUMP:
	{
		/*
		 * Stop profiling and write data to gmon.out.
		 */
		struct gmonhdr *h;
		struct rawarc *ra;
		char *filename = "gmon.out";
		int narcs, i;
		oskit_file_t *bmod;
		oskit_bufio_t *bmod_io;
		oskit_openfile_t *open_bmod;
		oskit_size_t out_actual;
		oskit_off_t offset;
		int rc;
		extern oskit_dir_t *bmod_root;

		/* XXX: only if no user or kernel profiling anymore. */
		pic_disable_irq(8);
		printf("FLUKE_PROF_DUMP\n");
		/*
		 * XXX: what about bmod_lock mutex?
		 */

		narcs = p->fromssize / sizeof(*p->froms);

		/* Create the gmon.out file; fail if it exists */
		rc = oskit_dir_create(bmod_root, filename, 1, 0622, &bmod);
		if (rc) {
			printf("File %s exists\n", filename);
			return 0;
		}

		rc = oskit_file_open(bmod, OSKIT_O_WRONLY|OSKIT_O_CREAT,
			&open_bmod);
		if (rc) {
			printf("open %s failed (rc = %p).\n", filename, rc);
			return 0;
		}
		/*
		 * The bmod returned is going to be NULL...
		 * no per-open state.
		 */
		assert(open_bmod == 0);

		/* query for a bufio object */
		rc = oskit_bufio_query(bmod, &oskit_bufio_iid, &bmod_io);
		assert(bmod_io);

		/*
		 * Allocate the space we'll need, rather than having
		 * the bmod continually realloc-ing the file.
		 */
		oskit_bufio_write(bmod_io, 0 /* buf */,
			(oskit_off_t)(sizeof(struct gmonhdr) + p->kcountsize +
				(narcs * sizeof(struct rawarc))),
			0, &out_actual);

		h = kalloc(sizeof(*h));

		/*
		 * The gmonhdr comes first in the profiling data file.
		 */
		memset(h, 0, sizeof(*h));
		h->lpc = p->lowpc;
		h->hpc = p->highpc;
		h->ncnt = p->kcountsize + sizeof(struct gmonhdr);
		h->version = GMONVERSION;
		h->profrate = OSKIT_PROFHZ; /* XXX clock hard-coded! */

		offset = 0;
		oskit_bufio_write(bmod_io, h, offset, sizeof(*h), &out_actual);
		offset+=sizeof(*h);

		/*
		 * Now we copy in all the tick buffer info.
		 */
		rc = oskit_bufio_write(bmod_io, p->kcount, offset,
			p->kcountsize, &out_actual);
		offset+=p->kcountsize;

#if 0
		for (i = 0; i < p->kcountsize/sizeof(int); i++) {
			if (p->kcount[i] != 0) {
				printf("NONZERO SAMPLE! %d: %d\n", 
				       i, p->kcount[i]);
			}
		}
#endif
		
		/*
		 * We write all the rawarc structures.  These are 
		 * massaged from the tos and froms.
		 */
		ra = kalloc(sizeof(*ra));

		for (i = 0; i < narcs; i++) {
			unsigned long frompc;
			int toindex;

			if (p->froms[i] == 0)
				continue;

			frompc = (unsigned long)p->lowpc +
				 (i * p->hashfraction * sizeof(*p->froms));

			for (toindex = p->froms[i]; toindex != 0;
			     toindex = p->tos[toindex].link) {
				ra->raw_frompc = frompc;
				ra->raw_selfpc = 
				    (unsigned long)p->tos[toindex].selfpc;
				ra->raw_count = p->tos[toindex].count;
#if 0
				printf("%d: frompc = %lx, selfpc = %lx, count = %lx\n", 
				       i, ra->raw_frompc, ra->raw_selfpc, ra->raw_count);
#endif

				rc = oskit_bufio_write(bmod_io, ra, offset,
					sizeof(*ra), &out_actual);
				offset+=sizeof(*ra);
			}
		}
		kfree(h, sizeof(*h));
		kfree(ra, sizeof(*ra));
		/* close(bmod); */
		printf("END FLUKE_PROF_DUMP\n");

	}
	}

	return 0;
}
#endif /* CONFIG_PROFILING */
