/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include <fluke/cond.h>
#include <fluke/exception.h>

#include "link.h"
#include "s_cond.h"
#include "s_mutex.h"
#include "s_task.h"
#include "s_thread.h"
#include "exception.h"
#include "debug.h"

int s_cond_create(
	oskit_addr_t		uva,
	fluke_hash_t		hash,
	s_cond_t		**out_cond)
{
	s_cond_t *cond;
	int rc;

	rc = task_ob_create(current_thread(), FLUKE_COND, uva, hash,
			    (s_ob_t **)&cond);
	if (rc)
		return (rc);

	cond_init(&cond->cond);
	cond->flags = 0;	/* XXX flags unused */
	link_init(&cond->inheritor_link);

	*out_cond = cond;
	return 0;
}

int s_cond_destroy(
	s_cond_t		*cond)
{
	assert(cond && cond->sob.active);
	assert(cond->sob.type == FLUKE_COND);
	assert_s_cond_locked(cond);

	/*
	 * XXX as with mutexes, user code should ensure that no one is
	 * waiting on a condition before it is destroyed.  However,
	 * generating an insanity trap is not appropriate for a memory
	 * manager attempting to clean memory.  So, we just wake everybody
	 * up and let them encounter an invalid condition object.
	 */
	cond_broadcast(&cond->cond);

	cond_destroy(&cond->cond);

	s_ob_destroy(&cond->sob);
	return 0;
}

int s_cond_get_state(
	s_cond_t		*cond,
	struct fluke_cond_state	*state,
	struct link		*inheritor_link)
{
	assert_s_ob_locked(&cond->sob);
	assert(cond->sob.type == FLUKE_COND);

	if (state) {
		state->hash = s_ob_gethash(&cond->sob);
		state->flags = cond->flags;
	}

	if (inheritor_link)
		link_create_copy(&cond->inheritor_link, inheritor_link);

	return 0;
}

int s_cond_set_state(
	s_cond_t		*cond,
	struct fluke_cond_state	*state,
	struct link		*inheritor_link)
{
	assert_s_ob_locked(&cond->sob);
	assert(cond->sob.type == FLUKE_COND);

	if (state)
		cond->flags = state->flags;

	if (inheritor_link)
		link_transfer(inheritor_link, &cond->inheritor_link);

	return 0;
}

int s_cond_wait(
	s_cond_t		*cond)
{
	link_t link;

	assert_s_cond_locked(cond);

#ifdef CONFIG_PRIO_INHERIT
	link_create_copy(&cond->inheritor_link, &link);
#else
	link_init(&link);
#endif

	return cond_wait(&cond->cond, &cond->sob.lock, &link);
}

int s_cond_signal(
	s_cond_t		*cond)
{
	assert_s_cond_locked(cond);

	cond_signal(&cond->cond);

	return 0;
}

int s_cond_broadcast(
	s_cond_t		*cond)
{
	assert_s_cond_locked(cond);

	cond_broadcast(&cond->cond);

	return 0;
}

#ifdef DEBUG
void
s_cond_dump(s_cond_t *cond, int indent)
{
	int i;

	for (i = 0; i < indent; i++) putchar(' ');
	printf("flags=%#x, inheritor=%p, &kcond=%p\n",
	       cond->flags, cond->inheritor_link.target, &cond->cond);
}
#endif
