/*
 * Copyright (c) 1997, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Syscall and event instrumentation functions.
 */
#include <stdio.h>
#include "debug.h"

#include "config_instrument.h"

/*
 * If CONFIG_INSTRUMENT isn't defined don't compile a thing.  Shouldn't
 * even need stubs, as everything is protected with ifdef's where its
 * used.   None of the functions should be even namedvisible
 */
#ifdef CONFIG_INSTRUMENT
#include "s_thread.h"
#include "kalloc.h"
#include "s_instrument.h"

/* Control of the instrumentation */
#undef INSTRUMENT_ALL_THREADS

#define CONFIG_INSTRUMENT_BUF_SIZE (8 * 4096)
#define INSTRUMENT_FLAG_ON  1
#define INSTRUMENT_FLAG_IPC 2

/*
 * Define the buffer elements that record the data we capture.
 */
#define BUF_TYPE_ENTRY 0
#define BUF_TYPE_EXIT  1
#define BUF_TYPE_EVENT 2
#define BUF_TYPE_EXCEPT 3
#define BUF_TYPE_TRAPRET 4
#define BUF_TYPE_USERTRAP 5
#define BUF_TYPE_PAGEFAULT 6

typedef struct buf_entry_struct {
      unsigned char type;
      union {
	    struct {
		  unsigned int eip;
	    } entry;
	    struct {
		  unsigned int rc;
		  unsigned int eip;
	    } exit;
	    struct {
		  unsigned int rc;
	    } trapret;
	    struct {
		  unsigned int exception;
	    } except;
	    struct {
		  unsigned int event;
		  unsigned int eip;
		  unsigned int aux;
	    } event;
      } data;
      struct
      {
	    unsigned long long stop;
	    unsigned long long start;
      } tsc;
} buf_entry_t;



void
s_instrument_thread_init(s_thread_t *th)
{
	th->instrument_buffer = (void *)0;
	th->instrument_flags  = 0;
	th->instrument_buffer_used = 0;

#ifdef INSTRUMENT_ALL_THREADS
	s_instrument_thread_begin(th);
#endif
}

void s_instrument_thread_begin(s_thread_t *th)
{
	/*
	 * reset the buffer, if one exists already.
	 */
	if (th->instrument_buffer)
		kfree(th->instrument_buffer,
		      CONFIG_INSTRUMENT_BUF_SIZE);

	th->instrument_buffer =
		kalloc(CONFIG_INSTRUMENT_BUF_SIZE);
	th->instrument_buffer_used = 0;
	th->instrument_flags = INSTRUMENT_FLAG_ON;
}


void
s_instrument_thread_end(s_thread_t *th)
{
	/*
	 * Shut down all the instrumentation for this thread.
	 */
	th->instrument_flags = 0;
	if (th->instrument_buffer)
	{
		s_instrument_thread_dump_buf(th);
		kfree(th->instrument_buffer,
		      CONFIG_INSTRUMENT_BUF_SIZE);
	}
}


#include <oskit/x86/proc_reg.h>
#include <oskit/x86/eflags.h>
#include <oskit/x86/pc/pic.h>

void
s_instrument_thread_dump_buf(s_thread_t *th)
{
	buf_entry_t *buf = (buf_entry_t *)th->instrument_buffer;
	unsigned int idx = 0;
	char *label = NULL;
	int intr_enabled = get_eflags() & EFL_IF;

	if (! buf)
		return;

	cli();
#ifdef DEBUG
	label = th->sob.label;
#endif
	printf("INSTRUMENT DUMP START: Thread %p (%s)\n", th, label);

	while(idx < th->instrument_buffer_used)
	{
		switch(buf[idx].type)
		{
		case BUF_TYPE_ENTRY:
			printf("+%x ",
			       buf[idx].data.entry.eip);
			break;
		case BUF_TYPE_EXIT:
			printf("-%x(%x) ",
			       buf[idx].data.exit.rc,
			       buf[idx].data.exit.eip);
			break;
		case BUF_TYPE_EVENT:
			printf("!%x(%x,%x) ",
			       buf[idx].data.event.event,
			       buf[idx].data.event.eip,
			       buf[idx].data.event.aux);
			break;
		case BUF_TYPE_USERTRAP:
			printf("+U ");
			break;
		case BUF_TYPE_EXCEPT:
			printf("+X %x ",
			       buf[idx].data.except.exception);
			break;
		case BUF_TYPE_PAGEFAULT:
			printf("+P ");
			break;
		case BUF_TYPE_TRAPRET:
			printf("-%x(X) ",
			       buf[idx].data.trapret.rc);
			break;
		}

		printf("%08x%08x:%08x%08x\n",
		       (unsigned)(buf[idx].tsc.stop >> 32),
		       (unsigned)(buf[idx].tsc.stop),
		       (unsigned)(buf[idx].tsc.start >> 32),
		       (unsigned)(buf[idx].tsc.start));
		idx++;
	}

	/* reset the buffer */
	th->instrument_buffer_used = 0;
	printf("INSTRUMENT DUMP END: Thread %p\n", th);
	if (intr_enabled)
		sti();
}


void
s_instrument_thread_destroy(s_thread_t *th)
{
	if (th->instrument_flags && th->instrument_buffer)
	{
		s_instrument_thread_dump_buf(th);

		kfree(th->instrument_buffer, CONFIG_INSTRUMENT_BUF_SIZE);
	}
}


void
s_instrument_syscall_entry(s_thread_t *cur_thread,
			   unsigned int eip,
			   unsigned long long tsc)
{
	buf_entry_t *buf = (buf_entry_t *)cur_thread->instrument_buffer;

	assert(cur_thread->instrument_flags);

	if (cur_thread->instrument_buffer_used * sizeof(buf_entry_t)
	    > CONFIG_INSTRUMENT_BUF_SIZE-sizeof(buf_entry_t))
	{
		s_instrument_thread_dump_buf(cur_thread);
		/* resets the buffer */
	}

	/* Point to next available entry */
	buf = buf + cur_thread->instrument_buffer_used;

	cur_thread->instrument_buffer_used++;
	buf->tsc.stop = tsc;
	buf->type = BUF_TYPE_ENTRY;
	buf->data.entry.eip = eip;

	/* Try to make this last.. */
	buf->tsc.start = get_tsc();
}

void
s_instrument_syscall_exit(s_thread_t *cur_thread,
			  unsigned int rc,
			  unsigned int eip,
			  unsigned long long tsc)
{
	buf_entry_t *buf = (buf_entry_t *)cur_thread->instrument_buffer;

	assert(cur_thread->instrument_flags);

	if (cur_thread->instrument_buffer_used * sizeof(buf_entry_t)
	    > CONFIG_INSTRUMENT_BUF_SIZE-sizeof(buf_entry_t))
	{
		s_instrument_thread_dump_buf(cur_thread);
		/* resets the buffer */
	}

	/* Point to next available entry */
	buf = buf + cur_thread->instrument_buffer_used;

	cur_thread->instrument_buffer_used++;
	buf->tsc.stop = tsc;
	buf->type = BUF_TYPE_EXIT;
	buf->data.exit.rc  = rc;
	buf->data.exit.eip = eip;

	/* Try to make this last.. */
	buf->tsc.start = get_tsc();
}

void
s_instrument_event(s_thread_t *cur_thread,
		   unsigned int event,
		   unsigned int aux,
		   unsigned int eip,
		   unsigned long long tsc)
{
	/*
	 * This function is called from arbitrary code in the
	 * kernel so we need to check instrument_flags.
	 *
	 * XXX Locking???
	 */
	if (cur_thread->instrument_flags)
	{
		buf_entry_t *buf = (buf_entry_t *)cur_thread->instrument_buffer;
		if (cur_thread->instrument_buffer_used * sizeof(buf_entry_t)
		    > CONFIG_INSTRUMENT_BUF_SIZE-sizeof(buf_entry_t))
		{
			s_instrument_thread_dump_buf(cur_thread);
			/* resets the buffer */
		}

		/* Point to next available entry */
		buf = buf + cur_thread->instrument_buffer_used;

		cur_thread->instrument_buffer_used++;
		buf->type = BUF_TYPE_EVENT;
		buf->tsc.stop = tsc;
		buf->data.event.event = event;
		buf->data.event.eip   = eip;
		buf->data.event.aux   = aux;

		/* Try to make this last.. */
		buf->tsc.start = get_tsc();
	}
}

void
s_instrument_usertrap(struct s_thread *cur_thread,
		      unsigned long long tsc)
{
	buf_entry_t *buf = (buf_entry_t *)cur_thread->instrument_buffer;

	assert(cur_thread->instrument_flags);

	if (cur_thread->instrument_buffer_used * sizeof(buf_entry_t)
	    > CONFIG_INSTRUMENT_BUF_SIZE-sizeof(buf_entry_t))
	{
		s_instrument_thread_dump_buf(cur_thread);
		/* resets the buffer */
	}

	/* Point to next available entry */
	buf = buf + cur_thread->instrument_buffer_used;

	cur_thread->instrument_buffer_used++;
	buf->tsc.stop = tsc;
	buf->type = BUF_TYPE_USERTRAP;

	/* Try to make this last.. */
	buf->tsc.start = get_tsc();
}


void
s_instrument_pagefault(struct s_thread *cur_thread,
		       unsigned long long tsc)
{
	buf_entry_t *buf = (buf_entry_t *)cur_thread->instrument_buffer;

	assert(cur_thread->instrument_flags);

	if (cur_thread->instrument_buffer_used * sizeof(buf_entry_t)
	    > CONFIG_INSTRUMENT_BUF_SIZE-sizeof(buf_entry_t))
	{
		s_instrument_thread_dump_buf(cur_thread);
		/* resets the buffer */
	}

	/* Point to next available entry */
	buf = buf + cur_thread->instrument_buffer_used;

	cur_thread->instrument_buffer_used++;
	buf->tsc.stop = tsc;
	buf->type = BUF_TYPE_PAGEFAULT;

	/* Try to make this last.. */
	buf->tsc.start = get_tsc();
}


void
s_instrument_except(struct s_thread *cur_thread,
		    unsigned int exception, /* KR_.* */
		    unsigned long long tsc)
{
	buf_entry_t *buf = (buf_entry_t *)cur_thread->instrument_buffer;

	assert(cur_thread->instrument_flags);

	if (cur_thread->instrument_buffer_used * sizeof(buf_entry_t)
	    > CONFIG_INSTRUMENT_BUF_SIZE-sizeof(buf_entry_t))
	{
		s_instrument_thread_dump_buf(cur_thread);
		/* resets the buffer */
	}

	/* Point to next available entry */
	buf = buf + cur_thread->instrument_buffer_used;

	cur_thread->instrument_buffer_used++;
	buf->tsc.stop = tsc;
	buf->type = BUF_TYPE_EXCEPT;
	buf->data.except.exception = exception;

	/* Try to make this last.. */
	buf->tsc.start = get_tsc();
}


void
s_instrument_trapret(struct s_thread *cur_thread,
		     unsigned int trap_rc, /* KR_.* */
		     unsigned long long tsc)
{
	buf_entry_t *buf = (buf_entry_t *)cur_thread->instrument_buffer;

	assert(cur_thread->instrument_flags);

	if (cur_thread->instrument_buffer_used * sizeof(buf_entry_t)
	    > CONFIG_INSTRUMENT_BUF_SIZE-sizeof(buf_entry_t))
	{
		s_instrument_thread_dump_buf(cur_thread);
		/* resets the buffer */
	}

	/* Point to next available entry */
	buf = buf + cur_thread->instrument_buffer_used;

	cur_thread->instrument_buffer_used++;
	buf->tsc.stop = tsc;
	buf->type = BUF_TYPE_TRAPRET;
	buf->data.trapret.rc  = trap_rc;

	/* Try to make this last.. */
	buf->tsc.start = get_tsc();
}

#endif
