/*
 * Copyright (c) 1997 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Hacks and functions for instrumenting syscalls and events in the
 * kernel.  There are calls in the entry layer (trap.S) and at various
 * points in the kernel where an "exception is raised" (i.e, 'return
 * KR_CANCEL'.)  Trace data is recorded per-thread, and stored in a
 * buffer associated with the thread.  See s_instrument.c for the details.
 */
#ifndef _KERNEL_S_INSTRUMENT_H
#define _KERNEL_S_INSTRUMENT_H

#ifndef CONFIG_INSTRUMENT

#define INSTRUMENT_EVENT(th, event, aux)

#else /* CONFIG_INSTRUMENT defined */

#include <oskit/x86/proc_reg.h> /* XXX x86 specific. */

/*
 * This macro is used most places to hide the get_eip() and
 * get_tsc().
 */
#define INSTRUMENT_EVENT(th, event, aux)	\
      s_instrument_event(th, event, aux,        \
			 get_eip(),		\
			 get_tsc());

/*
 * Flags used in calls to kentry_md_syscall_instrument().
 */
#define INSTRUMENT_SYSCALL_ENTRY     0x1
#define INSTRUMENT_SYSCALL_EXIT      0x2
#define INSTRUMENT_SYSCALL_EXCEPT    0x3
#define INSTRUMENT_SYSCALL_TRAPRET   0x4
#define INSTRUMENT_SYSCALL_USERTRAP  0x5
#define INSTRUMENT_SYSCALL_PAGEFAULT 0x6

/*
 * Flags used for s_instrument_event().  Additionally, 0 (KR_OK) and
 * the KR_* flags are accepted.
 */
#define INSTRUMENT_EVENT_SWITCH_IN   11
#define INSTRUMENT_EVENT_SWITCH_OUT  12
#define INSTRUMENT_EVENT_MIGHT_FAULT 13
#define INSTRUMENT_EVENT_USER_1      14
#define INSTRUMENT_EVENT_USER_2      15
#define INSTRUMENT_EVENT_USER_3      16
#define INSTRUMENT_EVENT_READY     17

#ifndef ASSEMBLER
struct s_thread;

/* Record entry to a syscall (in kernel mode). */
void 
s_instrument_syscall_entry(struct s_thread *cur_thread,
			   unsigned int eip,
			   unsigned long long tsc);

/* Record return from a syscall (in kernel mode) */
void 
s_instrument_syscall_exit(struct s_thread *cur_thread,
			  unsigned int rc,
			  unsigned int eip,
			  unsigned long long tsc);

/* Record an "event" of a type above or one of the KR_* */
void 
s_instrument_event(struct s_thread *cur_thread,
		   unsigned int event,
		   unsigned int aux,
		   unsigned int eip,
		   unsigned long long tsc);

/* Record a user mode trap (see trap.S) */
void 
s_instrument_usertrap(struct s_thread *cur_thread,
		      unsigned long long tsc);

/* Record a user mode page fault (see trap.S) */
void 
s_instrument_pagefault(struct s_thread *cur_thread,
		       unsigned long long tsc);

/* Record an exception (see trap.S) */
void 
s_instrument_except(struct s_thread *cur_thread,
		    unsigned int exception, /* KR_.* */
		    unsigned long long tsc);

/* Record return from any of the above exceptions. */
void 
s_instrument_trapret(struct s_thread *cur_thread,
		     unsigned int trap_rc, /* KR_.* */
		     unsigned long long tsc);

/* Init the instrumentation info stored with the thread. */
void s_instrument_thread_init(struct s_thread *th);

/* Begin instrumenting a thread (allocates buffer) */
void s_instrument_thread_begin(struct s_thread *th);

/* Stop instrumenting thread */
void s_instrument_thread_end(struct s_thread *th);

/* Destroy the buffer associated with the thread. */
void s_instrument_thread_destroy(struct s_thread *th);

/* Dump the buffer (to stdout) associated with the thread. */
void s_instrument_thread_dump_buf(struct s_thread *th);

#endif /* ASSEMBLER */
#endif /* CONFIG_INSTRUMENT defined */
#endif /* _KERNEL_S_INSTRUMENT_H_ */
