/*
 * Copyright (c) 1996 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#ifndef _FLOSS_KERNEL_FLUKE_S_MAPPING_H_
#define _FLOSS_KERNEL_FLUKE_S_MAPPING_H_

#include <oskit/queue.h>
#include <fluke/vm.h>
#include "s_ob.h"
#include "frag.h"	/* XXX */

/*
 * Fluke memory mapping implementation.
 *
 * A mapping makes part or all of a region accessible in a task at a
 * selected address.  Mappings are the mechanism by which tasks populate
 * their own or other task's address spaces.
 *
 * Given a region reference object, a task may map all or part of the
 * region at a given address with specified access permissions.
 *
 * Locking protocol:
 *
 * For mappings, each of the object, fault and unmap layers has its own lock.
 *
 * The start_va, size, prot, region_off, and task fields are read-only for
 * the fault and unmap layers but may be written by the object layer
 * (mapping_set_state).  The region field is read-only for the fault layer
 * and read-write for the object layer.  Hence, mapping_set_state should
 * hold both locks when modifying any of these fields.
 */
typedef struct s_mapping {
	/* object layer, protected by object lock */
	struct s_ob	sob;		/* supervisor object state */
	struct link	task;		/* link form of task reference */
	oskit_addr_t	region_off;	/* offset into the region */
	oskit_addr_t	start_va;	/* VA at which the region is mapped */
	oskit_size_t	size;		/* size of mapping */
	fluke_prot_t	prot;		/* protection attributes granted */

	/* fault layer */
	struct mutex	flock;		/* lock */
	queue_chain_t	tlink;		/* mappings for a task */
	struct link	region;		/* region reference */

	/* unmap layer */
	struct mutex	ulock;		/* lock */
	queue_chain_t	rlink;		/* mappings for a region */
	struct s_task	*taskp;		/* pointer form of task reference */
#ifdef USEAMM
	struct amm	fragments;	/* fragment map */
#else
	struct fragment	*fhint;		/* fragment hint */
	queue_head_t	fragments;	/* fragments for this mapping */
#endif
} s_mapping_t;

#ifdef USEAMM
#define AMMTOMAP(amm) \
	((s_mapping_t *)((void *)(amm) - (int)&((s_mapping_t *)0)->fragments))
#endif

/* XXX for now */
#define s_mapping_lock(m)	s_ob_lock(&(m)->sob)
#define s_mapping_unlock(m)	s_ob_unlock(&(m)->sob)
#define s_mapping_f_lock(m)	mutex_lock(&(m)->flock);
#define s_mapping_f_unlock(m)	mutex_unlock(&(m)->flock);
#define s_mapping_u_lock(m)	mutex_lock(&(m)->ulock);
#define s_mapping_u_unlock(m)	mutex_unlock(&(m)->ulock);

struct s_ref;
struct s_task;
struct s_region;
struct frag;
struct fluke_mapping_state;

int s_mapping_create(oskit_addr_t uva, fluke_hash_t hash,
		     s_mapping_t **out_mapping);
void s_mapping_destroy(s_mapping_t *mapping);
int s_mapping_get_state(s_mapping_t *mapping,
			struct fluke_mapping_state *state,
			struct link *region_link, struct link *task_ref);
int s_mapping_set_state(s_mapping_t *mapping,
			struct fluke_mapping_state *state,
			struct link *region_link, struct link *task_link);
void s_mapping_protect(s_mapping_t *mapping, fluke_prot_t prot);
void s_mapping_restrict(s_mapping_t *mapping, oskit_addr_t offset,
			oskit_size_t size, fluke_prot_t prot);
int s_mapping_insert_task(s_mapping_t *mapping, struct s_task *task);
void s_mapping_remove_task(s_mapping_t *mapping, struct s_task *task);
void s_mapping_insert_region(s_mapping_t *mapping, struct s_region *region);

void s_mapping_boot(void);

#ifdef DEBUG
/*
 * Dump state of thread to console
 */
void s_mapping_dump(s_mapping_t *map, int indent);
#endif

#if FLASK
void s_mapping_region_revoke(s_mapping_t *mapping,security_id_t tsid,
			     oskit_bool_t try_remove, 
			     access_vector_t *out_retained);

int s_mapping_region_reset(s_mapping_t *mapping);
#endif

#endif /* _FLOSS_KERNEL_FLUKE_S_MAPPING_H_ */

