/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include <fluke/mutex.h>
#include <fluke/exception.h>
#include <fluke/machine/syscall.h>
#include <oskit/machine/base_vm.h>

#include "s_mutex.h"
#include "s_task.h"
#include "s_thread.h"
#include "exception.h"
#include "debug.h"

#ifndef CONFIG_OPT_SYNCH

#define MUTEX_SETOWNER(m, th) { \
	s_thread_lock(th); \
	link_produce(&(m)->owner, &th->sob); \
	s_thread_unlock(th); \
}

#define MUTEX_CLROWNER(m) { \
	link_destroy(&(m)->owner); \
	link_init(&(m)->owner); \
}

#else

#define MUTEX_SETOWNER(m, th)
#define MUTEX_CLROWNER(m)

#endif

int s_mutex_create(
	oskit_addr_t		uva,
	fluke_hash_t		hash,
	s_mutex_t		**out_mutex)
{
	s_mutex_t *mutex;
	int rc;

	rc = task_ob_create(current_thread(), FLUKE_MUTEX, uva, hash,
			    (s_ob_t **)&mutex);
	if (rc)
		return (rc);

#ifndef CONFIG_OPT_SYNCH
	link_init(&mutex->owner);
	mutex->locked = 0;
#else
	/* Initialize the pointer to the user-level part.  */
	mutex->u_mutex = (u_mutex_t *) phystokv(mutex->sob.u_pa);
	MUTEX_INITLOCK(mutex);
#endif
	mutex->flags = 0;	/* XXX */

	cond_init(&mutex->cond);

	*out_mutex = mutex;
	return 0;
}

int s_mutex_destroy(
	s_mutex_t		*mutex)
{
	assert(mutex && mutex->sob.active);
	assert(mutex->sob.type == FLUKE_MUTEX);
	assert_s_mutex_locked(mutex);

	/*
	 * User code is supposed to guarantee
	 * that no one is holding a mutex when it is destroyed
	 * and hence we could generate an insanity trap here.
	 *
	 * However, a memory manager should be able to destroy any object
	 * in its managed memory without generating an insanity trap.
	 *
	 * Since we don't differentiate the cases, the users lose.
	 * We just unlock the mutex and wake everybody up where they
	 * will discover that the mutex is no longer valid.
	 */
	if (MUTEX_ISLOCKED(mutex))
	{
		MUTEX_CLROWNER(mutex);
		MUTEX_SETLOCKED(mutex, 0);
		cond_broadcast(&mutex->cond);
	}

#ifndef CONFIG_OPT_SYNCH
	assert(link_consume(&mutex->owner) == 0);
#endif

	cond_destroy(&mutex->cond);

	s_ob_destroy(&mutex->sob);
	return 0;
}

int s_mutex_get_state(
	s_mutex_t			*mutex,
	struct fluke_mutex_state	*state,
	struct link			*owner_link)
{
	assert_s_ob_locked(&mutex->sob);
	assert(mutex->sob.type == FLUKE_MUTEX);

	if (state) {
		state->hash = s_ob_gethash(&mutex->sob);
		state->flags = mutex->flags;
		state->locked = MUTEX_ISLOCKED(mutex);
	}

	if (owner_link) {
#ifndef CONFIG_OPT_SYNCH
		link_create_copy(&mutex->owner, owner_link);
#else
		link_init(owner_link); /* return null reference */
#endif
	}

	return 0;
}

int s_mutex_set_state(
	s_mutex_t			*mutex,
	struct fluke_mutex_state	*state,
	struct link			*owner_link)
{
	assert_s_ob_locked(&mutex->sob);
	assert(mutex->sob.type == FLUKE_MUTEX);

	if (state) {
		mutex->flags = state->flags;
		MUTEX_SETLOCKED(mutex, state->locked);
	}

#ifndef CONFIG_OPT_SYNCH
	if (owner_link)
		link_transfer(owner_link, &mutex->owner);
#endif

	/*
	 * XXX kernel assumes that if !mutex->locked then owner link is NULL.
	 * Make it so.
	 */
	if (!state->locked)
	{
		MUTEX_CLROWNER(mutex);
		cond_signal(&mutex->cond);
	}

	return 0;
}

#ifndef MD_S_MUTEX_LOCKOP
int s_mutex_lockop(
	s_mutex_t		*mutex,
	s_thread_t		*cur_th)
{
	int rc;

	assert_s_mutex_locked(mutex);
	assert(current_thread() == cur_th);

	if (MUTEX_ISLOCKED(mutex))
	{
		link_t link;

		/*
		 * Be nice and throw an exception if the thread attempts
		 * to acquire a mutex that it already holds.
		 */
		if (link_compare_ptr(&mutex->owner, &cur_th->sob))
		{
			EXC_RETURN_INSANITY(MUTEX_LOCKED);
		}

#ifdef CONFIG_PRIO_INHERIT
		link_create_copy(&mutex->owner, &link);
#else
		link_init(&link);
#endif

		rc = cond_wait(&mutex->cond, &mutex->sob.lock, &link);
		assert(rc != 0);
		return rc;
	}

	mutex->locked = 1;
	MUTEX_SETOWNER(mutex, cur_th);
	s_mutex_unlock(mutex);

	return 0;
}
#endif

#ifndef MD_SYS_MUTEX_TRYLOCK
int s_mutex_trylockop(
	s_mutex_t		*mutex)
{
	assert_s_mutex_locked(mutex);

	if (!mutex->locked)
	{
		mutex->locked = 1;
		MUTEX_SETOWNER(mutex, current_thread());
		return FLUKE_SUCCESS;
	}
	return FLUKE_MUTEX_HELD;
}
#endif

#ifndef MD_S_MUTEX_UNLOCKOP
int s_mutex_unlockop(
	s_mutex_t		*mutex,
	s_thread_t		*cur_th)
{
	assert_s_mutex_locked(mutex);

	if (!mutex->locked)
	{
		EXC_RETURN_INSANITY(MUTEX_NOT_LOCKED);
	}

#ifndef CONFIG_OPT_SYNCH
	if (!link_compare_ptr(&mutex->owner, &cur_th->sob)) {
		EXC_RETURN_INSANITY(MUTEX_NOT_OWNER);
	}
#endif
	/*
	 * Change the mutex's state to unlocked.
	 */
	MUTEX_CLROWNER(mutex);
	mutex->locked = 0;

	/*
	 * Wake up any other threads blocked on this mutex.
	 * We can't just pick a waiting thread off the queue
	 * and give it the lock, because we can't be sure that
	 * it hasn't lost access to the memory for this mutex
	 * sometime since it blocked.
	 */
	cond_broadcast(&mutex->cond);

	return 0;
}
#endif

#ifdef DEBUG
void
s_mutex_dump(s_mutex_t *mutex, int indent)
{
	int i;

	for (i = 0; i < indent; i++) putchar(' ');
#ifndef CONFIG_OPT_SYNCH
	printf(" owner=%p, locked=%d, &cond=%p\n",
	       mutex->owner.target, mutex->locked, &mutex->cond);
#else
	printf(" u_mutex state=");
	for (i = 0; i < sizeof(*mutex->u_mutex) / sizeof(int); ++i)
		printf("%#x ", ((int *)mutex->u_mutex)[i]);
	printf("&cond=%p\n", &mutex->cond);
#endif
}
#endif
