/*
 * Copyright (c) 1995, 1998, 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <oskit/page.h>
#include <fluke/exception.h>
#include <fluke/debug.h>

#include "s_ob.h"
#include "debug.h"
#include "exception.h"
#include "s_thread.h"	/* XXX needed for EXC_SET_INSANITY */
#include "kalloc.h"
#include "config_instrument.h"
#include "config_kalloc_sob.h"

#if FLASK
#include "avc.h"
#endif

#include "config_flob_check.h"

/*#define SOB_STATS*/

#define OBHASH_SIZE	1024

/*
 * In order to detect overlap, all object within a MAXOB_SIZE chunk must
 * hash to the same bucket and objects in that bucket are kept sorted by
 * start (physical) address.
 *
 * N.B. this hash should not be confused with the exported object hash value
 * which can be set by the application.
 */
#define s_ob_hash(pa) \
	((((pa) >> FLUKE_MAXOB_SHIFT) ^ ((pa) >> PAGE_SHIFT)) & (OBHASH_SIZE-1))

/* internal routines */
int s_ob_insert(struct s_ob *ob);
void s_ob_remove(struct s_ob *ob);

mutex_t all_obs_lock;
s_ob_t *all_obs[OBHASH_SIZE];

/* Default value for exported object hash value.  */
static fluke_hash_t	obhash;
#if 1
#define s_ob_init_hash()	0xd0000000
#endif

/*
 * Memory freelists for kernel objects.
 * We keep a linked list of memory for each object type.
 * We allocate power-of-two sized chunks for simplicity and to preserve
 * nice cache line alignment properties.
 *
 * XXX this will have to change if/when machine dependent object types
 * are added.
 */
struct s_ob_freelist {
	mutex_t	lock;
	void	*list;
	int	asize;
} s_ob_freelist[FLUKE_MI_MAX];

#ifdef MEMSTATS
int mstats_obs, mstats_maxobs, mstats_inactive;
unsigned long mstats_obbytes, mstats_maxobbytes;
struct mstats_ob {
	int num;
	int max;
	int inactive;
} mstats_ob[FLUKE_MI_MAX+1];
#endif

#include "config_inlining.h"
#ifdef CONFIG_INLINING
inline
#endif
int s_ob_enter(
	fluke_type_t		type,
	oskit_addr_t		u_pa,
	fluke_hash_t		u_hash,
	s_ob_t			*ob)
{
	int err;

	ob->type = type;
	ob->active = 1;
	acount_init(&ob->refs, 1);
	mutex_init(&ob->lock);
	ob->u_pa = u_pa;
	ob->u_hash = u_hash ? u_hash : obhash++;
#ifdef DEBUG
	ob->label = 0;
	ob->st = 0;
#endif

#if FLASK
	ob->sid = SECSID_NULL;
#endif

	OB_INFO_ENTER(ob);

	/*
	 * Lock the object before we go any further.
	 * This must be before s_ob_insert,
	 * because after that point, other threads may find the object.
	 */
	mutex_lock(&ob->lock);

	/*
	 * Insert the object into the global object list.
	 */
	mutex_lock(&all_obs_lock);
	err = s_ob_insert(ob);
	mutex_unlock(&all_obs_lock);

	return err;
}

int s_ob_create(
	fluke_type_t		type,
	oskit_addr_t		u_pa,
	fluke_hash_t		u_hash,
	s_ob_t			**out_ob)
{
	struct s_ob_freelist *list;
	s_ob_t *ob;
	int err;

	assert(type > 0 && type <= FLUKE_MI_MAX);
	assert(u_pa);

#ifdef CONFIG_KALLOC_SOB
	if (OBT_SALIGN(type) > 0)
		ob = kalloc_aligned(OBT_SSIZE(type), OBT_SALIGN(type));
	else
		ob = kalloc(OBT_SSIZE(type));
	if (!ob)
		return KR_NO_MEMORY;
#else
	list = &s_ob_freelist[type - 1];
	mutex_lock(&list->lock);
	if (list->list == 0) {
		void *mem, *emem;

		mem = kalloc_aligned(PAGE_SIZE, PAGE_SHIFT);
		if (mem == 0) {
			mutex_unlock(&list->lock);
			return KR_NO_MEMORY;
		}
		assert(list->asize > 0);
		emem = mem + PAGE_SIZE;
		do {
			emem -= list->asize;
			*(void **)emem = list->list;
			list->list = emem;
		} while (emem > mem);
	}
	ob = (s_ob_t *)list->list;
	list->list = *(void **)ob;
	mutex_unlock(&list->lock);
#endif

#ifdef DEBUG
	/* Help catch uninitialized variables quicker... */
	memset(ob, 0xa5, OBT_SSIZE(type));
#endif

	err = s_ob_enter(type, u_pa, u_hash, ob);
	if (err) {
		ob->active = 0;
		mutex_unlock(&ob->lock);
		s_ob_free(ob);
		return err;
	}

#ifdef MEMSTATS
	{
		int ix = type > FLUKE_MI_MAX ? 0 : type;

		if (++mstats_ob[ix].num > mstats_ob[ix].max)
			mstats_ob[ix].max = mstats_ob[ix].num;
		if (++mstats_obs > mstats_maxobs)
			mstats_maxobs = mstats_obs;
		mstats_obbytes += OBT_SSIZE(type);
		if (mstats_obbytes > mstats_maxobbytes)
			mstats_maxobbytes += mstats_obbytes;
	}
#endif

	*out_ob = ob;
	return 0;
}

#ifdef DEBUG
s_ob_t *undead_list;
#endif

void s_ob_destroy(
	struct s_ob		*ob)
{
	assert(ob);
	assert(ob->type > 0);
	assert(ob->u_pa);
	assert(ob->active);
	assert_s_ob_locked(ob);

	/*
	 * Mark the object inactive.
	 */
	ob->active = 0;
#ifdef MEMSTATS
	mstats_inactive++;
	mstats_ob[ob->type > FLUKE_MI_MAX ? 0 : ob->type].inactive++;
#endif

	/*
	 * Remove us from the global object list.
	 */
	mutex_lock(&all_obs_lock);
	s_ob_remove(ob);
#ifdef DEBUG
	ob->all_next = undead_list;
	undead_list = ob;
#endif
	mutex_unlock(&all_obs_lock);

	do_debug(ob->u_pa = 0);

#if FLASK
	do_debug(ob->sid = SECSID_NULL);
#endif

	/*
	 * Unlock and drop our active reference
	 * (which may result in object being freed).
	 */
	s_ob_unlock(ob);
	s_ob_decrref(ob);
}

/*
 * We were about to unlock an object,
 * but noticed that its reference count had dropped to zero.
 * Return the object to the appropriate free zone.
 * On entry, the object is unlocked.
 */
void
s_ob_free(
	struct s_ob		*ob)
{
	struct s_ob_freelist *list;

	assert(ob->active == 0);
	assert_mutex_unlocked(&ob->lock);

#ifdef DEBUG
	if (ob->label)
		kfree(ob->label, FLUKE_DEBUG_LABEL_STR_MAXLEN);
	if (ob->st)
		kfree(ob->st, sizeof *ob->st);
	{
		s_ob_t **nob;
		for (nob = &undead_list; *nob; nob = &(*nob)->all_next)
			if (*nob == ob) {
				*nob = ob->all_next;
				break;
			}
	}
#endif
#ifdef MEMSTATS
	mstats_ob[ob->type > FLUKE_MI_MAX ? 0 : ob->type].inactive--;
	mstats_inactive--;
	mstats_ob[ob->type > FLUKE_MI_MAX ? 0 : ob->type].num--;
	mstats_obs--;
	mstats_obbytes -= OB_SSIZE(ob);
#endif
#ifdef CONFIG_KALLOC_SOB
	kfree(ob, OB_SSIZE(ob));
#else
	list = &s_ob_freelist[ob->type - 1];
	mutex_lock(&list->lock);
	*(void **)ob = list->list;
	list->list = (void *)ob;
	mutex_unlock(&list->lock);
#endif
}

/*
 * Associates the supervisor part of an object with a new kernel VA.
 * Object should be locked on call.
 */
int s_ob_move(
	struct s_ob		*ob,
	oskit_addr_t		new_pa)
{
	oskit_addr_t old_pa;
	int err;

	assert(ob);
	assert(new_pa);
	assert(ob->type > 0);
	assert(ob->u_pa);
	assert(ob->active);
	assert_mutex_locked(&ob->lock);

	mutex_lock(&all_obs_lock);

	s_ob_remove(ob);
	old_pa = ob->u_pa;
	ob->u_pa = new_pa;
	err = s_ob_insert(ob);
	if (err) {
		ob->u_pa = old_pa;
		(void)s_ob_insert(ob);
	}

	mutex_unlock(&all_obs_lock);

	return (err);
}

/*
 * Lookup an object given a kernel virtual address for the user part.
 * Returns an error (and sets insanity code) if the object wasn't found.
 * If object is found, it is returned locked.
 */
int s_ob_lookup(
	oskit_addr_t		pa,
	struct s_ob		**obp)
{
	s_ob_t **nob;

	mutex_lock(&all_obs_lock);

	for (nob = &all_obs[s_ob_hash(pa)]; *nob; nob = &(*nob)->all_next) {
		if (pa == (*nob)->u_pa) {
			/*
			 * XXX problem: other routines lock object and then
			 * global list, we could deadlock here as we grab
			 * list then object.
			 */
			s_ob_lock(*nob);
			mutex_unlock(&all_obs_lock);
			*obp = *nob;
			return (0);
		}
		if (pa < (*nob)->u_pa)
			break;
	}

	mutex_unlock(&all_obs_lock);
	EXC_RETURN_INSANITY(NO_OBJECT);
}

/*
 * Insert an object on the global active object list.
 * Returns an error code if the object would overlap with an existing one.
 * Global object list should be locked on call.
 */
int s_ob_insert(
	struct s_ob		*ob)
{
	s_ob_t **nob;

	for (nob = &all_obs[s_ob_hash(ob->u_pa)]; *nob;
	     nob = &(*nob)->all_next) {
		if (ob->u_pa + OB_USIZE(ob) <= (*nob)->u_pa)
			break;
		if (ob->u_pa < (*nob)->u_pa + OB_USIZE(*nob)) {
			EXC_RETURN_INSANITY(OBJECT_EXISTS);
		}
	}
	ob->all_next = *nob;
	*nob = ob;

	return (0);
}

/*
 * Remove an object from the global object list.
 * The object must be there.
 * Global object list should be locked on call.
 */
void s_ob_remove(
	struct s_ob		*ob)
{
	s_ob_t **nob;

	for (nob = &all_obs[s_ob_hash(ob->u_pa)]; *nob != ob;
	     nob = &(*nob)->all_next)
		assert(*nob);
	*nob = ob->all_next;
}

/*
 * Find all objects who fall in the range [*pa - epa] and record them in
 * the provided oblist.  The global list lock is held during the search to
 * ensure that objects don't move or get destroyed.
 * If all objects found fit in the array, *pa is set to epa and 0 is returned.
 * If more objects remained after the array is filled, *pa is set to the
 * address of the first object that didn't fit and 1 is returned.
 * Note that calling with *lsizep == 0 will return 1 with *pa set to the
 * address of the first object found.
 *
 * Note also that successive calls are non-atomic w.r.t. the object list.
 * Hence, the object system had better be quiescent when making multiple calls.
 */
int s_ob_search(
	oskit_addr_t		*pa,
	oskit_addr_t		epa,
	fluke_obspec_t		*oblist,
	unsigned int		*lsizep)
{
	s_ob_t **nob;
	oskit_addr_t spa, lepa;
	int listix = 0;

	spa = *pa;
	if (spa >= epa) {
		*lsizep = 0;
		return (0);
	}

	mutex_lock(&all_obs_lock);

	while (spa < epa) {
		/*
		 * XXX note embedded knowledge of hash function to avoid
		 * hashing every possible address.
		 */
		lepa = (spa + FLUKE_MAXOB_SIZE) & ~(FLUKE_MAXOB_SIZE-1);
		if (lepa > epa)
			lepa = epa;
		for (nob = &all_obs[s_ob_hash(spa)]; *nob;
		     nob = &(*nob)->all_next) {
			if ((*nob)->u_pa < spa)
				continue;
			if ((*nob)->u_pa >= lepa)
				break;
			if (listix == *lsizep) {
				mutex_unlock(&all_obs_lock);
				*pa = (*nob)->u_pa;
				return (1);
			}

			/*
			 * task_ob_search needs this to do signature
			 * check on the user space part. So pass the
			 * s_ob_t pointer rather than the u_pa
			 */
			oblist[listix].ob = (void *)(*nob);
			oblist[listix].type = (*nob)->type;
			listix++;
		}
		spa = lepa;
	}

	mutex_unlock(&all_obs_lock);
	*pa = epa;
	*lsizep = listix;
	return (0);
}

/*
 * Summary of existing objects
 */
void s_ob_dump_all(void)
{
#ifdef SOB_STATS
	int obcount[16], obtcount;
	int obusize, obssize;
	oskit_addr_t loaddr, hiaddr;
	int minb, maxb, bcount;
	s_ob_t *ob;
	int i;

	for (i = 0; i < 16; i++)
		obcount[i] = 0;
	obtcount = 0;
	obusize = obssize = 0;
	loaddr = 0x80000000;
	hiaddr = 0;
	minb = 0x7fffffff;
	maxb = 0;

	mutex_lock(&all_obs_lock);
	for (i = 0; i < OBHASH_SIZE; i++) {
		bcount = 0;
		for (ob = all_obs[i]; ob; ob = ob->all_next) {
			if (ob->type > 15)
				obcount[0]++;
			else
				obcount[ob->type]++;
			obtcount++;
			obusize += OB_USIZE(ob);
			obssize += OB_SSIZE(ob);
			if (ob->u_pa < loaddr)
				loaddr = ob->u_pa;
			if (ob->u_pa + OB_USIZE(ob) > hiaddr)
				hiaddr = ob->u_pa + OB_USIZE(ob);
			bcount++;
		}
		if (bcount < minb)
			minb = bcount;
		if (bcount > maxb)
			maxb = bcount;
	}
	mutex_unlock(&all_obs_lock);

	printf("%d objects in [0x%08x-0x%08x] (1 per %d bytes)\n",
	       obtcount, loaddr, hiaddr, (hiaddr-loaddr) / obtcount);
	for (i = 0; i < 16; i++)
		if (obcount[i])
			printf("\ttype %d: %d\n", i, obcount[i]);
	printf("ssize: %d bytes (%d bytes per object)\n",
	       obssize, obssize / obtcount);
	printf("usize: %d bytes (%d bytes per object)\n",
	       obusize, obusize / obtcount);
	printf("%d hash buckets, %d min, %d max\n", OBHASH_SIZE, minb, maxb);
#endif
}


void s_ob_boot(void)
{
	int i;

	mutex_init(&all_obs_lock);

	/* MD hash init in case we do something like read a TOD clock.  */
	obhash = s_ob_init_hash();

	for (i = 0; i < FLUKE_MI_MAX; i++) {
		int bsize;

		mutex_init(&s_ob_freelist[i].lock);
		s_ob_freelist[i].list = 0;
		for (bsize = 1; bsize < OBT_SSIZE(i+1); bsize <<= 1)
			;
		s_ob_freelist[i].asize = bsize;
	}
}

#ifdef DEBUG
/*
 * Store the current user stack trace in the object
 */
void s_ob_store_trace(
	struct s_ob		*ob)
{
	ob->st = kalloc(sizeof *ob->st);
	if (ob->st) {
		ob->st->thread = &current_thread()->sob;
		s_thread_store_stacktrace(current_thread(), ob->st->buf,
					  sizeof(ob->st->buf)/
					  sizeof(ob->st->buf[0]));
	}
}

/*
 * Give object a label.  Should be locked on call.
 */
void s_ob_label(
	struct s_ob             *ob,
	char                    *label)
{
	int ct = 0;

	assert(ob);
	assert(label);
	assert(ob->type > 0);
	assert(ob->u_pa);
	assert(ob->active);
	assert_s_ob_locked(ob);

	if (ob->label)
		kfree(ob->label, FLUKE_DEBUG_LABEL_STR_MAXLEN);

	ob->label = kalloc(FLUKE_DEBUG_LABEL_STR_MAXLEN);

	while (*label && (ct < FLUKE_DEBUG_LABEL_STR_MAXLEN-1))
		ob->label[ct++] = *(label++);
	ob->label[ct] = '\0';
}

/*
 * Dump all thread objects
 */
static int s_ob_dump(s_ob_t *ob, void *arg)
{
#ifdef PREEMPTSTATS
	if (ob->type == FLUKE_THREAD)
		thread_disp_dump_preemptstats((s_thread_t *)ob);
#else
	s_debug_dump_flob(ob, 0);
#endif
	return 0;
}

void s_ob_dump_flobs(void)
{
#if 1
	s_ob_t *ob;
	int count;
	extern int s_ref_scan(s_ob_t *ob, void *arg);

	/*
	 * Dump the undead
	 */
	printf("\nObjects destroyed but not freed:\n");
	count = 0;
	for (ob = undead_list; ob; ob = ob->all_next)
		count++;
	printf("  before ref scan: %d\n", count);
	s_ob_map(1 << FLUKE_REF, s_ref_scan, 0);
	count = 0;
	for (ob = undead_list; ob; ob = ob->all_next)
		count++;
	printf("  after ref scan: %d\n", count);
	printf("  dump:\n");
	for (ob = undead_list; ob; ob = ob->all_next)
		s_debug_dump_flob(ob, 2);
#endif
	/*
	 * and the living
	 */
	printf("\nActive objects:\n");
	s_ob_map(1 << FLUKE_THREAD, s_ob_dump, 0);
	s_ob_map(1 << FLUKE_TASK, s_ob_dump, 0);
	s_ob_map(1 << FLUKE_MAPPING, s_ob_dump, 0);
	s_ob_map(1 << FLUKE_REGION, s_ob_dump, 0);
	s_ob_map(1 << FLUKE_COND, s_ob_dump, 0);
	s_ob_map(1 << FLUKE_MUTEX, s_ob_dump, 0);
	s_ob_map(1 << FLUKE_PSET, s_ob_dump, 0);
	s_ob_map(1 << FLUKE_PORT, s_ob_dump, 0);
	s_ob_map(1 << FLUKE_REF, s_ob_dump, 0);
}

void s_ob_dump_threads(void)
{
	s_ob_map(1 << FLUKE_THREAD, s_ob_dump, 0);
#if 1
	s_ob_map(1 << FLUKE_TASK, s_ob_dump, 0);
#endif
}
#endif DEBUG

#ifdef PREEMPTSTATS
void s_ob_dump_thread_preemptstats(void)
{
	s_ob_map(1 << FLUKE_THREAD, s_ob_dump, 0);
}
#endif

#ifdef MEMSTATS
void s_ob_dump_mstats()
{
	static char *obnames[] = {
		"UNKNOWN",
		"tasks   ", "threads ",
		"mutexes ", "conds   ",
		"ports   ", "psets   ",
		"catalogs", "refs    ",
		"mappings", "regions "
	};
	int ix;

	printf("  Kernel objects:    %8lu bytes (%8lu bytes max)\n",
	       mstats_obbytes, mstats_maxobbytes);
	printf("    object dump:     %d (%d/%d inactive/max)\n",
	       mstats_obs, mstats_inactive, mstats_maxobs);
	for (ix = 1; ix <= FLUKE_MI_MAX; ix++)
		if (mstats_ob[ix].max)
			printf("      %4d %s (%4d/%4d inactive/max): %8u (%8u/%8u) bytes\n",
			       mstats_ob[ix].num, obnames[ix],
			       mstats_ob[ix].inactive,
			       mstats_ob[ix].max,
			       mstats_ob[ix].num * OBT_SSIZE(ix),
			       mstats_ob[ix].inactive * s_ob_freelist[ix-1].asize,
			       mstats_ob[ix].num * s_ob_freelist[ix-1].asize);
	if (mstats_ob[0].max)
		printf("      %4d %s (%4d/%4d inactive/max)\n",
		       mstats_ob[0].num, obnames[0],
		       mstats_ob[0].inactive, mstats_ob[0].max);
}
#endif

#ifdef CONFIG_INSTRUMENT
#include "s_instrument.h"

static int s_instrument_thread_dump(s_ob_t *ob, void *arg)
{
	s_instrument_thread_dump_buf((s_thread_t *)ob);
	return 0;
}

void dump_instrument_threads()
{
	s_ob_map(1 << FLUKE_THREAD, s_instrument_thread_dump, 0);
}
#endif

/*
 * General iterator for the global object list.
 * Applies the given function to each object whose type is indicated in
 * obmask until done or until the function returns non-zero.
 *
 * XXX assumes no more than 32 object types.
 */
int s_ob_map(int obmask, int (*f)(s_ob_t *sob, void *args), void *args)
{
	s_ob_t *ob;
	int i, rc = 0;

	mutex_lock(&all_obs_lock);
	for (i = 0; i < OBHASH_SIZE; i++) {
		for (ob = all_obs[i]; ob; ob = ob->all_next) {
			if ((1 << ob->type) & obmask) {
				rc = f(ob, args);
				if (rc) {
					mutex_unlock(&all_obs_lock);
					return rc;
				}
			}
		}
	}
	mutex_unlock(&all_obs_lock);
	return rc;
}
