/*
 * Copyright (c) 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * Contributed by the Computer Security Research division,
 * INFOSEC Research and Technology Office, NSA.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#if FLASK

#include "s_ob.h"
#include "debug.h"
#include "s_thread.h"
#include "s_mapping.h"
#include "avc.h"


/*
 * Currently, we exclude kernel and security threads from examination for
 * revocation.  At the very least, we must be sure to avoid stopping the
 * current thread (the kserver thread which is performing the avc revoke or
 * reset operation) and we must be sure to avoid stopping the security server
 * threads since they must handle security faults generated during a reset.
 * This is bad, since such threads could very well be in a connection which
 * needs to be revoked.
 */
#define THREAD_REVOKE_CONDITION(ob) \
	((ob)->sid != SECINITSID_KERNEL && \
	 (ob)->sid != SECINITSID_SECURITY)

/*
 * Lock, stop and capture all threads, placing them back into the stopped
 * state so that they will be ready for examination during a revoke or reset.
 */
static int
stop_all_threads(s_ob_t * ob, void *args)
{
    int             rc;

    assert(ob->type == FLUKE_THREAD);
    if (THREAD_REVOKE_CONDITION(ob)) {
	struct s_thread *thread = (s_thread_t *) ob;
	s_thread_lock(thread);
	rc = s_thread_stop_th(thread);
	if (rc)
	    return rc;
	thread_wait_other(NULL, WAIT_STOPPED, thread);
	s_thread_unlock(thread);
    }
    return 0;
}

int
s_ob_stop_all_threads(void)
{
    return s_ob_map(1 << FLUKE_THREAD, stop_all_threads, 0);
}


/*
 * Release all threads which are waiting in the stopped state due to a revoke
 * or reset.
 */
static int
release_all_threads(s_ob_t * ob, void *args)
{
    int             rc;

    assert(ob->type == FLUKE_THREAD);
    if (THREAD_REVOKE_CONDITION(ob)) {
	struct s_thread *thread = (s_thread_t *) ob;
	s_thread_lock(thread);
	rc = s_thread_stop_th(thread);
	assert(rc == 0);
	s_thread_release(thread);
    }
    return 0;
}

int
s_ob_release_all_threads(void)
{
    return s_ob_map(1 << FLUKE_THREAD, release_all_threads, 0);
}


#define IPC_REVOKE_PERMS(x) ((x) & (SUBJECT_SUBJECT__CONNECT | \
					 SUBJECT_SUBJECT__CALL | \
					 SUBJECT_SUBJECT__SPECIFY_CLIENT | \
					 SUBJECT_SUBJECT__SPECIFY_SERVER))

typedef struct revoke_args {
    security_class_t sclass;
    security_id_t   ssid;
    security_class_t tclass;
    security_id_t   tsid;
    access_vector_t revoke;
    oskit_bool_t    try_remove;
    access_vector_t *retained;
}               revoke_args_t;

static int
revokeit(s_ob_t * ob, void *args)
{
    revoke_args_t  *rargs = (revoke_args_t *) args;
    security_class_t sclass = rargs->sclass;
    security_id_t   ssid = rargs->ssid;
    security_class_t tclass = rargs->tclass;
    security_id_t   tsid = rargs->tsid;
    access_vector_t revoke = rargs->revoke;
    oskit_bool_t    try_remove = rargs->try_remove;
    access_vector_t *retained = rargs->retained;
    int             rc = 0;

    assert(sclass == SECCLASS_SUBJECT);
    assert(tclass == SECCLASS_SUBJECT);

    if (try_remove && (*retained == revoke))
	/* no need to continue searching */
	return 0;


    if ((revoke & FLUKE_PROT_ALL) &&
	ob->type == FLUKE_TASK && ob->sid == ssid)
	s_task_segment_revoke((s_task_t *) ob, tsid, revoke, try_remove, retained);
    if ((revoke & SUBJECT_SUBJECT__TASK_KEEPER) &&
	ob->type == FLUKE_TASK && ob->sid == ssid)
	s_task_keeper_revoke((s_task_t *) ob, tsid, try_remove, retained);

    if (ob->type == FLUKE_THREAD &&
	THREAD_REVOKE_CONDITION(ob) &&
	(IPC_REVOKE_PERMS(revoke) ||
	 ((revoke & SUBJECT_SUBJECT__THREAD_SCHEDULER) &&
	  ob->sid == ssid))) {
	struct s_thread *thread = (s_thread_t *) ob;

	s_thread_lock(thread);
	rc = s_thread_stop_th(thread);
	assert(rc == 0);	/* because of s_ob_stop_all_threads */

	if (IPC_REVOKE_PERMS(revoke))
	    thread_ipc_revoke(thread, ssid, tsid, revoke, try_remove, retained);

	if ((revoke & SUBJECT_SUBJECT__THREAD_SCHEDULER) &&
	    ob->type == FLUKE_THREAD && ob->sid == ssid)
	    s_thread_scheduler_revoke(thread, tsid, try_remove, retained);

	thread_wait_other(NULL, WAIT_STOPPED, thread);
	s_thread_unlock(thread);
    }
    if ((revoke & SUBJECT_SUBJECT__MAP) &&
	ob->type == FLUKE_MAPPING && ob->sid == ssid)
	s_mapping_region_revoke((s_mapping_t *) ob, tsid, try_remove, retained);



    return rc;
}


int
s_ob_revoke(security_class_t sclass,
	    security_id_t ssid,
	    security_class_t tclass,
	    security_id_t tsid,
	    access_vector_t revoke)
{
    revoke_args_t   rargs;

    rargs.sclass = sclass;
    rargs.ssid = ssid;
    rargs.tclass = tclass;
    rargs.tsid = tsid;
    rargs.revoke = revoke;
    rargs.try_remove = FALSE;
    rargs.retained = 0;
    return s_ob_map(~0, revokeit, &rargs);
}


int
s_ob_try_remove(security_class_t sclass,
		security_id_t ssid,
		security_class_t tclass,
		security_id_t tsid,
		access_vector_t revoke,
		access_vector_t * retained)
{
    revoke_args_t   rargs;

    rargs.sclass = sclass;
    rargs.ssid = ssid;
    rargs.tclass = tclass;
    rargs.tsid = tsid;
    rargs.revoke = revoke;
    rargs.try_remove = TRUE;
    rargs.retained = retained;
    return s_ob_map(~0, revokeit, &rargs);
}


static int
resetit(s_ob_t * ob, void *args)
{
    int             rc = 0;


    switch (ob->type) {
    case FLUKE_TASK:
	rc = s_task_keeper_segment_reset((s_task_t *) ob);
	break;
    case FLUKE_THREAD:
	{
	    struct s_thread *thread = (s_thread_t *) ob;
	    oskit_bool_t    reset = FALSE;

	    if (THREAD_REVOKE_CONDITION(ob)) {
		s_thread_lock(thread);
		rc = s_thread_stop_th(thread);
		assert(rc == 0);/* because of s_ob_stop_all_threads */

		rc = thread_ipc_reset(thread, &reset);
		if (!rc)
		    rc = s_thread_scheduler_reset(thread, &reset);

		thread_wait_other(NULL, WAIT_STOPPED, thread);
		s_thread_unlock(thread);
	    }
	}
	break;
    case FLUKE_MAPPING:
	rc = s_mapping_region_reset((s_mapping_t *) ob);
	break;
    }

    return rc;
}


int
s_ob_reset(void)
{
    return s_ob_map(~0, resetit, 0);
}
#endif
