/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include <fluke/port.h>
#include <fluke/machine/syscall.h>

#include "exception.h"
#include "s_port.h"
#include "s_thread.h"
#include "debug.h"

#if FLASK
#include "avc.h"
#include "avc_ipc.h"
#endif 

int s_port_create(oskit_addr_t uva, fluke_hash_t hash, s_port_t **out_port)
{
	s_port_t *port;
	int err;

	err = task_ob_create(current_thread(), FLUKE_PORT, uva, hash,
			     (s_ob_t **)&port);
	if (err)
		return err;

	link_init(&port->pset_link);

	*out_port = port;
	return (0);
}


/*
 * Disconnect a port from any port set it is on,
 * and stir the port set to make sure no client threads calling this port
 * remain waiting on the old port set.
 * Port set is locked on entry and exit.
 * On exit, the port's pset link is uninitialized.
 */
static void s_port_disconnect(s_port_t *port)
{
	s_pset_t *pset;

	assert_s_ob_locked(&port->sob);
	assert(port->sob.type == FLUKE_PORT);

	pset = (s_pset_t*)link_consume(&port->pset_link);
	link_init(&port->pset_link);
	if (pset)
	{
		assert(pset->sob.type == FLUKE_PSET);
		s_pset_cancel_clients(pset);
		s_pset_unlock(pset);
	}
}

void s_port_destroy(s_port_t *port)
{
	assert(port && port->sob.active);
	assert_s_ob_locked(&port->sob);
	assert(port->sob.type == FLUKE_PORT);

	s_port_disconnect(port);

	s_ob_destroy(&port->sob);
}

int s_port_get_state(s_port_t *port, fluke_port_state_t *state,
		     link_t *pset_link)
{
	assert_s_ob_locked(&port->sob);
	assert(port->sob.type == FLUKE_PORT);

	if (state) {
		state->hash = s_ob_gethash(&port->sob);
		state->alias = port->alias;
	}

	if (pset_link)
		link_create_copy(&port->pset_link, pset_link);

	return 0;
}

#include "config_user_sched.h"

int s_port_set_state(s_port_t *port, fluke_port_state_t *state,
		     link_t *pset_link)
{
	assert_s_ob_locked(&port->sob);
	assert(port->sob.type == FLUKE_PORT);

	if (state)
		port->alias = state->alias;

	if (pset_link)
	{
		s_port_disconnect(port);
		link_create_copy(pset_link, &port->pset_link);
	}
#ifdef CONFIG_USER_SCHED
	else if (state) {
		/*
		 * The pset wasn't changed, but the alias was.
		 * If this is some thread's scheduler port,
		 * it may have cached the alias value before
		 * blocking on the pset.  Any such cached values
		 * are now invalid, so we have to wake up all clients
		 * blocked on the pset so any of them in scheduler
		 * wait will recache the new alias before reblocking.
		 */
		s_pset_t *pset = (s_pset_t*)link_follow(&port->pset_link);
		if (pset) {
			s_pset_cancel_clients(pset);
			s_pset_unlock(pset);
		}
	}
#endif

	return 0;
}

int s_port_capture_locked(
	struct s_port		*port,
	struct s_thread		**out_server)
{
	s_pset_t *pset;
	fluke_alias_t alias;
	int rc;


	assert_s_ob_locked(&port->sob);
	assert(port->sob.type == FLUKE_PORT);

	/* Find and lock the port set this port is connected to.  */
	pset = (s_pset_t*)link_follow(&port->pset_link);

	alias = port->alias;
	s_port_unlock(port);

	/* If it's not connected to any port, it's disabled.  */
	if (pset == 0) {
		*out_server = 0;
		return 0;
	}

	assert(pset->sob.type == FLUKE_PSET);

	/* Try to establish a connection through that port set.  */
	rc = s_pset_capture(pset, out_server);
	if (rc)
		return rc;

	/* Provide the port alias to the server thread. */
	EXC_SYSPARAM(*out_server, ipc, port_alias) = alias;

	return 0;
}

#if FLASK
int s_port_capture(s_thread_t 		*client, 
		   access_vector_t	permission,
		   security_id_t	client_sid,
		   s_thread_t 		**out_server)
#else
int s_port_capture(s_thread_t *client, s_thread_t **out_server)
#endif  
{
	s_port_t *port;
	int rc;
#if FLASK
        s_ref_t *ref;
#endif

	/*
	 * Look up the destination port reference
	 * and follow it to find a server thread.
	 */
#if FLASK
       	rc = task_ob_lookup_type(client, IPC_CONNECT_DEST(client), 
				 FLUKE_PROT_READ, FLUKE_REF,
                                (s_ob_t**)&ref);
       	if (rc)
               	return rc;
 
       	/* XXX major locking order violation? */
       	port = (s_port_t*)s_ref_follow(ref);
 
       	if (port && (port->sob.type != FLUKE_PORT))
       	{
               	s_ob_unlock((s_ob_t*)port);
               	port = 0;
       	}
#else
	rc = task_ob_lookup_ref_type(client, IPC_CONNECT_DEST(client),
				     FLUKE_PORT, (s_ob_t**)&port);
	if (rc)
		return rc;
#endif
	if (port == 0)
	{
		/* The reference is dead or is not a port reference.  */
		IPC_STATUS(client) = FLUKE_IPC_CONNECT_INVALID_DEST;
		EXC_SET_IP(client, fluke_nop__entry);
#if FLASK
		s_ref_unlock(ref);
#endif
		return KR_RESTART;
	}

#if FLASK
	assert(client_sid);
	if (client_sid != client->sob.sid)
	{
	    rc = avc_has_permission_impersonate(client->sob.sid,
					        client_sid,
					        SUBJECT_SUBJECT__SPECIFY_CLIENT,
						client,
					        FLUKE_IPC_CONNECT_PERMISSION_DENIED);
	    if (rc)
	    {
		*out_server = 0;
		s_port_unlock(port);
        	s_ref_unlock(ref);
		return rc;
	    }	    
	}

	rc = avc_has_permission_ipc(ref, client_sid, port->sob.sid,
				    permission, client,
				    FLUKE_IPC_CONNECT_PERMISSION_DENIED);

	if (rc)
	{
	    *out_server = 0;
	    s_port_unlock(port);
	    return rc;
	}
#endif 
	rc = s_port_capture_locked(port, out_server);
	if (rc)
		return rc;

	if (*out_server == 0) {
		IPC_STATUS(client) = FLUKE_IPC_CONNECT_INVALID_DEST;
		EXC_SET_IP(client, fluke_nop__entry);
		return KR_RESTART;
	}

	return 0;
}

#ifdef DEBUG
void
s_port_dump(s_port_t *port, int indent)
{
	int i;

	for (i = 0; i < indent; i++) putchar(' ');
	printf("alias=%#x, pset=%p\n", port->alias, port->pset_link.target);
}
#endif
