/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include <fluke/pset.h>
#include <fluke/exception.h>

#include "cond.h"
#include "exception.h"
#include "ipc.h"
#include "s_pset.h"
#include "s_thread.h"
#include "debug.h"
#if FLASK
#include "avc.h"
#include "avc_ipc.h"
#endif 

int s_pset_create(oskit_addr_t uva, fluke_hash_t hash, s_pset_t **out_pset)
{
	s_pset_t *pset;
	int err;

	err = task_ob_create(current_thread(), FLUKE_PSET, uva, hash,
			     (s_ob_t **)&pset);
	if (err)
		return (err);

	cond_init(&pset->server_cond);
	cond_init(&pset->client_cond);

	*out_pset = pset;
	return (0);
}

int s_pset_destroy(s_pset_t *pset)
{
	assert(pset && pset->sob.active);
	assert(pset->sob.type == FLUKE_PSET);
	assert_s_ob_locked(&pset->sob);

	/*
	 * The user is supposed to ensure that
	 * a port set is not destroyed while server threads are waiting on it.
	 * XXX we need to wake up all waiting client and server threads
	 * when a pset is unmapped.
	 */
	cond_destroy(&pset->server_cond);
	cond_destroy(&pset->client_cond);

	s_ob_destroy(&pset->sob);
	return 0;
}

int s_pset_get_state(s_pset_t *pset, fluke_pset_state_t *state)
{
	assert_s_ob_locked(&pset->sob);
	assert(pset->sob.type == FLUKE_PSET);

	if (state)
		state->hash = s_ob_gethash(&pset->sob);

	return 0;
}

int s_pset_set_state(s_pset_t *pset, fluke_pset_state_t *state)
{
	assert_s_ob_locked(&pset->sob);
	assert(pset->sob.type == FLUKE_PSET);

	if (state)
		;	/* XXX no state other than the hash */

	return 0;
}

void thread_pset_state_init(struct s_thread *thread)
{
	/* nothing to do */
}

void thread_pset_state_destroy(struct s_thread *thread)
{
	/* nothing to do */
}

int s_pset_capture(s_pset_t *pset, s_thread_t **out_server)
{
	s_thread_t *server;

	assert_s_ob_locked(&pset->sob);
	assert(pset->sob.type == FLUKE_PSET);

	/* First see if there's already a server thread available.  */
	server = cond_wakeup(&pset->server_cond);
	if (server)
	{
		s_pset_unlock(pset);
		*out_server = server;
		return 0;
	}

	/* If not, wait ourselves for a server thread to show up. */
	return cond_wait(&pset->client_cond, &pset->sob.lock, 0/*XXX*/);
}

#ifdef CONFIG_INLINING
/* this is a critical path, so it helps to inline this into
   s_pset_wait_receive */
#include "config_user_sched.h"
#ifndef CONFIG_USER_SCHED
static
#endif
inline
#endif
int s_pset_wait_receive_1(
	struct s_thread		*server,
	struct s_pset		**out_pset_locked)
{
	s_pset_t *pset;
	int rc;


	assert(server == current_thread());

	/* Make sure we're not involved in any IPC connection already. */
	rc = ipc_server_disconnect(server);
	if (rc)
		return rc;

	/* Look up the port set to receive from. */
	rc = task_ob_lookup_type(server, server->ipc_state.server_pset_uva,
				 FLUKE_PROT_READ, FLUKE_PSET,
				 (s_ob_t **)&pset);
	if (rc)
		return rc;
	assert(pset != 0);
	assert(pset->sob.type == FLUKE_PSET);
	assert_s_ob_locked(&pset->sob);

#if FLASK
	if (server->sob.sid != pset->sob.sid)
	{
	    rc = avc_has_permission_impersonate(server->sob.sid,
					        pset->sob.sid,
					        SUBJECT_SUBJECT__SPECIFY_SERVER,
					        server,
					        FLUKE_IPC_WAIT_PERMISSION_DENIED);
	    if (rc)
	    {
		s_ob_unlock(&pset->sob);
		return rc;
	    }
	}

	server->ipc_state.effective_server_sid = pset->sob.sid;
#endif 

	*out_pset_locked = pset;
	return 0;
}

int s_pset_wait_receive(
	struct s_thread		*server)
{
	int rc;
	s_pset_t *pset;

	rc = s_pset_wait_receive_1(server, &pset);
	if (rc)
		return rc;

	/*
	 * Wakeup all client threads waiting for service.
	 *
	 * XXX we cannot just signal a single thread here as there is
	 * no guarentee that the signalled thread will actually wakeup
	 * and "connect" with the service thread (e.g., a thread blocked
	 * in an idempotent region keeper request will retry the operation
	 * that caused the page fault and that operation may no longer fault).
	 * This means that any other clients waiting for service may be left
	 * stranded on the queue since the server thread may never re-enter
	 * this routine and signal someone else.
	 *
	 * The problem is that the client is responsible for initiating the
	 * action.  If, instead, the server were to actively seek out a
	 * client at this point, we could just wakeup a single client and
	 * avoid the problem.
	 */
	cond_broadcast(&pset->client_cond);

	/* Go to sleep waiting for a client to come pick us up. */
	return cond_wait(&pset->server_cond, &pset->sob.lock, 0);
}



void s_pset_cancel_clients(s_pset_t *pset)
{
	assert_s_ob_locked(&pset->sob);
	assert(pset->sob.type == FLUKE_PSET);

	cond_broadcast(&pset->client_cond);
}

#ifdef DEBUG
void
s_pset_dump(s_pset_t *pset, int indent)
{
	int i;

	for (i = 0; i < indent; i++) putchar(' ');
	printf("&clientkcond=%p, &serverkcond=%p\n",
	       &pset->client_cond, &pset->server_cond);
}
#endif
