/*
 * Copyright (c) 1995, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#ifndef _FLOSS_KERNEL_FLUKE_S_PSET_H_
#define _FLOSS_KERNEL_FLUKE_S_PSET_H_

#include "cond.h"
#include "s_ob.h"

#include "config_user_sched.h"


/*
 * Supervisor part of user-exported port set objects.
 * Port sets provide the rendezvous point
 * for client and server threads engaging in IPC.
 */
struct s_pset
{
	struct s_ob		sob;

	/* All of the following fields are protected by sob.lock.  */

	/* Server threads waiting for connections on this port set.  */
	cond_t			server_cond;

	/* Client threads waiting for server threads.  */
	cond_t			client_cond;
};
typedef struct s_pset s_pset_t;

#define s_pset_lock(pset)	s_ob_lock(&(pset)->sob)
#define s_pset_unlock(pset)	s_ob_unlock(&(pset)->sob)

struct s_thread;
struct fluke_pset_state;
struct fluke_ipc_params;

/*
 * Create a new port set.
 * If successful, returns with the port set locked.
 */
int s_pset_create(
	oskit_addr_t		uva,
	fluke_hash_t		hash,
	s_pset_t		**out_pset);

/*
 * Attempt to destroy a port set.
 * Note that this routine can fail due to interruptions and such.
 * The port set must be locked on entry, and is unlocked on exit
 * (regardless of whether the port set was successfully destroyed).
 */
int s_pset_destroy(
	s_pset_t *pset);

int s_pset_get_state(
	s_pset_t		*pset,
	struct fluke_pset_state	*state);

int s_pset_set_state(
	s_pset_t		*pset,
	struct fluke_pset_state	*state);

/*
 * Initialize and destroy the per-thread state used by the port set layer.
 */
void thread_pset_state_init(struct s_thread *thread);
void thread_pset_state_destroy(struct s_thread *thread);

/*
 * Capture a server thread waiting on the specified pset.
 * If a server thread isn't waiting yet, we wait for it.
 * The pset must be locked on entry, and is released on return.
 * On successful return, *out_server points to the captured thread.
 */
int s_pset_capture(s_pset_t *pset, struct s_thread **out_server);

/*
 * Wait as a server thread for incoming client connections
 * on the port set specified by its ipc_state.server_pset_uva.
 * If a client thread is already waiting to connect, it is woken up.
 */
int s_pset_wait_receive(
	struct s_thread		*server);

#ifdef CONFIG_USER_SCHED
/*
 * This does all of `s_pset_wait_receive' except for waking up
 * clients waiting to connect and actually waiting.
 * Returns the locked pset to be woken up and waited on.
 */
int s_pset_wait_receive_1(
	struct s_thread		*server,
	struct s_pset		**out_pset_locked);
#endif


/*
 * Wake up any client threads currently waiting on this port set,
 * so that they will back out to user mode and restart from scratch.
 * This is called from s_port.c, for example,
 * when a port changes to point to a different port set,
 * to ensure that client threads waiting on the old port set are moved properly.
 * Port set is locked on entry and exit.
 */
void s_pset_cancel_clients(s_pset_t *pset);

#ifdef DEBUG
void s_pset_dump(s_pset_t *pset, int indent);
#endif

#endif /* _FLOSS_KERNEL_FLUKE_S_PSET_H_ */
