/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include <fluke/ref.h>
#include "s_ref.h"
#include "s_task.h"
#include "s_thread.h"
#include "debug.h"
#ifdef CONFIG_OPT_REF
#include "u_ref.h"
#include <oskit/machine/base_vm.h>
#include <fluke/exception.h>
#endif

/*
 * Create a NULL reference.
 */
int s_ref_create(oskit_addr_t uva)
{
	s_ref_t *ref;
	int err;

	err = task_ob_create(current_thread(), FLUKE_REF, uva, 0 /* XXX hash */,
			     (s_ob_t **)&ref);
	if (err)
		return (err);

	link_init(&ref->link);
#if FLASK
	ref->ae = NULL;
	ref->aever = 0;
#endif
#ifdef CONFIG_OPT_REF
	assert(sizeof(u_ref_t) <= FLUKE_REF_SIZE);
	ref->u_ref = (u_ref_t *) phystokv(ref->sob.u_pa);
	ref->u_ref->hash = 0;
	ref->u_ref->handle = 0;
#endif

	s_ob_unlock(&ref->sob);
	return (0);
}


void s_ref_destroy(s_ref_t *ref)
{
	assert(ref && ref->sob.active);
	assert(ref->sob.type == FLUKE_REF);
	assert_s_ob_locked(&ref->sob);

	link_destroy(&ref->link);

	s_ob_destroy(&ref->sob);
}

/*
 * Internal routines
 */

#ifndef CONFIG_INLINING
/*
 * Return the object pointed to by a reference.
 * Returns the referenced object locked.
 */
struct s_ob *s_ref_follow(s_ref_t *ref)
{
	assert_s_ob_locked(&ref->sob);
	assert(ref->sob.type == FLUKE_REF);

	return (link_follow(&ref->link));
}

void
s_ref_create_link(s_ref_t *ref, link_t *link)
{
	assert_s_ob_locked(&ref->sob);
	assert(ref->sob.type == FLUKE_REF);

	link_create_copy(&ref->link, link);
}
#endif

/*
 * Associate a reference object with a target object.
 * Any old association is first severed.
 * Target object should be locked.
 */
void s_ref_associate(s_ref_t *ref, s_ob_t *target)
{
	assert_s_ob_locked(&ref->sob);
	assert(ref->sob.type == FLUKE_REF);

	if (target) {
		assert_s_ob_locked(target);
		link_destroy(&ref->link);
		link_produce(&ref->link, target);
#ifdef CONFIG_OPT_REF
		ref->u_ref->hash = target->u_hash;
		ref->u_ref->handle = REF_HANDLE(target);
#endif
	}
	else {
		link_reinit(&ref->link);
#ifdef CONFIG_OPT_REF
		ref->u_ref->hash = 0;
		ref->u_ref->handle = 0;
#endif
	}

#if FLASK
        ref->ae = NULL;
        ref->aever = 0;
#endif
}

#ifdef CONFIG_OPT_REF
void
s_ref_sanify_flob(struct s_ob *ref_sob)
{
	s_ref_t *const ref = (s_ref_t *)ref_sob;
	s_ob_t *target;

	s_ref_lock(ref);
	target = s_ref_follow(ref);

	if (target == 0) {
		ref->u_ref->hash = 0;
		ref->u_ref->handle = 0;
	}
	else {
		ref->u_ref->hash = target->u_hash;
		ref->u_ref->handle = REF_HANDLE(target);
		s_ob_unlock(target);
	}

	s_ref_unlock(ref);
}
#endif

void
s_ref_link_transfer(link_t *srclink, s_ref_t *ref)
{
	s_ob_t *target = link_follow(srclink);
	s_ref_associate(ref, target);
	if (target)
		s_ob_unlock(target);
}


int
s_ref_check(s_ref_t *ref, int *rc)
{
#ifdef CONFIG_OPT_REF
	u_ref_t *const u = ref->u_ref;
#endif
	link_t link;

	assert_s_ob_locked(&ref->sob);
	assert(ref->sob.type == FLUKE_REF);

	link_create_copy(&ref->link, &link);
	s_ref_unlock(ref);

	if (link.target == 0) {
#if defined CONFIG_OPT_REF && defined CONFIG_FLOB_CHECK
		/*
		 * This ref was already null.
		 * Its user space stuff should be fine.
		 */
		if (u->hash != 0 || u->handle != 0)
			EXC_RETURN_INSANITY(INVALID_OBJECT);
#endif
		*rc = 0;
	}
	else {
		struct s_ob *sob;
		sob = link_consume(&link);
		if (sob) {
			/*
			 * The referneced object is live now.
			 * It might die as soon as we unlock it,
			 * but that is "after" our syscall completes.
			 */
#if defined CONFIG_OPT_REF && defined CONFIG_FLOB_CHECK
			const fluke_hash_t sob_hash = sob->u_hash;
#endif
			s_ob_unlock(sob);

#ifdef CONFIG_OPT_REF
			/*
			 * The ref is live.
			 * Its user space stuff should be fine.
			 */
#ifdef CONFIG_FLOB_CHECK
			if (u->handle != REF_HANDLE(sob) ||
			    u->hash != sob_hash)
				EXC_RETURN_INSANITY(INVALID_OBJECT);
#endif
#endif
			*rc = 1;
		}
		else {
#ifdef CONFIG_OPT_REF
			/*
			 * The ref just went stale.
			 * Clear its user-space stuff to indicate a null ref.
			 */
			u->hash = 0;
			u->handle = 0;
#endif
			*rc = 0;
		}
	}

	return 0;
}


#ifdef DEBUG
/*
 * Clean up refs to dead objects.
 */
int
s_ref_scan(s_ob_t *ob, void *arg)
{
	struct s_ob *sobp;

	sobp = link_follow(&((s_ref_t *)ob)->link);
	if (sobp)
		s_ob_unlock(sobp);
	return 0;
}

void
s_ref_dump(s_ref_t *ref, int indent)
{
	extern char *fluke_type_name();
	struct s_ob *sobp;
	int i;

	for (i = 0; i < indent; i++) putchar(' ');
	sobp = link_follow(&ref->link);
	if (sobp) {
		s_ob_unlock(sobp);
		printf("reference to %s=%p\n",
		       fluke_type_name(sobp->type), sobp);
#if 0
		s_debug_dump_flob(sobp, indent+1);
#endif
	} else
		printf("NULL reference\n");
}
#endif
