/*
 * Copyright (c) 1996 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Fluke region object implementation.
 * Regions define a contiguous range of memory in a task that can be exported
 * to others tasks.
 */

#include <fluke/vm.h>
#include <fluke/region.h>
#include <fluke/exception.h>
#include <oskit/page.h>

#include "s_region.h"
#include "s_mapping.h"
#include "s_task.h"
#include "s_thread.h"
#include "debug.h"

/* XXX move these elsewhere */
#define min(a, b) ((a) <= (b) ? (a) : (b))
#define max(a, b) ((a) < (b) ? (b) : (a))

/*
 * Create an ``empty'' region.
 * Task and keeper references are NULL, so the region has no effect.
 * New region object is returned locked.
 */
int s_region_create(oskit_addr_t uva, fluke_hash_t hash, s_region_t **out_region)
{
	s_region_t *region;
	int err;

	err = task_ob_create(current_thread(), FLUKE_REGION, uva, hash,
			     (s_ob_t **)&region);
	if (err)
		return (err);

	link_init(&region->task);
	link_init(&region->keeper);
	region->start_va = 0;
	region->size = 0;
	region->prot = FLUKE_PROT_NONE;

	mutex_init(&region->ulock);
	queue_init(&region->mappings);

	*out_region = region;
	return (0);
}


/*
 * Destroy the indicated region.
 * Region object should be locked.
 *
 * This has the additional effect of invalidating any VtoP translations
 * covered by this region or descendent regions.
 */
void s_region_destroy(s_region_t *region)
{
	s_task_t *task;
	s_mapping_t *mapping;

	assert(region && region->sob.active);
	assert(region->sob.type == FLUKE_REGION);
	assert_s_ob_locked(&region->sob);

	/*
	 * Remove mapping from task list.
	 * This will invalidate descendent translations.
	 */
	task = (s_task_t *)link_consume(&region->task);
	if (task) {
		s_task_u_lock(task);
		s_region_remove_task(region, task);
		s_task_u_unlock(task);
		s_task_unlock(task);
	}

	/*
	 * Disassociate region from all mappings.
	 */
	s_region_u_lock(region);
	while (!queue_empty(&region->mappings)) {
		mapping = (s_mapping_t *)queue_first(&region->mappings);
		queue_remove(&region->mappings, mapping, s_mapping_t *, rlink);
	}
	s_region_u_unlock(region);

	link_destroy(&region->keeper);
	s_ob_destroy(&region->sob);
}

/*
 * Return the state of a region.
 * Region object should be locked on call and will be locked on return.
 */
int s_region_get_state(s_region_t *region, fluke_region_state_t *state,
		       link_t *task_link, link_t *keeper_link)
{
	assert(region && region->sob.active);
	assert(region->sob.type == FLUKE_REGION);
	assert_s_ob_locked(&region->sob);

	if (task_link)
		link_create_copy(&region->task, task_link);

	if (keeper_link)
		link_create_copy(&region->keeper, keeper_link);

	if (state) {
		state->hash = s_ob_gethash(&region->sob);
		state->start_va = region->start_va;
		state->size = region->size;
		state->prot = region->prot;
	}

	return (0);
}

/*
 * Set the state of a region.
 * Region object must be locked and is returned locked.
 */
int s_region_set_state(s_region_t *region, fluke_region_state_t *state,
		       link_t *task_link, link_t *keeper_link)
{
	s_task_t *task;
	oskit_bool_t flushed = FALSE;

	assert(region && region->sob.active);
	assert(region->sob.type == FLUKE_REGION);
	assert_s_ob_locked(&region->sob);

	/*
	 * Trivial state checks.
	 * Address must be page aligned and size a page multiple.
	 * Covered range must be within the AS bounds of a task.
	 */
	if (state) {
		if (!page_aligned(state->start_va) ||
		    trunc_page(state->size) != state->size)
			EXC_RETURN_INSANITY(NOT_ALIGNED);

		if (state->start_va < FLUKE_VM_MIN ||
		    state->start_va + state->size > FLUKE_VM_MAX)
			EXC_RETURN_INSANITY(INVALID_ADDRESS);
	}

	/*
	 * Handle task reference
	 * If the specified task is not the same as the current task
	 * unlink the region from that task, flush all translations
	 * for the region, and insert on the new task list.
	 */
	if (task_link && !link_compare(&region->task, task_link)) {
		task = (s_task_t *)link_consume(&region->task);
		link_init(&region->task);
		if (task) {
			s_task_u_lock(task);
			s_region_remove_task(region, task);
			s_task_u_unlock(task);
			s_task_unlock(task);
			flushed = TRUE;
		}
		task = (s_task_t *)link_follow(task_link);
		if (task) {
			s_task_u_lock(task);
			s_region_insert_task(region, task);
			s_task_u_unlock(task);
			link_produce(&region->task, &task->sob);
			s_task_unlock(task);
		}
	}

	/*
	 * Before modifying any other state fields, take the unmap lock
	 * to avoid races with unmap path.
	 */
	s_region_u_lock(region);

	/*
	 * Handle changes in start or size of region.
	 */
	if (state && (region->start_va != state->start_va ||
		      region->size != state->size)) {
		/*
		 * Flush all translations if it hasn't already been done.
		 *
		 * XXX there are a number of cases where this is overkill,
		 * for example:
		 *	- if the region is just growing, don't need to flush
		 *	- if the region has shrunk, only flush the excess
		 * Basically just need to flush the parts that don't overlap.
		 * If this is ever changed, watch out for the protection change
		 * case below.
		 */
		if (!flushed) {
			s_region_restrict(region, 0, region->size,
					  FLUKE_PROT_NONE);
			flushed = TRUE;
		}

		region->start_va = state->start_va;
		region->size = state->size;
	}

	/*
	 * Handle protection change.
	 */
	if (state && region->prot != state->prot) {
		/*
		 * If we have not already flushed everything and the protection
		 * is being reduced, flush all translations now.
		 */
		if (!flushed && (region->prot | state->prot) == region->prot)
			s_region_restrict(region, 0, region->size, state->prot);

		region->prot = state->prot;
	}

	s_region_u_unlock(region);

	/*
	 * Handle keeper.
	 * Note that we need to make sure that the task, range and protection
	 * changes don't cause any faults until the keeper has been changed.
	 * Otherwise an old keeper might get faults for the new state.  Since
	 * the fault path must acquire the region lock before it could do that,
	 * we are safe.
	 */
	if (keeper_link && !link_compare(&region->keeper, keeper_link))
		link_transfer(keeper_link, &region->keeper);

	return (0);
}

/*
 * Change the protection on a region.
 * This routine may either increase or decrease the protection
 * and affects all mappings and regions descended from this region.
 * Region object should be locked on call.
 */
void s_region_protect(s_region_t *region, fluke_prot_t prot)
{
	assert(region && region->sob.active);
	assert(region->sob.type == FLUKE_REGION);
	assert_s_ob_locked(&region->sob);
	assert((prot & FLUKE_PROT_ALL) == prot);

	s_region_u_lock(region);
	if (region->prot != prot) {
		/*
		 * Further restricting access, must do it now.
		 * Modify VtoP translations for every mapping descended from us.
		 */
		if ((region->prot | prot) == region->prot)
			s_region_restrict(region, 0, region->size, prot);

		/*
		 * If we are relaxing, let it happen on demand.
		 * In order to eagerly determine how much to relax the actual
		 * VtoP translation protections, we would have to trace down
		 * from the root to determine the maximum allowable value.
		 */
		else
			;

		region->prot = prot;
	}
	s_region_u_unlock(region);
}

/*
 * Remove all VtoP translations for all mappings derived from the indicated
 * area of the region.
 * Region unmap lock should be held on call and return.
 */
void s_region_restrict(s_region_t *region, oskit_addr_t offset, oskit_size_t size,
		       fluke_prot_t prot)
{
	s_mapping_t *mapping;
	oskit_addr_t moffset, msva, meva, rsva, reva;
	oskit_size_t msize;

	assert_mutex_locked(&region->ulock);

	/*
	 * If the region allows no stronger access than indicated,
	 * then none of the frags descended from mappings of ours can
	 * either, so there is no need to check them.
	 */
	if ((region->prot & prot) == region->prot)
		return;

	/*
	 * For every mapping over this region range,
	 * flush the relevant portion.
	 */
	rsva = region->start_va + offset;
	reva = rsva + size;
	assert(reva <= region->start_va + region->size);

	queue_iterate(&region->mappings, mapping, s_mapping_t *, rlink) {
		s_mapping_u_lock(mapping);

		/*
		 * These two ``mapping addresses'' are actually addresses
		 * in the region's task space from which the mapping is
		 * derived.
		 */
		msva = region->start_va + mapping->region_off;
		meva = msva + mapping->size;

		if (msva < reva && meva > rsva) {
			moffset = max(msva, rsva);
			msize = min(meva, reva) - moffset;
			moffset -= msva;
			s_mapping_restrict(mapping, moffset, msize, prot);
		}

		s_mapping_u_unlock(mapping);
	}
}

/*
 * Insert a region into a task's region list.
 * Task unmap lock should be held when called.
 */
void s_region_insert_task(s_region_t *region, s_task_t *task)
{
	assert(task);
	assert_mutex_locked(&task->ulock);
	assert(region);

	queue_enter(&task->regions, region, s_region_t *, tlink);
	task->version++;	/* XXX hack around race in s_mapping_restrict */
}

/*
 * Remove a region from a task's region list.
 * Task unmap lock should be held on call and return.
 */
void s_region_remove_task(s_region_t *region, s_task_t *task)
{
	assert(task);
	assert_mutex_locked(&task->ulock);
	assert(region);

	queue_remove(&task->regions, region, s_region_t *, tlink);
	task->version++;	/* XXX hack around race in s_mapping_restrict */

	s_region_u_lock(region);
	s_region_restrict(region, 0, region->size, FLUKE_PROT_NONE);
	s_region_u_unlock(region);
}

/*
 * Convert a region offset and size into task start and end virtual addresses.
 * Region lock should be held.
 */
int s_region_range(s_region_t *region, oskit_addr_t offset, oskit_size_t size,
		   oskit_addr_t *saddrp, oskit_addr_t *eaddrp)
{
	if (offset < 0 || offset + size > region->size)
		EXC_RETURN_INSANITY(INVALID_RANGE);

#if 0
/* XXX not really practical when repeated search call must be made */
	if (!page_aligned(offset) || trunc_page(size) != size)
		EXC_RETURN_INSANITY(NOT_ALIGNED);
#endif

	*saddrp = region->start_va + offset;
	*eaddrp = *saddrp + size;

	return (0);
}

#ifdef DEBUG
void
s_region_dump(s_region_t *reg, int indent)
{
	int i;

	for (i = 0; i < indent; i++) putchar(' ');
	printf("R%p: task=%p, keeper=%p, start_va=0x%08x, size=0x%08x, prot=0x%08x(%c%c%c%c)\n",
	       reg,
	       reg->task.target,
	       reg->keeper.target,
	       reg->start_va,
	       reg->size,
	       reg->prot,
	       (reg->prot&8)?'C':'-',
	       (reg->prot&4)?'r':'-',
	       (reg->prot&2)?'w':'-',
	       (reg->prot&1)?'x':'-');
	printf("\n");
#if 0
	for (i = 0; i < indent; i++) putchar(' ');
	s_debug_dump_flob(reg->task.target, indent+1);
#endif
}
#endif
