/*
 * Copyright (c) 1995, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#ifndef _FLOSS_KERNEL_FLUKE_S_TASK_H_
#define _FLOSS_KERNEL_FLUKE_S_TASK_H_

#include <oskit/queue.h>
#include "config_user_gprof.h"

#include "pmap.h"
#include "s_ob.h"
#include "cond.h"

#define OBTAB_SIZE		32	/* size of object lookup hash table */
#if FLUKE_MINOB_SHIFT != FLUKE_MAXOB_SHIFT
#define OBTAB_HASH(uva)	\
	((((uva)>>FLUKE_MINOB_SHIFT) ^ ((uva)>>FLUKE_MAXOB_SHIFT) ^ \
		((uva)>>PAGE_SHIFT)) & (OBTAB_SIZE-1))
#else
#define OBTAB_HASH(uva)	\
	((((uva)>>FLUKE_MINOB_SHIFT) ^ ((uva)>>PAGE_SHIFT)) & (OBTAB_SIZE-1))
#endif

/*
 * Entry in a task's object lookup cache.
 * Every user address direct-maps to an obtab_entry using a hash function;
 * a lookup hits the cached when the supplied address matches the entry's.
 */
struct obtab_entry
{
	oskit_addr_t		uva;	  /* address of object in task */
	fluke_prot_t		prot;	  /* fragment's VM protection */
	struct link		link;	  /* direct link to object */
};

struct s_task
{
	/* object layer, protected by object lock */
	struct s_ob		sob;	  /* supervisor object (must be 1st) */
	struct link		keeper;	  /* task's default pager port */
	queue_head_t		threads;  /* threads running in the task */
	struct cond		cond;	  /* used to synchronize with threads */
#if FLASK
	security_id_t		segment_sid; /* only if a segment task */
#endif 

	/* fault layer, uses object lock as well */
	struct s_mapping	*mhint;	  /* mapping lookup hint */
	queue_head_t		mappings; /* list of mappings forming our AS */

	/* unmap layer */
	struct mutex		ulock;	  /* lock */
	int			version;  /* XXX ugh, hack around race */
	queue_head_t		regions;  /* list of regions over our AS */
	struct obtab_entry	obtab[OBTAB_SIZE];/* object lookup hash table */

	/* Per-task state maintained by the pmap layer
	   (e.g. processor page table information).  */
	struct task_pmap_state	pmap_state;  /* PMAP info (e.g. page tables) */

#ifdef IOPL0
	/* XXX: x86-specific */
	/* unoptimized: sparse array copied on CSW */
	char	bitmask[65536/8 + 1];
#endif

#ifdef CONFIG_USER_GPROF
	int p_flag;	/* enabled? */
	struct uprof *p_prof;	/* profile info */
#endif
};
typedef struct s_task s_task_t;

/*
 * Per-thread state maintained by the task layer.
 */
struct task_state
{
	/*
	 * Pointer to the task this thread is running in, if any.
	 * If non-NULL, implies that we are on that task's threads list.
	 * Can only be modified by the thread itself,
	 * while holding the task's main object lock.
	 */
	s_task_t		*task;

	/*
	 * Chain on our task's threads list.
	 * Protected by the task's main lock.
	 */
	queue_chain_t		task_chain;

	/*
	 * When someone is trying to destroy a task,
	 * it sets the task_dying flags in all threads running in the task,
	 * and cancels all of those threads.
	 * When the threads wake up and notice this flag,
	 * they stop and remove themselves from the task's list.
	 */
	oskit_bool_t		task_dying;

	/*
	 * When a thread is stopped,
	 * this holds its task association instead of 'task'.
	 */
	link_t			task_link;
};


/* XXX for now */
#define s_task_lock(t)		s_ob_lock(&(t)->sob)
#define s_task_unlock(t)	s_ob_unlock(&(t)->sob)
#define s_task_f_lock(t)	s_ob_lock(&(t)->sob)
#define s_task_f_unlock(t)	s_ob_unlock(&(t)->sob)
#define s_task_u_lock(t)	mutex_lock(&(t)->ulock);
#define s_task_u_unlock(t)	mutex_unlock(&(t)->ulock);

#define assert_s_task_locked(t)	assert_s_ob_locked(&(t)->sob)

struct s_ref;
struct s_region;
struct fluke_task_state;
struct fragment;

int s_task_create(
	oskit_addr_t			uva,
	fluke_hash_t			hash,
	s_task_t			**out_task);

int s_task_destroy(
	s_task_t			*task);

int s_task_get_state(
	struct s_task			*task,
	struct fluke_task_state		*state,
	struct link			*keeper_link);

int s_task_set_state(
	struct s_task			*task,
	struct fluke_task_state		*state,
	struct link			*keeper_link);

/*
 * Initialize the per-thread state maintained by the task layer for a thread.
 * Thread must be locked throughout.
 * Initially, the thread isn't associated with any task.
 */
void thread_task_state_init(
	struct s_thread			*thread);

/*
 * Destroy the per-thread state maintained by the task layer for a thread.
 * Thread must be locked throughout.
 */
void thread_task_state_destroy(
	struct s_thread			*thread);

void thread_task_state_start(
	struct s_thread			*thread);

void thread_task_state_stop(
	struct s_thread			*thread);

/*
 * Get the specified thread's task association.
 * If it is currently attached to a task,
 * returns a link for that task.
 * Otherwise, returns a NULL link.
 * The target thread must be stopped and locked.
 */
void thread_task_state_get(
	struct s_thread			*thread,
	struct link			*task_link);

/*
 * Set the specified thread's task association
 * to refer to the task indicated by the given link.
 * Assumes that the thread is not currently associated with any task.
 * The target thread must be stopped and locked.
 */
int thread_task_state_set(
	struct s_thread			*thread,
	struct link			*task_link);

/*
 * This is called from the kernel entry layer
 * after a thread is cancelled and rolled back to its user mode entrypoint.
 * It checks to see if the thread's task state is stale, and if so, fixes it.
 */
void thread_task_canceled(
	struct s_thread			*cur_thread);

/*
 * Look up a virtual address in a thread's task address space.
 * Memory should allow the indicated access.  If successful, returns a
 * frag describing the physical memory with the frag's mapping locked.
 */
int task_lookup(
	struct s_thread			*thread,
	struct s_task			*task,
	oskit_addr_t			uva,
	fluke_prot_t			prot,
	struct fragment			**fragp);

/*
 * Called by the various s_foo_create routines to verify the user part
 * of an object and to create the supervisor part.
 * Returns the supervisor object locked.
 */
int task_ob_create(
	struct s_thread			*thread,
	fluke_type_t			type,
	oskit_addr_t			uva,
	fluke_hash_t			hash,
	s_ob_t				**out_sob);

/*
 * Lookup an object at the indicated user address with the intention of
 * destroying it.  If the object is of the correct type, we flush it from
 * the task object cache and return a pointer to the locked supervisor object.
 */
int task_ob_destroy(
	struct s_thread			*thread,
	oskit_addr_t			uva,
	fluke_type_t			type,
	struct s_ob			**out_sob);

/*
 * Lookup a supervisor object given a user virtual address, ensuring that
 * the specified thread has the indicated permission to that memory.
 * Normally 'thread' should be the current thread;
 * however, in some situations it can also be another thread
 * (e.g. during IPC when we're looking up refs in a remote task).
 * In that case, the target thread must currently be 'passivated' somehow
 * as described in xfer.h.
 * Returns the s_ob locked.
 */
int task_ob_lookup(
	struct s_thread			*thread,
	oskit_addr_t			uva,
	fluke_prot_t			prot,
	int				cache,
	struct s_ob			**out_sob);

/*
 * Lookup a supervisor object and verify that it's of a certain type.
 * If not, produces an appropriate insanity trap.
 */
int task_ob_lookup_type(
	struct s_thread			*thread,
	oskit_addr_t			uva,
	fluke_prot_t			prot,
	fluke_type_t			type,
	struct s_ob			**out_sob);

/*
 * Lookup a reference object at the specified address,
 * and then find and return the (locked) object pointed to by that reference.
 * Also check that the referenced object is of type 'type'.
 * If the reference is null (invalid) or of the wrong type,
 * returns 0, but with *out_sob set to null (and nothing locked).
 */
int task_ob_lookup_ref_type(
	struct s_thread			*thread,
	oskit_addr_t			uva,
	fluke_type_t			type,
	struct s_ob			**out_sob);

/*
 * Search an address range of a task for kernel objects.
 * On call, *obcountp contains the size of the uobs array.
 * On return, it is the number of objects placed in the array.
 * On call, *saddrp contains the start address of the search.
 * On return, it contains where the search terminated.  It will
 * be equal to eaddr if all objects were found, less than eaddr
 * if the array was not large enough to hold all objects.
 *
 * Routine return non-zero due to page faults or other memory faults.
 */
int task_ob_search(
	struct s_thread	*thread,
	struct s_task	*task,
	oskit_addr_t	*saddrp,
	oskit_addr_t	eaddr,
	fluke_obspec_t	*uobs,
	unsigned int	*obcountp);

/*
 * Move a kernel object from one user VA to another.
 * May return non-zero due to source or destination page faults,
 * other memory faults, or various object related insanity conditions.
 */
int task_ob_move(
	struct s_thread		*thread,
	oskit_addr_t		old_uva,
	oskit_addr_t		new_uva,
	fluke_type_t		type,
	oskit_size_t		usize,
	oskit_size_t		copy_usize, /* bytes of user data to copy */
	struct s_ob		**out_sob);

/*
 * Special helper routine for space-to-space copy and fluke_*_move
 * functions.  Used to lookup and lock two fragments at the same time
 * while avoiding potential deadlock.
 */
int task_lookup_two_frags(
	struct s_thread		*src_thread,
	struct s_thread		*dest_thread,
	oskit_addr_t		src_uva,
	oskit_addr_t		dest_uva,
	fluke_prot_t		src_prot,
	fluke_prot_t		dest_prot,
	struct fragment		**out_src_frag,
	struct fragment		**out_dest_frag,
	int			*out_remote_rc);

/*
 * Bootstrap initialization of the task layer.
 */
void s_task_boot(void);

void obcache_stats_dump(void);

#ifdef DEBUG
/*
 * Dump state of task to console
 */
void s_task_dump(s_task_t *task, int indent);
#endif

#if FLASK
void s_task_keeper_revoke(struct s_task *task,
			  security_id_t tsid,
			  oskit_bool_t try_remove,
			  access_vector_t *out_retained);

void s_task_segment_revoke(struct s_task *task,
			  security_id_t tsid, 
			  access_vector_t revoke,
                          oskit_bool_t try_remove,
			  access_vector_t *out_retained);

int s_task_keeper_segment_reset(struct s_task *task);
#endif

#endif /* _FLOSS_KERNEL_FLUKE_S_TASK_H_ */
