/*
 * Copyright (c) 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * Contributed by the Computer Security Research division,
 * INFOSEC Research and Technology Office, NSA.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#if FLASK

#include "s_task.h"
#include "s_mapping.h"
#include "debug.h"

#include "avc.h"

void
s_task_keeper_revoke(struct s_task * task,
		     security_id_t tsid,
		     oskit_bool_t try_remove,
		     access_vector_t * out_retained)
{
    int             rc;
    link_t          link;
    s_ob_t         *sob;


    link_init(&link);
    s_task_lock(task);

    sob = link_follow(&task->keeper);

    if (sob && sob->sid == tsid) {
	s_ob_unlock(sob);
	if (!try_remove) {
	    rc = s_task_set_state(task, 0, &link);
	    assert(rc == 0);
	} else {
	    *out_retained |= SUBJECT_SUBJECT__TASK_KEEPER;
	}
    } else if (sob)
	s_ob_unlock(sob);

    s_task_unlock(task);

    return;
}


void
s_task_segment_revoke(s_task_t * task,
		      security_id_t tsid,
		      access_vector_t revoke,
		      oskit_bool_t try_remove,
		      access_vector_t * out_retained)
{
    s_mapping_t    *mp;

    /*
     * Scan this task's mapping list for mappings which contain frags whose
     * segment_sid is equal to tsid, and revoke the specified permissions
     * from these frags.
     */

    assert(task->sob.type == FLUKE_TASK);
    s_task_lock(task);

    queue_iterate(&task->mappings, mp, s_mapping_t *, tlink) {
	s_mapping_f_lock(mp);
	s_mapping_u_lock(mp);
	mapping_frag_revoke(mp, tsid, revoke, try_remove, out_retained);
	s_mapping_u_unlock(mp);
	s_mapping_f_unlock(mp);
	if ((*out_retained & revoke) == revoke)
	    /* no need to continue searching */
	    break;
    }

    s_task_unlock(task);
    return;
}


int
s_task_keeper_segment_reset(struct s_task * task)
{
    s_mapping_t    *mp;
    s_ob_t         *sob;
    link_t          link;
    oskit_bool_t    revoke;
    int             rc;


    s_task_lock(task);

    sob = link_follow(&task->keeper);
    if (sob) {
	rc = avc_has_permission_reset(SECCLASS_SUBJECT,
				      task->sob.sid,
				      SECCLASS_SUBJECT,
				      sob->sid,
				      SUBJECT_SUBJECT__TASK_KEEPER,
				      &revoke);
	s_ob_unlock(sob);
	if (rc) {
	    s_task_unlock(task);
	    return rc;
	}
	if (revoke) {
	    link_init(&link);
	    rc = s_task_set_state(task, 0, &link);
	    assert(rc == 0);
	}
    }
    queue_iterate(&task->mappings, mp, s_mapping_t *, tlink) {
	s_mapping_f_lock(mp);
	s_mapping_u_lock(mp);
	rc = mapping_frag_reset(mp, task->sob.sid);
	s_mapping_u_unlock(mp);
	s_mapping_f_unlock(mp);
	if (rc) {
	    s_task_unlock(task);
	    return rc;
	}
    }

    s_task_unlock(task);

    return 0;
}
#endif
