/*
 * Copyright (c) 1995, 1997, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <errno.h>
#include <malloc.h>
#include <oskit/lmm.h>
#include <oskit/page.h>
#include <oskit/machine/base_vm.h>
#include <fluke/thread.h>
#include <fluke/exception.h>

#include "s_task.h"
#include "s_mapping.h"
#include "fault.h"
#include "s_region.h"
#include "s_thread.h"
#include "debug.h"
#if FLASK
#include "avc.h"
#endif
#include "sched.h"
#include "per_cpu.h"

#include "config_instrument.h"
#include "s_instrument.h"

/*#define BASE_DEBUG*/
#ifdef BASE_DEBUG
#ifdef SOURCE
#undef SOURCE
#endif
#define SOURCE "s_thread.c"
#define BD_BEGIN() DEBUG_BASE_BEGIN()
#define BD_END() DEBUG_BASE_END()
#define BD() DEBUG_BASE()
#define BD_ARGS(fmt, args...) DEBUG_BASE_ARGS(fmt,args)
#else
#define BD_BEGIN()
#define BD_END()
#define BD()
#define BD_ARGS(args...)
#endif

int sys_region_search_hack(s_thread_t *cur_thread);

static int s_thread_init(struct s_thread *);

extern link_t kernel_task_link;	/* s_task.c */

extern fluke_thread_t *pthread_boot(struct exc_state *);

void s_thread_boot()
{
	int rc;
	fluke_thread_t *startup_thread_flob;
	struct exc_state state;

	BD_BEGIN();
	if (sizeof(struct s_thread) > S_THREAD_SIZE)
		panic("s_thread struct too large");
	s_thread_md_boot();
	BD_END();

	/*
	 * Initialize pthread crapola needed when we become a kserver thread.
	 */
	pthread_init();

	/*
	 * Set up user-space stuff for the startup thread to behave
	 * as a normal pthread.
	 */
	startup_thread_flob = pthread_boot(&state);

	/*
	 * Eat the link's ref, since s_thread_init will clobber it (below).
	 */
	link_destroy(&startup_thread.task_state.task_link);

	/*
	 * Now initialize the startup_thread to be a proper thread structure.
	 * First make it a proper flob.
	 */
	rc = s_ob_enter(FLUKE_THREAD, kvtophys(startup_thread_flob), 0,
			&startup_thread.sob);
	assert_rc_0(rc);

#if FLASK
	/*
	 * Assign the kernel SID to the startup thread
	 */
	startup_thread.sob.sid = SECINITSID_KERNEL;
#endif

#ifdef DEBUG
	s_ob_label(&startup_thread.sob, "root thread [boot CPU]");
#endif

	/*
	 * Now initialize thread state.
	 * This clobbers a few words at the base of our stack (to
	 * to up the call to the normal user thread startup function),
	 * but we never return back that far anyway...
	 */
	rc = s_thread_init(&startup_thread);
	assert_rc_0(rc);

	/*
	 * Now we have a normal thread flob for the startup thread,
	 * as if it was just s_thread_create'd.  We need a task link and va.
	 */
	link_transfer(&kernel_task_link, &startup_thread.task_state.task_link);
	startup_thread.thread_va = (oskit_addr_t) startup_thread_flob;

	/*
	 * Give it the starting user register state that pthread_boot set.
	 */
	startup_thread.exc_state = state;

	/*
	 * We were initialized to FLUKE_STOPPED, but we ain't stoppin'.
	 * Assign the boot CPU to the startup thread.
	 */
	startup_thread.gflags = 0;
#if defined(CONFIG_SMP) || defined(CONFIG_USER_SCHED)
	startup_thread.disp_state.cpu = &boot_cpu_info;
#endif

	s_thread_unlock(&startup_thread); /* s_ob_enter left it locked */
}


static int
s_thread_init(
	struct s_thread		*th)
{
	assert(((oskit_addr_t)th & ((1 << S_THREAD_ALIGN) - 1)) == 0);

	/* Thread-specific state initialization for the lower layers */
	thread_csw_state_init(th);
	thread_exc_state_init(th);
	thread_disp_state_init(th);
	thread_wait_state_init(th);
	thread_task_state_init(th);
	thread_ipc_state_init(th);
	thread_pset_state_init(th);
	thread_sched_state_init(th);

#ifdef CONFIG_INSTRUMENT
	s_instrument_thread_init(th);
#endif
	/*
	 * Initialize the thread's machine-dependent user-mode state.
	 * This call will also set up the new thread's stack if necessary.
	 */
	s_thread_md_create(th);

	/* General thread layer initialization.  */
	cond_init(&th->stop_cond);
	cond_init(&th->death_cond);
	th->gflags = FLUKE_THREAD_STOPPED;
	th->th_lflags = 0;

#ifdef DEBUG
	memset(&th->kr_stats, 0, sizeof th->kr_stats);
#endif

	BD_ARGS("Return %p",th);
	return 0;
}

int s_thread_create(
	oskit_addr_t		uva,
	fluke_hash_t		hash,
	s_thread_t		**out_th)
{
	struct s_thread *th;
	int rc;
	BD();

	assert(out_th);

	rc = task_ob_create(current_thread(), FLUKE_THREAD, uva, hash,
			    (s_ob_t **)&th);
	if (rc) return rc;

	*out_th = th;
	return s_thread_init(th);
}


int s_thread_destroy(
	s_thread_t		*th)
{
	BD();
	assert_s_thread_locked(th);
	assert(th->sob.type == FLUKE_THREAD);

#ifdef CONFIG_INSTRUMENT
	s_instrument_thread_destroy(th);
#endif

	/* Tear down lower-level layer state. */
	thread_pset_state_destroy(th);
	thread_ipc_state_destroy(th);
	thread_task_state_destroy(th);
	thread_wait_state_destroy(th);
	thread_disp_state_destroy(th);
	thread_sched_state_destroy(th);

	/*
	 * Wake up all threads in thread_wait waiting for us to
	 * wake them up "by name" (for IPC interaction or something).
	 * When they run, they will restart their system calls and
	 * discover the invalidity of their links to us.
	 */
	cond_broadcast(&th->death_cond);

	/*
	 * Destroy and free the thread object.
	 */
	cond_destroy(&th->stop_cond);
	cond_destroy(&th->death_cond);

	s_ob_destroy(&th->sob);
	return 0;
}

int s_thread_stop_th(s_thread_t *target)
{
        struct s_thread *cur_th = current_thread();
        link_t link;

	assert_s_thread_locked(target);

        /* It's never OK to try to manipulate yourself... */
        if (target == cur_th) {
                s_thread_unlock(target);
                EXC_RETURN_INSANITY(CURRENT_THREAD);
        }

        /* If the thread is waiting in the stopped state, we're done. */
        if (thread_wakeup(target, WAIT_STOPPED))
        {
                return 0;
        }

        /* Otherwise, we have to prod it into stopping. */
        thread_cancel(target);
#ifdef CONFIG_PRIO_INHERIT
        link_produce(&link, &target->sob);
#else
	link_init(&link);
#endif
        return cond_wait(&target->stop_cond, &target->sob.lock, &link);
}

int s_thread_stop(oskit_addr_t target_uva, int dying, s_thread_t **out_target)
{
	struct s_thread *cur_th = current_thread();
	s_thread_t *target;
	int rc;

	BD_ARGS("target=%p",*out_target);
	/* Lookup and lock the target thread. */
	if (dying)
		rc = task_ob_destroy(cur_th, target_uva,
				     FLUKE_THREAD, (s_ob_t **)&target);
	else
		rc = task_ob_lookup_type(cur_th, target_uva,
					 FLUKE_PROT_WRITE, FLUKE_THREAD,
					 (s_ob_t **)&target);
	if (rc)
		return rc;
	assert_s_thread_locked(target);
	*out_target = target;

	return s_thread_stop_th(target);
}

void s_thread_release(s_thread_t *target)
{
	assert_s_thread_locked(target);

	/*
	 * If the thread's stopped flag has been cleared,
	 * then start the thread running.
	 */
	if (!(target->gflags & FLUKE_THREAD_STOPPED)) {
		thread_ready(target);
	} else {
		/* Put the thread back to sleep in the stopped state. */
		thread_wait_other(NULL, WAIT_STOPPED, target);

		/* Wake up the next thread waiting to control it. */
		cond_signal(&target->stop_cond);
	}
	s_thread_unlock(target);
}

void s_thread_get_state(
	s_thread_t		*target,
	fluke_thread_state_t	*state)
{
	BD_ARGS("target=%p",target);
	assert(target);
	assert_s_thread_locked(target);
	assert(state);

	state->hash = s_ob_gethash(&target->sob);
	state->thread_va = target->thread_va;

	/* Note that thread_ipc_get_state will OR the IPC flags into this. */
	assert((target->gflags & ~THREAD_GFLAGS) == 0);
#ifndef CONFIG_OPT_THREADS
	assert((target->lflags & ~THREAD_LFLAGS) == 0);
#endif
	state->flags = target->gflags | target->th_lflags;

	thread_exc_get_state(target, state);
	thread_ipc_get_state(target, state);
	s_thread_md_get_state(target, state);
	thread_sched_get_state(target, state);
}

#if FLASK
void s_thread_get_state_secure(
	s_thread_t			*target,
	fluke_thread_state_secure_t 	*state)
{
    thread_ipc_get_state_secure(target,state);
}
#endif

void s_thread_set_state(
	s_thread_t		*target,
	fluke_thread_state_t	*new_state)
{
	BD_BEGIN();
	assert(target);
	assert_s_thread_locked(target);
	assert(new_state);

	target->thread_va = new_state->thread_va;
	target->gflags = new_state->flags & THREAD_GFLAGS;
	target->th_lflags = new_state->flags & THREAD_LFLAGS;

	thread_exc_set_state(target, new_state);
	thread_ipc_set_state(target, new_state);
	thread_sched_set_state(target, new_state);
	s_thread_md_set_state(target, new_state);
	BD_END();
}

int s_thread_canceled(s_thread_t *cur_thread)
{
	s_thread_t *other;
	int rc;
	BD_ARGS("target=%p",cur_thread);


	/*
	 * Notify interested layers that we've been canceled,
	 * so they can update any state as necessary.
	 * These routines must not block or be interrupted or anything,
	 * because then the thread would be left "partly canceled",
	 * and important things might be left undone...
	 */
	s_thread_md_canceled(cur_thread);
	thread_ipc_canceled(cur_thread);
	thread_task_canceled(cur_thread);
	thread_sched_canceled(cur_thread);

	s_thread_lock(cur_thread);

	/*
	 * If an asynchronous exception needs to be dispatched to the thread,
	 * and exceptions are enabled, dispatch it now.
	 * This just involves twiddling register state, so can't fail.
	 */
	if ((cur_thread->th_lflags & FLUKE_THREAD_EXCEPTION_ENABLE) &&
	    (cur_thread->gflags & FLUKE_THREAD_ASYNC_PENDING))
		s_thread_md_async_exception(cur_thread);

	/* If we've been asked to stop, do so.  */
	other = cond_wakeup(&cur_thread->stop_cond);
	if (other != 0)
	{
		s_thread_unlock(cur_thread);

		/* Passivate all of our entangling associations.  */
		thread_task_state_stop(cur_thread);
		rc = thread_ipc_state_stop(cur_thread, other);
		if (rc)
			return rc;

		/* Atomically stop and handoff to whoever wants us stopped. */
		return thread_handoff(cur_thread, WAIT_STOPPED, other);
	}

	/*
	 * Now we've done everything necessary
	 * to account for all the reasons other threads might cancel us;
	 * we can safely clear the cancel condition.
	 */
	thread_cancel_clear(cur_thread);

	s_thread_unlock(cur_thread);

	return 0;
}

/* Called from trap.S.  */
int s_thread_exception(int rc)
{
	s_thread_t *cur_thread = current_thread();
	BD();

	switch (rc)
	{
		case KR_CANCEL:
		{
			return s_thread_canceled(cur_thread);
		}
#if FLASK
		case KR_SECURITY_FAULT:
	        {
		        return avc_security_fault(cur_thread->sclass,
						  cur_thread->ssid,
						  cur_thread->tclass,
						  cur_thread->tsid,
						  cur_thread->requested);
		}
#endif
		case KR_PAGE_FAULT:
		{
			oskit_addr_t va;
			fluke_prot_t prot;
			s_region_t *region;
			oskit_addr_t regoff;

			va = trunc_page(cur_thread->exc_state.subcode);
			prot = cur_thread->exc_state.code >> 16;
			rc = task_fault(cur_thread, 0, va, prot, 0,
					&region, &regoff);
			if (rc == KR_PAGE_FAULT)
				rc = s_thread_md_fault_rpc(cur_thread, region,
							   regoff, prot);
			return rc;
		}
		case KR_USER_EXCEPTION:
		{


			/*
			 * XXX handle bad memory references in region_search
			 */
			if (sys_region_search_hack(cur_thread) == 0)
				break;

			/*
			 * First try to dispatch the exception
			 * to the thread's own exception handler.
			 */
			s_thread_lock(cur_thread);
			if (cur_thread->th_lflags &
			    FLUKE_THREAD_EXCEPTION_ENABLE)
			{
				s_thread_md_sync_exception(cur_thread);
				s_thread_unlock(cur_thread);
				break;
			}
			s_thread_unlock(cur_thread);

			/*
			 * That didn't work,
			 * so make the thread perform an RPC to its keeper.
			 */
			return s_thread_md_exception_rpc(cur_thread);
		}
		case KR_NO_MEMORY:
		{
			kentry_md_syscall_error_return(cur_thread,
						       FLUKE_NO_MEMORY);
			break;
		}
		default:
			panic("s_thread_exception: %p: unknown code %x",
			      cur_thread, rc);
	}

	return 0;
}
