/*
 * Copyright (c) 1995-1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#ifndef _FLOSS_KERNEL_FLUKE_S_THREAD_H_
#define _FLOSS_KERNEL_FLUKE_S_THREAD_H_

#include "s_thread_md.h"

#include "config_opt_threads.h"
#include "config_instrument.h"

#if FLASK
#include <flask/security.h>
#endif 

#include "s_ob.h"
#include "csw.h"
#include "exception.h"
#include "dispatch.h"
#include "wait.h"
#include "s_task.h"
#include "ipc.h"
#include "s_pset.h"
#include "kentry.h"

#ifdef _FLUKE_KERNEL_SCHED_H
#error cannot #include sched.h before s_thread.h
#endif
#include "sched.h"

struct s_thread
{
	struct s_ob		sob;

	/*
	 * Virtual address of the user part of the thread object
	 * as mapped into the task the thread runs in.
	 */
	oskit_addr_t		thread_va;

	/*
	 * The remainder of this structure is simply an aggregation of
	 * all the per-thread state used in all of the functionality layers
	 * that require any per-thread state.
	 * Code in each functionality layer must only access
	 * that particular layer's thread-specific state area.
	 * This basically takes the place of pthread's thread-specific data;
	 * this approach is much simpler and more efficient
	 * given the fairly limited, static kernel environment.
	 */
	struct csw_state	csw_state;
	DECLARE_EXC_STATE	/* see exception.h */
	struct disp_state	disp_state;
	struct wait_state	wait_state;
	struct task_state	task_state;
	struct ipc_state	ipc_state;
	struct sched_state   	sched_state;

	/*
	 * When another thread is trying to stop and manipulate us,
	 * it cancels us and then waits on this condition variable.
	 */
	struct cond		stop_cond;

	/*
	 * When another thread is in a thread_wait waiting for us to
	 * wake it up "by name", it puts itself on this queue.
	 * When we are destroyed, we must wake all these threads
	 * so they can restart and see the invalidity of their links to us.
	 */
	struct cond		death_cond;

	/*
	 * Thread flags, as defined in fluke/x86/thread.h.
	 * 'gflags' stores the "global" flags accessible by other threads,
	 * this flags word is locked by the thread's sob lock.
	 * 'lflags' stores the "local" flags accessible only by this thread
	 * (or a thread that is holding this thread captive);
	 * accesses to this flags word do not need to be locked.
	 * The THREAD_GFLAGS symbol below defines the flags that are in gflags,
	 * and THREAD_LFLAGS defines the flags that are in lflags.
	 */
	unsigned		gflags;
#ifndef CONFIG_OPT_THREADS
	unsigned		lflags;
#define th_lflags lflags
#endif

#if FLASK
	security_class_t sclass;
	security_id_t ssid;
	security_class_t tclass;
	security_id_t tsid;
	access_vector_t	requested;
#endif 

#ifdef CONFIG_INSTRUMENT
        unsigned                instrument_flags;
        void *                  instrument_buffer;
        unsigned                instrument_buffer_used;
#endif

#ifdef DEBUG
	/* Kernel return stats */
	struct kr_stats {
		unsigned calls;
		unsigned uexceptions;
		unsigned pagefaults;
		unsigned cancels;
		unsigned restarts;
		unsigned secfaults;
	}			kr_stats;
#endif

	/* Machine-dependent thread layer data. */
	struct s_thread_md	md;
};
typedef struct s_thread s_thread_t;

#define s_thread_lock(th)	s_ob_lock(&(th)->sob)
#define s_thread_unlock(th)	s_ob_unlock(&(th)->sob)
#define assert_s_thread_locked(th) assert_s_ob_locked(&(th)->sob)

#define THREAD_GFLAGS	(FLUKE_THREAD_ASYNC_PENDING |	\
			 FLUKE_THREAD_STOPPED)

#define THREAD_LFLAGS	(FLUKE_THREAD_EXCEPTION_ENABLE |	\
			 FLUKE_THREAD_USES_FPU)


/*
 * Special statically-allocated pseudo-thread that we "run on" during startup.
 * Machine-dependent code must define this with proper alignment and such.
 */
extern struct s_thread startup_thread;


struct fluke_thread_state;


/*
 * Find the currently running thread and return a pointer to its s_thread.
 * Machine-dependent code must define this function or macro.
 */
struct s_thread *(current_thread)(void);


/* Bootstrap initialization of the threads layer. */
void s_thread_boot(void);

/*
 * Create a new thread but don't start it running yet.
 * Returns the newly created thread, still locked and captured.
 */
int s_thread_create(
	oskit_addr_t		uva,
	fluke_hash_t		hash,
	struct s_thread		**out_thread);

/*
 * Retrieve the simple state of the specified thread,
 * which must be locked and held captive.
 */
void s_thread_get_state(
	s_thread_t			*target,
	struct fluke_thread_state	*state);

#if FLASK
void s_thread_get_state_secure(
	s_thread_t			*target,
        fluke_thread_state_secure_t 	*state);
#endif

/*
 * Set the simple state of the specified thread,
 * which must be locked and held captive.
 */
void s_thread_set_state(
	struct s_thread			*target,
	struct fluke_thread_state	*new_state);

/*
 * Lock, stop, and capture the specified target threa.
 * On successful return, the thread is locked and captured.
 * On unsuccessful return, the target thread is unlocked and still free.
 */
int s_thread_stop(
	oskit_addr_t			target_uva,
	int	  			dying,
	struct s_thread			**out_target);

/*
 * Same as above, except a locked thread sob is provided
 * by the caller.
 */
int s_thread_stop_th(s_thread_t *target);

/*
 * Release a locked and captured thread,
 * placing it back into the WAIT_STOPPED state if it is still stopped,
 * or starting it running if the flags say it should now be runnable.
 */
void s_thread_release(s_thread_t *target);

/*
 * Destroy the specified thread,
 * which must be locked and captured using s_thread_stop().
 */
int s_thread_destroy(
	struct s_thread			*thread);



/*** Machine-dependent thread layer routines ***/

/* Boot-time initialization of the MD part of the thread layer. */
void s_thread_md_boot(void);

/* Initialize a new thread's machine-dependent user-mode state.  */
void s_thread_md_create(
	struct s_thread			*th);

/*
 * Get the current thread's machine-dependent state
 * into a fluke_thread_state structure provided by the user.
 */
void s_thread_md_get_state(
	struct s_thread			*cur_th,
	struct fluke_thread_state	*state);

/*
 * Set the current thread's machine-dependent state
 * from a fluke_thread_state structure provided by the user.
 */
void s_thread_md_set_state(
	struct s_thread			*cur_th,
	struct fluke_thread_state	*state);

/*
 * Deliver an asynchronous exception to the current thread.
 * Expects at least one of the thread's pending flags to be set.
 */
void s_thread_md_async_exception(struct s_thread *cur_th);

/*
 * Deliver a synchronous exception to the current thread.
 * Expects the exception code fields in the saved state to be valid.
 */
void s_thread_md_sync_exception(struct s_thread *cur_th);

/*
 * Deliver a synchronous exception as an idempotent IPC
 * to the current thread's keeper.
 * Expects the exception code fields in the saved state to be valid.
 */
int s_thread_md_exception_rpc(struct s_thread *cur_th);

int s_thread_md_fault_rpc(struct s_thread *cur_th, struct s_region *region,
			  oskit_addr_t region_off, fluke_prot_t prot);

/*
 * Change saved user state to reflect a syscall error return.
 */
void kentry_md_syscall_error_return(
	struct s_thread			*cur_th,
	int				err);

/*
 * Perform an exception RPC to the current thread's keeper.
 */
int kentry_md_exception_rpc(struct s_thread *cur_th);

int s_thread_canceled(s_thread_t *cur_thread);
void s_thread_md_canceled(s_thread_t *cur_thread);


#ifdef CONFIG_INSTRUMENT
/*
 * Called for each syscall if instrumentation is enabled.
 * eax is the EIP of the syscall on the way in and the
 * return value of the sys call on the way out.
 */
void kentry_md_syscall_instrument(
	int                syscall_flag,
	unsigned long long tsc,
	unsigned int       eax);
#endif

#ifdef DEBUG
/*
 * Dump state of thread to console
 */
void s_thread_dump(s_thread_t *thread, int indent);
#endif

#if FLASK
void s_thread_scheduler_revoke(s_thread_t *thread,
			       security_id_t tsid,
			       oskit_bool_t try_remove,
			       access_vector_t *out_retained);

int s_thread_scheduler_reset(struct s_thread *thread, oskit_bool_t *out_reset);
#endif

#endif /* _FLOSS_KERNEL_FLUKE_S_THREAD_H_ */
