/*
 * Copyright (c) 1997 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include "config_stride_sched.h"
#ifdef CONFIG_STRIDE_SCHED

#include "config_hier_sched.h"
#include "config_hier_sched_set_parent.h"

#include "s_thread.h"
#include "sched.h"
#include "stride.h"
#include <oskit/x86/pc/pit.h>
#include <oskit/x86/pio.h>
#include <oskit/x86/proc_reg.h>
#include <stddef.h>
#include "debug.h"
#include "preempt.h"


#define MAX_PASS 0xffffffff

/* Number of base tickets for hierarchical scheduling. 
   STRIDE1 has to be bigger than BASE_TICKETS */
#define BASE_TICKETS 1024*4

/* If too big can result in overflow */
#define STRIDE1 (1 << 15)

/* 
 * Determines resolution of a quantum. It also indicates the number of hours 
 * of execution without overflow. 
 */
#define TICKS_PER_TIME_UNIT 1000

#define QUANTUM (PIT_HZ/100) /* It should be coherent with clock_md.c XXX*/
#define SCALED_QUANTUM (QUANTUM/TICKS_PER_TIME_UNIT)

/* It is possible to choose BASE DEBUG only or BASE_DEBUG together with 
   _STRD_TEST. Idenpedentrly D2 can be turned on or of with any of
   the options.*/
#define _STRD_TEST

/*#define BASE_DEBUG
#define DEBUG_LEVEL_2*/

#ifdef SOURCE
#undef SOURCE
#define SOURCE "stride"
#endif
#ifdef BASE_DEBUG
#ifdef _STRD_TEST
/* BASE_DEBUG && _STRD_TEST DEFINED*/
#define BD_BEGIN() \
	if (check_debug_on()) DEBUG_BASE_BEGIN()
#define BD_END() \
	if (check_debug_on()) DEBUG_BASE_END()
#define BD() \
	if (check_debug_on()) DEBUG_BASE()
#define BD_ARGS(fmt, args...) \
	if (check_debug_on()) DEBUG_BASE_ARGS(fmt,args)
#define BD_BEGIN_ARGS(fmt, args...) \
	if (check_debug_on()) DEBUG_BASE_BEGIN_ARGS(fmt,args)
#define BD_END_ARGS(fmt, args...) \
	if (check_debug_on()) DEBUG_BASE_END_ARGS(fmt,args)
#ifdef DEBUG_LEVEL_2
#define D2(x) if (check_debug_on()) {x;}
#else 
#define D2(x)
#endif
#else
/* ONLY BASE_DEBUG DEFINED */
#define SOURCE "stride"
#define BD_BEGIN() DEBUG_BASE_BEGIN()
#define BD_END() DEBUG_BASE_END()
#define BD() DEBUG_BASE()
#define BD_ARGS(fmt, args...) DEBUG_BASE_ARGS(fmt,args)
#define BD_BEGIN_ARGS(fmt, args...) DEBUG_BASE_BEGIN_ARGS(fmt,args)
#define BD_END_ARGS(fmt, args...) DEBUG_BASE_END_ARGS(fmt,args)
#ifdef DEBUG_LEVEL_2
#define D2(x) {x;}
#else 
#define D2(x)
#endif 
#endif
#else
/* NONE OF THEM DEFINED*/
#define BD_BEGIN_ARGS(args...)
#define BD_END_ARGS(args...)
#define BD_BEGIN()
#define BD_END()
#define BD() 
#define BD_ARGS(args...)
#ifdef DEBUG_LEVEL_2
#define D2(x) {x;}
#else 
#define D2(x)
#endif 
#endif

#ifdef _STRIDE_PROFILE
#define STRD_PROF(x) x
#else
#define STRD_PROF(x)
#endif

#ifdef _STRD_TEST
/*Only for testing stride scheduling*/
#define TICKETS_TEST1 200
#define TICKETS_TEST2 400
#endif 

#define MAX(a,b) ((a)>(b)?(a):(b))
#define STRDS(t) ((t)->sched_state)
#define PRIO(t) ((t)->disp_state.pri)

/*Lock that must be set when using this interface*/
spin_lock_t sched_lock;

/* Current client */
static struct s_thread * current;

/* Queue of clients */
static prio_queue_t queue;

#ifdef CONFIG_HIER_SCHED
/* Currency tree */
currency_t currency;
#endif

/*Global amounts*/
static unsigned global_tickets = 0;
static unsigned global_stride = 0;
static unsigned global_pass = 0;

/* static functions prototypes*/
static void global_pass_update(void);
static void global_tickets_update(unsigned tickets);
static void stride_client_leave(struct s_thread *t);
static void current_thread_pass_update(void);
static void stride_client_init(struct s_thread *t, unsigned tickets);
static void stride_client_join(struct s_thread *t);
static void global_pass_adjust(unsigned pass);

static unsigned 
get_time()
{
#ifdef DETERMINISTIC_EXECUTION
	static time = 0;
	return (time = time + SCALED_QUANTUM / 10);
#else
	extern unsigned millitime(void);
	return (millitime()/(1000/TICKS_PER_TIME_UNIT));
#endif
}

/*
 * Debugging 
 */

void
thread_sched_dump(s_thread_t *t)
{
	stride_dump_thread(&STRDS(t).link);
}

void static 
stride_dump_thread(prio_queue_node_t *s)
{
	s_thread_t *t = SCHED_LINK_TO_THREAD(s);
#ifdef _STRIDE_PROFILE
	printf("\tthread=%p,prio=%u,tickets=%u,pass=%u,stride=%u,remain=%d,"
		"gpass=%u,gstride=%u,gtickets=%u,total_ticks=%u\n",
		t, PRIO(t), STRDS(t).tickets, STRDS(t).pass,
		STRDS(t).stride, STRDS(t).remain, global_pass, global_stride,
		global_tickets, STRDS(t).total_ticks);
#else
	printf("\tthread=%p,prio=%u,tickets=%u,pass=%u,stride=%u,remain=%d,"
		"gpass=%u,gstride=%u,gtickets=%u\n",
		t, PRIO(t), STRDS(t).tickets, STRDS(t).pass,
		STRDS(t).stride, STRDS(t).remain, global_pass, global_stride,
		global_tickets);
#endif
}


#ifdef _STRD_TEST
#define DEBUG_COUNT 5000;

static int debug_active = 0;
static int debug_count = DEBUG_COUNT;

static void 
debug_on(void)
{
	debug_active = 1;
}

static int 
check_debug_on(void)
{
	if (debug_active && debug_count-- <= 0)
		debug_active = 0;
	return debug_active;
}
#endif


#if defined(CONFIG_HIER_SCHED) && defined(DEBUG)
void currency_dump_tree(currency_t *c);

void 
stride_dump_tree(void)
{ 
	currency_dump_tree(&currency);
}

int 
stride_check_tree(void)
{ 
	return currency_check_tree(&currency);
}
#endif

void 
stride_state_dump()
{
	printf("Current thread\n");
	stride_dump_thread(&STRDS(current_thread()).link);
	printf("Priority queue\n");
	prio_queue_map(&queue, stride_dump_thread);
#ifdef CONFIG_HIER_SCHED
	stride_dump_tree();
#endif
}


/*
 * Hook functions for the scheduler related to threads (s_thread.c)
 */

void 
thread_sched_set_state(struct s_thread *target, 
		       fluke_thread_state_t *new_state)
{
#ifdef CONFIG_KPREEMPT
	int flag;
#endif
	BD_BEGIN();
	save_preemption_enable(flag); 
	disable_preemption();
	/* I assumed it is not in the queue. */
  
	/* XXX */
	if (PRIO(target) == 2) {
		new_state->tickets = 1;
#ifdef CONFIG_HIER_SCHED
		new_state->own_tickets = 1;
#endif
	}

#ifdef _STRD_TEST
	if (new_state->tickets == TICKETS_TEST1 || 
	    new_state->tickets == TICKETS_TEST2)
		debug_on();
#endif
#ifdef CONFIG_HIER_SCHED
	currency_set_state(&currency, target, new_state);
#else
	stride_client_modify(target, new_state->tickets);
#endif 
	restore_preemption_enable(flag);
	BD_END();
}

void 
thread_sched_get_state(struct s_thread *target, fluke_thread_state_t *state)
{
	BD_BEGIN();
#ifdef CONFIG_HIER_SCHED
	currency_get_state(&currency, target, state);
#else
	state->tickets = STRDS(target).tickets; 
#endif 
	BD_END();
}

void 
thread_sched_state_init(struct s_thread *th)
{
	BD_BEGIN();
	/* The number of tickets dosen't really matter. They will 
	   do a set_state anyway */
	stride_client_init(th, BASE_TICKETS);
	STRD_PROF(STRDS(th).start_exec_time = get_time());
#ifdef CONFIG_HIER_SCHED
#ifdef CONFIG_HIER_SCHED_SET_PARENT
	/* Initialize the parent link */
	link_init(&(STRDS(th).sched_link));
#endif
	STRDS(th).kserver = 0;
	currency_state_init(th);
#endif
	BD_END();
}

void 
thread_sched_state_destroy(struct s_thread *th)
{
#ifdef CONFIG_KPREEMPT
	int flag;
#endif
	BD_BEGIN();
	save_preemption_enable(flag);
	disable_preemption();
#ifdef _STRD_TEST
	if (STRDS(th).tickets == TICKETS_TEST1 || 
	    STRDS(th).tickets == TICKETS_TEST2) {
		STRD_PROF(printf("Thread %p END. CPU TIME=%u, EXEC TIME=%u\n"
				  ,th, STRDS(th).total_ticks,
				  get_time()-STRDS(th).start_exec_time));
		STRD_PROF(stride_dump_thread((void *)th));
		STRD_PROF(stride_dump_thread((void *)current));
		STRD_PROF(prio_queue_map(&queue, stride_dump_thread));
	}
#endif
#ifdef CONFIG_HIER_SCHED
	/* Destroy the link */
#ifdef CONFIG_HIER_SCHED_SET_PARENT
	link_destroy(&(STRDS(th).sched_link));
#endif
	currency_thread_destroy(&currency, th);
#endif
	enable_preemption();
	BD_END();
}


void 
thread_sched_canceled(struct s_thread *th)
{
	/* We don't have to do anything */
	BD_BEGIN();
	BD_END();
}


/*
 * Hook functions for the scheduler related to the dispacher (dispatch.c)
 */

static int 
idle_thread(struct s_thread *t)
{
	return (t->disp_state.pri == DISP_PRIIDLE);
}

/*
 * Boot time initialization.
 * If using hierarchical scheduling we initialize together with the first thread
 */
void 
sched_boot(void)
{
	BD_BEGIN();
	prio_queue_init(&queue);
	spin_lock_init(&sched_lock);
	BD_END();
}

/*
 * Marks the specified thread as runnable and returns non-zero if that
 * thread is higher priority than the current thread. 
 */
int
sched_thread_ready(struct s_thread *t)
{
	int ret;

	BD_BEGIN();
	stride_client_join(t);
	ret = (PRIO(t) < PRIO(current_thread()));
	BD_END();
	return ret;
}

/* Is there any thread more suitable to run in the queue? */
int 
sched_thread_yield(void)
{  
	s_thread_t *cur_thread = current_thread();

	BD_BEGIN();
	current_thread_pass_update();

	if (!prio_queue_empty(&queue)) {
		s_thread_t *t = SCHED_LINK_TO_THREAD(prio_queue_head(&queue));
		unsigned prio_head = PRIO(t);
		unsigned prio_cur = PRIO(cur_thread);
		unsigned pass_head = STRDS(t).pass;
		unsigned pass_cur = STRDS(cur_thread).pass;

		if (prio_head < prio_cur || 
		    (prio_head == prio_cur && pass_head <= pass_cur)) {
			D2(prio_queue_map(&queue, stride_dump_thread));
			prio_queue_insert_2p(&queue, &STRDS(cur_thread).link,
					     prio_cur, pass_cur);
			current = NULL;
			BD_END();
			return 1;
		}
		else {
			global_pass_adjust(pass_cur);
			BD_END();
			return 0;
		}
	} else 
		global_pass_adjust(STRDS(cur_thread).pass);
	BD_END();
	return 0;
}

/* Function called from dispatcher to get the new thread to run */
struct s_thread * 
sched_thread_next(void)
{
	BD_BEGIN();
	D2(prio_queue_map(&queue, stride_dump_thread););
	if (current == NULL) { 
		/* Yield has been called before. 
		   Getting a new thread to run */
	} else { 
		/* Current thread is not in the run queue anymore */
		current_thread_pass_update();
		stride_client_leave(current);
	}
	assert(!prio_queue_empty(&queue));
	current = SCHED_LINK_TO_THREAD(prio_queue_remove_head(&queue));
	STRDS(current).start_time = get_time();
	global_pass_adjust(STRDS(current).pass);
#if 0
	if (idle_thread(current))
		panic("thread %p: empty ready queue in disp_yield\n", current);
#endif
	BD_END();
	return current;
}

struct s_thread * 
sched_thread_idle(void)
{
	assert(current == NULL);

	current = SCHED_LINK_TO_THREAD(
			prio_queue_remove_head_prio(&queue, DISP_PRIIDLE));
	assert(current);
	STRDS(current).start_time = get_time();
	global_pass_adjust(STRDS(current).pass);
	return current;
}

int 
sched_ready_queue_empty(void)
{
	int ret;
	BD_BEGIN();
	ret = prio_queue_empty(&queue);
	BD_END();
	return ret;
}

void 
sched_first_thread(struct s_thread *t)
{
	BD_BEGIN_ARGS("t=%p",t);
	global_pass_update();
	/* Inmidiatly after this call there is a set state with a logical 
	   number of tickets */
	stride_client_init(t, BASE_TICKETS);
	global_tickets_update(STRDS(t).tickets);
	STRDS(t).pass = global_pass;
	STRDS(t).start_time = get_time();
	current = t;
#ifdef CONFIG_HIER_SCHED
	/* We have the first thread. We initialize the currency tree */
	/* The root thread is not affected by any further set state */
	currency_init(&currency, t, BASE_TICKETS);
#endif
	BD_END();
}




/*
 * Stride algorithm functions
 */

/* Test some adjusting. For now it only makes things worse */
void 
global_pass_adjust(unsigned pass)
{
#ifdef GLOBAL_PASS_ADJUST
	global_pass = MAX(pass, global_pass);
#endif
}

static void 
current_thread_pass_update(void)
{
	unsigned elapsed = get_time();

	assert(elapsed >= STRDS(current).start_time);
	elapsed -= STRDS(current).start_time;

#if 0
	if (elapsed > 2*SCALED_QUANTUM)
		printf("!!!!!!!!!!!!!!!elapsed in quant=%d\n",
		       elapsed / SCALED_QUANTUM);
#endif
	assert(elapsed == 0 ||
	       ((STRDS(current).stride*elapsed)/elapsed)
	       == STRDS(current).stride);
	STRDS(current).pass +=
		(STRDS(current).stride * elapsed) / SCALED_QUANTUM;

#ifdef ANALISE_POLICY
	if (abs(STRDS(current).pass-global_pass) > 2*STRDS(current).stride)
		printf("%d\t", ((int)STRDS(current).pass -
				(int)global_pass)/(int)STRDS(current).stride);
#endif
	STRD_PROF(STRDS(current).total_ticks += elapsed);
	STRDS(current).start_time = get_time();
}

static void 
global_pass_update(void) 
{
	static unsigned last_update = 0;
	unsigned elapsed;
	unsigned time;

	BD_BEGIN();
	time = get_time();
	elapsed = time - last_update;
	last_update = time;
#if 0
	if (elapsed > 2*SCALED_QUANTUM)
		printf("!!elapsed in quant=%d\n", elapsed / SCALED_QUANTUM);
#endif
#if DEBUG
	if (elapsed && !((global_stride * elapsed) / elapsed == global_stride))
		printf("!! global_stride=%u, elapsed=%u\n",
		       global_stride, elapsed);
#endif
	assert(elapsed == 0 ||
	       ((global_stride*elapsed)/elapsed) == global_stride);
	global_pass += (global_stride * elapsed) / SCALED_QUANTUM;
	BD_END_ARGS("gtime=%u, gpass=%u, gstride=%u",
		    time, global_pass, global_stride);
}

static void 
global_tickets_update(unsigned tickets) 
{
	BD_BEGIN();
	global_tickets += tickets;
	assert(global_tickets);
	global_stride = STRIDE1 / global_tickets;
	BD_END();
}

static void 
stride_client_init(struct s_thread *t, unsigned tickets)
{
	BD_BEGIN();
	assert(tickets);
	STRDS(t).tickets = tickets;
#if 0
	if (tickets == 0)
		STRDS(t).stride = STRIDE1;
#endif
	STRDS(t).stride = STRIDE1 / tickets;
	STRDS(t).remain = STRDS(t).stride;
	STRD_PROF(STRDS(t).total_ticks = 0);
	BD_END();
}

static void 
stride_client_join(struct s_thread *t)
{
	BD_BEGIN_ARGS("s_thread=%p, tickets=%u", t, STRDS(t).tickets);
	D2(prio_queue_map(&queue, stride_dump_thread););
	global_pass_update();
	STRDS(t).pass = global_pass + STRDS(t).remain;
	global_tickets_update(STRDS(t).tickets);
	prio_queue_insert_2p(&queue, &STRDS(t).link, PRIO(t), STRDS(t).pass);
	D2(prio_queue_map(&queue, stride_dump_thread););
	BD_END();
}

static void 
stride_client_leave(struct s_thread *t)
{
	/* PRE-CONDITION: it is not in the ready queue */
	BD_BEGIN();
	global_pass_update();
	STRDS(t).remain = STRDS(t).pass - global_pass;
	global_tickets_update(-(STRDS(t).tickets));
	BD_END();
}

void 
stride_client_modify(struct s_thread *t, int tickets)
{
	unsigned stride;
	int remain;
	int in_queue;

	BD_BEGIN();
	in_queue = prio_queue_find(&queue, &STRDS(t).link);
	if (in_queue) {
		prio_queue_remove(&queue, &STRDS(t).link);
		stride_client_leave(t);
	}
	assert(tickets);
	assert(STRDS(t).stride);
	if (current == t) {
		global_tickets_update(tickets-(STRDS(t).tickets));
	}
	stride = STRIDE1 / tickets;
	remain = (STRDS(t).remain*stride) / STRDS(t).stride;
	STRDS(t).tickets = tickets;
	STRDS(t).stride = stride;
	STRDS(t).remain = remain;
	if (in_queue) {
		stride_client_join(t);
	}
	BD_END();
}
#endif


