/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * System calls for manipulating condition variable objects.
 */

#include <fluke/cond.h>
#include <fluke/machine/syscall.h>

#include "s_thread.h"
#include "s_mutex.h"
#include "s_cond.h"
#include "xfer.h"
#include "debug.h"

int sys_cond_create_hash()
{
	struct s_thread *cur_th = current_thread();
	s_cond_t *new_cond;
	int rc;

	rc = s_cond_create(EXC_SYSPARAM(cur_th, cond_create_hash, cond),
			   EXC_SYSPARAM(cur_th, cond_create_hash, hash),
			   &new_cond);

	if (rc)
		return rc; /* XXX no memory */

	EXC_SYSPARAM(cur_th, cond_create_hash, hash) =
		s_ob_gethash(&new_cond->sob);

	s_cond_unlock(new_cond);

	EXC_SYSPARAM(cur_th, cond_create_hash, rc) = 0;
	return 0;
}

int sys_cond_destroy()
{
	struct s_thread *cur_th = current_thread();
	s_cond_t *cond;
	int rc;

	rc = task_ob_destroy(cur_th, EXC_SYSPARAM(cur_th, cond_destroy, cond),
			     FLUKE_COND, (s_ob_t **)&cond);
	if (rc) return rc;

	return s_cond_destroy(cond);
}

int sys_cond_get_state()
{
	struct s_thread *cur_th = current_thread();
	s_cond_t *cond = 0;
	fluke_cond_state_t state, *statep = 0;
	struct link inheritor_link, *inheritor_linkp = 0;
	int rc;

	/*
	 * Determine which values we want
	 */
	if (EXC_SYSPARAM(cur_th, cond_get_state, state))
		statep = &state;
	if (EXC_SYSPARAM(cur_th, cond_get_state, inheritor))
		inheritor_linkp = &inheritor_link;

	/*
	 * Lookup the cond
	 */
	rc = task_ob_lookup_type(cur_th,
				 EXC_SYSPARAM(cur_th, cond_get_state, cond),
				 FLUKE_PROT_READ, FLUKE_COND,
				 (s_ob_t **)&cond);
	if (rc) return rc;

	/*
	 * Get its state
	 */
	rc = s_cond_get_state(cond, statep, inheritor_linkp);
	s_cond_unlock(cond);
	if (rc) return rc;

	/*
	 * Copyout state
	 */
	if (statep) {
		rc = copyout_local(statep,
				   EXC_SYSPARAM(cur_th, cond_get_state, state),
				   sizeof(state) / sizeof(oskit_reg_t));
		if (rc) goto out;
	}
	if (inheritor_linkp)
	{
		rc = kentry_ref_associate(cur_th,
				EXC_SYSPARAM(cur_th, cond_get_state, inheritor),
				inheritor_linkp);
		inheritor_linkp = 0;
		if (rc) goto out;
	}
out:
	if (inheritor_linkp)
		link_destroy(inheritor_linkp);
	return rc;
}

int sys_cond_set_state()
{
	struct s_thread *cur_th = current_thread();
	s_cond_t *cond = 0;
	fluke_cond_state_t state, *statep = 0;
	struct link inheritor_link, *inheritor_linkp = 0;
	int rc;

	/*
	 * Copy in the simple state structure if there is one.
	 */
	if (EXC_SYSPARAM(cur_th, cond_set_state, state))
	{
		rc = copyin_local(EXC_SYSPARAM(cur_th, cond_set_state, state),
				  &state, sizeof(state) / sizeof(oskit_reg_t));
		if (rc) return rc;
		statep = &state;
	}

	/*
	 * Look up the references to insert into the cond.
	 */
	if (EXC_SYSPARAM(cur_th, cond_set_state, inheritor))
	{
		rc = kentry_ref_extract_type(cur_th,
				EXC_SYSPARAM(cur_th, cond_set_state, inheritor),
				FLUKE_THREAD, &inheritor_link);
		if (rc) goto out;
		inheritor_linkp = &inheritor_link;
	}

	/*
	 * Lookup and lock the cond
	 */
	rc = task_ob_lookup_type(cur_th,
				 EXC_SYSPARAM(cur_th, cond_set_state, cond),
				 FLUKE_PROT_WRITE, FLUKE_COND,
				 (s_ob_t **)&cond);
	if (rc) goto out;

	rc = s_cond_set_state(cond, statep, inheritor_linkp);

	out:
	if (cond)
		s_cond_unlock(cond);
	if (inheritor_linkp)
		link_destroy(&inheritor_link);

	return rc;
}

int sys_cond_wait()
{
	extern int sys_mutex_lock(void);
	struct s_thread *cur_th = current_thread();
	s_cond_t *cond;
	s_mutex_t *mutex;
	int rc;

	/*
	 * First look up the condition variable object.
	 */
	rc = task_ob_lookup_type(cur_th, EXC_SYSPARAM(cur_th, cond_wait, cond),
				 FLUKE_PROT_WRITE, FLUKE_COND,
				 (s_ob_t **)&cond);
	if (rc) return rc;

	/*
	 * Then the mutex object, which is farther down the locking hierarchy.
	 */
	rc = task_ob_lookup_type(cur_th, EXC_SYSPARAM(cur_th, cond_wait, mutex),
				 FLUKE_PROT_WRITE, FLUKE_MUTEX,
				 (s_ob_t **)&mutex);
	if (rc)
	{
		s_cond_unlock(cond);
		return rc;
	}

	/*
	 * Unlock the associated mutex.
	 * We can do this safely now because the cond is still locked,
	 * so no threads will be able to signal before we wait.
	 */
	rc = s_mutex_unlockop(mutex, cur_th);
	if (rc)
	{
		s_mutex_unlock(mutex);
		s_cond_unlock(cond);
		return rc;
	}
	s_mutex_unlock(mutex);

	/*
	 * Now wait on the condition variable.
	 * If this wait gets interrupted or something,
	 * then change the current thread's program counter
	 * to point to the mutex_lock() entrypoint
	 * so that the lock will be reacquired properly on restart.
	 */
	EXC_SYSPARAM(cur_th, mutex_lock, mutex) =
		EXC_SYSPARAM(cur_th, cond_wait, mutex);
	EXC_SET_IP(cur_th, fluke_mutex_lock__entry);

	return s_cond_wait(cond);
}

int sys_cond_signal()
{
	struct s_thread *cur_th = current_thread();
	s_cond_t *cond;
	int rc;

	rc = task_ob_lookup_type(cur_th, EXC_SYSPARAM(cur_th, cond_signal, cond),
				 FLUKE_PROT_WRITE, FLUKE_COND,
				 (s_ob_t **)&cond);
	if (rc) return rc;

	rc = s_cond_signal(cond);

	s_cond_unlock(cond);

	return rc;
}

int sys_cond_broadcast()
{
	struct s_thread *cur_th = current_thread();
	s_cond_t *cond;
	int rc;

	rc = task_ob_lookup_type(cur_th, EXC_SYSPARAM(cur_th, cond_broadcast, cond),
				 FLUKE_PROT_WRITE, FLUKE_COND,
				 (s_ob_t **)&cond);
	if (rc) return rc;

	rc = s_cond_broadcast(cond);

	s_cond_unlock(cond);

	return rc;
}

int sys_cond_move()
{
	struct s_thread *cur_th = current_thread();
	s_cond_t *cond;
	int rc;

	rc = task_ob_move(cur_th, EXC_SYSPARAM(cur_th, cond_move, oaddr),
			  EXC_SYSPARAM(cur_th, cond_move, naddr),
			  FLUKE_COND, FLUKE_COND_SIZE, 0, (s_ob_t **)&cond);
	if (rc == 0)
		s_cond_unlock(cond);
	return (rc);
}

int sys_cond_reference()
{
	struct s_thread *cur_th = current_thread();

	return kentry_ob_reference(cur_th,
				   EXC_SYSPARAM(cur_th, cond_reference, cond),
				   EXC_SYSPARAM(cur_th, cond_reference, ref),
				   FLUKE_COND);
}
