/*
 * Copyright (c) 1996, 1997 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Special debugging entrypoints for our kernel implementation
 */

#ifdef DEBUG
#include <fluke/machine/syscall.h>
#include <fluke/debug.h>
#include <stdio.h>

#include "s_ob.h"
#include "s_thread.h"
#include "s_task.h"
#include "s_mutex.h"
#include "s_cond.h"
#include "s_ref.h"
#include "s_mapping.h"
#include "s_region.h"
#include "s_port.h"
#include "s_pset.h"
#include "xfer.h"
#include "debug.h"
#include "config_nonblocking.h"
#include "config_hier_sched.h"
#if defined(CONFIG_HIER_SCHED) && defined(DEBUG)
#include "stride.h"
#endif
/* XXX for asserts */
#include "preempt.h"
#include "intr.h"

static struct types {
	int	code;
	char	*str;
} fluketypes[] = {
	{ FLUKE_NULL,	"NULL" },
	{ FLUKE_TASK,	"TASK" },
	{ FLUKE_THREAD,	"THREAD" },
	{ FLUKE_MUTEX,	"MUTEX" },
	{ FLUKE_COND,	"COND" },
	{ FLUKE_PORT,	"PORT" },
	{ FLUKE_PSET,	"PSET" },
	{ FLUKE_CATALOG,"CATALOG" },
	{ FLUKE_REF,	"REF" },
	{ FLUKE_MAPPING,"MAPPING" },
	{ FLUKE_REGION,	"REGION" },
};


char *fluke_type_name(fluke_type_t type)
{
	int i;

	for (i = 0; i < sizeof fluketypes / sizeof fluketypes[0]; i++)
		if (fluketypes[i].code == type)
			return fluketypes[i].str;
	return "Unknown";
}

/*
 * XXX do we bother moving this into s_debug.c (which we'd have to create)?
 * XXX should this lock the sob?
 */
int s_debug_dump_flob(s_ob_t *sobp, int indent)
{
	int i;

	for (i = 0; i < indent; i++) putchar(' ');
	printf("%s %p: ", fluke_type_name(sobp->type), sobp);
	if (sobp->label)
		printf("label=`%s'", sobp->label);
	else
		printf("no label");
#ifdef CONFIG_FLOB_CHECK
	printf(", %schecked", sobp->flob_debug ? "" : "un");
#endif
	printf(", %sactive, %d links, hash=%#x, pa=%#x",
	       sobp->active ? "" : "in", sobp->refs.counter,
	       sobp->u_hash, sobp->u_pa);
#if 0 /* XXX don't bother printing this since we have the lock */
	if (sobp->lock.owner)
		printf(" locked (owner=%p)", sobp->lock.owner);
	else
		printf(" unlocked");
#endif
#if FLASK
	printf(", sid=%ld", sobp->sid);
#endif
	putchar('\n');
	for (i = 0; i < indent; i++) putchar(' ');
	printf("creation: thread=%p\n", sobp->st->thread);
	for (i = 0; i < indent; i++) putchar(' ');
	printf("          create_st:");
	s_thread_print_stacktrace((s_thread_t *)sobp->st->thread,
				  sobp->st->buf,
				  sizeof(sobp->st->buf)/
				  sizeof(sobp->st->buf[0]));

	switch(sobp->type) {
	case FLUKE_THREAD:
		s_thread_dump((s_thread_t *)sobp, indent+1);
		break;
	case FLUKE_TASK:
		s_task_dump((s_task_t *)sobp, indent+1);
		break;
	case FLUKE_MUTEX:
		s_mutex_dump((s_mutex_t *)sobp, indent+1);
		break;
	case FLUKE_COND:
		s_cond_dump((s_cond_t *)sobp, indent+1);
		break;
	case FLUKE_REF:
		s_ref_dump((s_ref_t *)sobp, indent+1);
		break;
	case FLUKE_MAPPING:
		s_mapping_dump((s_mapping_t *)sobp, indent+1);
		break;
	case FLUKE_REGION:
		s_region_dump((s_region_t *)sobp, indent+1);
		break;
	case FLUKE_PORT:
		s_port_dump((s_port_t *)sobp, indent+1);
		break;
	case FLUKE_PSET:
		s_pset_dump((s_pset_t *)sobp, indent+1);
		break;
	default:
		break;
	}

	return 0;
}

int sys_debug_dump_flob()
{
	struct s_thread *cur_th = current_thread();
	oskit_addr_t sob_ua;
	s_ob_t *sobp;
	int rc;
	
	sob_ua = EXC_SYSPARAM(cur_th, debug_dump_flob, sob);

	if (sob_ua == 0) {
		s_ob_dump_threads();
		return 0;
	}

	/* Get the kernel s_ob--locks the ob */
	rc = task_ob_lookup(cur_th, sob_ua, FLUKE_PROT_READ, 1, &sobp);

	if (rc)
		return rc;

	rc = s_debug_dump_flob(sobp, 0);

	s_ob_unlock(sobp);

	return rc;
}

int sys_debug_puts()
{
	struct s_thread *cur_th = current_thread();
	char str[FLUKE_DEBUG_PUTS_STR_LEN];
	int rc;
	
	/* If NULL str, return */
	if (!EXC_SYSPARAM(cur_th, debug_puts, str))
		return 0;
	
	if (EXC_SYSPARAM(cur_th, debug_puts, len) >= FLUKE_DEBUG_PUTS_STR_LEN) {
		/* XXX */
		printf("--fluke_debug_puts() str too long--\n");
		return 0;
	}

	rc = copyin_local(EXC_SYSPARAM(cur_th, debug_puts, str),
			  str, FLUKE_DEBUG_PUTS_STR_LEN/sizeof(oskit_reg_t));
			   
	if (rc)
		return rc;

	/* XXX assumes str is null terminated... */
	printf(str);

	return 0;
}

int sys_debug_label_flob()
{
	struct s_thread *cur_th = current_thread();
	oskit_addr_t sob_ua;
	s_ob_t *sobp;
	char label[FLUKE_DEBUG_LABEL_STR_MAXLEN];
	int rc;

	sob_ua = EXC_SYSPARAM(cur_th, debug_label_flob, sob);
	
	/* Get the kernel s_ob--locks the ob */
	rc = task_ob_lookup(cur_th, sob_ua, FLUKE_PROT_READ, 1, &sobp);

	if (rc)
		return rc;

	rc = copyin_local(EXC_SYSPARAM(cur_th, debug_label_flob, str),
			  label, 
			  FLUKE_DEBUG_LABEL_STR_MAXLEN/sizeof(oskit_reg_t));
	if (rc)
		return rc;
	
	s_ob_label(sobp, label);

	/* Unlock ob */
	s_ob_unlock(sobp);

	return 0;
}

/*
 * Called before every syscall.
 * Caller register state has been saved in exc_state.
 * Invoking syscall EIP is in entry_eax (same as exc_state.eip).
 *
 * NOTE: entry_eax is not a standard call-by-value parameter.
 * Its value is restored from the stack and used to invoke the handler,
 * so don't screw with it unless you really mean it!
 */
void sys_debug_pre_syscall(unsigned int entry_eax)
{
	s_thread_t *cur_thread = current_thread();

	cur_thread->kr_stats.calls++;

	/*
	 * Record a backtrace leading up to the syscall.
	 */
	s_thread_store_stacktrace(cur_thread, cur_thread->md.syscall_st,
				  S_THREAD_SYSCALL_ST_SIZE);

	/*
	 * Do assertions after the trace has been recorded so we have some
	 * debug info (panic dumps the thread_state).
	 */
#ifndef IOPL0
	/*
	 * XXX some syscalls are special
	 */
	switch (entry_eax & 0xfff) { 
	case 0xfe0:	/* special: for do_in_supervisor_mode */
		return;
	default:
		break;
	}
#endif

#if defined(CONFIG_HIER_SCHED) && defined(DEBUG)
	if (stride_check_tree()) {
		printf("sys_debug.c: failed while checking sched tree.\n");
		sys_debug_dump_flob();
		s_ob_dump_threads();
		stride_dump_tree();
		panic("ERROR checking the hierarchical scheduling tree\n");
	}
#endif

	assert_interrupts_enabled();
	assert_preemption_enabled();
}

/*
 * Called after every syscall.
 * Fluke kernel return value (KR_*) is in exit_eax.
 *
 * NOTE: exit_eax is not a standard call-by-value parameter.
 * Its value is restored from the stack and used to determine if
 * the syscall should restart, an exception occured, etc,
 * so don't screw with it unless you really mean it!
 */
void sys_debug_post_syscall(unsigned int exit_eax)
{
	s_thread_t *cur_thread = current_thread();

	switch (exit_eax) {
	case KR_USER_EXCEPTION:
		cur_thread->kr_stats.uexceptions++;
		break;
	case KR_PAGE_FAULT:
		cur_thread->kr_stats.pagefaults++;
		break;
	case KR_CANCEL:
		cur_thread->kr_stats.cancels++;
		break;
	case KR_RESTART:
		cur_thread->kr_stats.restarts++;
		break;
#if FLASK
	case KR_SECURITY_FAULT:
		cur_thread->kr_stats.secfaults++;
		break;
#endif
	}
	/* interrupts are no longer disabled in user mode */
#ifndef IOPL0
	/*
	 * XXX some syscalls are special
	 */
	switch (cur_thread->exc_state.eip & 0xfff) {
	case 0xfe0:	/* special: for do_in_supervisor_mode */
		return;
	default:
		break;
	}
#endif

#if defined(CONFIG_HIER_SCHED) && defined(DEBUG)
	if (stride_check_tree()) {
		sys_debug_dump_flob();
		s_ob_dump_threads();
		stride_dump_tree();
		panic("ERROR checking the hierarchical scheduling tree\n");
	}
#endif
	assert_interrupts_enabled();
	assert_preemption_enabled();
}

#else /* not DEBUG */

struct s_ob;

int s_debug_dump_flob(struct s_ob *sobp)      { return 0; }
int sys_debug_dump_flob() 	              { return 0; }
int sys_debug_puts()		              { return 0; }
int sys_debug_label_flob()	              { return 0; }

#endif /* not DEBUG */

#ifdef MEMSTATS

int sys_debug_dump_memuse()
{
	extern void kalloc_dump_mstats();
	extern void frag_dump_mstats();
	extern void s_ob_dump_mstats();
	extern void kserver_dump_mstats();

	kalloc_dump_mstats();
	frag_dump_mstats();
	s_ob_dump_mstats();
	kserver_dump_mstats();

	return 0;
}
#else

int sys_debug_dump_memuse()	              { return 0; }

#endif

/* Every kernel should have a _real_ honest-to-goodness null system call! */
int sys_debug_null_syscall()			{ return 0; }

#include "pmcount.h"
#ifndef DEBUG
#include <fluke/machine/syscall.h>
#include "s_ob.h"
#include "s_thread.h"
#endif

#ifdef PMCOUNT
/*
 * Set values in the two P6 performance monitoring counters
 */
int sys_debug_set_monregs()
{
	s_thread_t *cur_th = current_thread();
	int reg0, reg1;

	/* XXX check to make sure this is a PPro/PII */

	reg0 = EXC_SYSPARAM(cur_th, debug_set_monregs, reg0);
	reg1 = EXC_SYSPARAM(cur_th, debug_set_monregs, reg1);

	/* XXX don't set APIC interrupt enable */
	reg0 &= ~PMR_SELINT;
	reg1 &= ~PMR_SELINT;

	/* reset the counters */
	set_msr(PMR_CNT0, 0);
	set_msr(PMR_CNT1, 0);

	set_msr(PMR_SEL1, reg1);
	set_msr(PMR_SEL0, reg0);

	/* enable/disable user mode access to counters */
	if (reg0 == 0 && reg1 == 0)
		set_cr4(get_cr4() & ~CR4_PCE);
	else
		set_cr4(get_cr4() | CR4_PCE);

	return 0;
}
#endif
