/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * System calls for manipulating mapping objects.
 */

#include <fluke/mapping.h>
#include <fluke/exception.h>
#include <fluke/machine/syscall.h>

#include "s_thread.h"
#include "s_mapping.h"
#include "xfer.h"
#include "debug.h"
#if FLASK
#include "avc.h"
#endif

int sys_mapping_create_hash()
{
	struct s_thread *cur_th = current_thread();
	s_mapping_t *new_mapping;
	int rc;

	rc = s_mapping_create(EXC_SYSPARAM(cur_th, mapping_create_hash, mapping),
			      EXC_SYSPARAM(cur_th, mapping_create_hash, hash),
			      &new_mapping);
	if (rc)
		return rc; /* XXX no memory */

	EXC_SYSPARAM(cur_th, mapping_create_hash, hash) =
		s_ob_gethash(&new_mapping->sob);

	s_mapping_unlock(new_mapping);

	EXC_SYSPARAM(cur_th, mapping_create_hash, rc) = 0;
	return 0;
}

int sys_mapping_destroy()
{
	struct s_thread *cur_th = current_thread();
	s_mapping_t *mapping;
	int rc;

	/*
	 * Look up the mapping
	 */
	rc = task_ob_destroy(cur_th,
			     EXC_SYSPARAM(cur_th, mapping_destroy, mapping),
			     FLUKE_MAPPING, (s_ob_t **)&mapping);
	if (rc) return rc;

	s_mapping_destroy(mapping);

	return (0);
}

int sys_mapping_protect()
{
	struct s_thread *cur_th = current_thread();
	s_mapping_t *mapping;
	int rc;

	rc = task_ob_lookup_type(cur_th,
				 EXC_SYSPARAM(cur_th, mapping_protect, mapping),
				 FLUKE_PROT_WRITE, FLUKE_MAPPING,
				 (s_ob_t **)&mapping);
	if (rc) return rc;

	s_mapping_protect(mapping, EXC_SYSPARAM(cur_th, mapping_protect, prot));

	s_mapping_unlock(mapping);
	return 0;
}

int sys_mapping_get_state()
{
	struct s_thread *cur_th = current_thread();
	s_mapping_t *mapping = 0;
	fluke_mapping_state_t state, *statep = 0;
	struct link region_link, *region_linkp = 0;
	struct link task_link, *task_linkp = 0;
	int rc;

	/*
	 * Determine which values we want
	 */
	if (EXC_SYSPARAM(cur_th, mapping_get_state, state))
		statep = &state;
	if (EXC_SYSPARAM(cur_th, mapping_get_state, region))
		region_linkp = &region_link;
	if (EXC_SYSPARAM(cur_th, mapping_get_state, task))
		task_linkp = &task_link;

	/*
	 * Lookup the mapping
	 */
	rc = task_ob_lookup_type(cur_th,
				 EXC_SYSPARAM(cur_th, mapping_get_state, mapping),
				 FLUKE_PROT_READ, FLUKE_MAPPING,
				 (s_ob_t **)&mapping);
	if (rc) return rc;

	/*
	 * Get its state
	 */
	rc = s_mapping_get_state(mapping, statep, region_linkp, task_linkp);
	s_mapping_unlock(mapping);
	if (rc) return rc;

	/*
	 * Copyout state
	 */
	if (statep) {
		rc = copyout_local(statep,
				   EXC_SYSPARAM(cur_th, mapping_get_state, state),
				   sizeof(state) / sizeof(oskit_reg_t));
		if (rc) goto out;
	}
	if (region_linkp)
	{
		rc = kentry_ref_associate(cur_th,
				EXC_SYSPARAM(cur_th, mapping_get_state, region),
				region_linkp);
		region_linkp = 0;
		if (rc) goto out;
	}
	if (task_linkp)
	{
		rc = kentry_ref_associate(cur_th,
				EXC_SYSPARAM(cur_th, mapping_get_state, task),
				task_linkp);
		task_linkp = 0;
		if (rc) goto out;
	}
out:
	if (region_linkp)
		link_destroy(region_linkp);
	if (task_linkp)
		link_destroy(task_linkp);
	return rc;
}

int sys_mapping_set_state()
{
	struct s_thread *cur_th = current_thread();
	s_mapping_t *mapping = 0;
	fluke_mapping_state_t state, *statep = 0;
	struct link task_link, *task_linkp = 0;
	struct link region_link, *region_linkp = 0;
	int rc;


	/*
	 * Copy in the simple state structure if there is one.
	 */
	if (EXC_SYSPARAM(cur_th, mapping_set_state, state))
	{
		rc = copyin_local(EXC_SYSPARAM(cur_th, mapping_set_state, state),
				  &state, sizeof(state) / sizeof(oskit_reg_t));
		if (rc) return rc;
		statep = &state;

		/*
		 * XXX this check used to be in s_mapping_set_state but
		 * it fails when called from the kernel context (i.e.,
		 * when creating the initial kernel mapping) so we do the
		 * check here.
		 */
		if (statep->start_va < FLUKE_ENTRY_MAX &&
		    statep->start_va + statep->size > FLUKE_ENTRY_MIN)
			EXC_RETURN_INSANITY(INVALID_ADDRESS);
	}

	/*
	 * Look up the references to insert into the mapping.
	 */
	if (EXC_SYSPARAM(cur_th, mapping_set_state, task))
	{
		rc = kentry_ref_extract_type(cur_th,
				EXC_SYSPARAM(cur_th, mapping_set_state, task),
				FLUKE_TASK, &task_link);
		if (rc) goto out;
		task_linkp = &task_link;
	}
	if (EXC_SYSPARAM(cur_th, mapping_set_state, region))
	{
		rc = kentry_ref_extract_type(cur_th,
				EXC_SYSPARAM(cur_th, mapping_set_state, region),
				FLUKE_REGION, &region_link);
		if (rc) goto out;
		region_linkp = &region_link;
	}

	/*
	 * Lookup and lock the actual mapping to modify.
	 */
	rc = task_ob_lookup_type(cur_th,
				 EXC_SYSPARAM(cur_th, mapping_set_state, mapping),
				 FLUKE_PROT_WRITE, FLUKE_MAPPING,
				 (s_ob_t **)&mapping);
	if (rc) goto out;

#if FLASK
	if (task_linkp)
	{
	    struct s_ob *sob;

	    sob = link_follow(task_linkp);
	    if (sob)
	    {
		rc = avc_sid_equiv(FLUKE_MAPPING,
				   mapping->sob.sid,
				   FLUKE_TASK,
				   sob->sid);
		if (rc)
		{
		    s_ob_unlock(sob);
		    goto out;
		}
		s_ob_unlock(sob);
	    }
	}

	if (region_linkp)
	{
	    struct s_ob *sob;


	    sob = link_follow(region_linkp);
	    if (sob)
	    {
		rc = avc_has_permission(SECCLASS_SUBJECT,
				    	mapping->sob.sid,
				     	SECCLASS_SUBJECT,
					sob->sid,
					SUBJECT_SUBJECT__MAP);
		if (rc)
		{
		    s_ob_unlock(sob);
		    goto out;
		}
		s_ob_unlock(sob);
	    }
	}
#endif

	rc = s_mapping_set_state(mapping, statep, region_linkp, task_linkp);

	out:
	if (mapping)
		s_mapping_unlock(mapping);
	if (task_linkp)
		link_destroy(&task_link);
	if (region_linkp)
		link_destroy(&region_link);

	return rc;
}

int sys_mapping_move()
{
	struct s_thread *cur_th = current_thread();
	s_mapping_t *mapping;
	int rc;

	rc = task_ob_move(cur_th, EXC_SYSPARAM(cur_th, mapping_move, oaddr),
			  EXC_SYSPARAM(cur_th, mapping_move, naddr),
			  FLUKE_MAPPING, FLUKE_MAPPING_SIZE, 0,
			  (s_ob_t **)&mapping);
	if (rc == 0)
		s_mapping_unlock(mapping);
	return (rc);
}
