/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * System calls for manipulating mutex objects.
 */

#include <fluke/mutex.h>
#include <fluke/machine/syscall.h>
#include <oskit/machine/base_vm.h>

#include "s_thread.h"
#include "s_mutex.h"
#include "xfer.h"
#include "debug.h"

int sys_mutex_create_hash()
{
	struct s_thread *cur_th = current_thread();
	s_mutex_t *new_mutex;
	int rc;

	rc = s_mutex_create(EXC_SYSPARAM(cur_th, mutex_create_hash, mutex),
			    EXC_SYSPARAM(cur_th, mutex_create_hash, hash),
			    &new_mutex);
	if (rc)
		return rc; /* XXX no memory */

	EXC_SYSPARAM(cur_th, mutex_create_hash, hash) =
		s_ob_gethash(&new_mutex->sob);

	s_mutex_unlock(new_mutex);

	EXC_SYSPARAM(cur_th, mutex_create_hash, rc) = 0;
	return 0;
}

int sys_mutex_destroy()
{
	struct s_thread *cur_th = current_thread();
	s_mutex_t *mutex;
	int rc;

	rc = task_ob_destroy(cur_th, EXC_SYSPARAM(cur_th, mutex_destroy, mutex),
			     FLUKE_MUTEX, (s_ob_t **)&mutex);
	if (rc) return rc;

	return s_mutex_destroy(mutex);
}

int sys_mutex_get_state()
{
	struct s_thread *cur_th = current_thread();
	s_mutex_t *mutex = 0;
	fluke_mutex_state_t state, *statep = 0;
	struct link owner_link, *owner_linkp = 0;
	int rc;

	/*
	 * Determine which values we want
	 */
	if (EXC_SYSPARAM(cur_th, mutex_get_state, state))
		statep = &state;
	if (EXC_SYSPARAM(cur_th, mutex_get_state, owner))
		owner_linkp = &owner_link;

	/*
	 * Lookup the mutex
	 */
	rc = task_ob_lookup_type(cur_th,
				 EXC_SYSPARAM(cur_th, mutex_get_state, mutex),
				 FLUKE_PROT_READ, FLUKE_MUTEX,
				 (s_ob_t **)&mutex);
	if (rc) return rc;

	/*
	 * Get its state
	 */
	rc = s_mutex_get_state(mutex, statep, owner_linkp);
	s_mutex_unlock(mutex);
	if (rc) return rc;

	/*
	 * Copyout state
	 */
	if (statep) {
		rc = copyout_local(statep,
				   EXC_SYSPARAM(cur_th, mutex_get_state, state),
				   sizeof(state) / sizeof(oskit_reg_t));
		if (rc) goto out;
	}
	if (owner_linkp)
	{
		rc = kentry_ref_associate(cur_th,
				EXC_SYSPARAM(cur_th, mutex_get_state, owner),
				owner_linkp);
		owner_linkp = 0;
		if (rc) goto out;
	}
out:
	if (owner_linkp)
		link_destroy(owner_linkp);
	return rc;
}

int sys_mutex_set_state()
{
	struct s_thread *cur_th = current_thread();
	s_mutex_t *mutex = 0;
	fluke_mutex_state_t state, *statep = 0;
	struct link owner_link, *owner_linkp = 0;
	int rc;


	/*
	 * Copy in the simple state structure if there is one.
	 */
	if (EXC_SYSPARAM(cur_th, mutex_set_state, state))
	{
		rc = copyin_local(EXC_SYSPARAM(cur_th, mutex_set_state, state),
				  &state, sizeof(state) / sizeof(oskit_reg_t));
		if (rc) return rc;
		statep = &state;
	}

	/*
	 * Look up the references to insert into the mutex.
	 */
	if (EXC_SYSPARAM(cur_th, mutex_set_state, owner))
	{
		rc = kentry_ref_extract_type(cur_th,
				EXC_SYSPARAM(cur_th, mutex_set_state, owner),
				FLUKE_THREAD, &owner_link);
		if (rc) goto out;
		owner_linkp = &owner_link;
	}

	/*
	 * Lookup and lock the mutex
	 */
	rc = task_ob_lookup_type(cur_th,
				 EXC_SYSPARAM(cur_th, mutex_set_state, mutex),
				 FLUKE_PROT_WRITE, FLUKE_MUTEX,
				 (s_ob_t **)&mutex);
	if (rc) goto out;

	rc = s_mutex_set_state(mutex, statep, owner_linkp);

	out:
	if (mutex)
		s_mutex_unlock(mutex);
	if (owner_linkp)
		link_destroy(&owner_link);

	return rc;
}

int sys_mutex_lock()
{
	struct s_thread *cur_th = current_thread();
	s_mutex_t *mutex;
	int rc;

	rc = task_ob_lookup_type(cur_th, EXC_SYSPARAM(cur_th, mutex_lock, mutex),
				 FLUKE_PROT_WRITE, FLUKE_MUTEX,
				 (s_ob_t **)&mutex);
	if (rc) return rc;

	rc = s_mutex_lockop(mutex, cur_th);

	return rc;
}

#ifndef MD_SYS_MUTEX_TRYLOCK
int sys_mutex_trylock()
{
	struct s_thread *cur_th = current_thread();
	s_mutex_t *mutex;
	int rc;

	rc = task_ob_lookup_type(cur_th, EXC_SYSPARAM(cur_th, mutex_trylock, mutex),
				 FLUKE_PROT_WRITE, FLUKE_MUTEX,
				 (s_ob_t **)&mutex);
	if (rc) return rc;

	rc = s_mutex_trylockop(mutex);

	s_mutex_unlock(mutex);

	if (rc < KR_MIN)
	{
		EXC_SYSPARAM(cur_th, mutex_trylock, rc) = rc;
		rc = 0;
	}
	return rc;
}
#endif

int sys_mutex_unlock()
{
	struct s_thread *cur_th = current_thread();
	s_mutex_t *mutex;
	int rc;

	rc = task_ob_lookup_type(cur_th, EXC_SYSPARAM(cur_th, mutex_unlock, mutex),
				 FLUKE_PROT_WRITE, FLUKE_MUTEX,
				 (s_ob_t **)&mutex);
	if (rc) return rc;

	rc = s_mutex_unlockop(mutex, cur_th);

	s_mutex_unlock(mutex);

	return rc;
}

int sys_mutex_move()
{
	struct s_thread *cur_th = current_thread();
	s_mutex_t *mutex;
	int rc;

	rc = task_ob_move(cur_th, EXC_SYSPARAM(cur_th, mutex_move, oaddr),
			  EXC_SYSPARAM(cur_th, mutex_move, naddr),
			  FLUKE_MUTEX, FLUKE_MUTEX_SIZE, 0, (s_ob_t **)&mutex);
	if (rc == 0) {
#ifdef CONFIG_OPT_SYNCH
		/*
		 * Update the pointer to the user-level flob data.
		 */
		mutex->u_mutex = (u_mutex_t *) phystokv(mutex->sob.u_pa);
#endif
		s_mutex_unlock(mutex);
	}
	return (rc);
}

int sys_mutex_reference()
{
	struct s_thread *cur_th = current_thread();

	return kentry_ob_reference(cur_th,
				   EXC_SYSPARAM(cur_th, mutex_reference, mutex),
				   EXC_SYSPARAM(cur_th, mutex_reference, ref),
				   FLUKE_MUTEX);
}
