/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * System calls for manipulating port objects.
 */

#include <fluke/port.h>
#include <fluke/machine/syscall.h>

#include "s_thread.h"
#include "s_port.h"
#include "xfer.h"
#if FLASK
#include "avc.h"
#endif         

int sys_port_create_hash()
{
	struct s_thread *cur_th = current_thread();
	s_port_t *new_port;
	int rc;

	rc = s_port_create(EXC_SYSPARAM(cur_th, port_create_hash, port),
			   EXC_SYSPARAM(cur_th, port_create_hash, hash),
			   &new_port);
	if (rc)
		return rc; /* XXX no memory */

	EXC_SYSPARAM(cur_th, port_create_hash, hash) =
		s_ob_gethash(&new_port->sob);

	s_port_unlock(new_port);

	EXC_SYSPARAM(cur_th, port_create_hash, rc) = 0;
	return 0;
}

int sys_port_destroy()
{
	struct s_thread *cur_th = current_thread();
	s_port_t *port;
	int rc;

	/*
	 * Look up the port
	 */
	rc = task_ob_destroy(cur_th, EXC_SYSPARAM(cur_th, port_destroy, port),
			     FLUKE_PORT, (s_ob_t **)&port);
	if (rc) return rc;

	s_port_destroy(port);

	return (0);
}

int sys_port_get_state()
{
	struct s_thread *cur_th = current_thread();
	s_port_t *port = 0;
	fluke_port_state_t state, *statep = 0;
	struct link pset_link, *pset_linkp = 0;
	int rc;

	/*
	 * Determine which values we want
	 */
	if (EXC_SYSPARAM(cur_th, port_get_state, state))
		statep = &state;
	if (EXC_SYSPARAM(cur_th, port_get_state, pset))
		pset_linkp = &pset_link;

	/*
	 * Lookup the port
	 */
	rc = task_ob_lookup_type(cur_th,
				 EXC_SYSPARAM(cur_th, port_get_state, port),
				 FLUKE_PROT_READ, FLUKE_PORT,
				 (s_ob_t **)&port);
	if (rc) return rc;

	/*
	 * Get its state
	 */
	rc = s_port_get_state(port, statep, pset_linkp);
	s_port_unlock(port);
	if (rc) return rc;

	/*
	 * Copyout state
	 */
	if (statep) {
		rc = copyout_local(statep,
				   EXC_SYSPARAM(cur_th, port_get_state, state),
				   sizeof(state) / sizeof(oskit_reg_t));
		if (rc) goto out;
	}
	if (pset_linkp)
	{
		rc = kentry_ref_associate(cur_th,
				EXC_SYSPARAM(cur_th, port_get_state, pset),
				pset_linkp);
		pset_linkp = 0;
		if (rc) goto out;
	}
out:
	if (pset_linkp)
		link_destroy(pset_linkp);
	return rc;
}

int sys_port_set_state()
{
	struct s_thread *cur_th = current_thread();
	s_port_t *port = 0;
	fluke_port_state_t state, *statep = 0;
	struct link pset_link, *pset_linkp = 0;
	int rc;


	/*
	 * Copy in the simple state structure if there is one.
	 */
	if (EXC_SYSPARAM(cur_th, port_set_state, state))
	{
		rc = copyin_local(EXC_SYSPARAM(cur_th, port_set_state, state),
				  &state, sizeof(state) / sizeof(oskit_reg_t));
		if (rc) return rc;
		statep = &state;
	}

	/*
	 * Look up the references to insert into the port.
	 */
	if (EXC_SYSPARAM(cur_th, port_set_state, pset))
	{
		rc = kentry_ref_extract_type(cur_th,
				EXC_SYSPARAM(cur_th, port_set_state, pset),
				FLUKE_PSET, &pset_link);
		if (rc) goto out;
		pset_linkp = &pset_link;
	}

	/*
	 * Lookup and lock the port
	 */
	rc = task_ob_lookup_type(cur_th,
				 EXC_SYSPARAM(cur_th, port_set_state, port),
				 FLUKE_PROT_WRITE, FLUKE_PORT,
				 (s_ob_t **)&port);
	if (rc) goto out;

#if FLASK
	if (pset_linkp)
	{
	    struct s_ob *sob;

	    sob = link_follow(pset_linkp);
	    if (sob)
	    {
		rc = avc_sid_equiv(FLUKE_PORT,
				   port->sob.sid,
				   FLUKE_PSET,
				   sob->sid); 
		if (rc)
		{
		    s_ob_unlock(sob);
		    goto out;
		}
		s_ob_unlock(sob);
	    }	    
	}	
#endif 

	rc = s_port_set_state(port, statep, pset_linkp);

	out:
	if (port)
		s_port_unlock(port);
	if (pset_linkp)
		link_destroy(&pset_link);

	return rc;
}

int sys_port_move()
{
	struct s_thread *cur_th = current_thread();
	s_port_t *port;
	int rc;

	rc = task_ob_move(cur_th, EXC_SYSPARAM(cur_th, port_move, oaddr),
			  EXC_SYSPARAM(cur_th, port_move, naddr),
			  FLUKE_PORT, FLUKE_PORT_SIZE, 0, (s_ob_t **)&port);
	if (rc == 0)
		s_port_unlock(port);
	return (rc);
}

int sys_port_reference()
{
	struct s_thread *cur_th = current_thread();

	return kentry_ob_reference(cur_th,
				   EXC_SYSPARAM(cur_th, port_reference, port),
				   EXC_SYSPARAM(cur_th, port_reference, ref),
				   FLUKE_PORT);
}
