/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * System calls for manipulating reference objects.
 *
 * XXX do we need a call to NULL out a reference?  Or is destroy/recreate
 * sufficient?  We need this internally for get_state where we might fail
 * after some references have already been copied out.  In that case we
 * really don't want to return failure and yet have some references valid.
 * In this case, we cannot destroy them and we cannot put them back to their
 * original state, so we just NULL them out.  Might there be a higher-level
 * "get state" protocol that will need similar capabilities?
 */

#include <fluke/ref.h>
#include <fluke/machine/syscall.h>
#include <oskit/machine/base_vm.h>

#include "s_thread.h"
#include "s_ref.h"
#include "xfer.h"

#ifdef CONFIG_OPT_REF
#include "u_ref.h"
#endif

int sys_ref_create()
{
	struct s_thread *cur_th = current_thread();
	int rc;

	rc = s_ref_create(EXC_SYSPARAM(cur_th, ref_create, ref));
	if (rc)
		return rc; /* XXX no memory */

	EXC_SYSPARAM(cur_th, ref_create, rc) = 0;
	return 0;
}

int sys_ref_destroy()
{
	struct s_thread *cur_th = current_thread();
	s_ref_t *ref;
	int rc;

	/*
	 * Look up the reference
	 */
	rc = task_ob_destroy(cur_th, EXC_SYSPARAM(cur_th, ref_destroy, ref),
			     FLUKE_REF, (s_ob_t **)&ref);
	if (rc) return rc;

	s_ref_destroy(ref);

	return (0);
}

int sys_ref_move()
{
	struct s_thread *cur_th = current_thread();
	s_ref_t *ref;
	int rc;

	rc = task_ob_move(cur_th, EXC_SYSPARAM(cur_th, ref_move, oaddr),
			  EXC_SYSPARAM(cur_th, ref_move, naddr),
			  FLUKE_REF, FLUKE_REF_SIZE,
#ifdef CONFIG_OPT_REF
			  sizeof(u_ref_t),
#else
			  0,
#endif
			  (s_ob_t **)&ref);
	if (rc == 0) {
#ifdef CONFIG_OPT_REF
		/*
		 * Update the pointer to the user-level flob data.
		 */
		ref->u_ref = (struct u_ref *) phystokv(ref->sob.u_pa);
#endif

		s_ref_unlock(ref);
	}
	return (rc);
}

int sys_ref_check()
{
	struct s_thread *cur_th = current_thread();
	s_ref_t *ref;
	int rc;

	rc = task_ob_lookup_type(cur_th, EXC_SYSPARAM(cur_th, ref_check, ref),
				 FLUKE_PROT_READ, FLUKE_REF, (s_ob_t **)&ref);
	if (rc) return rc;
	rc = s_ref_check(ref, &EXC_SYSPARAM(cur_th, ref_check, result));
	/* that unlocks ref */

	return rc;
}

#ifndef CONFIG_OPT_REF
int sys_ref_compare()
{
	struct s_thread *cur_th = current_thread();
	s_ref_t *ref;
	link_t link1, link2;
	int rc;

	rc = task_ob_lookup_type(cur_th, EXC_SYSPARAM(cur_th, ref_compare, ref1),
				 FLUKE_PROT_READ, FLUKE_REF, (s_ob_t **)&ref);
	if (rc)
		return rc;
	s_ref_create_link(ref, &link1);
	s_ref_unlock(ref);

	rc = task_ob_lookup_type(cur_th, EXC_SYSPARAM(cur_th, ref_compare, ref2),
				 FLUKE_PROT_READ, FLUKE_REF, (s_ob_t **)&ref);
	if (rc)
	{
		link_destroy(&link1);
		return rc;
	}
	s_ref_create_link(ref, &link2);
	s_ref_unlock(ref);

	rc = link_compare(&link1, &link2);
	link_destroy(&link1);
	link_destroy(&link2);

	EXC_SYSPARAM(cur_th, ref_compare, result) = rc;
	return 0;
}
#endif

int sys_ref_copy()
{
	struct s_thread *cur_th = current_thread();
	s_ref_t *ref;
	link_t link;
	int rc;

	rc = task_ob_lookup_type(cur_th, EXC_SYSPARAM(cur_th, ref_copy, fref),
				 FLUKE_PROT_READ, FLUKE_REF, (s_ob_t **)&ref);
	if (rc) return rc;
	s_ref_create_link(ref, &link);
	s_ref_unlock(ref);

	rc = task_ob_lookup_type(cur_th, EXC_SYSPARAM(cur_th, ref_copy, tref),
				 FLUKE_PROT_WRITE, FLUKE_REF, (s_ob_t **)&ref);
	if (rc)
	{
		link_destroy(&link);
		return rc;
	}
	s_ref_link_transfer(&link, ref);
	s_ref_unlock(ref);

	link_destroy(&link);
	return 0;
}

int
sys_ref_type()
{
	int rc;
	s_ob_t *ob;
	link_t link;
	s_ref_t *ref;
	struct s_thread *cur_th = current_thread();


	rc = task_ob_lookup_type(cur_th, EXC_SYSPARAM(cur_th, ref_type, ref),
				 FLUKE_PROT_READ, FLUKE_REF, (s_ob_t **)&ref);
	if (rc)
		return (rc);
	s_ref_create_link(ref, &link);
	s_ref_unlock(ref);
	ob = link_consume(&link);
	if (ob) {
		EXC_SYSPARAM(cur_th, ref_type, result_type) = ob->type;
		s_ob_unlock(ob);
	} else
		EXC_SYSPARAM(cur_th, ref_type, result_type) = FLUKE_NULL;
	return (0);
}

#ifndef CONFIG_OPT_REF
int
sys_ref_hash()
{
	int rc;
	s_ob_t *ob;
	link_t link;
	s_ref_t *ref;
	struct s_thread *cur_th = current_thread();


	rc = task_ob_lookup_type(cur_th, EXC_SYSPARAM(cur_th, ref_hash, ref),
				 FLUKE_PROT_READ, FLUKE_REF, (s_ob_t **)&ref);
	if (rc)
		return rc;

	s_ref_create_link(ref, &link);
	s_ref_unlock(ref);
	ob = link_consume(&link);
	if (ob) {
		EXC_SYSPARAM(cur_th, ref_hash, result_hash) = ob->u_hash;
		s_ob_unlock(ob);
	} else
		EXC_SYSPARAM(cur_th, ref_hash, result_hash) = 0;

	return 0;
}
#endif

#if FLASK
int
sys_ref_sid()
{
	int rc;
	s_ob_t *ob;
	link_t link;
	s_ref_t *ref;
	struct s_thread *cur_th = current_thread();


	rc = task_ob_lookup_type(cur_th, EXC_SYSPARAM(cur_th, ref_sid, ref),
				 FLUKE_PROT_READ, FLUKE_REF, (s_ob_t **)&ref);
	if (rc)
		return (rc);

	link_create_copy(&ref->link, &link);
	s_ref_unlock(ref);
	ob = link_consume(&link);
	if (ob) {
		EXC_SYSPARAM(cur_th, ref_sid, result_sid) = ob->sid;
		s_ob_unlock(ob);
	} else {
		EXC_SYSPARAM(cur_th, ref_sid, result_sid) = SECSID_NULL;
	}

	return (0);
}


/* XXX doesn't belong here */
extern link_t security_port_link;

int sys_security_reference(void)
{
    struct s_thread *cur_th = current_thread();
    link_t link;

    /* XXX permission check? */

    link_create_copy(&security_port_link, &link);
    return kentry_ref_associate(cur_th,
				EXC_SYSPARAM(cur_th, security_reference, ref),
				&link);
}
#endif 
