/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * System calls relating to threads.
 */

#include <string.h>
#include <fluke/thread.h>
#include <fluke/exception.h>
#include <fluke/machine/syscall.h>
#include <oskit/machine/base_vm.h>

#include "s_thread.h"
#include "s_ref.h"
#include "xfer.h"
#include "wait.h"
#include "debug.h"

/*#define BASE_DEBUG*/
#ifdef BASE_DEBUG
#undef SOURCE
#define SOURCE "sys_thread"
#define BD_BEGIN() DEBUG_BASE_BEGIN()
#define BD_END() DEBUG_BASE_END()
#define BD_BEGIN_ARGS(fmt, args...) DEBUG_BASE_BEGIN_ARGS(fmt,args)
#define BD_END_ARGS(fmt, args...) DEBUG_BASE_END_ARGS(fmt,args)
#define BD() DEBUG_BASE()
#define BD_ARGS(fmt, args...) DEBUG_BASE_ARGS(fmt,args)
#else
#define BD_BEGIN()
#define BD_END()
#define BD()
#define BD_ARGS(args...)
#endif

#include "config_hier_sched.h"
#include "config_hier_sched_set_parent.h"

#if FLASK
#include "avc.h"
#endif 

int sys_thread_create_hash()
{
	struct s_thread *cur_th = current_thread();
	struct s_thread *new_th;
	int rc;

	rc = s_thread_create(EXC_SYSPARAM(cur_th, thread_create_hash, thread),
			     EXC_SYSPARAM(cur_th, thread_create_hash, hash),
			     &new_th);
	if (rc)
		return rc;

	EXC_SYSPARAM(cur_th, thread_create_hash, hash) =
		s_ob_gethash(&new_th->sob);

	/* Release the thread; it'll remain waiting in the STOPPED state. */
	s_thread_release(new_th);

	EXC_SYSPARAM(cur_th, thread_create_hash, rc) = 0;
	return 0;
}

int sys_thread_destroy(void)
{
	struct s_thread *cur_th = current_thread();
	s_thread_t *thread;
	int rc;

	/* Look up, lock, and capture the target thread. */
	rc = s_thread_stop(EXC_SYSPARAM(cur_th, thread_destroy, thread), 1,
			   &thread);
	if (rc)
		return rc;

	return s_thread_destroy(thread);
}

struct get_set_state_params
{
	oskit_addr_t thread;
	oskit_addr_t state;
	oskit_addr_t task_ref;
	oskit_addr_t scheduler_ref;
	oskit_addr_t waiting_for_ref;
	oskit_addr_t client_ref;
	oskit_addr_t server_ref;
};

#define COPYOUT_REF(name, link)						\
	if (params.name##_ref) {					\
		link_t tlink;						\
		link_create_copy(&sth->link, &tlink);			\
		rc = kentry_ref_associate(cur_th, params.name##_ref,	\
					  &tlink);			\
		if (rc) goto out;					\
	}

#if FLASK
int sys_thread_get_state_common(oskit_bool_t secure)
#else
int sys_thread_get_state(void)
#endif 
{
	struct s_thread *cur_th = current_thread();
	struct get_set_state_params params;
#if FLASK
	struct fluke_thread_state_secure state;
#else
	struct fluke_thread_state state;
#endif
	s_thread_t *sth;
	int rc;

	/*
	 * Retrieve the system call parameters from the user's stack.
	 * XXX x86 specific.
	 */
	rc = copyin_local(cur_th->exc_state.esp + 4,
			  &params, sizeof(params) / 4);
	if (rc) return rc;

	/* Look up, lock, and capture the target thread. */
	rc = s_thread_stop(params.thread, 0, &sth);
	if (rc)
		return rc;
	assert_s_thread_locked(sth);

	/*
	 * Copyout state and references
	 */
	if (params.state) {

#if FLASK
		s_thread_get_state(sth, &state.state);
		if (secure)
		{
		    s_thread_get_state_secure(sth, &state);
		    rc = copyout_local(&state, params.state,
				       sizeof(state) / sizeof(oskit_reg_t));
		}
		else
		{
		    rc = copyout_local(&state.state, params.state,
			       sizeof(state.state) / sizeof(oskit_reg_t));
		}
#else
		s_thread_get_state(sth, &state);
		rc = copyout_local(&state, params.state,
				   sizeof(state) / sizeof(oskit_reg_t));
#endif
		if (rc) goto out;
	}
	COPYOUT_REF(task, task_state.task_link);
#if defined(CONFIG_HIER_SCHED) && defined(CONFIG_HIER_SCHED_SET_PARENT)
	COPYOUT_REF(scheduler, sched_state.sched_link);
#endif
	/* COPYOUT_REF(waiting_for);*/
	COPYOUT_REF(client, ipc_state.client_link);
	COPYOUT_REF(server, ipc_state.server_link);

	/*
	 * Set the STOPPED flag in the thread's state
	 * so future calls to thread_get_state will return an indication
	 * that the thread was already stopped.
	 */
	sth->gflags |= FLUKE_THREAD_STOPPED;

	rc = 0;

	out:
	s_thread_release(sth);
	return rc;
}
#undef COPYOUT_REF

#if FLASK
int sys_thread_get_state(void)
{
    return sys_thread_get_state_common(FALSE);
}

int sys_thread_get_state_secure(void)
{
    return sys_thread_get_state_common(TRUE);
}
#endif 

#define COPYIN_REF(name, link_name)					\
	if (params.name##_ref != 0) {					\
		s_ref_t *ref;						\
		rc = task_ob_lookup_type(cur_th, params.name##_ref,	\
					FLUKE_PROT_READ, FLUKE_REF,	\
					(s_ob_t**)&ref);		\
		if (rc) goto out;					\
		link_transfer(&ref->link, &sth->link_name);		\
		s_ref_unlock(ref);					\
	}

#if FLASK
#define COPYIN_REF_OBEQUIV(name, link_name)				\
	if (params.name##_ref != 0) {					\
		struct s_ob *sob;					\
		s_ref_t *ref;						\
		rc = task_ob_lookup_type(cur_th, params.name##_ref,	\
					FLUKE_PROT_READ, FLUKE_REF,	\
					(s_ob_t**)&ref);		\
		if (rc) goto out;					\
		sob = link_follow(&ref->link);				\
		if (sob)						\
		{							\
		    rc = avc_sid_equiv(FLUKE_THREAD,                 \
				       sth->sob.sid, 			\
				       sob->type, 			\
				       sob->sid);			\
		    if (rc)						\
		    {							\
			s_ob_unlock(sob);				\
			s_ref_unlock(ref);				\
			goto out;					\
  		    }							\
		    s_ob_unlock(sob);					\
		}							\
		link_transfer(&ref->link, &sth->link_name);		\
		s_ref_unlock(ref);					\
	}
#endif 

#if FLASK
int sys_thread_set_state_common(oskit_bool_t secure)
#else
int sys_thread_set_state(void)
#endif 
{
	struct s_thread *cur_th = current_thread();
	struct get_set_state_params params;
#if FLASK
	struct fluke_thread_state_secure state;
	oskit_bool_t reset = FALSE;
#else
	struct fluke_thread_state state;
#endif
	s_thread_t *sth;
	int rc;
	BD_ARGS("thread=%p",cur_th);

	/*
	 * Retrieve the system call parameters from the user's stack.
	 * XXX x86 specific.
	 */
	rc = copyin_local(cur_th->exc_state.esp + 4,
			  &params, sizeof(params) / 4);
	if (rc) return rc;

	/*
	 * Copy in the simple state structure if there is one.
	 */
	if (params.state)
	{
#if FLASK
	        if (secure)
		{
		    rc = copyin_local(params.state, &state, sizeof(state) / 4);
		}
		else
		{
		    rc = copyin_local(params.state, &state.state, 
				      sizeof(state.state) / 4);
		}
#else
		rc = copyin_local(params.state, &state, sizeof(state) / 4);
#endif
		if (rc) return rc;
	}

	/* Look up, lock, and capture the target thread. */
	rc = s_thread_stop(params.thread, 0, &sth);
	if (rc)
		return rc;
	assert_s_thread_locked(sth);

	/*
	 * Make sure the thread is stopped.
	 * The caller is supposed to ensure this.
	 */
	if (!(sth->gflags & FLUKE_THREAD_STOPPED))
	{
		s_thread_release(sth);
		EXC_RETURN_INSANITY(THREAD_NOT_STOPPED);
	}

	/*
	 * Copy in any references to be set.
	 */

#if FLASK
	/* require SID equivalency for a thread and its task */
	COPYIN_REF_OBEQUIV(task, task_state.task_link);
#else
	COPYIN_REF(task, task_state.task_link);
#endif

#if defined(CONFIG_HIER_SCHED) && defined(CONFIG_HIER_SCHED_SET_PARENT)
	COPYIN_REF(scheduler, sched_state.sched_link);
	/* XXX HACK. To distinguish between user threads and
	   kserver threads */
	sth->sched_state.kserver = 0;
#endif

	/*COPYIN_REF(waiting_for);*/
	COPYIN_REF(client, ipc_state.client_link);
	COPYIN_REF(server, ipc_state.server_link);

#if FLASK
	if (secure && params.state)
	{
	    sth->ipc_state.server_effective_server_sid = 
		    state.server_effective_server_sid;
	    sth->ipc_state.client_effective_client_sid = 
		    state.client_effective_client_sid;
	    sth->ipc_state.effective_server_sid = 
		    state.effective_server_sid;
	    sth->ipc_state.effective_client_sid =
		    state.effective_client_sid;
	}
	else
	{
	    sth->ipc_state.server_effective_server_sid = SECSID_NULL;
	    sth->ipc_state.client_effective_client_sid = SECSID_NULL;
            sth->ipc_state.effective_server_sid = sth->sob.sid;
            sth->ipc_state.effective_client_sid = sth->sob.sid;
	}

        
        /*
         * Validate the IPC state of this thread.
         */
	rc = thread_ipc_reset(sth, &reset);
        if (rc) goto out;
        if (reset)
        {
	    exc_set_insanity(cur_th, FLUKE_INSANITY_PERMISSION_DENIED);
            rc = KR_USER_EXCEPTION;
            goto out;
        }

        /*
         * Validate the scheduler state of this thread.
         */
        rc = s_thread_scheduler_reset(sth, &reset);
        if (rc) goto out;
        if (reset)
        {
            exc_set_insanity(cur_th, FLUKE_INSANITY_PERMISSION_DENIED);
            rc = KR_USER_EXCEPTION;        
            goto out;        
        }
#endif

	/*
	 * Finally, set the simple state.  This must be done at the end
	 * so that the FLUKE_THREAD_STOPPED flag will only be cleared,
	 * allowing the thread to run, if the entire set_state was successful.
	 */
	if (params.state)
#if FLASK
		s_thread_set_state(sth, &state.state);
#else
		s_thread_set_state(sth, &state);
#endif

	rc = 0;

out:
	s_thread_release(sth);
	BD_ARGS("%s","End");
	return rc;
}
#undef COPYIN_REF

#if FLASK
#undef COPYIN_REF_OBEQUIV

int sys_thread_set_state(void)
{
    return sys_thread_set_state_common(FALSE);
}

int sys_thread_set_state_secure(void)
{
    return sys_thread_set_state_common(TRUE);
}
#endif 

#ifndef CONFIG_OPT_THREADS
int sys_thread_self()
{
	struct s_thread *cur_th = current_thread();

	EXC_SYSPARAM(cur_th, thread_self, thread) = cur_th->thread_va;

	return 0;
}
#endif

#if 1/*ndef CONFIG_OPT_THREADS */
int sys_thread_enable_exceptions()
{
	struct s_thread *cur_th = current_thread();

	s_ob_lock(&cur_th->sob);
	cur_th->th_lflags |= FLUKE_THREAD_EXCEPTION_ENABLE;

	/*
	 * Was an interrupt sent while interrupts were disabled?
	 * If so, go ahead and process it now, since I can.
	 */
	if (cur_th->gflags & FLUKE_THREAD_INTERRUPT_PENDING) {
		thread_cancel(cur_th);
	}
	s_ob_unlock(&cur_th->sob);

	return 0;
}

int sys_thread_disable_exceptions()
{
	struct s_thread *cur_th = current_thread();

	cur_th->th_lflags &= ~FLUKE_THREAD_EXCEPTION_ENABLE;

	return 0;
}
#endif /* not CONFIG_OPT_THREADS */

#if FLASK
int sys_thread_get_client_secure(void)
{
	s_thread_t *cur_th = current_thread();
	unsigned flags;
	link_t link;
	int rc;

	if (EXC_SYSPARAM(cur_th, thread_get_client, cur_thread)
	    != cur_th->thread_va)
		EXC_RETURN_INSANITY(NOT_CURRENT_THREAD);

	/* Pickle any connection we are the server for */
	rc = ipc_server_pickle(cur_th, cur_th, &link, &flags);
	if (rc)
		return rc;

	/* Return the link and flags */
	rc = kentry_ref_associate(cur_th,
				EXC_SYSPARAM(cur_th, thread_get_client, client),
				&link);
	EXC_SYSPARAM(cur_th, thread_get_client, flags) = flags;

/* FLASK */
	EXC_SYSPARAM(cur_th, thread_get_client, essid) =
		cur_th->ipc_state.effective_server_sid;
	EXC_SYSPARAM(cur_th, thread_get_client, cecsid) =
		cur_th->ipc_state.client_effective_client_sid;
/* FLASK */

	return rc;
}
#endif 

int sys_thread_get_client()
{
	s_thread_t *cur_th = current_thread();
	unsigned flags;
	link_t link;
	int rc;

	if (EXC_SYSPARAM(cur_th, thread_get_client, cur_thread)
	    != cur_th->thread_va)
		EXC_RETURN_INSANITY(NOT_CURRENT_THREAD);

	/* Pickle any connection we are the server for */
	rc = ipc_server_pickle(cur_th, cur_th, &link, &flags);
	if (rc)
		return rc;

	/* Return the link and flags */
	rc = kentry_ref_associate(cur_th,
				EXC_SYSPARAM(cur_th, thread_get_client, client),
				&link);
	EXC_SYSPARAM(cur_th, thread_get_client, flags) = flags;
	return rc;
}

#if FLASK
int sys_thread_set_client_common(oskit_bool_t secure)
#else
int sys_thread_set_client()
#endif
{
        s_thread_t *cur_th = current_thread();
	link_t link;
#if FLASK
        oskit_bool_t reset = FALSE;
#endif
	int rc;

	if (EXC_SYSPARAM(cur_th, thread_set_client, cur_thread)
	    != cur_th->thread_va)
		EXC_RETURN_INSANITY(NOT_CURRENT_THREAD);

	/* Lookup the new link and ensure it is the correct type */
	rc = kentry_ref_extract_type(cur_th,
				EXC_SYSPARAM(cur_th, thread_set_client, client),
				FLUKE_THREAD, &link);
	if (rc)
		return rc;

	/* Ensure our client link is pickled */
	rc = ipc_server_pickle(cur_th, cur_th, 0, 0);
	if (rc)
		return rc;

	/* XXX layering */
	link_transfer(&link, &cur_th->ipc_state.client_link);
	link_destroy(&link);
	cur_th->ipc_state.flags =
		(cur_th->ipc_state.flags & ~FLUKE_THREAD_SERVER_SENDER)
			| (EXC_SYSPARAM(cur_th, thread_get_client, flags)
			   & FLUKE_THREAD_SERVER_SENDER);

#if FLASK
	if (secure)
	 {
	   cur_th->ipc_state.client_effective_client_sid =
	     EXC_SYSPARAM(cur_th, thread_get_client, cecsid); 
	   cur_th->ipc_state.effective_server_sid =
	     EXC_SYSPARAM(cur_th, thread_get_client, essid);
	 }
	else
	  {
	    cur_th->ipc_state.client_effective_client_sid = SECSID_NULL;
            cur_th->ipc_state.effective_server_sid = cur_th->sob.sid;
	  }


        /*
         * Validate the IPC state of this thread.
         */
        rc = thread_ipc_reset(cur_th, &reset);
        if (rc) return rc; 
        if (reset)
            EXC_RETURN_INSANITY(PERMISSION_DENIED);
#endif

#if 0	/* XXX do this on demand? */
	/* Attempt to connect with the new client.  May block. */
	rc = ipc_server_unpickle(cur_th, 0);
	if (rc)
		return rc;
#endif

	return 0;
}

#if FLASK
int sys_thread_set_client_secure()
{
  return sys_thread_set_client_common(TRUE);
}


int sys_thread_set_client()
{
  return sys_thread_set_client_common(FALSE);
}
#endif


#if FLASK
int sys_thread_get_server_secure(void)
{
	s_thread_t *cur_th = current_thread();
	unsigned flags;
	link_t link;
	int rc;

	if (EXC_SYSPARAM(cur_th, thread_get_server, cur_thread)
	    != cur_th->thread_va)
		EXC_RETURN_INSANITY(NOT_CURRENT_THREAD);

	/* Pickle any connection we are the client for */
	rc = ipc_client_pickle(cur_th, cur_th, &link, &flags);
	if (rc)
		return rc;

	/* Return the link and flags */
	rc = kentry_ref_associate(cur_th,
				EXC_SYSPARAM(cur_th, thread_get_server, server),
				&link);
	EXC_SYSPARAM(cur_th, thread_get_server, flags) = flags;

/* FLASK */
	EXC_SYSPARAM(cur_th, thread_get_server, ecsid) =
		cur_th->ipc_state.effective_client_sid;
	EXC_SYSPARAM(cur_th, thread_get_server, sessid) =
		cur_th->ipc_state.server_effective_server_sid;
/* FLASK */

	return rc;
}
#endif 

int sys_thread_get_server()
{
	s_thread_t *cur_th = current_thread();
	unsigned flags;
	link_t link;
	int rc;

	if (EXC_SYSPARAM(cur_th, thread_get_server, cur_thread)
	    != cur_th->thread_va)
		EXC_RETURN_INSANITY(NOT_CURRENT_THREAD);

	/* Pickle any connection we are the client for */
	rc = ipc_client_pickle(cur_th, cur_th, &link, &flags);
	if (rc)
		return rc;

	/* Return the link and flags */
	rc = kentry_ref_associate(cur_th,
				EXC_SYSPARAM(cur_th, thread_get_server, server),
				&link);
	EXC_SYSPARAM(cur_th, thread_get_server, flags) = flags;
	return rc;
}

#if FLASK
int sys_thread_set_server_common(oskit_bool_t secure)
#else
int sys_thread_set_server()
#endif
{
	s_thread_t *cur_th = current_thread();
	link_t link;
#if FLASK
	oskit_bool_t reset = FALSE;
#endif
	int rc;

	if (EXC_SYSPARAM(cur_th, thread_set_server, cur_thread)
	    != cur_th->thread_va)
		EXC_RETURN_INSANITY(NOT_CURRENT_THREAD);

	/* Lookup the new link and ensure it is the correct type */
	rc = kentry_ref_extract_type(cur_th,
				EXC_SYSPARAM(cur_th, thread_set_server, server),
				FLUKE_THREAD, &link);
	if (rc)
		return rc;

	/* Ensure our server link is pickled */
	rc = ipc_client_pickle(cur_th, cur_th, 0, 0);
	if (rc)
		return rc;

	/* XXX layering */
	link_transfer(&link, &cur_th->ipc_state.server_link);
	link_destroy(&link);
	cur_th->ipc_state.flags =
		(cur_th->ipc_state.flags & ~FLUKE_THREAD_CLIENT_SENDER)
			| (EXC_SYSPARAM(cur_th, thread_get_server, flags)
			   & FLUKE_THREAD_CLIENT_SENDER);

#if FLASK
	if (secure)
	  {
	    cur_th->ipc_state.server_effective_server_sid =
	      EXC_SYSPARAM(cur_th, thread_get_server, sessid); 
	    cur_th->ipc_state.effective_client_sid =
	      EXC_SYSPARAM(cur_th, thread_get_server, ecsid);
	  }
	else
	  {
	    cur_th->ipc_state.server_effective_server_sid = SECSID_NULL;
            cur_th->ipc_state.effective_client_sid = cur_th->sob.sid;
	  }


        /* 
         * Validate the IPC state of this thread.
         */
        rc = thread_ipc_reset(cur_th, &reset);
        if (rc) return rc; 
        if (reset)
            EXC_RETURN_INSANITY(PERMISSION_DENIED);
#endif

#if 0	/* XXX do this on demand? */
	/* Attempt to connect with the new server.  May block. */
	rc = ipc_client_unpickle(cur_th, 0);
	if (rc)
		return rc;
#endif

	return 0;
}

#if FLASK
int sys_thread_set_server_secure()
{
  return sys_thread_set_server_common(TRUE);
}

int sys_thread_set_server()
{
  return sys_thread_set_server_common(FALSE);
}
#endif


int sys_thread_interrupt()
{
	struct s_thread *cur_th = current_thread();
	struct s_thread *dest_th;
	int rc;

	rc = task_ob_lookup_type(cur_th,
				 EXC_SYSPARAM(cur_th, thread_interrupt, thread),
				 FLUKE_PROT_WRITE, FLUKE_THREAD,
				 (s_ob_t**)&dest_th);
	if (rc) return rc;

	if (dest_th == cur_th)
	{
		s_thread_unlock(dest_th);
		EXC_RETURN_INSANITY(CURRENT_THREAD);
	}

	if (!(dest_th->gflags & FLUKE_THREAD_INTERRUPT_PENDING))
	{
		/*
		 * Flag the pending interrupt and wake up the thread.
		 */
		dest_th->gflags |= FLUKE_THREAD_INTERRUPT_PENDING;
		thread_cancel(dest_th);
	}

	s_ob_unlock(&dest_th->sob);

	return 0;
}

int sys_thread_move()
{
	struct s_thread *cur_th = current_thread();
	s_thread_t *thread;
	int rc;

	rc = task_ob_move(cur_th, EXC_SYSPARAM(cur_th, thread_move, oaddr),
			  EXC_SYSPARAM(cur_th, thread_move, naddr),
			  FLUKE_THREAD, FLUKE_THREAD_SIZE,
#ifdef CONFIG_OPT_THREADS
			  sizeof(u_thread_t),
#else
			  0,
#endif
			  (s_ob_t **)&thread);

	/* XXX what about thread_va? */

	if (rc == 0) {
#ifdef CONFIG_OPT_THREADS
		/*
		 * Update the pointer to the user-level flob data.
		 */
		thread->md.u_thread = (u_thread_t *)
			phystokv(thread->sob.u_pa);
#endif

		s_thread_unlock(thread);
	}
	return (rc);
}

int sys_thread_reference()
{
	struct s_thread *cur_th = current_thread();

	return kentry_ob_reference(cur_th,
				   EXC_SYSPARAM(cur_th, thread_reference, thread),
				   EXC_SYSPARAM(cur_th, thread_reference, ref),
				   FLUKE_THREAD);
}

int sys_thread_yield()
{
	struct s_thread *cur_th = current_thread();

	if (EXC_SYSPARAM(cur_th, thread_yield, cur_thread)
	    != cur_th->thread_va)
		EXC_RETURN_INSANITY(NOT_CURRENT_THREAD);

	EXC_SET_IP(cur_th, fluke_nop__entry);
	return thread_yield(cur_th);
}

