/*
 * Copyright (c) 1996-1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include "s_thread.h"
#include "preempt.h"
#include "wait.h"
#include "debug.h"
#include "config_instrument.h"
#include "config_smp.h"
#include "s_instrument.h"
#include "per_cpu.h"
#include "smp.h"

void pd_thread_cancel_locked(
	struct s_thread		*thread);


/*
 * Put thread into the given wait state.
 * If state requires wakeup on IPC peer death, set that up.
 */
static inline void
thread_enter_wait_state(
	struct s_thread *thread,
	wait_val_t state,
	struct s_thread *other_th)
{
	thread->wait_state.val = state;
	if ((state & WAIT_IPC_PEER_DEATH) == WAIT_IPC_PEER_DEATH) {
		/*
		 * We are entering a wait state from which we should
		 * wake when other_th dies, so put us on his death_cond queue.
		 */
		thread->wait_state.waiting_on = &other_th->death_cond;
		wait_queue_add(&other_th->death_cond.queue, thread);
	}
	else {
		assert(state == WAIT_STOPPED); /* only non-IPC thread wait */
		thread->wait_state.waiting_on = 0;
	}
}

/*
 * Common second half shared by thread_wait and thread_handoff (see below).
 * Called with other_th->wait_state.lock locked.
 */
static int
thread_wait_finish(
	struct s_thread *cur_th,
	wait_val_t state,
	struct s_thread *other_th)
{
	spin_unlock(&other_th->wait_state.lock);

	/* By default, when we wake up, just restart from user space. */
	cur_th->wait_state.resume_rc = KR_RESTART;

	/*
	 * If a cancel was already pending,
	 * return an error immediately without switching threads.
	 */
	if (cur_th->wait_state.cancel_pending) {
		if ((state & WAIT_CANCELABLE) != WAIT_CANCELABLE) {
			/*
			 * This is an uncancelable wait,
			 * used when actually handling a cancellation.
			 * Go ahead and sleep, but when we wake up,
			 * restart the cancellation handling
			 * rather than restarting normal execution.
			 */
			cur_th->wait_state.resume_rc = KR_CANCEL;
		} else {
			spin_unlock(&cur_th->wait_state.lock);
			enable_preemption();
			INSTRUMENT_EVENT(cur_th, KR_CANCEL, 0);
			return KR_CANCEL;
		}
	}

	/*
	 * We don't need to add the current thread to any wait queue,
	 * because the wait state itself defines the "ownership" of the thread.
	 */
	thread_enter_wait_state(cur_th, state, other_th);

	cur_th = pd_thread_dispatch(cur_th, &cur_th->wait_state.lock,
				    other_th, 0);
	enable_preemption();

	INSTRUMENT_EVENT(cur_th, cur_th->wait_state.resume_rc, 0);
	return cur_th->wait_state.resume_rc;
}

int thread_wait(
	struct s_thread		*cur_th,
	wait_val_t		state,
	struct s_thread		*other_th,
	wait_val_t		other_state)
{
	assert(cur_th == current_thread());
	assert(cur_th != other_th);
	assert(state != WAIT_NONE);
	assert(state != WAIT_ON_COND);

	disable_preemption();

	/* Lock both threads. */
	if (cur_th < other_th) {
		spin_lock(&cur_th->wait_state.lock);
		spin_lock(&other_th->wait_state.lock);
	} else {
		spin_lock(&other_th->wait_state.lock);
		spin_lock(&cur_th->wait_state.lock);
	}

	assert(cur_th->wait_state.val == WAIT_NONE);

	/* Perform the requested atomic action on the other thread... */
	if (other_state == 0)
	{
		/* Just cancel the other thread atomically with the sleep.  */
		pd_thread_cancel_locked(other_th);
	}
	else if (other_state == -1) /* XXX horrible hack for ipc unpickle */
	{
		pd_mutex_unlock(cur_th, &other_th->sob.lock);
	}
	else if ((other_th->wait_state.val & other_state) == other_state)
	{
		/*
		 * The other thread has entered other_state,
		 * so return without waiting.
		 */
		spin_unlock(&other_th->wait_state.lock);
		spin_unlock(&cur_th->wait_state.lock);
		enable_preemption();
		INSTRUMENT_EVENT(cur_th, KR_RESTART, 0);
		return KR_RESTART;
	}

	return thread_wait_finish(cur_th, state, other_th);
}

wait_val_t pd_thread_wakeup_locked(
	struct s_thread		*thread,
	wait_val_t		required_state)
{
	wait_val_t rc = 0;

	assert(required_state != WAIT_NONE);
	assert(required_state != WAIT_ON_COND);

	if ((thread->wait_state.val & required_state) == required_state)
	{
		if ((thread->wait_state.val & WAIT_QUEUED_ON_COND)
		    == WAIT_QUEUED_ON_COND)
		{
			struct cond *c = thread->wait_state.waiting_on;
			assert(c != 0);
			if (wait_queue_remove_specific(&c->queue, thread))
			{
				do_debug(thread->wait_state.waiting_on = 0);
				rc = thread->wait_state.val;
				thread->wait_state.val = WAIT_NONE;
			}
			else
			{
				/*
				 * Somebody's already removed the thread,
				 * but hasn't yet cleared its wait state.
				 * We don't need to clear the wait,
				 * since we know they'll do it for us
				 * (see pd_cond_wakeup).
				 */
			}
		}
		else if (thread->wait_state.val != WAIT_NONE)
		{
			assert(thread->wait_state.waiting_on == 0);
			rc = thread->wait_state.val;
			thread->wait_state.val = WAIT_NONE;
		}
	}

	return rc;
}

wait_val_t thread_wakeup(
	struct s_thread		*thread,
	wait_val_t		required_state)
{
	wait_val_t rc;

	disable_preemption();
	spin_lock(&thread->wait_state.lock);

	rc = pd_thread_wakeup_locked(thread, required_state);

	spin_unlock(&thread->wait_state.lock);
	enable_preemption();

	return rc;
}

void thread_wait_other(
	struct s_thread		*cur_th,
	wait_val_t		state,
	struct s_thread		*thread)
{
	assert(thread != current_thread());


	disable_preemption();
	spin_lock(&thread->wait_state.lock);

	assert(thread->wait_state.val == WAIT_NONE);
	if (thread->wait_state.cancel_pending)
	{
		thread->wait_state.resume_rc = KR_CANCEL;
		if ((state & WAIT_CANCELABLE) != WAIT_CANCELABLE) {
			assert((state & WAIT_IPC_PEER_DEATH)
			       != WAIT_IPC_PEER_DEATH);
			thread->wait_state.val = state;
			do_debug(thread->wait_state.waiting_on = 0);
		}
		else {
			thread_ready(thread);
		}
	}
	else
		thread_enter_wait_state(thread, state, cur_th);

	spin_unlock(&thread->wait_state.lock);
	enable_preemption();
}

int thread_handoff(
	struct s_thread		*cur_th,
	wait_val_t		state,
	struct s_thread		*other_th)
{
	assert(cur_th == current_thread());
	assert(cur_th != other_th);
	assert(state != WAIT_NONE);
	assert(state != WAIT_ON_COND);

	disable_preemption();

	/* Lock both threads. */
	if (cur_th < other_th) {
		spin_lock(&cur_th->wait_state.lock);
		spin_lock(&other_th->wait_state.lock);
	} else {
		spin_lock(&other_th->wait_state.lock);
		spin_lock(&cur_th->wait_state.lock);
	}

	assert(cur_th->wait_state.val == WAIT_NONE);

	/* We can ready the destination thread now that it's locked. */
	/* XXX should be optimized. */
	thread_ready(other_th);

	return thread_wait_finish(cur_th, state, other_th);
}

/*
 * XXX This was supposed to wait rather than block,
 * but it turns out its "wait" isn't a real wait
 * because the wait_val and queue aren't set...
 */
int thread_yield(
	struct s_thread		*cur_thread)
{
	/* Just restart normally when we get back */
	cur_thread->wait_state.resume_rc = KR_RESTART;

	disable_preemption();
	cur_thread = pd_disp_yield(cur_thread);
	enable_preemption();

	/* We might be in the same or a different thread now */
	return cur_thread->wait_state.resume_rc;
}

int thread_run_preempters(void)
{
	struct s_thread *cur_thread = current_thread();

	/* Just restart normally when we get back */
	cur_thread->wait_state.resume_rc = KR_RESTART;

#ifdef PREEMPTSTATS
	{
		extern void u_run_preempters(void);
		u_run_preempters();
	}
#else
	run_preempters();
#endif

	/* We might be in the same or a different thread now */
	return cur_thread->wait_state.resume_rc;
}

void pd_thread_cancel_locked(
	struct s_thread		*thread)
{
	assert(!thread->wait_state.cancel_pending ||
	       (thread->wait_state.val & WAIT_CANCELABLE) != WAIT_CANCELABLE);

	thread->wait_state.cancel_pending = 1;
	if (pd_thread_wakeup_locked(thread, WAIT_CANCELABLE)) {
		thread->wait_state.resume_rc = KR_CANCEL;
		thread_ready(thread);
	}
#ifdef CONFIG_SMP
	else {
		/*
		 * This lock protects us from other processors
		 * switching into or out of this thread.
		 */
		spin_lock(&thread->disp_state.lock);
		if (thread->disp_state.cpu)
			/*
			 * This thread is running right now on another CPU.
			 * (A CPU holds the disp_state.lock while setting
			 * the cpu pointer, until it is done switching.)
			 */
			smp_thread_cancel(thread->disp_state.cpu);
		spin_unlock(&thread->disp_state.lock);
	}
#endif
}

void thread_cancel(
	struct s_thread		*thread)
{
	disable_preemption();
	spin_lock(&thread->wait_state.lock);

	pd_thread_cancel_locked(thread);

	spin_unlock(&thread->wait_state.lock);
	enable_preemption();
}

oskit_bool_t thread_cancel_pending(
	  struct s_thread	*thread)
{
	oskit_bool_t rc;

	disable_preemption();
	spin_lock(&thread->wait_state.lock);
	rc = thread->wait_state.cancel_pending;
	spin_unlock(&thread->wait_state.lock);
	enable_preemption();

	return rc;
}

void thread_wait_state_init(
	struct s_thread		*th)
{
	spin_lock_init(&th->wait_state.lock);
	th->wait_state.val = WAIT_NONE;
	th->wait_state.cancel_pending = 0;
	th->wait_state.resume_rc = KR_RESTART;
	do_debug(th->wait_state.waiting_on = 0);
}

void thread_wait_state_destroy(
	struct s_thread		*th)
{
	/*
	 * The target thread should be captured by the caller (destroyer)
	 * and thus not waiting for anything.
	 */
	assert(th->wait_state.val == WAIT_NONE);
	assert(th->wait_state.waiting_on == 0);
}
