/*
 * Copyright (c) 1996-1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Support for interruptible, "stateless" thread waits.
 * This is the only kind of thread sleep done in nonblocking kernels;
 * in blocking kernels, threads can block as well (duhh) with kernel state.
 *
 * Two kinds of waits are supported by the wait layer:
 * thread waits and condition waits.
 * Thread waits are done when a thread wants to go to sleep
 * and simply wait for another thread to wake it up "by name";
 * condition waits are done when a thread wants to wait for a condition
 * which other threads might be waiting for as well;
 * in this case, the wakeup is directed at the condition,
 * and causes one thread or all threads waiting on that condition to wake up.
 * Thread waits are implemented in this module;
 * condition waits are implemented by cond.c/h, also in this layer.
 */
#ifndef _FLUKE_KERNEL_WAIT_H_
#define _FLUKE_KERNEL_WAIT_H_

#include "spin_lock.h"
#include "dispatch.h"

/*
 * The wait states defined below define a "tree" of states.  Leaves are
 * more specific version of branch states.  Here is the tree:
 *
 * WAIT_NONE
 * WAIT_ANYTHING ->
 *	WAIT_STOPPED
 *	WAIT_QUEUED_ON_COND ->
 *		WAIT_IPC_PICKLE
 *		WAIT_CANCELABLE ->
 *			WAIT_ON_COND
 *			WAIT_IPC_SRV_DSCN
 *			WAIT_IPC_SRV_RECEIVER ->
 *				WAIT_IPC_SRV_RECV
 *				WAIT_IPC_SRV_ASEND
 *			WAIT_IPC_SRV_SENDER ->
 *				WAIT_IPC_SRV_ORECV
 *				WAIT_IPC_SRV_SEND
 *			WAIT_IPC_CLI_DSCN
 *			WAIT_IPC_CLI_RECEIVER ->
 *				WAIT_IPC_CLI_RECV
 *				WAIT_IPC_CLI_ASEND
 *			WAIT_IPC_CLI_SENDER ->
 *				WAIT_IPC_CLI_ORECV
 *				WAIT_IPC_CLI_SEND
 *			WAIT_IPC_UNPICKLE
 *			WAIT_IPC_IDEMPOTENT ->
 *				WAIT_IPC_EXCEPTION
 *				WAIT_IPC_CALL
 *				WAIT_IPC_SECURITY
 *
 * NOTE: Leaves are not the only valid wait states, threads can be
 * blocked in several of the internal states.
 *
 * NOTE: All of the values defined satisfy the property that
 * (NODE & <NODE or any child>) == NODE.  Ie, a bitwise AND of
 * the wait state and itself or any of its "child" states will
 * return itself.  Thus the hierarchy above.  So, if you see
 * thread_wakeup(thread, WAIT_IPC_SRV_SENDER), then it will
 * return non-zero if the thread is in WAIT_IPC_SRV_SENDER,
 * WAIT_IPC_SRV_ORECV, or WAIT_IPC_SRV_SEND.
 *
 * NOTE: The tree above is a lie, but it's pretty.  Really it is a DAG,
 * almost a tree but not quite: the WAIT_IPC_PEER_DEATH state is included
 * both in WAIT_IPC_PICKLE (which is not WAIT_CANCELABLE) and in all the
 * other * WAIT_IPC_* states (which are WAIT_CANCELABLE).  All states
 * except WAIT_ON_COND, WAIT_SCHEDULER and WAIT_STOPPED are children of the
 * WAIT_IPC_PEER_DEATH state branch(sic). Noone ever waits specifically for
 * this branch(sic) of states; the thread_wait functions test for entering
 * a state on that branch, and set the waiting_on cond when doing so.
 * Since WAIT_IPC_PEER_DEATH is a child of the WAIT_QUEUED_ON_COND state
 * branch, all states on that branch(sic) indicate dequeuing from the cond
 * on wakeup.
 */
#define WAIT_NONE		0x00000000 /* Ready, running, or blocked */
#define WAIT_ANYTHING		0x00000010 /* Waiting at any clean point */
#define WAIT_CANCELABLE		0x00000018 /* Waiting at any clean point */
#define WAIT_QUEUED_ON_COND	0x00000014 /* on waiting_on cond's queue */
#define WAIT_ON_COND		0x0000003c /* Waiting on waiting_on cond */
#define WAIT_SCHEDULER		0x0000001c /* waiting_on is scheduler's pset */
#define WAIT_IPC_PEER_DEATH	0x00000016 /* waiting_on is peer death_cond */
#define WAIT_IPC_SRV_DSCN	0x0000005e /* in ipc_server_disconnect */
#define WAIT_IPC_SRV_RECEIVER	0x0000009e /* as the receiver */
#define WAIT_IPC_SRV_SENDER	0x0000011e /* as the sender */
#define WAIT_IPC_SRV_ORECV	0x0000031e /* in ipc_server_over_receive */
#define WAIT_IPC_SRV_RECV	0x0000049e /* in ipc_server_receive */
#define WAIT_IPC_SRV_ASEND	0x0000089e /* in ipc_server_ack_send */
#define WAIT_IPC_SRV_SEND	0x0000111e /* in ipc_server_send */
#define WAIT_IPC_CLI_DSCN	0x0000201e /* in ipc_client_disconnect */
#define WAIT_IPC_CLI_RECEIVER	0x0000401e /* in ipc_client_over_receive */
#define WAIT_IPC_CLI_SENDER	0x0000801e /* in ipc_client_over_receive */
#define WAIT_IPC_CLI_ORECV	0x0001801e /* in ipc_client_over_receive */
#define WAIT_IPC_CLI_RECV	0x0002401e /* in ipc_client_receive */
#define WAIT_IPC_CLI_ASEND	0x0004401e /* in ipc_client_ack_send */
#define WAIT_IPC_CLI_SEND	0x0008801e /* in ipc_client_send */
#define WAIT_IPC_PICKLE		0x00100016 /* for my conns to be pickled */
#define WAIT_IPC_UNPICKLE	0x0020001e /* for conns to be activated */
#define WAIT_IPC_IDEMPOTENT	0x0040001e /* making any idempotent request */
#define WAIT_IPC_CALL		0x00c0001e /* explicit idempotent ipc_call */
#define WAIT_IPC_EXCEPTION	0x0140001e /* idempotent exception IPC */
#if FLASK
#define WAIT_IPC_SECURITY	0x0240001e /* security exception IPC */
#endif
#define WAIT_STOPPED		0x04000010 /* Thread is stopped */


typedef unsigned wait_val_t;


/* Per-thread state maintained by the cancellation support code.  */
struct wait_state
{
	/* Lock protecting the following fields.  */
	spin_lock_t		lock;

	/* Current thread waiting state value, defined above. */
	wait_val_t		val;

	/*
	 * Cancel pending flag.
	 * Set by thread_cancel(); cleared by thread_clear_cancel().
	 */
	oskit_bool_t		cancel_pending;

	/*
	 * KR_* code indicating what to do on wakeup.
	 * Set to KR_RESTART when a thread is woken normally.
	 * Set to KR_CANCEL when a thread is woken by cancellation,
	 * and can be set to KR_USER_EXCEPTION by the IPC code
	 * to cause a receiver thread to take an exception.
	 */
	int			resume_rc;

	/*
	 * Whenever this thread is waiting on a condition variable,
	 * this variable points to the condition the thread is waiting on.
	 * At all other times, this pointer is null.
	 */
	struct cond		*waiting_on;
};
typedef struct wait_state wait_state_t;


struct s_thread;

typedef int unblock_func_t(struct s_thread *cur_thread, int rc);


/*
 * Put the current thread to sleep in some wait state,
 * not waiting on any condition variable,
 * but instead just waiting for some other thread
 * to come along, notice our state, and wake us up.
 * An action can be performed on another thread atomically with the wait:
 * if other_state == 0, the other_thread is canceled;
 * if other_state != 0, the other_thread's current state is checked,
 * and if it is waiting in that state, the wait returns immediately.
 * This atomic action is analogous to the mutex unlock on cond_wait,
 * and is required for exactly the same reason:
 * to avoid race conditions causing missed events.
 */
int thread_wait(
	struct s_thread		*cur_thread,
	wait_val_t		state,
	struct s_thread		*other_thread,
	wait_val_t		other_state);

/*
 * Handoff directly from one thread to another (captured) thread.
 * Atomically puts us to sleep and starts the other thread.
 */
int thread_handoff(
	struct s_thread		*cur_th,
	wait_val_t		state,
	struct s_thread		*other_th);

/*
 * Give up the processor to another thread if there is one.
 * The current thread will be waiting, rather than blocked,
 * so it can be interrupted by other user-mode threads and such.
 */
int thread_yield(
	struct s_thread		*cur_thread);

/*
 * Wake up and capture the specified other thread
 * if its current state is at least 'required_state'.
 */
wait_val_t thread_wakeup(
	struct s_thread		*thread,
	wait_val_t		required_state);

/*
 * Same as thread_wakeup, but preemption is disabled
 * and thread->wait_state.lock is already held (and stays held on return).
 */
wait_val_t pd_thread_wakeup_locked(
	struct s_thread		*thread,
	wait_val_t		required_state);


/*
 * Put a captured thread "back" to sleep waiting in the specified state.
 * It goes back on waits_on_th's death_cond queue; waits_on_th is null
 * only for state == WAIT_STOPPED.
 */
void thread_wait_other(
	struct s_thread		*waits_on_th, /* thread other_th waits on */
	wait_val_t		state,
	struct s_thread		*other_th);

/* Returns true if a cancel is pending on the indicated thread.  */
oskit_bool_t thread_cancel_pending(
	struct s_thread		*thread);		    

/* Clear the cancel-pending flag in the indicated thread.  */
#define thread_cancel_clear(th)	((th)->wait_state.cancel_pending = 0)

/* Cancel another thread.  */
void thread_cancel(
	struct s_thread		*thread);

/* Initialize the thread-specific cancel state in a new thread.  */
void thread_wait_state_init(
	struct s_thread		*thread);

/* Destroy the cancel state in a thread.  */
void thread_wait_state_destroy(
	struct s_thread		*thread);


#endif /* _FLUKE_KERNEL_WAIT_H_ */
