/*
 * Copyright (c) 1996, 1997 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include "wait_queue.h"
#include "dispatch.h" /* XXX disp_state.next */
#include "s_thread.h"
#include "debug.h"

void wait_queue_init(
	struct wait_queue	*queue)
{
	spin_lock_init(&queue->lock);
	queue->head = 0;
	queue->tail = &queue->head;
}

void wait_queue_destroy(
	struct wait_queue	*queue)
{
	assert(queue->head == 0);
	assert(queue->tail == &queue->head);
}

void wait_queue_add_head_locked(
	struct wait_queue	*queue,
	struct s_thread		*thread)
{
	thread->disp_state.next = queue->head;
	queue->head = thread;
#ifndef LIFO_QUEUEING
	if (queue->tail == &queue->head)
		queue->tail = &thread->disp_state.next;
#endif
}

void wait_queue_add_head(
	struct wait_queue	*queue,
	struct s_thread		*thread)
{
	wait_queue_lock(queue);
	wait_queue_add_locked(queue, thread);
	wait_queue_unlock(queue);
}

void wait_queue_add_tail_locked(
	struct wait_queue	*queue,
	struct s_thread		*thread)
{
#ifdef LIFO_QUEUEING
	struct s_thread *nth;

	if (queue->head) {
		nth = queue->head;
		while (nth->disp_state.next)
			nth = nth->disp_state.next;
		nth->disp_state.next = thread;
	} else
		queue->head = thread;
	thread->disp_state.next = 0;
#else
	assert(queue->tail != 0);
	thread->disp_state.next = 0;
	*queue->tail = thread;
	queue->tail = &thread->disp_state.next;
	assert(queue->head != 0);
#endif
}

void wait_queue_add_tail(
	struct wait_queue	*queue,
	struct s_thread		*thread)
{
	wait_queue_lock(queue);
	wait_queue_add_tail_locked(queue, thread);
	wait_queue_unlock(queue);
}

s_thread_t *wait_queue_remove_locked(
	struct wait_queue	*queue)
{
	struct s_thread *thread;

	thread = queue->head;
	if (thread == 0)
		return 0;

	queue->head = thread->disp_state.next;
#ifndef LIFO_QUEUEING
	if (queue->head == 0)
		queue->tail = &queue->head;
#endif

	return thread;
}

s_thread_t *wait_queue_remove(
	struct wait_queue	*queue)
{
	struct s_thread *thread;

	wait_queue_lock(queue);
	thread = wait_queue_remove_locked(queue);
	wait_queue_unlock(queue);

	return thread;
}

s_thread_t *wait_queue_remove_specific(
	struct wait_queue	*queue,
	struct s_thread		*thread)
{
	struct s_thread **thp;

	wait_queue_lock(queue);

	/* Remove the specified thread from the wait queue.  */
	thp = &queue->head;
	while (*thp != thread)
	{
		if (*thp == 0)
		{
			wait_queue_unlock(queue);
			return 0;
		}
		thp = &(*thp)->disp_state.next;
	}
	*thp = thread->disp_state.next;
#ifndef LIFO_QUEUEING
	if (*thp == 0)
		queue->tail = thp;
#endif

	wait_queue_unlock(queue);

	return thread;
}

