/*
 * Copyright (c) 1997, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * This is inter-processor communication support code for SMP
 */

#include "config_smp.h"

#ifdef CONFIG_SMP

#include <oskit/x86/paging.h>
#include "debug.h"

#include <oskit/x86/eflags.h>
#include <oskit/x86/proc_reg.h>

#include "spin_lock.h"
#include "per_cpu.h"
#include "intr.h"
#include "dispatch.h"

#include <oskit/smp.h>
#include <oskit/x86/smp.h>

#define flush_tlb() set_pdbr(get_pdbr())


spin_lock_t ipi_lock = SPIN_LOCK_INITIALIZER;


/*
 * This is the C code called as a result of an IPI.
 */
void
IPI_handler(void)
{
#define test_clear(member) ({						      \
	int flag;							      \
	asm("xchgl %0,%1" /* atomic exchange */				      \
	    : "=r" (flag), "=m" (my_cpu.member)				      \
	    : "0" (0));							      \
	flag; })

	/*
	 * What this should do is de-multiplex the `implicit'
	 * message from the sender and call the OS-provided
	 * callback routine.
	 */
	assert_interrupts_disabled();

	/*
	 * If a thread-cancel was posted, it will be shown in the
	 * cancelled thread's wait_state and checked before we
	 * return to user mode.  Now that we've gotten the the IPI,
	 * we can clear the per-CPU flag immediately to let the
	 * other processor cancelling us continue on his way.
	 */
	my_cpu.cancel_pending = 0;

	if (test_clear(inval_tlb)) {
		/*
		 * There is a TLB invalidation to be done.
		 */
		flush_tlb();
	}

#if !defined(CONFIG_NOPREEMPT) && !defined(CONFIG_USER_SCHED)
	/*
	 * Check for a request to reschedule posted by
	 * the CPU receiving timer interrupts.
	 */
	if (test_clear(reschedule)) {
		intr_preempter(&my_cpu.disp_preempter, disp_yield);
	}
#endif

#ifdef CONFIG_USER_SCHED
	if (test_clear(yield_pending)) {
		intr_preempter(&current_thread()->disp_state.disp_preempter,
			       disp_yield);
	}
#endif

#undef test_clear
	/* we are done in the handler; ack the interrupt */
	smp_apic_ack();
}


void
apic_init()
{
	int rc;

	rc = irq_alloc_ku(SMP_IPI_VECTOR, IPI_handler, IPI_handler, 0);
	assert(!rc);	/* Must get IPI vector... */
}


/*
 * These are convenience functions.  They handle the higher-level
 * operations using the low-level smp_message_pass() function.
 */

#ifndef CONFIG_NOPREEMPT
/*
 * This shows just how easy it is: set flag telling cpu what to
 * do and send it an interrupt.  We aren't going to block here,
 * because the other CPU can reschedule `whenever'.
 */
void
smp_reschedule()
{
	fluke_cpu_info_t *cpu;

	for (cpu = my_cpu.next; cpu != &my_cpu; cpu = cpu->next) {
		cpu->reschedule = 1;	/* atomic store */
		smp_message_pass(cpu->number);
	}
}
#endif

/*
 * If a thread is running on processor B, this sends an IPI to
 * processor B.  When returning from the ipi handlers (which does
 * nothing), it will handle the cancel I have already posted to
 * the thread (before calling smp_thread_cancel).
 * The semantics of thread_cancel/thread_interrupt are that no
 * more user-mode code will be executed after the fluke_thread_interrupt
 * system call returns until after it has been handled.
 */
void
smp_thread_cancel(fluke_cpu_info_t *const cpu)
{
	/*
	 * wait_state.cancel_pending was set before this was called.
	 * All I have to do is make sure the thread gets into the
	 * kernel at some point, so it will handle the cancel on the
	 * way out.  Giving it an interrupt is good enough.
	 */

	cpu->cancel_pending = 1;
	smp_message_pass(cpu->number);

	while (cpu->cancel_pending); /* spin */
}

#endif CONFIG_SMP
