/*
 * Copyright (c) 1997-1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * Contributed by the Computer Security Research division,
 * INFOSEC Research and Technology Office, NSA.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#if FLASK

#include <fluke/ipc.h>
#include <fluke/machine/syscall.h>
#include <fluke/cp/rpc.h>

#include "avc.h"
#include "s_ref.h"
#include "s_port.h"
#include "ipc.h"
#include "debug.h"
#include "xfer.h"

#include "config_instrument.h"
#include "s_instrument.h"

extern link_t security_port_link;

int avc_security_fault(security_class_t sclass,
		       security_id_t ssid,
		       security_class_t tclass,
		       security_id_t tsid,
		       access_vector_t requested)
{
    struct s_thread *client = current_thread();
    struct s_thread *server;
    fluke_ipc_params_t sparams;
    fluke_ipc_word_t send_buf[5];
    s_port_t *security_port;    
    s_ob_t *sob;
    int rc, nwords; 
    

    /* Break any existing idempotent connection. */
    if (client->ipc_state.idemp_server) {
	rc = ipc_client_idemp_disconnect(client);
	if (rc)
		return rc;
	return KR_RESTART;
    }

    /* Look up the SID port. */  
    sob = (s_ob_t *) link_follow(&security_port_link);
    if (!sob)
    {
	/*
	 * The security port has not yet been set.
	 */	
#ifdef DEBUG
	avc_debug_handle_fault(sclass,ssid,tclass,tsid,requested);
	return KR_RESTART;
#else
	panic("avc:  no security policy server");
#endif DEBUG	
    }

    assert(sob->type == FLUKE_PORT);
    security_port = (s_port_t *) sob;

    /* Try to capture a server thread. */
    rc = s_port_capture_locked(security_port, &server);
    if (rc)
	    return rc;
    
    if (server == 0)
    {
	/*
	 * The security port is not connected to a port set;
	 * either the security policy server is not yet operational,
	 * in which case the access vector cache initial state needs
	 * to be extended, or the security policy server has died.
	 */
#ifdef DEBUG
	avc_debug_handle_fault(sclass,ssid,tclass,tsid,requested);
	return KR_RESTART;
#else
	panic("avc:  no security policy server");
#endif DEBUG	
    }


    /* Extract the receive parameters from the server thread state. */
    sparams.recv_buf.words = IPC_RECV_BUF_WORDS(server);
    sparams.recv_buf.buf = (fluke_ipc_word_t*)IPC_RECV_BUF_BUF(server);

    if (!server->ipc_state.server_wants_client_sid)
    {
	sparams.recv_buf.words -= FLUKE_MIN_MSG_WORDS;
	sparams.recv_buf.buf += FLUKE_MIN_MSG_SIZE;
    }

    if (server->ipc_state.server_wants_client_sid)
    {
	send_buf[0] = req_Security_1_0_compute_access_vector_idemp;
	send_buf[1] = sclass | (tclass << 16);
	send_buf[2] = ssid;
	send_buf[3] = tsid;
	send_buf[4] = requested;	    
	nwords = 5;
    }
    else
    {
	send_buf[0] = ssid;
	send_buf[1] = tsid;
	send_buf[2] = requested;	    
	nwords = 3;
    }

    assert(sparams.recv_buf.words >= nwords);
    nwords = ((nwords > sparams.recv_buf.words) ? 
	      sparams.recv_buf.words : nwords); 
	
    rc = copyout_remote(server,
			(void *)send_buf,
			(oskit_addr_t)sparams.recv_buf.buf,
			nwords);
    if (rc)
    {
	server->wait_state.resume_rc = rc;
	thread_ready(server);
	INSTRUMENT_EVENT(client, KR_RESTART, 0);
	return KR_RESTART;
    }

    sparams.recv_buf.words -= nwords;

    /* The transfer succeeded, so update the receiver's register state. */    
    client->ipc_state.idemp_server = server;
    server->ipc_state.idemp_client = client;
    IPC_RECV_BUF_WORDS(server) = sparams.recv_buf.words;
    if (server->ipc_state.server_wants_client_sid)
    {
	server->ipc_state.md.min_msg[0] = SECINITSID_KERNEL;
	server->ipc_state.md.min_msg[1] = 0xdeadbeef;
    }
    else
    {
	server->ipc_state.md.min_msg[0] = 
		req_Security_1_0_compute_access_vector_idemp;
	server->ipc_state.md.min_msg[1] = sclass | (tclass << 16);
    }
    assert(FLUKE_MIN_MSG_WORDS == 2);
    IPC_FINISH_RECEIVE(server, FLUKE_IPC_WAIT_IDEMPOTENT);
    return thread_handoff(client, WAIT_IPC_SECURITY, server);
}

#endif 
