/*
 * Copyright (c) 1996-1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Machine dependent clock related functions.
 * Loosely based on BSD clock code.
 */
#include <stdio.h>
#include <oskit/c/sys/time.h>

#include <oskit/x86/pio.h>
#include <oskit/x86/seg.h>
#include <oskit/x86/pc/pic.h>
#include <oskit/x86/pc/pit.h>
#include <oskit/x86/base_trap.h>
#include <oskit/x86/eflags.h>
#include <sys/types.h>

#include "config_profiling.h"
#include "config_fdev.h"
#include "config_nopreempt.h"
#include "config_kmodule.h"
#include "config_user_gprof.h"
#include "per_cpu.h"
#include "smp.h"

#include "spin_lock.h"
#include "exception.h"
#include "intr.h"
#include "preempt.h"
#include "debug.h"
#include "dispatch.h"


/*
 * Set the value non-zero to cause a panic to occur after that number
 * of seconds.  Useful when your kernel has wedged but the clock is still
 * ticking.
 */
#define SECONDS_TIL_PANIC	0


#ifdef CONFIG_FDEV
#include <oskit/dev/dev.h>
void fdev_timer_intr(void);
#endif

extern int stathz;
extern int profhz;
extern volatile int ticks;

extern volatile struct timeval time;

#ifdef PREEMPTSTATS
#include "s_thread.h"
#include "preempt.h"

int pdticks;

#define HZ		10000	/* ticks per second */
#define PREEMPT_TICKS	100	/* check for preemption every N ticks */
#else
#define HZ		100	/* ticks per second */
#define PREEMPT_TICKS	1	/* check for preemption every N ticks */
#endif

volatile int     tick = 1000000 / HZ;

/*
 * Bump a timeval by a small number of usec's.
 */
#define BUMPTIME(t, usec) { \
	register volatile struct timeval *tp = (t); \
	register long us; \
 \
	tp->tv_usec = us = tp->tv_usec + (usec); \
	if (us >= 1000000) { \
	tp->tv_usec = us - 1000000; \
	tp->tv_sec++; \
	} \
}


#ifdef CONFIG_FDEV
struct preempter fdev_timer_preempter;

static void softclock(struct preempter *p)
{
	assert_interrupts_enabled();
	assert_preemption_disabled();

	fdev_timer_intr();
}
#endif

static void hardclock(void)
{
	/*
	 * The interrupt has been disabled & acknowledged at the PIC.
	 */
	pic_enable_irq(0);

	/* Update the system time */
	ticks++;
	BUMPTIME(&time, tick);
#if SECONDS_TIL_PANIC > 0
	if (ticks > SECONDS_TIL_PANIC * HZ) {
#ifdef DEBUG
		s_ob_dump_flobs();
#endif
		panic("\n!!! STUCK !!!\n");
	}
#endif

#ifdef CONFIG_FDEV
#ifdef CONFIG_KMODULE
	/*
	 * Fdev code runs in supervisor mode so we cannot delay softclock
	 * processing til we return to user mode!
	 *
	 * XXX gak, this still isn't enough.  Using this, the linux
	 * driver set autoconfigs some things, but hangs when probing
	 * SCSI (or IDE?) devices.  See hack-II in fdev_intr.c.
	 */
	if (preempt_enable_flag) {
		disable_preemption();
		enable_interrupts();
		fdev_timer_intr();
		disable_interrupts();
		enable_preemption();
	} else {
		enable_interrupts();
		fdev_timer_intr();
		disable_interrupts();
	}
#else
	/*
	 * Schedule any registered fdev timer handlers to be called;
	 * call it immediately if we're preemptible now.
	 */
	intr_preempter(&fdev_timer_preempter, softclock);
#endif
#endif

#ifdef HEARTBEAT_THREAD
	{
		extern struct preempter heartbeat_preempter;
		extern void heartbeat_func();
#ifdef PREEMPTSTATS
		extern unsigned int hbqueued, hbqtime;

		if (preempt_enable_flag == 0)
			pdticks++;
		if (heartbeat_preempter.func)
			hbqueued++;
		else
			asm volatile("rdtsc" : "=a" (hbqtime) :: "eax", "edx");
#endif
		intr_preempter(&heartbeat_preempter, heartbeat_func);
	}
#endif

#ifndef CONFIG_NOPREEMPT
	if ((ticks % PREEMPT_TICKS) == 0) {
		extern struct preempter disp_preempter;

		intr_preempter(&disp_preempter, disp_yield);
		/* have other processors reschedule on this tick */
		smp_reschedule();
	}
#endif
}


static int
u_clock(void *data)
{
	/* This came from user mode, so we can do any kernel calls */
	assert_preemption_enabled();
	assert_interrupts_disabled();

	hardclock();
	return 0;
}


/*
 * XXX This really should be in the machine independent code.
 * However, for that to happen we can't just pass in a trap_state
 * struct.
 */
static void
k_clock(void *data, struct trap_state *ts)
{
	assert_interrupts_disabled();

	hardclock();
}

void
clock_init(void)
{
#if 1
#ifdef CONFIG_FDEV
	osenv_timer_init();
#endif

	disable_interrupts();

	pit_set(0, PIT_RATEGEN, PIT_HZ / (HZ));
	irq_alloc_ku(0, k_clock, u_clock, NULL);
	pic_enable_irq(0);

	/* irq_alloc_ku already did this ... */
	enable_interrupts();
#endif
}

#include "config_stride_sched.h"
#ifdef CONFIG_STRIDE_SCHED
/*
 * Millisecond timer for stride scheduler.
 *
 * XXX a more general microsecond timer would be better but I don't want
 * to deal with overflow issues (or long long arithmetic) right now.
 */
unsigned
millitime(void)
{
#if HZ >= 1000
	return ticks / (HZ / 1000);
#else
	static unsigned lastms, lasttickval, lastpitval;
	unsigned ms, tickval, pitval;

	tickval = ticks;
	pitval = (PIT_HZ / HZ) - pit_read(0);
	if (tickval != ticks) {
		tickval = ticks;
		pitval = (PIT_HZ / HZ) - pit_read(0);
	}

	/*
	 * XXX this routine is typically (always?) called with interrupts
	 * disabled so we need to check if the PIT has wrapped but ticks
	 * has not yet been incremented.
	 */
	if (lasttickval > tickval ||
	    (lasttickval == tickval && lastpitval > pitval))
		tickval++;

	ms = (tickval * (1000 / HZ)) + (pitval / (PIT_HZ / 1000));
#if 1
	if (lastms > ms) {
		printf("lastms=%u, (ticks=%u, pitval=%u), ms=%u (ticks=%u, pitval=%u)\n",
		       lastms, lasttickval, lastpitval, ms, ticks, pitval);
	}
#endif
	assert(lastms <= ms);

	lasttickval = tickval;
	lastpitval = pitval;
	lastms = ms;

	return ms;
#endif
}
#endif
