/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * This file implements simple console I/O routines
 * for "hardwired" display, serial, or GDB-serial consoles.
 * XXX currently not thread- or MP-safe.
 */

#include <stdio.h>

#include <oskit/c/termios.h>
#include <oskit/tty.h>
#include <oskit/gdb.h>
#include <oskit/gdb_serial.h>
#include <oskit/x86/pio.h>
#include <oskit/x86/seg.h>
#include <oskit/x86/eflags.h>
#include <oskit/x86/base_idt.h>
#include <oskit/x86/base_trap.h>
#include <oskit/x86/base_vm.h>
#include <oskit/x86/base_cpu.h>
#include <oskit/x86/base_paging.h>
#include <oskit/x86/pc/pic.h>
#include <oskit/x86/pc/base_console.h>
#include <oskit/x86/pc/direct_cons.h>
#include <oskit/x86/pc/com_cons.h>
#include <oskit/x86/pc/base_irq.h>
#include <oskit/x86/pc/reset.h>
#include <oskit/debug.h>
#include <oskit/base_critical.h>

#include "config_smp.h"
#include "config_fdev.h"
#include "config_gdb_serial.h"
#include "config_instrument.h"
#include "per_cpu.h"

#ifdef CONFIG_FDEV
#include <oskit/dev/dev.h>
#endif

#include "gdt.h"
#include "exception.h"
#include "prog.h"
#include "intr.h"
#include "avc.h"	/* avd_dump() prototype */

int enable_kdb;
int enable_killswitch;

static int cons_com_port = 1;  /*XXX*/
static int gdb_com_port = 2;  /*XXX*/
static int gdb_com_irq;
static int fast_com_ports = 0;  /*XXX*/

static int
u_gdb_interrupt(void *data)
{
	pic_enable_irq(gdb_com_irq);
	com_cons_enable_receive_interrupt(gdb_com_port);

	/*
	 * We interrupted user mode, but we're debugging the kernel!
	 * Cause a (recursive) kernel trap
	 * so the debugger will get a sensible state frame.
	 */
	gdb_breakpoint();
	return 0;
}

static void
k_gdb_interrupt(void *data, struct trap_state *ts)
{
	pic_enable_irq(gdb_com_irq);
	com_cons_enable_receive_interrupt(gdb_com_port);

	gdb_trap(ts);
}

static int
u_comkill_interrupt(void *data)
{
	panic("Killed by the kill switch!");
}

static void
k_comkill_interrupt(void *data, struct trap_state *ts)
{
	printf("Killed by the kill switch!\n");
	trap_dump_panic(ts);
}

void
gdb_intr_boot()
{
	if (enable_kdb) {
		extern void gdb_intr();
		gdb_com_irq = gdb_com_port & 1 ? 4 : 3;

		/*
		 * Set up the appropriate COM port interrupt
		 * so the remote debugger can interrupt us.
		 */
		irq_alloc_ku(gdb_com_irq,
			     k_gdb_interrupt, u_gdb_interrupt, NULL);
		com_cons_enable_receive_interrupt(gdb_com_port);
	}
	else if (enable_killswitch && gdb_com_port != cons_com_port) {
		int com_irq = gdb_com_port & 1 ? 4 : 3;
	
		/* Set up the appropriate COM port interrupt */
		irq_alloc_ku(com_irq,
			     k_comkill_interrupt, u_comkill_interrupt, NULL);

		com_cons_enable_receive_interrupt(gdb_com_port);
	}
}


/*
 * These take the gdb-style function calls, and add the port number
 * for the `new' OS Kit interface that supports more than one COM port
 */
int
gdb_cons_getchar()
{
	return com_cons_getchar(gdb_com_port);
}

void 
gdb_cons_putchar(int ch)
{
	com_cons_putchar(gdb_com_port, ch);
}




/*
 * This function overrides the default base_console_init() in libkern;
 * it is called during bootup just before main() is called.
 */
void
base_console_init(int argc, char **argv)
{
	unsigned i;
	extern int askname;

	for (i = 1; i < argc; i++) {
		if (strcmp(argv[i], "-a") == 0)
			askname = 1;
		else if (strcmp(argv[i], "-d") == 0)
			enable_kdb = 1;
		else if (strcmp(argv[i], "-f") == 0)
			fast_com_ports = 1;
		else if (strcmp(argv[i], "-k") == 0)
			enable_killswitch = 1;
	}

	/* can have both now... */
	if (fast_com_ports) {
		base_cooked_termios.c_ispeed = B115200;
		base_cooked_termios.c_ospeed = B115200;
	}
	com_cons_init(cons_com_port, &base_cooked_termios);

#ifdef CONFIG_GDB_SERIAL
	if (enable_kdb)
	{
		if (gdb_com_port == cons_com_port)
			askname = 0;	/* -a and -d mutually exclusive */

		/* Set up the appropriate function pointers. */
		gdb_serial_recv = gdb_cons_getchar;
		gdb_serial_send = gdb_cons_putchar;
		gdb_signal = gdb_serial_signal;

		/* Initialize the serial port. */
		if (fast_com_ports) {
			base_raw_termios.c_ispeed = B115200;
			base_raw_termios.c_ospeed = B115200;
		}
		com_cons_init(gdb_com_port, &base_raw_termios);

		/*
		 * Take an immediate breakpoint trap here
		 * so that the debugger can take control from the start.
		 */
		printf ("base_raw_termios.c_ispeed=%d,base_raw_termios.c_ospeed=%d\n",base_raw_termios.c_ispeed,base_raw_termios.c_ospeed);
		asm("int $3");	/* XXX */
	}
	
	/*
	 * This gross hack, allows the second tipline (gdb line) to be
	 * used as a kill switch to force the kernel into a panic. Very
	 * useful when the kernel is looping someplace and interrupts are
	 * still enabled. Simply cat some characters into the special
	 * device file (/dev/tip/foo-gdb).
	 */
	if (enable_killswitch && !enable_kdb &&
	    gdb_com_port != cons_com_port) {

		/*
		 * Initialize the serial port with base_raw_termios.
		 */
		if (fast_com_ports) {
			base_raw_termios.c_ispeed = B115200;
			base_raw_termios.c_ospeed = B115200;
		}
		com_cons_init(gdb_com_port, &base_raw_termios);

		/*
		 * Drain it.
		 */
		while (com_cons_trygetchar(gdb_com_port) != -1)
			;
	}
#endif /* CONFIG_GDB_SERIAL */
}


#include "config_fdev_console.h"
#include "config_nopreempt.h"
#include "config_kpreempt.h"

static int
do_getchar(void *foo)
{
	int *c = (int *) foo;

#ifdef CONFIG_GDB_SERIAL
	if (enable_kdb && cons_com_port == gdb_com_port) {
		*c = gdb_serial_getchar();
		return 0;
	}
#endif
#if !defined(CONFIG_FDEV_CONSOLE) && \
    !defined(CONFIG_NOPREEMPT) && !defined(CONFIG_KPREEMPT)
	/*
	 * XXX big ole' steamin' hack.
	 * Only attempt to read chars until we get one or a preemption
	 * event (i.e., reschedule) occurs.  Gives us cheap interruptible
	 * input.
	 */
	do {
		*c = com_cons_trygetchar(cons_com_port);
		if (*c != -1) {
			return 0;
		}
	} while (!preempt_pending());
	return KR_RESTART;
#else
	*c = com_cons_getchar(cons_com_port);
	return 0;
#endif
}

static int
do_putchar(void *args)
{
	int c=(int)args;

#ifdef CONFIG_GDB_SERIAL
	if (enable_kdb && (cons_com_port==gdb_com_port))
		gdb_serial_putchar(c);
	else
#endif
		com_cons_putchar(cons_com_port, c);	/* serial line */
	direct_cons_putchar(c);	/* monitor */
	return 0;
}


static int
do_puts(void *args)
{
	char *s = args;

	/* XXX -- added this to avoid re-locking spinlock on every char! */
	base_critical_enter();

#ifdef CONFIG_GDB_SERIAL
	if (enable_kdb && (cons_com_port==gdb_com_port)) {
		gdb_serial_puts(s);
		while (*s)
			direct_cons_putchar(*s++);
		direct_cons_putchar('\n');
	} else
#endif
	{
#ifdef CONFIG_SMP
		{
			putchar(smp_find_cur_cpu() + '0');
			putchar(':');
		}
#endif
		while (*s) {
			putchar(*s++);
		}
		putchar('\n');
	}
	base_critical_leave();

	return 0;
}

struct putbytes_args
{
	const char *str;
	int len;
};

static int
do_putbytes(void *args)
{
	struct putbytes_args *a = args;
	const char *s = a->str;
	int len = a->len;

	/* XXX -- added this to avoid re-locking spinlock on every char! */
	base_critical_enter();
	while (len) {
		putchar(*s++);
		len--;
	}
	base_critical_leave();
	return 0;
}

static int
do_exit(void *args)
{
	extern oskit_addr_t return_address;
	int rc=(int)args;

#if FLASK
#ifdef DEBUG
	avc_dump("kernel");
#else
#if AVC_KEEP_STATS
	avc_dump_stats("kernel");
#endif
#endif
#endif

#ifdef CONFIG_INSTRUMENT
	dump_instrument_threads();
#endif
#ifdef CONFIG_SMP
	printf("_exit(%d) called by %d; rebooting...\n",rc, smp_find_cur_cpu());
#else
	printf("_exit(%d) called; rebooting...\n",rc);
#endif CONFIG_SMP
	if (return_address)
		printf("Have return_address 0x%08x\n\n", return_address);
#ifdef CONFIG_GDB_SERIAL
	if (enable_kdb) {
		gdb_serial_exit(rc);
		com_cons_flush(gdb_com_port);
	}
#endif
	com_cons_flush(cons_com_port);

	if (return_address) {
		/*
		 * The cleanup needs to be done here instead of in the
		 * returned-to code because the return address may not
		 * be accessible with our current paging and segment
		 * state.
		 * The order is important here: paging must be disabled
		 * after we reload the gdt.
		 */
		cli();
		phys_mem_va = 0;
		linear_base_va = 0;
		base_gdt_init();
		/* Reload all since we changed linear_base_va. */
		base_cpu_load();
		paging_disable();
		((void (*)(void))return_address)();
	}
	else
		pc_reset();

	otsan();
	return 0;
}

/*
 * These are the entry points that call the kernel-mode
 * functions to handle I/O.
 */

int
console_getchar()
{
        int c;

	do_in_supervisor_mode(do_getchar, (void *)&c);
	return c;
}

int
console_putchar(int c)
{
	do_in_supervisor_mode(do_putchar, (void *)c);
	return 0;
}

int
console_puts(const char *s)
{
	do_in_supervisor_mode(do_puts, (void *)s);
	return 0;
}

int
console_putbytes(const char *s, int len)
{
	struct putbytes_args args;

	args.str = s;
	args.len = len;
	
	do_in_supervisor_mode(do_putbytes, (void *)&args);
	return 0;
}

void
_exit(int rc)
{
	do_in_supervisor_mode(do_exit, (void *)rc);
}

