/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Implementation of context switching code for x86 for blocking kernel.
 */

#include <oskit/x86/paging.h>
#include <oskit/x86/base_tss.h>
#include <oskit/x86/base_paging.h>
#include <oskit/x86/base_stack.h>
#include "per_cpu.h"

#include "csw.h"
#include "s_thread.h"
#include "debug.h"
#include "per_cpu.h"

#include "config_opt_xfer.h"

struct s_thread;

void csw_boot(void)
{
}

void thread_csw_state_setup_stack(
	struct s_thread		*thread,
	int			(*init_func)(void),
	void			(*start_func)(void),
	void			*start_sp)
{
	/*
	 * Push the start_func address on the stack
	 * as the return value for init_func.
	 */
	((int*)start_sp)[-1] = (int)start_func;

	/*
	 * Now initialize the jmp_buf to jump to init_func on startup.
	 */
	thread->csw_state.state[3] = 0; /* clear ebp to terminate backtraces */
	thread->csw_state.state[4] = (int)start_sp - 4;
	thread->csw_state.state[5] = (int)init_func;
}

#include "fdev/bit.h"
struct s_thread *csw_switch(
	struct s_thread		*cur_thread,
	struct s_thread		*next_thread)
{
	oskit_addr_t pdir_pa = next_thread->csw_state.pdir_pa;

	assert(cur_thread);
	assert(next_thread);

	/*
	 * Switch to the target thread's address space,
	 * if it's different from the current thread's.
	 */
	if (get_pdbr() != pdir_pa) {
		set_pdbr(pdir_pa);

#ifdef CONFIG_OPT_XFER
		/*
		 * Since we've just flushed the TLB,
		 * our cross-space copy window is now clean.
		 * XXX should be per-processor
		 */
		*pdir_find_pde(pdir_pa, LIN_XFER_MIN) = 0;
#endif

#ifdef IOPL0
		/*
		 * Update the per-task info (IO space bitmask),
		 * but only if this thread has a task.
		 * The task pointer is lazily updated as the result
		 * of a trap on the first instruction it executes.
		 */
		if (next_thread->task_state.task)
			bcopy(next_thread->task_state.task->bitmask, my_bitmap,
				65536/8);
#endif

	}

	/*
	 * Change the ESP0 pointer in the current processor's TSS
	 * so that on reentry from user mode
	 * we will come back on the current thread's supervisor stack.
	 */
	assert(&my_cpu != 0);
	my_tss.esp0 = (oskit_addr_t)(&next_thread->exc_state + 1);

	/*
	 * Context switch the kernel stack and register state.
	 * This is only necessary in the blocking model,
	 * where there is one stack per thread.
	 */
	if (setjmp(cur_thread->csw_state.state) == 0) {
#ifdef PREEMPTSTATS
#ifdef HEARTBEAT_THREAD
		extern s_thread_t *hbthread;
		extern unsigned int hbstartt, hbendt, hbruns, hbcycles;

		if (cur_thread == hbthread) {
			asm volatile("rdtsc" : "=a" (hbendt) :: "eax", "edx");
			hbruns++;
			hbcycles += (hbendt - hbstartt);
		}
#endif
#endif
		longjmp(next_thread->csw_state.state, 1);
	}

#ifdef PREEMPTSTATS
	{
#ifdef HEARTBEAT_THREAD
		extern s_thread_t *hbthread;
		extern unsigned int hbstartt;
		extern unsigned int hbqtime, hbdcycles;
#endif
		int tmp, qtmp;
		struct disp_state *ds = &cur_thread->disp_state;


		asm volatile("rdtsc" : "=a" (tmp) :: "eax", "edx");
#ifdef HEARTBEAT_THREAD
		if (cur_thread == hbthread) {
			hbstartt = tmp;
			hbdcycles += (tmp - hbqtime);
		}
#endif
		qtmp = tmp;
		if (ds->valid) {
			tmp -= ds->stamp;
			if (tmp < ds->latmin)
				ds->latmin = tmp;
			if (tmp > ds->latmax)
				ds->latmax = tmp;
#if 1
			if (!(ds->lattotal + tmp > ds->lattotal))
				printf("WARNING: latency counter wrap on %p\n", cur_thread);
#else
			assert(ds->lattotal + tmp > ds->lattotal);
#endif
			ds->lattotal += tmp;
			ds->scheds++;
			ds->valid = 0;
		}
	}
#endif
	return cur_thread;
}

#include "preempt.h"

void
csw_start_user(
	struct s_thread		*cur_thread)
{
	my_tss.esp0 = (oskit_addr_t)(&cur_thread->exc_state + 1);

	assert_preemption_enabled();

	/*
	 * Bypass the init_func and go directly to the start func.
	 */
	cur_thread->csw_state.state[5] = (int) &&ret_insn;
	longjmp(cur_thread->csw_state.state, KR_RESTART);

 ret_insn: asm volatile("ret");
}
