/*
 * Copyright (c) 1996-1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Timer interrupt support.
 * XXX: This will go away with the user-mode device driver fixes.
 *
 * This is straight out of the OSKit default implementation,
 * except timer_intr() has been renamed, and we don't set the PIT and
 * register for a handler.
 */

#include "config_fdev.h"
#ifdef CONFIG_FDEV

#include <oskit/dev/dev.h>
#include <oskit/debug.h>
#include <oskit/x86/pio.h>
#include <oskit/x86/proc_reg.h>
#include <oskit/x86/pc/pit.h>
#include <oskit/x86/eflags.h>


#define NTIMERS		10
#define TIMER_FREQ	100
#define TIMER_VALUE	(PIT_HZ / TIMER_FREQ)

struct timer_handler {
	void	(*func)(void);
	struct	timer_handler *next;
};

static struct timer_handler *timer_head, *timer_free;
static struct timer_handler timer_data[NTIMERS];

/*
 * Generic timer interrupt dispatcher for fdev devices.
 * This is called by the clock interrupt handler in clock_md.c.
 */
void
fdev_timer_intr()
{
	struct timer_handler *th;

	for (th = timer_head; th; th = th->next)
		(*th->func)();
}


static int initialized = 0;	/* guard against multiple invocations */
/*
 * Initialize timers.
 */
void
osenv_timer_init()
{
	int i;

	printf("In %s\n", __FUNCTION__);

	if (initialized)
		return;

	initialized = 1;

	for (i = 0; i < NTIMERS; i++)
		timer_data[i].next = timer_data + i + 1;
	timer_data[--i].next = 0;
	timer_free = &timer_data[0];

	/*
	 * We get called by the Fluke kernel at HZ frequency == 100
	 * We don't install a handler; it is done automatically.
	 */
}

/*
 * XXX should this do anything?
 */
void
osenv_timer_shutdown()
{
}

/*
 * Register a timer handler to be called at the specified frequency.
 */
void
osenv_timer_register(void (*func)(void), int freq)
{
	struct timer_handler *th;
	unsigned int flags;

	assert(initialized);	/* osenv_timer_init must have been called */
	if (timer_free == 0)
		panic("%s:%d: ran out of entries", __FILE__, __LINE__);
	/*
	 * XXX
	 */
	if (freq != TIMER_FREQ)
		panic("%s:%d: bad freq", __FILE__, __LINE__);
	if (func == 0)
		panic("%s:%d: null function", __FILE__, __LINE__);

	flags = get_eflags();
	cli();
	th = timer_free;
	timer_free = th->next;
	th->next = timer_head;
	timer_head = th;
	th->func = func;
	set_eflags(flags);
}

void
osenv_timer_unregister(void (*func)(void), int freq)
{
	struct timer_handler *th, **lastp;
	unsigned flags;

	flags = get_eflags();
	cli();
	for (th = timer_head, lastp = &timer_head;
	     th && th->func != func;
	     lastp = &th->next, th = th->next)
		;
	assert(th);
	*lastp = th->next;
	th->next = timer_free;
	timer_free = th;
	set_eflags(flags);
}

static int readpit()
{
	int enb = get_eflags() & EFL_IF;
	int value;

	cli();
	value = pit_read(0);
	if (enb)
		sti();
	return value;
}

/*
 * Spin for a small amount of time, specified in nanoseconds,
 * without blocking or enabling interrupts.
 */
void
osenv_timer_spin(long nanosec)
{
	int prev_val = readpit();

	while (nanosec > 0) {
		int val = readpit();
		if (val > prev_val)
			prev_val -= TIMER_VALUE;
		nanosec -= (prev_val - val) * PIT_NS;
		prev_val = val;
	}
}

#endif /* CONFIG_FDEV */
