/*
 * Copyright (c) 1996 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Hardware interrupt management for uniprocessor x86 PC's.
 */

#include <stdio.h>

#include <oskit/x86/pio.h>
#include <oskit/x86/pc/pic.h>
#include <oskit/x86/base_idt.h>
#include <oskit/x86/base_gdt.h>
#include <oskit/x86/gate_init.h>
#include <oskit/x86/pc/base_irq.h>

#include "config_gdb_serial.h"
#include "config_fdev.h"

#include "s_thread.h"
#include "preempt.h"
#include "exception.h"
#include "dispatch.h"
#include "intr.h"
#include "kalloc.h"
#include "debug.h"

#define NHWI 16

static void std_preempter(struct preempter *p);

struct int_handler {
	void			(*func)(void *data);
	void			*data;
	struct int_handler	*next;
};

struct int_vector {

	/*
	 * Prempter used to defer normal interrupt processing
	 * to a point at which it is safe to wake up threads and such.
	 * Code below assumes that this is first in the int_vector.
	 */
	struct preempter	preempter;

	/*
	 * Chain of normal asynchronous interrupt handlers
	 * that don't care about the state of the system on entry.
	 * Normal interrupt handlers can share an interrupt vector
	 * only if the first one on that vector sets 'shareable'.
	 */
	struct int_handler	*handlers;
	int			shareable;

	/*
	 * Special interrupt handlers
	 * that are sensitive to the state of the system on entry.
	 * The entry-from-kernel-mode function is given a trap_state
	 * containing the saved kernel-mode processor state;
	 * it can be examined and modified freely before returning.
	 * The entry-from-user-mode function is entered
	 * just like a system call or trap handler:
	 * the saved state is in the current_thread(),
	 * and the return value indicates what to do next.
	 * The from-user-mode handler can do anything
	 * that a trap or system call handler can do,
	 * including context switch to another thread.
	 */
	void	(*k_func)(void *data, struct trap_state *ts);
	int	(*u_func)(void *data);
	void	*data;
} vector[NHWI];

/*
 * This routine is called whenever a hardware interrupt occurs in user mode.
 * It disables the interrupt in the PIC,
 * then wakes up the thread waiting for the appropriate interrupt.
 * Assumes the processor interrupt flag is disabled.
 *
 * XXX interrupt handling could be sped up a lot
 * just be inlining and specializing this PIC munging code
 * into the individual entry stubs in interrupt.S.
 */
int 
u_interrupt(unsigned irq)
{
	int rc = 0;
	struct int_vector *v = &vector[irq];

	assert(irq >= 0);
	assert(irq < NHWI);

	pic_disable_irq(irq);
	pic_ack(irq);

	assert_preemption_enabled();
	assert_interrupts_disabled();

	if (v->handlers)
		intr_preempter(&v->preempter, std_preempter);

	/* call special handler, if exists */
	if (vector[irq].u_func)
		vector[irq].u_func(vector[irq].data);

	/* interrupts are still disabled */
	return rc;
}

/*
 * Corresponding handler routine for interrupts from supervisor mode.
 * The trap_state in standard oskit format is on the stack.
 */
void 
k_interrupt(unsigned irq, ... /* trap_state */)
{
	struct trap_state *ts = (struct trap_state*)(&irq+1);
	struct int_vector *v = &vector[irq];

	assert(irq >= 0);
	assert(irq < NHWI);

	assert_interrupts_disabled();
	pic_disable_irq(irq);
	pic_ack(irq);

	if (v->handlers)
		intr_preempter(&v->preempter, std_preempter);

	/* call special handler, if exists */
	if (vector[irq].k_func)
		vector[irq].k_func(vector[irq].data, ts);
}


/*
 * This is the routine called by the preempter code when preemption
 * is allowed.  It is responsible for calling the registered handlers.
 */
static void std_preempter(struct preempter *p)
{
	struct int_vector *v = (struct int_vector*)p;
	struct int_handler *h;
	int irq = v - vector;

	assert(irq >= 0);
	assert(irq < NHWI);
	assert_preemption_disabled();

#ifdef DEBUG
	if (v->handlers == NULL) {
		static int lastirq = -1;

		if (irq != lastirq) {
			printf("unexpected interrupt: %d\n", irq);
			lastirq = irq;
		}
		/* XXX make sure it remains disabled at the PIC? */
	}
#endif
	/* Call the interrupt handlers */
	for (h = v->handlers; h; h = h->next) {
		h->func(h->data);
	}

	disable_interrupts();

#if 0
	/* Now that we've run all the handlers we can reenable the IRQ. */
	/* XXX: with fdev interrupt handlers in user mode, they
	 * haven't actually been run yet...
	 * XXX: We now enable the interrupt in the user-mode handler so a
	 * non-shared level-triggered interrupt will work.
	 */
	pic_enable_irq(irq);
#endif

	assert_preemption_disabled();
}


/*
 * Attach a normal device interrupt handler to an IRQ line.
 * Normal handlers can potentially share the same interrupt line,
 * but cannot be sensitive to the state
 * of the interrupted user- or kernel-mode activity.
 */
int
irq_alloc(int irq, void (*func), void *data, int shareable)
{
	struct int_vector *v = &vector[irq];
	struct int_handler *h;
	int ret = 0;

	assert(irq >= 0); assert(irq < NHWI);

	assert_interrupts_enabled();
	disable_interrupts();

	if (v->handlers != NULL && (!shareable || !v->shareable)) {
		ret = -1;
		goto out;
	}

	h = kalloc(sizeof(*h));
	if (h == NULL) {
		ret = -1;
		goto out;
	}

	h->func = func;
	h->data = data;
	h->next = v->handlers;
	v->handlers = h;
	v->shareable = shareable;

	/* XXX */
	pic_enable_irq(irq);
out:
	enable_interrupts();

	return ret;
}

/* Use a handle so know WHICH interrupt handler to remove (func & data). */
void
irq_free(int irq, void *func, void *data)
{
	struct int_vector *v = &vector[irq];
	struct int_handler *h, **hp;

	assert(irq >= 0); assert(irq < NHWI);

	assert_interrupts_enabled();
	disable_interrupts();

	for (hp = &v->handlers; ; hp = &h->next) {
		h = *hp;
		assert(h != NULL);	/* no failure allowed.. */
		if ((h->func == func) && (h->data == data)) {
			break;
		}
	}

	*hp = h->next;
	if (v->handlers == NULL) {
		pic_disable_irq(irq);
	}

	enable_interrupts();
	kfree(h, sizeof(*h));
}

/*
 * Attach a special kernel/user interrupt handler to an IRQ line.
 * The IRQ line cannot be shared with any other special handlers.
 */
int 
irq_alloc_ku(int irq,
	     void (*k_func)(void *data, struct trap_state *ts),
	     int (*u_func)(void *data),
	     void *data)
{
	assert(irq >= 0); assert(irq < NHWI);

	if (vector[irq].u_func)
		return -1;	/* already in use */

	disable_interrupts();

	vector[irq].k_func = k_func;
	vector[irq].u_func = u_func;
	vector[irq].data = data;

	pic_enable_irq(irq);

	enable_interrupts();

	return 0;
}

void
irq_free_ku(int irq)
{
	assert(irq >= 0); assert(irq < NHWI);
	assert(vector[irq].k_func);
	assert(vector[irq].u_func);

	disable_interrupts();

	if (!vector[irq].handlers)
		pic_disable_irq(irq);

	vector[irq].k_func = NULL;
	vector[irq].u_func = NULL;
	vector[irq].data = (void*)irq;

	enable_interrupts();
}

int 
interrupt_wait(unsigned irq)
{
otsan();
return 0;
#if 0
	struct s_thread *th = current_thread();
	int rc;

	disable_interrupts();

	/* Assert that we are about to do a cancelable wait
	   on the hardware interrupt's wait queue.  */
	assert(irq < 16);
	rc = assert_cancelable_wait(th, &intr_queue[irq]);
	if (rc)
	{
		enable_interrupts();
		return rc;
	}

	pic_enable_irq(irq);

	/* Go to sleep waiting on the appropriate condition.  */
	rc = do_cancelable_wait(th, &intr_queue[irq], 0, XXX_function?);

	return rc;
#endif
}

extern struct gate_init_entry base_int_inittab[];

void
interrupt_boot(void)
{
	int i;
#ifdef CONFIG_GDB_SERIAL
	void gdb_intr_boot(void);
#endif

        disable_interrupts();

        gate_init(base_idt, base_int_inittab, KERNEL_CS);

	for (i = 0; i < NHWI; i++) {
		vector[i].k_func = NULL;
		vector[i].u_func = NULL;
		vector[i].data = (void*)i;
	}

        pic_init(BASE_IRQ_MASTER_BASE, BASE_IRQ_SLAVE_BASE);
        pic_disable_all();

        enable_interrupts();

#ifdef CONFIG_GDB_SERIAL
        /* Enable the GDB interrupt so that CTRL-C works. */
        gdb_intr_boot();
#endif
}

