/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * This file implements the special POSIX threads interface
 * for use by kernel-loaded modules when creating and using (kernel) threads.
 * It is necessary to provide this higher-level threads interface for them,
 * rather than just having them use the normal user-level pthreads library
 * which is built upon the underlying fluke_thread_* operations,
 * because those operations create _user-level_ threads
 * even when called from kernel-loaded modules.
 * (And in any case, it's faster and simpler this way.)
 *
 * This file overrides kernel/x86/pthread.c, when kmodule is enabled.
 * It is supplemented by the pthread code in kernel/x86/pthread_shared.c
 */

#include <pthread.h>
#include "pthread_impl.h"
#include <malloc.h>
#include <string.h>
#include <strings.h>	/* bzero */
#include <errno.h>
#include <fluke/task.h>
#include <fluke/thread.h>

#include "s_task.h"
#include "s_thread.h"
#include "kentry.h"
#include "debug.h"
#include "kalloc.h"




static void
kmodule_pthread_start_func(void)
{
	s_thread_t *cur_thread = current_thread();
	void *(*start_func)(void *arg);
	void *start_arg;

	/*
	 * We need to do this so flobs and IPC work.
	 * Normally it's done by faulting in user mode.
	 */
	thread_task_state_start(cur_thread);

	/*
	 * Call the appropriate start function,
	 * which was stashed in the "user-level" register state area.
	 */
	start_func = (void*)cur_thread->exc_state.eax;
	start_arg = (void*)cur_thread->exc_state.ebx;

	pthread_exit(start_func(start_arg));
}


/*
 * Replacement for kernel/kserver/x86/pthread.c:pthread_create.  This
 * function creates a kernel-mode thread, and sets things up so that
 * the start_routine is executed as soon as the new thread is switched
 * to.  (The kserver implementation sets things up so that start_routine
 * is executed upon resumption of user-mode execution.)
 *
 * This returns the address of the user mode portion of the thread
 * in the pthread_t address.  Icky.
 */
int
pthread_create(pthread_t *out_thread, const pthread_attr_t *attr,
		void *(*start_routine)(void *), void *start_arg)
{
	struct pthread *uth;
	s_thread_t *sth;
	int rc;

	/* XXX: still ignores stacksize_attr */

	assert(pthread_can_exit);

	assert(sizeof(fluke_thread_t) == FLUKE_THREAD_SIZE);
	uth = kalloc_aligned(sizeof(*uth), FLUKE_THREAD_SHIFT);
	if (uth == 0)
		return ENOMEM;

	rc = s_thread_create((oskit_addr_t)uth, 0, &sth);
	assert_rc_0(rc);	/* XXX */
	assert_s_thread_locked(sth);

	sth->thread_va = (oskit_addr_t)uth;

	/*
	 * Attach this thread to the kernel task,
	 * so that IPC memory transfers and page faults will work for it.
	 */
	link_transfer(&kernel_task_link, &sth->task_state.task_link);

	/*
	 * Change the csw state to run our own start routine,
	 * rather than the usual one which drops into user mode.
	 */
	sth->exc_state.eax = (unsigned)start_routine;
	sth->exc_state.ebx = (unsigned)start_arg;
	thread_disp_state_setup_stack(sth, kmodule_pthread_start_func,
		&sth->exc_state);

	sth->gflags &= ~FLUKE_THREAD_STOPPED;

	thread_ready(sth);

	s_thread_unlock(sth);

	pthread_uth_init(uth);

	*out_thread = (pthread_t)uth;
	return 0;
}
