/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Implementation of context switching code for x86 in nonblocking kernel.
 */

#include <oskit/x86/paging.h>
#include <oskit/x86/base_tss.h>
#include <oskit/x86/base_paging.h>
#include <oskit/x86/base_stack.h>

#include "csw.h"
#include "s_thread.h"
#include "debug.h"

struct s_thread;

extern oskit_addr_t null_pdir_pa; /* XXX layering violation */

void csw_boot(void)
{
	/*
	 * Initialize the current thread pointer in the base stack
	 * (see x86/startup_thread.S).
	 */
	current_thread() = &startup_thread;

	/*
	 * Set the kernel stack pointer for trap and interrupt handling.
	 */
	base_tss.esp0 = (oskit_addr_t)base_stack_end;
}


struct s_thread *csw_switch(
	struct s_thread		*cur_thread,
	struct s_thread		*next_thread)
{
	oskit_addr_t pdir_pa = next_thread->csw_state.pdir_pa;

	assert(cur_thread);
	assert(next_thread);

	/*
	 * Switch to the target thread's address space,
	 * if it's different from the current thread's.
	 */
	if (get_pdbr() != pdir_pa) {
		set_pdbr(pdir_pa);

#ifdef CONFIG_OPT_XFER
		/*
		 * Since we've just flushed the TLB,
		 * our cross-space copy window is now clean.
		 * XXX should be per-processor
		 */
		*pdir_find_pde(pdir_pa, LIN_XFER_MIN) = 0;
#endif

#ifdef IOPL0
		/*
		 * Update the per-task info (IO space bitmask),
		 * but only if this thread has a task.
		 * The task pointer is lazily updated as the result
		 * of a trap on the first instruction it executes.
		 */
		if (next_thread->task_state.task)
			bcopy(next_thread->task_state.task->bitmask, my_bitmap,
				65536/8);
#endif

	}

	/* Update the notion of the current thread.  */
	current_thread() = next_thread;

#ifdef PREEMPTSTATS
	{
#ifdef HEARTBEAT_THREAD
		extern s_thread_t *hbthread;
		extern unsigned int hbstartt, hbendt, hbruns, hbcycles;
		extern unsigned int hbqtime, hbdcycles;
#endif
		int tmp, qtmp;
		struct disp_state *ds = &next_thread->disp_state;

		asm volatile("rdtsc" : "=a" (tmp) :: "eax", "edx");
#ifdef HEARTBEAT_THREAD
		if (cur_thread == hbthread) {
			hbendt = tmp;
			hbruns++;
			hbcycles += (hbendt - hbstartt);
		} else if (next_thread == hbthread) {
			hbstartt = tmp;
			hbdcycles += (tmp - hbqtime);
		}
#endif
		qtmp = tmp;
		if (ds->valid) {
			tmp -= ds->stamp;
			if (tmp < ds->latmin)
				ds->latmin = tmp;
			if (tmp > ds->latmax)
				ds->latmax = tmp;
#if 1
			if (!(ds->lattotal + tmp > ds->lattotal))
				printf("WARNING: latency counter wrap on %p\n", next_thread);
#else
			assert(ds->lattotal + tmp > ds->lattotal);
#endif
			ds->lattotal += tmp;
			ds->scheds++;
			ds->valid = 0;
		}
	}
#endif
	return next_thread;
}
