/*
 * Copyright (c) 1996-1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include <oskit/x86/asm.h>
#include <oskit/x86/trap.h>
#include <oskit/x86/eflags.h>
#include <oskit/x86/seg.h>
#include <oskit/x86/gate_init.h>
#include <oskit/x86/base_gdt.h>
#include <oskit/x86/anno.h>

#include <fluke/exception.h>
#include <fluke/x86/syscall.h>

#include "x86.h"
#include "idt.h"
#include "debug.h"

#include "config_opt_threads.h"
#include "config_gdb_serial.h"
#include "config_instrument.h"

	.text

#ifdef CONFIG_INSTRUMENT
#include "s_instrument.h"
/*
 * Push the flag indicating the type of syscall,
 * then push the actual entrypoint, then the timestamp
 * counter.  Call kentry_md_syscall_instrument() to record all
 * this.
 *
 */
#define INSTRUMENT_SYSCALL(flag)					\
	pushl   %edx		;					\
	pushl	%eax		; 					\
	rdtsc		/* Put tsc into edx:eax */ ; 			\
	pushl   %edx		; 					\
	pushl   %eax		; 					\
	pushl	$flag		;					\
	call	EXT(kentry_md_syscall_instrument) ; 			\
	addl	$12,%esp /* nuke the tsc and flag on the stack */ ; 	\
	popl    %eax     /* get original %eax back off the stack */ ;	\
	popl	%edx
#else
#define INSTRUMENT_SYSCALL(flag)
#endif

/*
 * Macro to load the standard segment registers
 * on entry to a hardware interrupt handler.
 * Uses ebx as a scratch register.
 */
#define LOAD_SEGS			\
	movl	$KERNEL_DS,%ebx		;\
	movw	%bx,%ds			;\
	movw	%bx,%es			;\
	cld


/*
 * General exception handler macro used by all of the following.
 * An error subcode of some kind has already been pushed
 * (zero in the case of traps where the processor doesn't provide one).
 */
#define	EXCEP_GEN(n,class)						 \
	testb	$3,8(%esp)						;\
	jz	1f							;\
	pushl	%ebx							;\
	LOAD_SEGS							;\
	movl	28-PROCESSOR_STACK_SIZE(%esp),%ebx			;\
	movl	$(n)|((class)<<8),TH_EXC_STATE+EXC_CODE(%ebx)		;\
	jmp	all_u_traps						;\
1:	pushl	$(n)	/* simple trap number for trap_state frame */	;\
	jmp	all_k_traps

/*
 * Privileged trap vector; error code has been pushed by the processor.
 */
#define	EXCEP_ERR(n,name,class)						 \
	GATE_ENTRY(n,name,ACC_PL_K|ACC_TRAP_GATE)			;\
name:	EXCEP_GEN(n,class)

/*
 * Privileged trap vector; no error code has been pushed.
 */
#define	EXCEPTION(n,name,class)						 \
	GATE_ENTRY(n,name,ACC_PL_K|ACC_TRAP_GATE)			;\
name:	pushl	$0							;\
	EXCEP_GEN(n,class)

/*
 * User-accessible exception.  Otherwise, same as above.
 */
#define	EXCEP_USR(n,name,class)						 \
	GATE_ENTRY(n,name,ACC_PL_U|ACC_TRAP_GATE)			;\
name:	pushl	$0							;\
	EXCEP_GEN(n,class)

/*
 * Special interrupt code: dispatches to a unique entrypoint,
 * not defined automatically here.
 */
#define	EXCEP_SPC(n,name,pl)  \
	GATE_ENTRY(n,name,pl|ACC_TRAP_GATE)


GATE_INITTAB_BEGIN(base_trap_inittab)

EXCEPTION(0x00, t_zero_div,	FLUKE_EXC_ARITHMETIC)
EXCEPTION(0x01, t_debug,	FLUKE_EXC_BREAKPOINT)
EXCEP_USR(0x03, t_int3,		FLUKE_EXC_BREAKPOINT)
EXCEP_USR(0x04, t_into,		FLUKE_EXC_ARITHMETIC)
EXCEP_USR(0x05, t_bounds,	FLUKE_EXC_ARITHMETIC)
EXCEPTION(0x06, t_invop,	FLUKE_EXC_BAD_INSTRUCTION)
EXCEPTION(0x07, t_nofpu,	FLUKE_EXC_BAD_INSTRUCTION)
EXCEPTION(0x08, a_dbl_fault,	FLUKE_EXC_BAD_ACCESS)
EXCEPTION(0x09, a_fpu_over,	FLUKE_EXC_BAD_ACCESS)
EXCEP_ERR(0x0a, a_inv_tss,	FLUKE_EXC_BAD_ACCESS)
EXCEP_ERR(0x0b, t_segnp,	FLUKE_EXC_BAD_ACCESS)
EXCEP_ERR(0x0c, t_stack_fault,	FLUKE_EXC_BAD_ACCESS)
EXCEP_ERR(0x0d, t_gen_prot,	FLUKE_EXC_BAD_ACCESS)
EXCEP_SPC(0x0e, t_page_fault,	ACC_PL_K)
EXCEPTION(0x0f, t_trap_0f,	FLUKE_EXC_BREAKPOINT)
EXCEPTION(0x10, t_fpu_err,	FLUKE_EXC_ARITHMETIC)
EXCEPTION(0x11, t_align_chk,	FLUKE_EXC_BAD_ACCESS)
EXCEPTION(0x12, t_machine_chk,	FLUKE_EXC_BREAKPOINT)
EXCEPTION(0x13, t_trap_13,	FLUKE_EXC_BREAKPOINT)
EXCEPTION(0x14, t_trap_14,	FLUKE_EXC_BREAKPOINT)
EXCEPTION(0x15, t_trap_15,	FLUKE_EXC_BREAKPOINT)
EXCEPTION(0x16, t_trap_16,	FLUKE_EXC_BREAKPOINT)
EXCEPTION(0x17, t_trap_17,	FLUKE_EXC_BREAKPOINT)
EXCEPTION(0x18, t_trap_18,	FLUKE_EXC_BREAKPOINT)
EXCEPTION(0x19, t_trap_19,	FLUKE_EXC_BREAKPOINT)
EXCEPTION(0x1a, t_trap_1a,	FLUKE_EXC_BREAKPOINT)
EXCEPTION(0x1b, t_trap_1b,	FLUKE_EXC_BREAKPOINT)
EXCEPTION(0x1c, t_trap_1c,	FLUKE_EXC_BREAKPOINT)
EXCEPTION(0x1d, t_trap_1d,	FLUKE_EXC_BREAKPOINT)
EXCEPTION(0x1e, t_trap_1e,	FLUKE_EXC_BREAKPOINT)
EXCEPTION(0x1f, t_trap_1f,	FLUKE_EXC_BREAKPOINT)
EXCEP_SPC(SYSCALL_INT, t_syscall, ACC_PL_U)

GATE_INITTAB_END


/*
 * Kernel page fault handler.
 * The processor's special page fault error code has been pushed,
 * but nothing else.
 */
k_page_fault:

	/* Push the trap number */
	pushl	$T_PAGE_FAULT

	/* fall through... */

/*
 * Handler for traps from supervisor mode (kernel traps).
 * In this case, all exception state is saved on the kernel stack.
 * The trap number and error code have been pushed, but nothing else.
 * Note that whereas traps from user mode use the exc_state structure,
 * traps from kernel mode use the oskit's trap_state.
 */
all_k_traps:

	/* Save registers */
	pusha
	pushl	%ds
	pushl	%es
	pushl	%fs
	pushl	%gs

	LOAD_SEGS
	CLEAR_EBP

trap_from_kernel:
	/* Search the trap annotation table for a recovery routine.  */
	pushl	TS_EIP(%esp)
	pushl	$EXT(anno_trap)
	call	EXT(anno_find_exact)
	addl	$8,%esp
	orl	%eax,%eax
	jz	1f
	movl	4(%eax),%ebx
	jmp	*%ebx
1:

#ifdef CONFIG_GDB_SERIAL
	/* Allow the remote debugger to handle the trap.  */
	pushl	%esp
	call	EXT(gdb_trap)
	popl	%edx
	orl	%eax,%eax
	jz	EXT(return_from_k_trap)
#endif

	/* Unexpected trap: just dump the state and panic. */
	pushl	%esp
	call	EXT(trap_dump_panic)
	.data
1:	.string "Unexpected trap in kernel"
	.text

	/* Return from a kernel trap and restart execution.  */
ENTRY(return_from_k_trap)
	popl	%gs
	popl	%fs
	popl	%es
	popl	%ds
	popa
	addl	$8,%esp
	iret



/*
 * Generic trap handler entrypoint for most traps from user mode.
 * The current thread pointer is in ebx, and the kernel seg regs are loaded.
 * The kernel stack looks like this:
 *
 *	esp ->	saved ebx
 *		exception subcode
 *		eip
 *		cs
 *		eflags
 *		esp
 *		ss
 */
	P2ALIGN(TEXT_ALIGN)
all_u_traps:

	/* Save the general registers in the thread state */
	movl	%eax,TH_EXC_STATE+EXC_EAX(%ebx)
	movl	%ecx,TH_EXC_STATE+EXC_ECX(%ebx)
	movl	%edx,TH_EXC_STATE+EXC_EDX(%ebx)
	popl	TH_EXC_STATE+EXC_EBX(%ebx)
	movl	%ebp,TH_EXC_STATE+EXC_EBP(%ebx)
	movl	%esi,TH_EXC_STATE+EXC_ESI(%ebx)
	movl	%edi,TH_EXC_STATE+EXC_EDI(%ebx)

	/* Move important stuff from the kernel stack to the thread state */
	popl	TH_EXC_STATE+EXC_SUBCODE(%ebx)
	movl	(%esp),%eax
	movl	8(%esp),%ecx
	movl	12(%esp),%edx
	movl	%eax,TH_EXC_STATE+EXC_EIP(%ebx)
	movl	%ecx,TH_EXC_STATE+EXC_EFLAGS(%ebx)
	movl	%edx,TH_EXC_STATE+EXC_ESP(%ebx)

	CLEAR_EBP

	/* Record the call into the kernel */
	INSTRUMENT_SYSCALL(INSTRUMENT_SYSCALL_USERTRAP)

	/* Call the standard exception handler for a user exception.  */
	pushl	$KR_USER_EXCEPTION
	call	EXT(s_thread_exception)
	addl	$4,%esp
	INSTRUMENT_SYSCALL(INSTRUMENT_SYSCALL_TRAPRET)
	jmp	EXT(return_to_user)


/*
 * Page faults are handled somewhat differently from other traps.
 * First, we need to save the page fault linear address from CR2.
 * Also, we need to call a different high-level handler routine
 * for traps from user mode,
 * and we need to check for fault handlers for traps from supervisor mode.
 * The processor's page fault error code has been pushed on the stack.
 */
	ALIGN_TEXT
t_page_fault:

	/* Check whether the trap was from user or kernel mode */
	testb	$3,8(%esp)
	jz	k_page_fault

	/* Trapped from user mode - save ebx and find the current thread */
	pushl	%ebx
	LOAD_SEGS
	movl	28-PROCESSOR_STACK_SIZE(%esp),%ebx

	/* Save the general registers in the thread state */
	movl	%eax,TH_EXC_STATE+EXC_EAX(%ebx)
	movl	%ecx,TH_EXC_STATE+EXC_ECX(%ebx)
	movl	%edx,TH_EXC_STATE+EXC_EDX(%ebx)
	popl	     TH_EXC_STATE+EXC_EBX(%ebx)
	movl	%ebp,TH_EXC_STATE+EXC_EBP(%ebx)
	movl	%esi,TH_EXC_STATE+EXC_ESI(%ebx)
	movl	%edi,TH_EXC_STATE+EXC_EDI(%ebx)

	CLEAR_EBP

	/* Pop the processor's page fault error code off the stack */
	popl	%eax

	/* Move important stuff from the kernel stack to the thread state */
	movl	(%esp),%esi
	movl	8(%esp),%ecx
	movl	12(%esp),%edx
	movl	%esi,TH_EXC_STATE+EXC_EIP(%ebx)
	movl	%ecx,TH_EXC_STATE+EXC_EFLAGS(%ebx)
	movl	%edx,TH_EXC_STATE+EXC_ESP(%ebx)

	/*
	 * The special error code pushed by the processor
	 * only includes one bit that we need to save: the read/write bit.
	 * Produce the protection mask in bits 16-20 of the exception code
	 * and use the faulting address instead as the subcode.
	 * Also check for page faults occuring in supervisor mode,
	 * and jump off to our special kernel page fault handling.
	 */
	movl	%cr2,%ecx
#ifndef CONFIG_NORWOX
	/* Linear VAs above 0x40000000 indicate execution faults */
	cmpl	$LIN_USERX_OFF,%ecx
	jle	2f
	subl	$LIN_USERX_OFF,%ecx
	movl $(FLUKE_PROT_EXEC<<16)|(FLUKE_EXC_PAGE_FAULT<<8)|T_PAGE_FAULT,%edx
	movl	%ecx,TH_EXC_STATE+EXC_SUBCODE(%ebx)
	jmp	1f
2:
#endif
	movl	%ecx,TH_EXC_STATE+EXC_SUBCODE(%ebx)
	movl $(FLUKE_PROT_WRITE<<16)|(FLUKE_EXC_PAGE_FAULT<<8)|T_PAGE_FAULT,%edx
	testb	$2,%al
	jnz	1f
	movl $(FLUKE_PROT_READ<<16)|(FLUKE_EXC_PAGE_FAULT<<8)|T_PAGE_FAULT,%edx
1:	movl	%edx,TH_EXC_STATE+EXC_CODE(%ebx)

	/* Record the call into the kernel */
	INSTRUMENT_SYSCALL(INSTRUMENT_SYSCALL_PAGEFAULT)

	/* Call the high-level page fault handler.  */
	pushl	$KR_PAGE_FAULT
	call	EXT(s_thread_exception)
	addl	$4,%esp
	INSTRUMENT_SYSCALL(INSTRUMENT_SYSCALL_TRAPRET)
	jmp	EXT(return_to_user)


/*
 * System call entrypoint.
 * Nothing has been pushed except the processor's basic interrupt stack frame.
 * We just assume we're coming in from user mode;
 * the kernel had better not make system calls!
 */
t_syscall:

	/* Find the current thread (ebx) */
	pushl	%ebx
	LOAD_SEGS
	movl	24-PROCESSOR_STACK_SIZE(%esp),%ebx

	/* Save the general registers in the thread state */
	movl	%eax,TH_EXC_STATE+EXC_EAX(%ebx)
	movl	%ecx,TH_EXC_STATE+EXC_ECX(%ebx)
	movl	%edx,TH_EXC_STATE+EXC_EDX(%ebx)
	popl	     TH_EXC_STATE+EXC_EBX(%ebx)
	movl	%ebp,TH_EXC_STATE+EXC_EBP(%ebx)
	movl	%esi,TH_EXC_STATE+EXC_ESI(%ebx)
	movl	%edi,TH_EXC_STATE+EXC_EDI(%ebx)

	CLEAR_EBP

	/*
	 * Move important stuff from the kernel stack to the thread state.
	 * The saved EIP tells us what entrypoint was called.
	 * Reset it to point to the start of the entrypoint code,
	 * so the system call can be restarted if an error occurs.
	 * Since all entrypoints are aligned on 16-byte boundaries
	 * and none of our entrypoint stubs are that big,
	 * we simply need to mask off the low-order four bits.
	 * XX no need to save eflags at all?
	 */
	movl	(%esp),%eax
	movl	8(%esp),%ecx
	movl	12(%esp),%edx
	andb	$0xf0,%al
	movl	%eax,TH_EXC_STATE+EXC_EIP(%ebx)
	movl	%ecx,TH_EXC_STATE+EXC_EFLAGS(%ebx)
	movl	%edx,TH_EXC_STATE+EXC_ESP(%ebx)

	/* Record the syscall being entered */
	INSTRUMENT_SYSCALL(INSTRUMENT_SYSCALL_ENTRY)

	/* Now look up the kernel routine that handles the system call.
	   The handler pointer is stored at entrypoint+12.
	   However, make sure the top-level part of the address
	   points into our system call stub page;
	   otherwise user code could spoof us into using a bad pointer.  */
	andl	$0x00000ff0,%eax
	call	*EXT(u_entry_page+12)(%eax)

	/* Record the return from the syscall */
	INSTRUMENT_SYSCALL(INSTRUMENT_SYSCALL_EXIT)
	/*
	 * If the syscall routine returns zero, indicating success,
	 * then we need to send the user EIP to the fluke_nop entrypoint,
	 * so that the return address can be popped safely in user mode.
	 * If the syscall routine returned nonzero,
	 * handle it as an exception that occurred during the syscall.
	 *
	 * XX It might be better just to make the syscall routines
	 * handle this and avoid the extra complexity here.
	 */
	orl	%eax,%eax
	jnz	handle_exception

	/*
	 * Return to the fluke_nop entrypoint,
	 * causing the system call to "finish" instead of restart.
	 */
	movl	$fluke_nop__entry,TH_EXC_STATE+EXC_EIP(%ebx)


/*
 * This exception return entrypoint is used
 * when returning from an exception that occured while in user mode.
 * The return code from the system call or exception handler is in eax.
 * A pointer to the current thread on entry into the kernel is in ebx;
 * however, it can only be used as a current thread pointer if eax is zero,
 * since if eax is nonzero a thread switch may have occurred.
 * The kernel stack looks like this:
 *
 *	esp ->	eip	(invalid)
 *		cs	(USER_CS)
 *		eflags	(invalid)
 *		esp	(invalid)
 *		ss	(USER_DS)
 */
	.globl	EXT(return_to_user)
LEXT(return_to_user)

	/*
	 * Check the return code from the system call or trap handler.
	 * If it is zero, just return to user mode and resume execution.
	 * Otherwise, we must perform some special processing
	 * for the exceptional situation.
	 */
	orl	%eax,%eax
	jnz	handle_exception

	/*
	 * Before returning to user mode, check for a pending cancel condition,
	 * which is Fluke's equivalent of an AST or pending signal.
	 * XXX need to disable interrupts here to close the window?
	 */
	movb	TH_CANCEL_PEND(%ebx),%al
	orb	%al,%al
	jnz	handle_cancel_pending


	/*
	 * XXX combine the tests for cancellation and preemption,
	 * make them use the processor's trace flag,
	 * and perhaps decompose cancel_pending into multiple bits...
	 */
	/* Do not preempt if nobody else is ready.  */
#ifdef CONFIG_SMP
	movl	-TH_EXC_STATE+TH_CPU(%esp),%eax
	cmpl	$0,CPU_PREEMPTERS(%eax)
#else
	cmpl	$0,CPU_PREEMPTERS+EXT(boot_cpu_info)
#endif
	jne	handle_preempt

return_to_user2:
	/* Set the new eip, esp, and eflags */
	movl	TH_EXC_STATE+EXC_EIP(%ebx),%eax
	movl	TH_EXC_STATE+EXC_EFLAGS(%ebx),%ecx
	movl	TH_EXC_STATE+EXC_ESP(%ebx),%edx
	movl	%eax,(%esp)
	movl	%ecx,8(%esp)
	movl	%edx,12(%esp)

	/* Restore the general-purpose registers */
	movl	TH_EXC_STATE+EXC_EAX(%ebx),%eax
	movl	TH_EXC_STATE+EXC_ECX(%ebx),%ecx
	movl	TH_EXC_STATE+EXC_EDX(%ebx),%edx
	movl	TH_EXC_STATE+EXC_EBP(%ebx),%ebp
	movl	TH_EXC_STATE+EXC_ESI(%ebx),%esi
	movl	TH_EXC_STATE+EXC_EDI(%ebx),%edi
	movl	TH_EXC_STATE+EXC_EBX(%ebx),%ebx

	/* Restore the user-mode segment registers */
	movw	EXC_SS-EXC_EIP(%esp),%ds
	movw	EXC_SS-EXC_EIP(%esp),%es

	/* Return from the trap */
	iret


handle_cancel_pending:
	movl	$KR_CANCEL,%eax

handle_exception:
	/* Find the (new) current thread pointer.  */
	movl	20-PROCESSOR_STACK_SIZE(%esp),%ebx

#if 0	/* XXX now handled in dispatch.c */
#ifdef CONFIG_OPT_THREADS
	/* Set the user-visible current thread pointer */
	movl	TH_THREAD_VA(%ebx),%ecx
	movl	%ecx,%gs:GS_CUR_THREAD

	/* XXX Eventually, make only the current _task_ pointer
	be held on the base of the kernel stack -
	always keep the current _thread_ pointer in user space. */
#endif
#endif

	/*
	 * KR_RESTART "exceptions" are handled
	 * just by returning to user mode immediately -
	 * it is used in various situations to indicate
	 * that the current operation should be "restarted"
	 * rather than "continued" or "finished".
	 */
	cmpl	$KR_RESTART,%eax
	je	handle_restart

handle_generic_exception:
	/* Record call into kernel */
	INSTRUMENT_SYSCALL(INSTRUMENT_SYSCALL_EXCEPT)

	/* Call the generic exception handler.  */
	pushl	%eax
	call	EXT(s_thread_exception)
	addl	$4,%esp
	INSTRUMENT_SYSCALL(INSTRUMENT_SYSCALL_TRAPRET)
	jmp	EXT(return_to_user)

handle_restart:
	xorl	%eax,%eax
	jmp	EXT(return_to_user)

handle_preempt:
	call	thread_run_preempters
	jmp	EXT(return_to_user)


/*
 * Drop into a user-mode thread for the first time on a processor.
 * The single argument is the current thread pointer.
 */
ENTRY(csw_start_user)
	movl	4(%esp),%ebx
	movl	$USER_CS,4(%esp)
	movl	$USER_DS,16(%esp)
	xorl	%eax,%eax
	/*
	 * XXX cannot jump to return_to_user.
	 * The clock is running and can schedule preempters.
	 * Thread_run_preempters would return KR_RESTART which would
	 * cause us to attempt to reload the thread pointer from
	 * the stack before we are on a "real" thead stack.
	 * So make sure we skip the exception/cancel/preempter checks.
	 */
	jmp	return_to_user2
