/*
 * Copyright (c) 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * This file defines the PC-lusring (that is, rollback/roll-forward)
 * that needs to be done for the various stages in the middle of the
 * u_synch_page.S calls that a user thread could be.
 */

#include "s_thread.h"
#include "s_mutex.h"
#include "u_entry.h"
#include <oskit/x86/eflags.h>


/*
 * %eax and %ecx may have been clobbered, but that's all.  Roll back.
 */
void
pclsr_u_mutex_trylock(struct s_thread *sth)
{
	sth->exc_state.eip = fluke_mutex_trylock__entry;
}

/*
 * He performed the cmpxchg, and the condition codes indicate what happened.
 */
void
pclsr_u_mutex_trylock_cmpxchg(struct s_thread *sth)
{
	/* EFL_ZF is 0x40: set eax to 0 if that bit is set, 1 if not.  */
	sth->exc_state.eax = ((sth->exc_state.eflags & 0x40) >> 2) ^ 1;
	sth->exc_state.eip = fluke_nop__entry;
}


/*
 * %eax and %ecx may have been clobbered, but that's all.  Roll back.
 */
void
pclsr_u_mutex_lock(struct s_thread *sth)
{
	sth->exc_state.eip = fluke_mutex_lock__entry;
}

/*
 * He performed the cmpxchg, and the condition codes indicate what happened.
 */
void
pclsr_u_mutex_lock_cmpxchg(struct s_thread *sth)
{
	if (sth->exc_state.eflags & EFL_ZF)
		/*
		 * He got the lock.  Roll forward to "finished syscall" state.
		 */
		sth->exc_state.eip = fluke_nop__entry;
	else
		/*
		 * He didn't get the lock.  Roll back.
		 */
		sth->exc_state.eip = fluke_mutex_lock__entry;
}


/*
 * %eax and %ecx may have been clobbered, but that's all.  Roll back.
 */
void
pclsr_u_mutex_unlock(struct s_thread *sth)
{
	sth->exc_state.eip = fluke_mutex_unlock__entry;
}

/*
 * He performed the cmpxchg, and the condition codes indicate what happened.
 */
void
pclsr_u_mutex_unlock_cmpxchg(struct s_thread *sth)
{
	if (sth->exc_state.eflags & EFL_ZF) {
		/*
		 * The lock was not contested and he completed the unlock.
		 * Roll forward to "finished syscall" state.
		 */
		sth->exc_state.eip = fluke_nop__entry;
	}
	else {
		/*
		 * The lock was contested and so he needs to enter the
		 * kernel to complete the unlock.  Roll back.
		 */
		sth->exc_state.eip = fluke_mutex_unlock__entry;
	}
}
