/*
 * Copyright (c) 1996-1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include <oskit/x86/types.h>
#include <oskit/x86/base_gdt.h>
#include <oskit/x86/base_paging.h>

#include "xfer.h"
#include "debug.h"
#include "pmap.h"
#include "s_thread.h"
#include "config_kmodule.h"

int copyin_local(
	oskit_addr_t		src_uva,
	void			*dest_kva,
	oskit_size_t		nwords)
{
	int rc;
	assert(nwords > 0);

	asm volatile("
		movw	%w5,%%ds
	1:	rep
		movsl
	2:	movw	%w6,%%ds
		.section .anno,\"aw\",@progbits
		.long	1b,opt_xfer_recover,2b,anno_trap
		.text
	" : "=a" (rc)
	  : "a" (0),
	    "c" (nwords),
	    "S" (src_uva),
	    "D" (dest_kva),
	    "r" (USER_DS),
	    "r" (KERNEL_DS)
	  : "ecx", "esi", "edi");

	return rc;
}

int copyout_local(
	void			*src_kva,
	oskit_addr_t		dest_uva,
	oskit_size_t		nwords)
{
	int rc;
	assert(nwords > 0);

	asm volatile("
		movw	%w5,%%es
	1:	rep
		movsl
	2:	movw	%w6,%%es
		.section .anno,\"aw\",@progbits
		.long	1b,opt_xfer_recover,2b,anno_trap
		.text
	" : "=a" (rc)
	  : "a" (0),
	    "c" (nwords),
	    "S" (src_kva),
	    "D" (dest_uva),
	    "r" (USER_DS),
	    "r" (KERNEL_DS)
	  : "ecx", "esi", "edi");

	return rc;
}

/*
 * Set up the data transfer window in our page map.
 * Returns the kva through which we should access the remote thread's space.
 */
oskit_addr_t
opt_xfer_setup_window(s_thread_t *rmt_thread, oskit_addr_t rmt_uva)
{
	s_thread_t *cur_thread;
	oskit_addr_t cur_pdir_pa, rmt_pdir_pa;
	oskit_addr_t *cur_pde, *rmt_pde;

	rmt_pdir_pa = rmt_thread->csw_state.pdir_pa;

#ifdef CONFIG_KMODULE
	if (rmt_pdir_pa == base_pdir_pa) {
		/*
		 * The remote thread is a kernel-mode thread,
		 * so its addresses are really in the high half of the AS.
		 */
		assert(rmt_uva < LIN_KERNEL_MIN);
		rmt_uva += LIN_KERNEL_MIN;
	}
#endif
	rmt_pde = pdir_find_pde(rmt_pdir_pa, rmt_uva);

	cur_thread = current_thread();
	cur_pdir_pa = cur_thread->csw_state.pdir_pa;
	cur_pde = pdir_find_pde(cur_pdir_pa, LIN_XFER_MIN);

	/* If the window is dirty, flush the TLB to clean it */
	if ((*cur_pde != 0) && (*cur_pde != *rmt_pde)) {
		set_pdbr(cur_pdir_pa);
	}

	/*
	 * Copy the first two PDEs in this position (8MB worth)
	 * from the destination thread's page directory to our own.
	 * If someone actually tries to do a transfer bigger than this,
	 * then a fault will occur on the destination address,
	 * causing the copy operation to be interrupted;
	 * but then the page fault handler will see nothing wrong
	 * and will therefore just return and restart the copy
	 * this time using a new 8MB window.
	 */
	cur_pde[0] = rmt_pde[0];
	cur_pde[1] = rmt_pde[1];

	return lintokv(LIN_XFER_MIN) + (rmt_uva & SUPERPAGE_MASK);
}
