/*
 * Copyright (c) 1996 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include <oskit/x86/asm.h>
#include <oskit/x86/trap.h>
#include <oskit/x86/anno.h>
#include <oskit/x86/base_gdt.h>
#include <fluke/exception.h>

#include "x86.h"
#include "s_thread_asm.h"

/*
 * Common trap recovery handler for optimized data transfer routines.
 * On entry, eax points to our annotation table entry,
 * and esp points to the standard kernel trap_state frame.
 * val3 in the annotation table entry contains the recovery address
 * (the place to jump after setting up the exception state and return code).
 */
ENTRY(opt_xfer_recover)

	/* Find the current thread (ebx). */
	CURRENT_THREAD(%ebx)

	/* Check the trap type, if not page fault just return USER_EXCEPTION */
	movl	TS_TRAPNO(%esp),%edx
	cmpl	$14,%edx
	je	1f
	orl	$(FLUKE_EXC_PAGE_FAULT<<8),%edx
	movl	%edx,TH_EXC_STATE+EXC_CODE(%ebx)
	movl	TS_ERR(%esp),%edx
	movl	%edx,TH_EXC_STATE+EXC_SUBCODE(%ebx)
	movl	$KR_USER_EXCEPTION,TS_EAX(%esp)
	jmp	3f
1:
	/* Set the exception code and subcode in the user exception state. */
	movl $(FLUKE_PROT_WRITE<<16)|(FLUKE_EXC_PAGE_FAULT<<8)|T_PAGE_FAULT,%edx
	testb	$2,TS_ERR(%esp)
	jnz	2f
	movl $(FLUKE_PROT_READ<<16)|(FLUKE_EXC_PAGE_FAULT<<8)|T_PAGE_FAULT,%edx
2:	movl	%edx,TH_EXC_STATE+EXC_CODE(%ebx)
	movl	%cr2,%edx
	movl	%edx,TH_EXC_STATE+EXC_SUBCODE(%ebx)

	/* Return to the recovery point with an error code. */
	movl	$KR_PAGE_FAULT,TS_EAX(%esp)
3:
	movl	8(%eax),%ebx
	movl	%ebx,TS_EIP(%esp)
	jmp	return_from_k_trap


/*
 * Fast cross-address-space copy routine.
 * XXX should do appropriate processor-specific speedup tricks.
 *
 * Stack frame:
 *	esp ->	saved ebx
 *	+4	saved esi
 *	+8	saved edi
 *	+12	return address
 *	+16	dest_thread
 *	+20	src_uva
 *	+24	dest_uva
 *	+28	&inout_nwords
 *	+32	&out_remote_rc
 */
ENTRY(copyto_remote)
	pushl	%edi
	pushl	%esi
	pushl	%ebx	/* XXX unnecessary */

	/* Setup the transfer window */
	pushl	24(%esp)	/* dest_uva */
	pushl	4+16(%esp)	/* dest_thread */
	call	opt_xfer_setup_window
	addl	$8,%esp

	/* Set up for the transfer */
	movl	20(%esp),%esi	/* src_uva */
	movl	%eax,%edi	/* dest_kva */
	movl	28(%esp),%edx	/* &inout_nwords */
	movl	(%edx),%ecx

	/* Return success unless something bad happens */
	movl	32(%esp),%edx	/* &out_remote_rc */
	xorl	%eax,%eax
	movl	%eax,(%edx)

	movl	$USER_DS,%edx
	movw	%dx,%ds

	/* Copy the data */
	ANNO_TRAP(copyto_fault, 0)
	rep
	movsl

copyto_recover:
	movl	$KERNEL_DS,%edx
	movw	%dx,%ds

	popl	%ebx
	popl	%esi
	popl	%edi
	ret

	/* Fault handler for copyto_remote */
copyto_fault:

	/* Return to the designated recovery point */
	movl	$copyto_recover,TS_EIP(%esp)

	/* Find the faulting address (edx) */
	movl	%cr2,%edx

	/* Adjust the number of words successfully transferred */
	movl	TR_KSIZE+28(%esp),%eax	/* &inout_nwords */
	movl	TS_ECX(%esp),%ecx	/* number of words not copied */
	subl	%ecx,(%eax)

	/* See what type of a fault it was */
	/* XXX what if it was a SEGV rather than a page fault? */
	testb	$2,TS_ERR(%esp)
	jnz	1f

	/* Read fault - blame it on the source (current) thread */
	CURRENT_THREAD(%ebx)
	movl	$(FLUKE_PROT_READ<<16)|(FLUKE_EXC_PAGE_FAULT<<8)|T_PAGE_FAULT,TH_EXC_STATE+EXC_CODE(%ebx)
	movl	%edx,TH_EXC_STATE+EXC_SUBCODE(%ebx)
	movl	$KR_PAGE_FAULT,TS_EAX(%esp)
	jmp	return_from_k_trap

	/* Write fault - blame it on the destination thread */
1:
	movl	TR_KSIZE+16(%esp),%ebx
	movl	$(FLUKE_PROT_WRITE<<16)|(FLUKE_EXC_PAGE_FAULT<<8)|T_PAGE_FAULT,TH_EXC_STATE+EXC_CODE(%ebx)

	/*
	 * Fix up the fault address with the high bits of dest_uva.
	 * Note that the fault can occur anywhere in the 8MB window.
	 */
	andl	$0x00ffffff,%edx
	movl	TR_KSIZE+24(%esp),%eax	/* dest_uva */
	andl	$0xffc00000,%eax
	addl	%eax,%edx
	movl	%edx,TH_EXC_STATE+EXC_SUBCODE(%ebx)

	movl	TR_KSIZE+32(%esp),%ebx	/* &out_remote_rc */
	movl	$KR_PAGE_FAULT,(%ebx)

	jmp	return_from_k_trap

