/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include <malloc.h>
#include <oskit/lmm.h>
#include <oskit/x86/base_cpu.h>
#include <oskit/x86/base_paging.h>
#include <oskit/x86/pc/phys_lmm.h>

#include "config_kmodule.h"
#include "config_smp.h"
#include "config_opt_threads.h"
#include "config_opt_synch.h"
#include "config_opt_ref.h"
#include "config_norwox.h"

#include "gdt.h"
#include "pmap.h"
#include "s_task.h"
#include "s_thread.h"
#include "exception.h"
#include "csw.h"
#include "debug.h"
#include "per_cpu.h"
#include "kalloc.h"


/*
 * Special reserved page directory, exactly like the base_pdir,
 * except that there are no mappings in the user linear address range,
 * only in the kernel linear address range.
 * All threads that don't have a task use this page directory,
 * which makes them fault immediately if they try to execute user code.
 */
oskit_addr_t null_pdir_pa;

/*
 * This is a page table used to map the Fluke API entrypoint code
 * into user address spaces.
 * This page table gets inserted in every task
 * in the second page directory entry, i.e. at 4MB.
 */
static oskit_addr_t entry_ptab_pa;

/*
 * End of currently used kernel address space.
 * Just bump this upwards to allocate kernel address space.
 * All kernel mappings must be entered within pmap_boot(),
 * after base_paging_init() but before it creates the null_pdir
 * by copying the top half of the base_pdir.
 */
static oskit_addr_t kva_max;

oskit_addr_t pmap_map_phys_mem(oskit_addr_t start_pa, oskit_size_t size,
			       oskit_bool_t writable)
{
	oskit_addr_t kva;

	assert(kva_max);
	assert(base_pdir_pa);

	kva = kva_max;
	pdir_map_range(base_pdir_pa, kvtolin(phystokv(kva)),
		       start_pa, size,
		       INTEL_PDE_VALID | INTEL_PDE_USER
		       | (writable ? INTEL_PDE_WRITE : 0));

	kva_max += round_page(size);
	assert(kva_max <= FLUKE_VM_MAX);
	return kva;
}

static pt_entry_t pmap_global_ext;

void pmap_boot()
{
	extern const char u_entry_page[];
#ifdef CONFIG_OPT_THREADS
	extern char u_thread_page[];
#endif
#ifdef CONFIG_OPT_SYNCH
	extern char u_synch_page[];
#endif
#ifdef CONFIG_OPT_REF
	extern char u_ref_page[];
#endif
#ifdef CONFIG_KMODULE
	extern const char kmodule_entry_page[];
	oskit_addr_t km_entry_ptab_pa;
#endif

	pd_entry_t *pde;
	pt_entry_t *pte;
	unsigned i;
	int mustcopy = 0;
	int rc;

	/*
	 * Limit the kernel code segment to our text segment,
	 * just in case we somehow tried to execute bogus code.
	 */
	extern void _start();
	extern char etext[];
	fill_descriptor(&base_gdt[KERNEL_CS >> 3],
			kvtolin(_start),
#ifdef CONFIG_KMODULE
			kvtolin(FLUKE_ENTRY_MAX),
#else
			kvtolin(etext),
#endif
			ACC_PL_K | ACC_CODE_R, SZ_32);


	/*
	 * The Pentium Pro and later support global PDE/PTE entries,
	 * which are not removed from the TLB when we switch tasks.
	 */
	pmap_global_ext = (base_cpuid.feature_flags & CPUF_PAGE_GLOBAL_EXT)
		? INTEL_PTE_GLOBAL : 0;

	/*
	 * Determine the address at which we can start mapping
	 * APICs and other devices located at high physical addresses.
	 */
	kva_max = round_page(phys_mem_max);
	if (kva_max < round_superpage(FLUKE_ENTRY_MAX))
		kva_max = round_superpage(FLUKE_ENTRY_MAX);

	/*
	 * Set up a basic page directory that direct-maps all physical memory.
	 */
	base_paging_init();

#ifdef CONFIG_SMP
	smp_init_paging();
#endif

	base_paging_load();
	/* XXX smp enable paging */

	/*
	 * Copy the direct physical memory mappings
	 * from user linear space to kernel linear space.
	 * Make sure all the high mappings are inaccessible to user mode.
	 * (Probably shouldn't actually matter, though,
	 * since segmentation is used to protect the high area anyway.)
	 */
	pde = pdir_find_pde(base_pdir_pa, 0);
	for (i = 0; i < NPDES/2; i++)
		if (pde[i] != 0)
			pde[i + NPDES/2]
				= (pde[i] & ~INTEL_PDE_USER) | pmap_global_ext;

	/*
	 * Relocate the kernel segments to refer to the high linear region.
	 * To do this, just change where we "think" linear memory appears
	 * and then reinitialize the GDT.
	 * Since the init and load routines use kvtolin()
	 * to determine what values to plug into the segment descriptors,
	 * they get adjusted for the new linear addresses automatically.
	 */
	linear_base_va = -LIN_KERNEL_MIN;
	base_gdt_init();

	/*
	 * Re-load the GDT, IDT, and base TSS into the processor.
	 * Although the contents of the IDT and TSS themselves
	 * don't contain any linear addresses affected by this change,
	 * the address of the IDT itself is represented as a linear address,
	 * and the processor caches the linear address of a loaded TSS,
	 * and so these need to be reloaded as well.
	 */
	base_cpu_load();

#ifdef CONFIG_OPT_THREADS
	/*
	 * Initialize the pdir/ptab for the LIN_GSPAGE area.
	 * This sets the pdirs for the area in base_pdir_pa;
	 * we copy them from there into other tasks.
	 */
	gspage_init_ptab();

	gspage_setup_cpu(&boot_cpu_info); /* set up boot CPU's page & seg */
	set_gs(USER_GS);	/* switch to the segment */
#endif

#if defined(CONFIG_SMP) || defined(IOPL0)
	/*
	 * Now load everything back into the CPU for the last time.
	 */
	per_cpu_load(&boot_cpu_info);
#endif

	/*
	 * Now we're fully out of the user linear region
	 * and we can map real user address spaces there.
	 * We still keep both mappings around in the base_pdir
	 * for making BIOS calls and booting other processors and such.
	 */

	/*
	 * Create the kernel entrypoint page table.
	 * It simply contains a whole slew of identical read-only mappings
	 * pointing to our u_entry_page.
	 */
	rc = ptab_alloc(&entry_ptab_pa);
	assert(rc == 0);
	pte = ptab_find_pte(entry_ptab_pa, 0);
	for (i = 0; i < (LIN_ENTRY_MAX - LIN_ENTRY_MIN)/PAGE_SIZE; i++) {
		pte[i] = (kvtophys(u_entry_page) | INTEL_PTE_VALID
			  | INTEL_PTE_USER | pmap_global_ext);
	}

	/*
	 * Ensure that the contents of the kernel entrypoint page
	 * also appears at FLUKE_ENTRY_MIN in the root task's address space
	 * so that the root task code which runs in the kernel's address space
	 * can make Fluke system calls successfully.
	 *
	 * On a machine with 64MB or less physical memory,
	 * we can do this with no memory loss
	 * simply by mapping the kernel entrypoint page table at 64MB.
	 * On machines with more memory,
	 * we reserve the physical memory pages at those addresses
	 * and copy the contents of the u_entry_page into them;
	 * this causes us to waste some physical memory, but oh well.
	 *
	 * Note that it's not enough just to change the _fluke_* symbols
	 * to point directly into u_entry_page.S
	 * instead of using their normal absolute addresses defined in libfluke,
	 * because then although system calls would initially work,
	 * any system call that gets interrupted or restarted with a new EIP
	 * would screw up because the standard absolute address will be used
	 * when setting the new system call EIP.
	 * Thus, we really do need the entrypoint area to appear
	 * at the proper location.
	 */

#ifndef CONFIG_KMODULE
#define kentry_ptab_pa		entry_ptab_pa
#define kentry_page		u_entry_page
#define KENTRY_PDE_USER		INTEL_PDE_USER
#define LIN_KENTRY_MIN		LIN_ENTRY_MIN
#define LIN_KENTRY_MAX		LIN_ENTRY_MAX
#ifndef CONFIG_NORWOX
	/*
	 * Map the kernel text at LIN_USER_MIN+LIN_USERX_OFF so that
	 * kserver threads execute properly.  Note that this limits physical
	 * memory to LIN_USER_MIN+LIN_USERX_OFF (1GB) of memory.
	 *
	 * XXX since the kernel (i.e., low physical memory) may be mapped
	 * with superpages, we simply copy all PDEs covering any part of
	 * the kernel text.  Most likely this means that we will end up
	 * mapping the entire kernel (and then some) with CS, but so it goes.
	 * We just want kserver threads to work, we are not trying to detect
	 * running in non-exec code like real user threads.  If you want to
	 * do that, you will have to allocate actual PT pages.
	 */
	if (phys_mem_max > LIN_USER_MIN+LIN_USERX_OFF)
		panic("Too much memory (0x%x bytes max)"
		      " send excess to fluke@flux.cs.utah.edu",
		      LIN_USER_MIN+LIN_USERX_OFF);
	else {
		oskit_addr_t ktstart = trunc_superpage(_start);
		oskit_size_t ktend = ktstart + round_superpage(etext);

		while (ktstart < ktend) {
			pde = pdir_find_pde(base_pdir_pa,
					    ktstart + LIN_USERX_OFF);
			assert(*pde == 0);
			*pde = *pdir_find_pde(base_pdir_pa, ktstart);
			ktstart += SUPERPAGE_SIZE;
		}
	}
	/*
	 * Map the kentry page as well
	 */
	pde = pdir_find_pde(base_pdir_pa, LIN_KENTRY_MIN + LIN_USERX_OFF);
	assert(*pde == 0);
	*pde = kentry_ptab_pa | INTEL_PDE_VALID | KENTRY_PDE_USER;
#endif
#else
	/*
	 * The "user" threads running in the kernel address space
	 * will actually run in kernel mode, so when they make
	 * Fluke API calls they will actually jump to LIN_KERNEL_MIN+entry.
	 * We map there a different entry-point page that does the
	 * right thing for kernel-mode threads (see kmodule/kmodule_kentry.S).
	 */
#define kentry_ptab_pa		km_entry_ptab_pa
#define kentry_page		kmodule_entry_page
#define KENTRY_PDE_USER		pmap_global_ext
#define LIN_KENTRY_MIN		(LIN_KERNEL_MIN + FLUKE_ENTRY_MIN)
#define LIN_KENTRY_MAX		(LIN_KERNEL_MIN + FLUKE_ENTRY_MAX)
	rc = ptab_alloc(&km_entry_ptab_pa);
	assert(rc == 0);
	pte = ptab_find_pte(km_entry_ptab_pa, 0);
	for (i = 0; i < (LIN_ENTRY_MAX - LIN_ENTRY_MIN)/PAGE_SIZE; i++) {
		pte[i] = kvtophys(kmodule_entry_page)
			| INTEL_PTE_VALID | pmap_global_ext;
	}
#endif


	pde = pdir_find_pde(base_pdir_pa, LIN_KENTRY_MIN);
	if (*pde == 0) {
		*pde = kentry_ptab_pa | INTEL_PDE_VALID | KENTRY_PDE_USER;
	} else {
		/*
		 * Allocate the required physical memory range.
		 */
		void *mem;
		/* XXX: do not have the kmalloc lock */
		mem = lmm_alloc_gen(&malloc_lmm,
				    LIN_KENTRY_MAX - LIN_KENTRY_MIN,
				    0, 0, 0,
				    LIN_KENTRY_MIN, LIN_KENTRY_MAX);
		if (mem == 0)
			panic("pmap_boot: phys mem from %#x-%#x already used",
			      LIN_KENTRY_MIN, LIN_KENTRY_MAX);
		assert(mem == (void*)LIN_KENTRY_MIN);

		/*
		 * Replicate the u_entry_page into this range.
		 */
		while (mem < (void*)LIN_KENTRY_MAX) {
			memcpy(mem, kentry_page, PAGE_SIZE);
			mem += PAGE_SIZE;
		}
#ifndef CONFIG_KMODULE
		mustcopy = 1;
#endif
	}
#undef kentry_page
#undef kentry_ptab_pa

	/*
	 * Create the null page directory used by taskless threads,
	 * by copying the top half of the base_pdir
	 * but leaving the bottom half empty of mappings.
	 * XX We could map the kernel entry area, if that's desirable...
	 */
	rc = ptab_alloc(&null_pdir_pa);
	assert(rc == 0);
	memcpy(pdir_find_pde(null_pdir_pa, LIN_KERNEL_MIN),
	       pdir_find_pde(base_pdir_pa, LIN_KERNEL_MIN),
	       sizeof(pd_entry_t) * NPDES/2);

	/* Set up the standard user-mode code and data segment descriptors.  */
#ifndef CONFIG_NORWOX
	fill_descriptor(&base_gdt[USER_CS >> 3],
			LIN_USER_MIN + LIN_USERX_OFF,
			LIN_USER_MAX - LIN_USER_MIN,
			ACC_PL_U | ACC_CODE, SZ_32);
#else
	fill_descriptor(&base_gdt[USER_CS >> 3],
			LIN_USER_MIN, LIN_USER_MAX - LIN_USER_MIN,
			ACC_PL_U | ACC_CODE_R, SZ_32);
#endif
	fill_descriptor(&base_gdt[USER_DS >> 3],
			LIN_USER_MIN, LIN_USER_MAX - LIN_USER_MIN,
			ACC_PL_U | ACC_DATA_W, SZ_32);

#ifdef CONFIG_OPT_THREADS
	/*
	 * Map the u_thread_page instead of the u_entry_page
	 * as the first page of the kernel entrypoint area.
	 */
	*ptab_find_pte(entry_ptab_pa, 0) = kvtophys(u_thread_page) |
		INTEL_PTE_VALID | INTEL_PTE_USER | pmap_global_ext;

	/*
	 * If necessary, we also copy the contents of the page into the
	 * entrypoint range of the kernel's address space.
	 */
	if (mustcopy)
		memcpy((void *)LIN_KENTRY_MIN, u_thread_page, PAGE_SIZE);
#endif /* CONFIG_OPT_THREADS */

#ifdef CONFIG_OPT_SYNCH
	/*
	 * Map the u_synch_page instead of the u_entry_page
	 * as the second page of the kernel entrypoint area.
	 */
	*ptab_find_pte(entry_ptab_pa, 1*PAGE_SIZE) = kvtophys(u_synch_page) |
		INTEL_PTE_VALID | INTEL_PTE_USER | pmap_global_ext;

	/*
	 * If necessary, we also copy the contents of the page into the
	 * entrypoint range of the kernel's address space.
	 */
	if (mustcopy)
		memcpy((void *)(LIN_KENTRY_MIN + 1*PAGE_SIZE),
		       u_synch_page, PAGE_SIZE);
#endif

#ifdef CONFIG_OPT_REF
	/*
	 * Map the u_ref_page instead of the u_entry_page
	 * as the third page of the kernel entrypoint area.
	 */
	*ptab_find_pte(entry_ptab_pa, 2*PAGE_SIZE) = kvtophys(u_ref_page) |
		INTEL_PTE_VALID | INTEL_PTE_USER | pmap_global_ext;

	/*
	 * If necessary, we also copy the contents of the page into the
	 * entrypoint range of the kernel's address space.
	 */
	if (mustcopy)
		memcpy((void *)(LIN_KENTRY_MIN + 2*PAGE_SIZE),
		       u_ref_page, PAGE_SIZE);
#endif
}


#ifdef CONFIG_OPT_THREADS
#include <oskit/x86/base_cpu.h>
#include "kalloc.h"
#include "gspage.h"
#include <stddef.h>		/* for offsetof */

oskit_addr_t gspage_ptab_pa;

void
gspage_init_ptab(void)
{
	int rc;

	/* Allocate a page table for the per-processor page mappings */
	rc = ptab_alloc(&gspage_ptab_pa);
	assert_rc_0(rc);

	*pdir_find_pde(base_pdir_pa, LIN_GSPAGE_MIN) = gspage_ptab_pa
		| INTEL_PDE_VALID | INTEL_PDE_USER | pmap_global_ext;
}


/*
 * Allocate and map the page itself, and set up the segment.
 * This is called only on the boot CPU.
 */
void
gspage_setup_cpu(fluke_cpu_info_t *cpu)
{
	static oskit_addr_t lin;
	void *page;
	pt_entry_t *pte = ptab_find_pte(gspage_ptab_pa, lin);
	assert(*pte == 0);

	page = kalloc_aligned(PAGE_SIZE, PAGE_SHIFT);
	assert(page);

	pte[0] = kvtophys(page) | INTEL_PTE_VALID | pmap_global_ext |
		INTEL_PTE_USER | INTEL_PTE_WRITE;
	pte[1] = kvtophys(page) | INTEL_PTE_VALID | pmap_global_ext |
		INTEL_PTE_USER;

	fill_descriptor(&cpu->gdt[USER_GS >> 3],
			(LIN_GSPAGE_MIN + lin +
			 PAGE_SIZE - sizeof(struct gspage_u)),
			sizeof(struct gspage_seg),
			ACC_PL_U | ACC_DATA_W, SZ_32);

	/*
	 * The special page mappings for the next processor will
	 * be above this processor's pages.
	 */
	lin += 2 * PAGE_SIZE;
}
#endif


/* XXX */

int task_pmap_state_init(struct s_task *task)
{
	static int first_task = 0;

	mutex_init(&task->pmap_state.lock);

	/* XXX horrible kludge to detect when creating the kernel task */
	if (first_task == 0) {
		first_task = 1;

		task->pmap_state.phys_pdir = base_pdir_pa;

	} else {

		/* Allocate a page directory for this task */
		if (ptab_alloc(&task->pmap_state.phys_pdir))
			return (KR_NO_MEMORY);

		/* Map the kernel entrypoint area into the task.  */
#ifndef CONFIG_NORWOX
		*pdir_find_pde(task->pmap_state.phys_pdir,
			       LIN_ENTRY_MIN+LIN_USERX_OFF)
			= (entry_ptab_pa |
			   INTEL_PDE_VALID | INTEL_PDE_USER | pmap_global_ext);
#endif
		*pdir_find_pde(task->pmap_state.phys_pdir, LIN_ENTRY_MIN)
			= (entry_ptab_pa |
			   INTEL_PDE_VALID | INTEL_PDE_USER | pmap_global_ext);

		/*
		 * Copy the kernel part of the page directory.
		 *
		 * XXX really assumes the values of LIN_KERNEL_*
		 * (i.e., that the kernel occupies the upper half of the LAS).
		 */
		memcpy(pdir_find_pde(task->pmap_state.phys_pdir,
				     LIN_KERNEL_MIN),
		       pdir_find_pde(base_pdir_pa, LIN_KERNEL_MIN),
		       sizeof(pd_entry_t) * NPDES/2);
	}

	return 0;
}

void task_pmap_state_destroy(struct s_task *task)
{
	/*
	 * Clean the user part of the page table.  As with
	 * task_pmap_state_init, it is assuming the location of the kernel.
	 */
	pdir_clean_range(task->pmap_state.phys_pdir, 0, LIN_KERNEL_MIN);

#ifdef DEBUG
	/*
	 * Ensure that the AS is empty.  */
	{
		oskit_addr_t va;

		for (va = LIN_USER_MIN; va < LIN_USER_MAX; va += SUPERPAGE_SIZE)
			if (va != LIN_ENTRY_MIN &&
#ifndef CONFIG_NORWOX
			    va != LIN_ENTRY_MIN+LIN_USERX_OFF &&
#endif
			    *pdir_find_pde(task->pmap_state.phys_pdir, va))
				panic("pages in [%x-%x] still valid",
				      va, va + SUPERPAGE_SIZE);
	}
#endif

	ptab_free(task->pmap_state.phys_pdir);

	mutex_destroy(&task->pmap_state.lock);
}

void pmap_enter(struct s_task *task, oskit_addr_t va, oskit_addr_t pa,
		oskit_size_t size, fluke_prot_t prot)
{
	int rc;
	pt_entry_t ptflags;

	assert(task);
#ifndef CONFIG_NORWOX
	assert(va < LIN_USER_MAX);
#endif

	ptflags = INTEL_PTE_USER | INTEL_PTE_VALID;
	if (prot & FLUKE_PROT_WRITE)
		ptflags |= INTEL_PTE_WRITE;

	mutex_lock(&task->pmap_state.lock);

#ifdef CONFIG_KMODULE
	/*
	 * For kernel kserver threads, mappings must be entered in the
	 * high half of the address space.
	 */
	if (task->pmap_state.phys_pdir == base_pdir_pa) {
		assert(va < LIN_KERNEL_MIN);
		va += LIN_KERNEL_MIN;
	}
#endif
	rc = pdir_map_range(task->pmap_state.phys_pdir, va, pa,
			    size, ptflags);
#ifndef CONFIG_NORWOX
	if (rc == 0 && va < LIN_KERNEL_MIN && (prot & FLUKE_PROT_EXEC)) {
		va += LIN_USERX_OFF;
		rc = pdir_map_range(task->pmap_state.phys_pdir, va, pa,
				    size, ptflags);
	}
#endif

	mutex_unlock(&task->pmap_state.lock);

	assert(rc == 0);
}


/*
 * Invalidate TLB entries for the given range of pages.
 */
static void
pmap_inval_tlb(struct s_task *task, oskit_addr_t va, oskit_size_t size)
{
#ifdef CONFIG_SMP
	/*
	 * Tell other CPUs to flush.
	 */
	fluke_cpu_info_t *cpu;
	int left;
	left = 0;
	for (cpu = my_cpu.next; cpu != &my_cpu; cpu = cpu->next) {
		cpu->inval_tlb = 1;
		smp_message_pass(cpu->number);
		left |= cpu->inval_tlb; /* see if he got it yet */
	}
#endif
#define flush_tlb() set_pdbr(get_pdbr())

	/*
	 * Now invalidate our own TLB entries.
	 */
	if (current_thread()->task_state.task == task) {
		/*
		 * We are running in the affected task, so our
		 * virtual addresses hit the TLB entries now invalid.
		 * We can use the limited TLB flush instruction
		 * on just these pages.  Note that we have to subtract
		 * the kernel segment base from the linear address we
		 * want because `invlpg' will add the segment base even
		 * it won't check the segmentation boundaries.
		 */
		assert(size > 0);
		do {
			asm volatile("invlpg %0" : :
				     "m" (*(int *) (va - LIN_KERNEL_MIN)));
			va += PAGE_SIZE;
			size -= PAGE_SIZE;
		} while (size > 0);
	}
	else
		flush_tlb();

#ifdef CONFIG_SMP
	/*
	 * Spin until we're sure all other CPUs have begun their flushes.
	 */
	while (left) {
		left = 0;
		for (cpu = my_cpu.next; cpu != &my_cpu; cpu = cpu->next) {
			left |= cpu->inval_tlb;
		}
	}
#endif
}


void pmap_remove(struct s_task *task, oskit_addr_t va, oskit_size_t size)
{
#ifndef CONFIG_NORWOX
	assert(va < LIN_USER_MAX);
#endif

	mutex_lock(&task->pmap_state.lock);

#ifdef CONFIG_KMODULE
	/*
	 * For kernel kserver threads, mappings must be entered in the
	 * high half of the address space.
	 */
	if (task->pmap_state.phys_pdir == base_pdir_pa) {
		assert(va < LIN_KERNEL_MIN);
		va += LIN_KERNEL_MIN;
	}
#endif
	pdir_unmap_range(task->pmap_state.phys_pdir, va, size);
	pmap_inval_tlb(task, va, size);
#ifndef CONFIG_NORWOX
	if (va < LIN_KERNEL_MIN) {
		oskit_size_t nsize = size;

		va += LIN_USERX_OFF;
#if 1
		/*
		 * XXX to avoid PDE/PTE valid checks in pdir_unmap_range
		 * we have to iterate through every page.
		 */
		while (nsize) {
			pdir_unmap_page(task->pmap_state.phys_pdir, va);
			nsize -= PAGE_SIZE;
			va += PAGE_SIZE;
		}
#else
		pdir_unmap_range(task->pmap_state.phys_pdir, va, size);
#endif
		pmap_inval_tlb(task, va, size);
	}
#endif

	mutex_unlock(&task->pmap_state.lock);
}

void pmap_protect(struct s_task *task, oskit_addr_t va, oskit_size_t size,
		  fluke_prot_t old_prot, fluke_prot_t new_prot)
{
	pt_entry_t ptflags;

#ifndef CONFIG_NORWOX
	assert(va < LIN_USER_MAX);
#endif

	mutex_lock(&task->pmap_state.lock);

#ifdef CONFIG_KMODULE
	/*
	 * For kernel kserver threads, mappings must be entered in the
	 * high half of the address space.
	 */
	if (task->pmap_state.phys_pdir == base_pdir_pa) {
		assert(va < LIN_KERNEL_MIN);
		va += LIN_KERNEL_MIN;
	}
#endif
	if (new_prot == FLUKE_PROT_NONE) {
		pdir_unmap_range(task->pmap_state.phys_pdir, va, size);
		pmap_inval_tlb(task, va, size);
#ifndef CONFIG_NORWOX
		if (va < LIN_KERNEL_MIN && (old_prot & FLUKE_PROT_EXEC)) {
			va += LIN_USERX_OFF;
			pdir_unmap_range(task->pmap_state.phys_pdir, va, size);
			pmap_inval_tlb(task, va, size);
		}
#endif
		mutex_unlock(&task->pmap_state.lock);
		return;
	}

	ptflags = INTEL_PTE_USER | INTEL_PTE_VALID;
	if (new_prot & FLUKE_PROT_WRITE)
		ptflags |= INTEL_PTE_WRITE;

	pdir_prot_range(task->pmap_state.phys_pdir, va, size, ptflags);

	if (old_prot &~ new_prot)
		/*
		 * We are revoking access, so invalidate the TLB.
		 * There is no need when granting new access,
		 * since any faults will resolve anyhow.
		 */
		pmap_inval_tlb(task, va, size);

#ifndef CONFIG_NORWOX
	if (va < LIN_KERNEL_MIN && ((old_prot|new_prot) & FLUKE_PROT_EXEC)) {
		va += LIN_USERX_OFF;

		/* Adding execute permission, enter CS mapping */
		if ((old_prot & FLUKE_PROT_EXEC) == 0) {
			oskit_addr_t pa;
			int rc;

			pa = pte_to_pa(
				*pdir_find_pte(task->pmap_state.phys_pdir,
					       va - LIN_USERX_OFF));
			assert(pa);
			rc = pdir_map_range(task->pmap_state.phys_pdir,
					    va, pa, size, ptflags);
			assert(rc == 0);
		}
		/* removing execute permission, remove CS mapping */
		else if ((new_prot & FLUKE_PROT_EXEC) == 0) {
			pdir_unmap_range(task->pmap_state.phys_pdir, va, size);
			pmap_inval_tlb(task, va, size);
		}
		/* changing other permissions, change CS mapping */
		else {
			pdir_prot_range(task->pmap_state.phys_pdir,
					va, size, ptflags);
			if (old_prot &~ new_prot)
				pmap_inval_tlb(task, va, size);
		}
	}
#endif

	mutex_unlock(&task->pmap_state.lock);
}

/*
 * Should ensure that pmap software structure is in sync with the hardware
 * (e.g., perform any delayed TLB).
 */
void pmap_update(struct s_task *task)
{
}

void pmap_activate(struct s_thread *cur_thread, struct s_task *task)
{
	csw_set_pdir(cur_thread, task->pmap_state.phys_pdir);
}

void pmap_deactivate(struct s_thread *cur_thread, struct s_task *task)
{
	csw_set_pdir(cur_thread, null_pdir_pa);
}

/*
 * Override OSKit page table allocation routines to ensure correct locking.
 */

#ifdef MEMSTATS
unsigned int mstats_ptpages, mstats_maxptpages;

void pmap_dump_mstats(void)
{
	printf("  Page tables:       %10u bytes (%10u bytes max)\n",
	       mstats_ptpages * PAGE_SIZE, mstats_maxptpages * PAGE_SIZE);
}
#endif

int ptab_alloc(oskit_addr_t *out_ptab_pa)
{
	void *ptab_va;

	ptab_va = kalloc_aligned(PAGE_SIZE, PAGE_SHIFT);
	if (ptab_va == 0)
		return -1;

#ifdef MEMSTATS
	if (++mstats_ptpages > mstats_maxptpages)
		mstats_maxptpages = mstats_ptpages;
#endif

	/* Clear it out to make sure all entries are invalid.  */
	memset(ptab_va, 0, PAGE_SIZE);

	*out_ptab_pa = kvtophys(ptab_va);
	return 0;
}

void ptab_free(oskit_addr_t ptab_pa)
{
	assert(ptab_pa);
	assert(ptab_pa != entry_ptab_pa);

#ifdef MEMSTATS
	mstats_ptpages--;
	assert((int)mstats_ptpages > 0);
#endif

	kfree((void *)phystokv(ptab_pa), PAGE_SIZE);
}
