/*
 * Copyright (c) 1996-1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include <pthread.h>
#include "pthread_impl.h"
#include <string.h>
#include <strings.h>	/* bzero */
#include "debug.h"

#include "s_thread.h"
#include "debug.h"
#include "prog.h"		/* XXX for do_in_supervisor_mode */
#include "kalloc.h"

/* pthread_init() needs these 3 */
#include <fluke/machine/syscall.h>
#include "sched.h"
#include "config_hier_sched.h"
#ifdef CONFIG_HIER_SCHED
#include <fluke/thread.h>
#endif

#ifdef BASE_DEBUG
#include "debug.h"
#define BD_BEGIN() DEBUG_BASE_BEGIN()
#define BD_END() DEBUG_BASE_END()
#define BD() DEBUG_BASE()
#define BD_ARGS(fmt, args...) DEBUG_BASE_ARGS(fmt,args)
#else
#define BD_BEGIN()
#define BD_END()
#define BD()
#define BD_ARGS(args...)
#endif


/*
 * This file contains a (very basic) pthread_create() that creates
 * a user thread in the kernel's address space from user mode.
 * Much of the pthread functionality is now in pthread_shared.c
 * which is also used by the kmodule code.
 */


struct pthread_create_args
{
	pthread_t *out_thread;
	const pthread_attr_t *attr;
	void *(*start_routine)(void *);
	void *start_arg;
	int *error;
};


/*
 * Called from supervisor mode, this creates a user-mode thread,
 * although in the kernel's address space.
 */
int
supervisor_pthread_create(pthread_t *out_thread,
			  const pthread_attr_t *attr,
			  void *(*start_routine)(void *), void *start_arg)
{
	struct pthread *uth;
	int rc;
        s_thread_t *thread;
        fluke_thread_state_t tstate;
	char *thread_stack;
	BD();

	if (attr == NULL)
		attr = &kthread_attr_default;

	uth = kalloc_aligned(sizeof *uth, FLUKE_THREAD_SHIFT);
	assert(uth);
	bzero(uth, sizeof(*uth));

	*out_thread = (pthread_t) uth;

	thread_stack = kalloc_aligned(attr->stacksize_attr, 4);
	assert(thread_stack);
	uth->stack = thread_stack;
	uth->stacksize = attr->stacksize_attr;

        rc = s_thread_create((oskit_addr_t)uth, 0, &thread);
        assert(rc == 0);
	assert_s_thread_locked(thread);

	/* XXX violation! */
	thread->sched_state.pri = attr->prio;

        memset(&tstate, 0, sizeof(tstate));
        tstate.thread_va = (oskit_addr_t)uth;
	tstate.ebp = 0;
        tstate.eip = (unsigned)start_routine;
        tstate.esp = (unsigned)thread_stack + attr->stacksize_attr;

	/* Push on the arg */
	tstate.esp -= sizeof(oskit_reg_t);
	*(oskit_reg_t *)tstate.esp = (oskit_reg_t)start_arg;

	/* Allocate space for the fake return IP */
	/* Some code just returns from start_routine */
	tstate.esp -= sizeof(oskit_reg_t);
	*(oskit_reg_t *)tstate.esp = (oskit_addr_t)pthread_exit;

	/* Initialize thread-specific data */
	pthread_uth_init(uth);

#ifdef CONFIG_HIER_SCHED
	/* Horrible hack until I think in a better way to do it. To
	   distinguish between this call and a sys_call. In this case we
	   have to set the parent in a different way. */
	thread->sched_state.kserver = 1;
#endif
	tstate.tickets = tstate.own_tickets = FLUKE_THREAD_DEFAULT_TICKETS;

        link_transfer(&kernel_task_link, &thread->task_state.task_link);

        s_thread_set_state(thread, &tstate);

	thread_ready(thread);

	s_thread_unlock(thread);

        /* I now have a USER-mode thread, created in kernel-mode */
	return(0);
}



/*
 * This is called by pthread_create using a do_in_supervisor_mode call
 */
int do_pthread_create(void *args)
{
        struct pthread_create_args *a = args;
	int *error = a->error;

	*error = supervisor_pthread_create(a->out_thread, a->attr,
					   a->start_routine, a->start_arg);
	return 0;
}


/*
 * This can be called by user-level threads in the kernel's address
 * space to easily create threads.
 */
int
pthread_create(pthread_t *out_thread, const pthread_attr_t *attr,
		   void *(*start_routine)(void *), void *start_arg)
{
        int error = 0;
	struct pthread_create_args args;

	assert(pthread_can_exit);

	args.out_thread = out_thread;
	args.attr = attr;
	args.start_routine = start_routine;
	args.start_arg = start_arg;
	args.error = &error;

	do_in_supervisor_mode(do_pthread_create, &args);

	return error;
}
