/*
 * Copyright (c) 1996-1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * This file contains pthread code for manipulating a `user' thread in the
 * kernel's address space.  This is used in the kserver in user mode and
 * as a kernel module (KMODULE).
 */

#include <pthread.h>
#include "pthread_impl.h"
#include <string.h>
#include <strings.h>	/* bzero */
#include "debug.h"

#include "debug.h"
#include "kalloc.h"
#include "round_robin.h"	/* SCHED_PRIUSER */
#include "s_thread.h"
#include "s_mutex.h"
#include "s_cond.h"
#include <oskit/x86/base_stack.h>

/* Key for accessing the per-thread errno value */
pthread_key_t __errno_key;

/* per-thread data destructor function pointer table */
pthread_destructor_t per_thread_destr[PTHREAD_KEYS_MAX];

/*
 * Default attributes for a kernel pthread.
 */
pthread_attr_t kthread_attr_default = {
	SCHED_RR,
	SCHED_PRIUSER,		/* XXX kernel specific */
	PTHREAD_CREATE_JOINABLE,
	NULL,
	NULL,
	NULL,
	KTHREAD_STACK_SIZE	/* XXX kernel specific */
};

int
pthread_attr_init(pthread_attr_t *attr)
{
	*attr = kthread_attr_default;
	return 0;
}

int     
pthread_attr_setstacksize(pthread_attr_t *attr, size_t size)
{
	if (size < PTHREAD_STACK_MIN)
		return EINVAL;

	attr->stacksize_attr = size;
	return 0;
}


fluke_mutex_t pthread_mutex;
fluke_cond_t pthread_cond;
struct pthread *killme;
int pthread_can_exit = 0;


/*
 * Called in kernel mode to prepare for kserver threads,
 * which are kernel threads in CONFIG_KMODULE and user threads otherwise.
 */
void
pthread_init(void)
{
	extern link_t kernel_task_link;
	int rc;
	s_ob_t *sob;

	/*
	 * This macro used in supervisor mode approximates
	 * `type_create(flobptr);' in user mode.
	 */
#define S_CREATE(type, flobptr)						      \
	rc = s_##type##_create((oskit_addr_t) (flobptr), 0,		      \
			       (s_##type##_t **) &sob);			      \
	assert_rc_0(rc);						      \
	s_ob_unlock(sob);

	assert(current_thread() == &startup_thread);

	/*
	 * We must set our task_state up for the flob lookups in
	 * the following fluke calls to work.
	 */
	thread_task_state_init(&startup_thread);
	link_transfer(&kernel_task_link, &startup_thread.task_state.task_link);
	thread_task_state_start(&startup_thread);
	assert(startup_thread.task_state.task != 0);

	S_CREATE(mutex, &pthread_mutex);
	S_CREATE(cond, &pthread_cond);
	killme = NULL;

	/* zero out destructors for per-thread data */
	bzero(per_thread_destr, sizeof(*per_thread_destr));

	pthread_key_create(&__errno_key, NULL);
	pthread_can_exit = 1;
}


/*
 * Allocate a space for flob and stack and initialize the pthread
 * data structure, but do not create any flob.  Set's the *state
 * stack pointer to the newly allocated stack, and returns the address
 * allocated for the thread flob.
 */
fluke_thread_t *
pthread_boot(struct exc_state *state)
{
	struct pthread *uth;

	uth = kalloc_aligned(sizeof *uth, FLUKE_THREAD_SHIFT);
	assert(uth);

	thread_exc_state_init(&startup_thread);
	*state = startup_thread.exc_state;

#ifdef CONFIG_KMODULE
	state->esp = (oskit_addr_t)base_stack_end;
#else
	{
		void *thread_stack = kalloc_aligned(KTHREAD_STACK_SIZE, 4);
		assert(thread_stack);
		uth->stack = thread_stack;
		uth->stacksize = KTHREAD_STACK_SIZE;
		state->esp = ((int) uth->stack + uth->stacksize);
	}
#endif

	pthread_uth_init(uth);

	return &uth->thread;
}

void
pthread_uth_init(struct pthread *uth)
{

	/* init per-thread info */
	uth->per_thread_size = 0;
	uth->per_thread = (void **)NULL;

	/* set up for errno macro */
	uth->__errno = 0;
	pthread_setspecific_thread(uth, __errno_key, &(uth->__errno));
}


void OSKIT_NORETURN
pthread_exit(void *arg)
{
	int i;
	struct pthread *self;

	assert(pthread_can_exit);

	self = pthread_self();

	/* per-thread data destructors */
	for (i = 0; i < self->per_thread_size; i++)
	if ((self->per_thread[i]) && (per_thread_destr[i]))
		per_thread_destr[i](self->per_thread[i]);

	fluke_mutex_lock(&pthread_mutex);
	if (killme) {
		/*
		 * There is another exitting thread waiting to die.
		 * We kill it and then wait for somebody to kill us.
		 */
		fluke_thread_destroy(&killme->thread);
#ifndef CONFIG_KMODULE
		kfree(killme->stack, killme->stacksize);
#endif
		free(killme->per_thread);
		kfree(killme, sizeof *killme);
	}

	killme = self;

	/*
	 * Block until the next thread exits and reaps us.
	 * We shouldn't get a spurious wakeup, but to be safe:
	 */
	while (1)
		fluke_cond_wait(&pthread_cond, &pthread_mutex);

	/* never reached ... */
	otsan();
}


/*
 * Support for per-thread data.
 */
static pthread_key_t key_val = 0;	/* unsigned int */

int
pthread_key_create(pthread_key_t *key, pthread_destructor_t destructor)
{

	if (key_val >= PTHREAD_KEYS_MAX)
		return EAGAIN;

	per_thread_destr[key_val] = destructor;
	*key = key_val;
	key_val ++;

	return 0;
}

int
pthread_key_delete(pthread_key_t key)
{
	if (key >= key_val)
		return EAGAIN;

	per_thread_destr[key] = NULL;	/* don't call destructors */

	return 0;
}


/*
 * XXX: This is so that pthread_create doesn't have to run an
 * intermediate function that sets the thread-specific data and
 * then calls the start function.
 * This is not currently thread-safe.
 */
int
pthread_setspecific_thread(struct pthread *thread, pthread_key_t key,
		const void *value)
{
	if (key >= PTHREAD_KEYS_MAX)
		return EINVAL;

	/*
	 * Expand the array if necessary, to the number of supported keys
	 * This may allocate more space than we actually need.
	 */
	if (key >= thread->per_thread_size) {
		thread->per_thread = realloc(thread->per_thread,
			sizeof(void *) * key_val);
		assert(thread->per_thread);
		bzero((void *)thread->per_thread +
				thread->per_thread_size * sizeof(void *),
			(key_val - thread->per_thread_size) * sizeof(void *));
		thread->per_thread_size = key_val;
	}

	thread->per_thread[key] = (void *)value;
	return 0;
}

int
pthread_setspecific(pthread_key_t key, const void *value)
{
	return pthread_setspecific_thread(pthread_self(), key, value);
}


void *
pthread_getspecific(pthread_key_t key)
{
	struct pthread *self = pthread_self();

	if (key > self->per_thread_size)
		return NULL;

	return self->per_thread[key];
}



/*
 * More kserver-support for pthreads.
 * Not exactly a stellar implementation, but it isn't any worse
 * than what libsac does.
 */

int
pthread_equal(pthread_t t1, pthread_t t2)
{
	return (t1 == t2);
}

int
pthread_detach(pthread_t t)
{
	/*
	 * XXX: this isn't implemented; it is just here to avoid
	 * rewriting the MOM support code.
	 */
	return 0;
}


pthread_t
pthread_self(void)
{
	return (pthread_t)fluke_thread_self();
}


int
pthread_join(pthread_t p, void **stat)
{
	otsan();
	return 0;
}


/*
 * These don't do anything, but are here to prevent libsac's pthreads
 * from being pulled in by mistake (which does nothing as well).
 */
void
pthread_yield(void)
{
}

int
pthread_setschedparam(pthread_t p, int policy, struct sched_param *param)
{
	return 0;
}

int
pthread_getschedparam(pthread_t p, int *policy, struct sched_param *param)
{
	return 0;
}

int
pthread_kill(struct pthread *p, int sig)
{
	return 0;
}

int
pthread_signal(int sig, void (*handler)(int))
{
	return 0;
}

pthread_t *__pthread_client_alert(pthread_t cur_th)
{
	panic("__pthread_client_alert");
	return 0;
}
