/*
 * Copyright (c) 1995, 1997, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>

#include <oskit/x86/seg.h>
#include <oskit/x86/eflags.h>
#include <oskit/x86/base_vm.h>
#include <oskit/x86/base_gdt.h>
#include <fluke/thread.h>
#include <fluke/cp/rpc.h>
#include <fluke/machine/syscall.h>

#include "config_kstackcheck.h"
#include "config_instrument.h"
#include "s_instrument.h"
#include "gdt.h"
#include "s_thread.h"
#include "s_region.h"
#include "xfer.h"
#include "debug.h"

#if FLASK
#include "avc.h"
#endif

#include "config_opt_threads.h"
#include "config_opt_synch.h"
#if defined CONFIG_OPT_THREADS || defined CONFIG_OPT_SYNCH
#define PCLSR 1
#include "anno.h"
#include "u_entry.h"
struct anno_table anno_fluke_entry;
#endif

/* Bootup inititalization of the machine-dependent part of the thread layer.  */
void s_thread_md_boot(void)
{
}

/*
 * Initialize a new thread's machine-dependent user-mode state.
 * We only need to initialize the elements that need to be nonzero,
 * since the whole structure was initialized to already
 * by our caller, kentry_thread_create().
 */
void s_thread_md_create(s_thread_t *th)
{
#ifndef CONFIG_NONBLOCKING

	extern void thread_start(void);

	/*
	 * Initialize the dispatcher and context switch layers of the thread.
	 * In blocking (stack-per-thread) kernels,
	 * this will set up the thread's kernel stack and csw state
	 * so that the thread will jump to `thread_start'
	 * as soon as it is resumed from its WAIT_STOPPED state,
	 * and it will take an immediate page fault in user mode
	 * because it doesn't have a task yet.
	 */
	thread_disp_state_setup_stack(th, thread_start, &th->exc_state);

#endif /* CONFIG_NONBLOCKING */

#ifdef CONFIG_OPT_THREADS
	/* Initialize the pointer to the user-level part */
	th->md.u_thread = (u_thread_t*)phystokv(th->sob.u_pa);
#endif
}

/*
 * Get a thread's machine-dependent state
 * into a fluke_thread_state structure provided by the user.
 */
void s_thread_md_get_state(
	s_thread_t		*th,
	fluke_thread_state_t	*state)
{
	assert_s_thread_locked(th);
	assert(state);

	state->trap_handler = th->th_trap_handler;
	state->interrupt_handler = th->th_interrupt_handler;
	state->client_alert_handler = th->th_client_alert_handler;
	state->server_alert_handler = th->th_server_alert_handler;

	state->saved_eax = th->th_saved_eax;
	state->saved_ecx = th->th_saved_ecx;
	state->saved_edx = th->th_saved_edx;
	state->saved_eip = th->th_saved_eip;
}

/*
 * Set the current thread's machine-dependent state
 * from a fluke_thread_state structure provided by the user.
 */
void s_thread_md_set_state(
	s_thread_t		*th,
	fluke_thread_state_t	*state)
{
	assert_s_thread_locked(th);
	assert(state);

	th->th_trap_handler = state->trap_handler;
	th->th_interrupt_handler = state->interrupt_handler;
	th->th_client_alert_handler = state->client_alert_handler;
	th->th_server_alert_handler = state->server_alert_handler;

	th->th_saved_eax = state->saved_eax;
	th->th_saved_ecx = state->saved_ecx;
	th->th_saved_edx = state->saved_edx;
	th->th_saved_eip = state->saved_eip;
	th->th_saved_eflags = state->saved_eflags;
}

/*
 * Deliver an asynchronous exception to the current thread.
 * Expects at least one of the thread's pending flags to be set.
 */
void s_thread_md_async_exception(s_thread_t *cur_th)
{
	/*
	 * The current thread needs to be locked
	 * in order to manipulate the gflags safely.
	 */
	assert_s_thread_locked(cur_th);

	/*
	 * Save the registers we'll be modifying.
	 */
	cur_th->th_saved_eax = cur_th->exc_state.eax;
	cur_th->th_saved_ecx = cur_th->exc_state.ecx;
	cur_th->th_saved_edx = cur_th->exc_state.edx;
	cur_th->th_saved_eip = cur_th->exc_state.eip;
	cur_th->th_saved_eflags = cur_th->exc_state.eflags;

	/*
	 * We always supply the current thread pointer in eax,
	 * in order to give the thread a place to stand
	 * (i.e. a place to save additional state,
	 * or an alternate stack to switch to, or whatever.)
	 */
	cur_th->exc_state.eax = cur_th->thread_va;

	/*
	 * Dispatch a single pending async exception.
	 * We don't initialize ecx or edx
	 * because they only apply to traps.
	 */
	if (cur_th->gflags & FLUKE_THREAD_INTERRUPT_PENDING)
	{
		cur_th->exc_state.eip = cur_th->th_interrupt_handler;
		cur_th->gflags &= ~FLUKE_THREAD_INTERRUPT_PENDING;
	}
	else if (cur_th->gflags & FLUKE_THREAD_CLIENT_ALERT_PENDING)
	{
		cur_th->exc_state.eip = cur_th->th_client_alert_handler;
		cur_th->gflags &= ~FLUKE_THREAD_CLIENT_ALERT_PENDING;
	}
	else
	{
		assert(cur_th->gflags & FLUKE_THREAD_SERVER_ALERT_PENDING);
		cur_th->exc_state.eip = cur_th->th_client_alert_handler;
		cur_th->gflags &= ~FLUKE_THREAD_SERVER_ALERT_PENDING;
	}

	/*
	 * Clear the exception enable flag
	 * so that the exception handler won't be called again
	 * before it's ready.
	 */
#ifndef CONFIG_OPT_THREADS
	assert(cur_th->th_lflags & FLUKE_THREAD_EXCEPTION_ENABLE);
#endif
	cur_th->th_lflags &= ~FLUKE_THREAD_EXCEPTION_ENABLE;
}

/*
 * Deliver a synchronous exception to the current thread.
 * Expects the trap code and subcode in the saved state frame to be valid.
 */
void s_thread_md_sync_exception(s_thread_t *cur_th)
{
	/*
	 * The current thread needs to be locked
	 * in order to manipulate the flags safely.
	 */
	assert_s_thread_locked(cur_th);

	/*
	 * Save the registers we'll be modifying.
	 */
	cur_th->th_saved_eax = cur_th->exc_state.eax;
	cur_th->th_saved_ecx = cur_th->exc_state.ecx;
	cur_th->th_saved_edx = cur_th->exc_state.edx;
	cur_th->th_saved_eip = cur_th->exc_state.eip;
	cur_th->th_saved_eflags = cur_th->exc_state.eflags;

	/*
	 * Set the registers to their new values.
	 */
	cur_th->exc_state.eax = cur_th->thread_va;
	cur_th->exc_state.ecx = cur_th->exc_state.code;
	cur_th->exc_state.edx = cur_th->exc_state.subcode;
	cur_th->exc_state.eip = cur_th->th_trap_handler;
	cur_th->exc_state.eflags &= ~EFL_TF;

	/*
	 * Clear the exception enable flag
	 * so that the trap handler won't be called again
	 * before it's ready.
	 */
#ifndef CONFIG_OPT_THREADS
	assert(cur_th->th_lflags & FLUKE_THREAD_EXCEPTION_ENABLE);
#endif
	cur_th->th_lflags &= ~FLUKE_THREAD_EXCEPTION_ENABLE;
}

#ifdef CONFIG_INSTRUMENT
/*
 * Called for each syscall if instrumentation is enabled.
 * eax is the EIP of the syscall on the way in and the
 * return value of the sys call on the way out.
 */
void kentry_md_syscall_instrument(
	int                syscall_flag,
	unsigned long long tsc,
	unsigned int       eax)
{
	s_thread_t *cur_thread = current_thread();

	s_thread_lock(cur_thread); /* XXX */

	if (cur_thread->instrument_flags)
	{
		switch(syscall_flag)
		{
		case INSTRUMENT_SYSCALL_ENTRY:
			s_instrument_syscall_entry(cur_thread,
						   eax, /* entrypoint */
						   tsc);
			break;
		case INSTRUMENT_SYSCALL_EXIT:
			s_instrument_syscall_exit(cur_thread,
						  eax, /* syscall rc */
						  EXC_GET_IP(cur_thread),
						  tsc);
			break;
		case INSTRUMENT_SYSCALL_USERTRAP:
			/*
			 * Don't care so much about these, as it means
			 * something really unusual happened, and tracing
			 * is silly.
			 *
			 * eax isn't meaningful.
			 */
			s_instrument_usertrap(cur_thread, tsc);
			break;
		case INSTRUMENT_SYSCALL_PAGEFAULT:
			/* eax isn't meaningful. */
			s_instrument_pagefault(cur_thread, tsc);
			break;
		case INSTRUMENT_SYSCALL_EXCEPT:
			s_instrument_except(cur_thread,
					    eax, /* exception code */
					    tsc);
			break;
		case INSTRUMENT_SYSCALL_TRAPRET:
			s_instrument_trapret(cur_thread,
					     eax, /* trap rc */
					     tsc);
			break;
		default:
			break;
		}
	}

	s_thread_unlock(cur_thread);
}
#endif

/*
 * Change saved user state to reflect a syscall error return.
 * XXX
 */
void kentry_md_syscall_error_return(s_thread_t *cur_th, int err)
{
	cur_th->exc_state.eax = err;
	cur_th->exc_state.eip = fluke_nop__entry;
}


#include <fluke/machine/syscall.h>

#include "s_port.h"

/*
 * Send an idempotent exception message to a region keeper.
 * Exception message includes the offset in the region at which the fault
 * occured as well as the access type that caused the fault.
 * Region should be locked on call and is returned unlocked.
 *
 * XXX could/should be merged with exception_rpc?
 */
int
s_thread_md_fault_rpc(s_thread_t *client, s_region_t *region,
		      oskit_addr_t region_off, fluke_prot_t prot)
{
	struct s_thread *server;
	s_port_t *keeper;
	int rc;
	fluke_ipc_params_t sparams;

	assert(region != 0);
	assert_s_ob_locked(&region->sob);
	assert(client == current_thread());

	/* Break any existing idempotent connection. */
	if (client->ipc_state.idemp_server) {
		s_region_unlock(region);
		rc = ipc_client_idemp_disconnect(client);
		if (rc)
			return rc;
		return KR_RESTART;
	}

	/* Look up the region's keeper port. */
	keeper = (s_port_t *)link_follow(&region->keeper);
	s_region_unlock(region);
	if (keeper == 0)
		return (KR_USER_EXCEPTION);
	assert(keeper->sob.type == FLUKE_PORT);

	/* Try to capture a server thread. */
	rc = s_port_capture_locked(keeper, &server);
	if (rc)
		return rc;
	if (server == 0)
		return (KR_USER_EXCEPTION);

	/* Extract the receive parameters from the server thread state. */
#if FLASK
	sparams.recv_buf.words = IPC_RECV_BUF_WORDS(server);
	sparams.recv_buf.buf = (fluke_ipc_word_t*)IPC_RECV_BUF_BUF(server);

	if (!server->ipc_state.server_wants_client_sid)
	{
	    sparams.recv_buf.words -= FLUKE_MIN_MSG_WORDS;
	    sparams.recv_buf.buf += FLUKE_MIN_MSG_SIZE;
	}
#else
	sparams.recv_buf.words = IPC_RECV_BUF_WORDS(server) -
		FLUKE_MIN_MSG_WORDS;
	sparams.recv_buf.buf = (fluke_ipc_word_t*)IPC_RECV_BUF_BUF(server) +
		FLUKE_MIN_MSG_WORDS;
#endif
	sparams.recv_buf_count = IPC_RECV_BUF_COUNT(server);
	sparams.recv_buf_tab = (fluke_ipc_buf_t*)IPC_RECV_BUF_TAB(server);
	sparams.recv_ref_count = IPC_RECV_REF_COUNT(server);
	sparams.recv_ref_tab = (struct fluke_ref**)IPC_RECV_REF_TAB(server);

	if (sparams.recv_buf.words > 0) {
#if FLASK
		fluke_ipc_word_t send_buf[3];
#else
		fluke_ipc_word_t send_buf[1];
#endif
		int nwords;

#if FLASK
		if (server->ipc_state.server_wants_client_sid)
		{
		    send_buf[0] = req_RegionKeeper_fault;
		    send_buf[1] = region_off;
		    send_buf[2] = prot;
		}
		else
			send_buf[0] = prot;
#else
		send_buf[0] = prot;
#endif

		/*
		 * XXX only send as many words as the recieve buffer
		 * has space, i.e., potentially truncate the message.
		 * Is there something else we should do instead?
		 */
		nwords = sparams.recv_buf.words;
		if (nwords > sizeof(send_buf) / sizeof(send_buf[0]))
			nwords = sizeof(send_buf) / sizeof(send_buf[0]);

		/*
		 * The first two words go in the min_msg buffer,
		 * but the rest must be copied into user memory.
		 */
		rc = copyout_remote(server, send_buf,
				    (oskit_addr_t)sparams.recv_buf.buf, nwords);
		if (rc)
		{
			server->wait_state.resume_rc = rc;
			thread_ready(server);
			INSTRUMENT_EVENT(client, KR_RESTART, 0);
			return KR_RESTART;
		}
		sparams.recv_buf.words -= nwords;
	}

	/* Update the receiver's register state. */
	client->ipc_state.idemp_server = server;
	server->ipc_state.idemp_client = client;
	IPC_RECV_BUF_WORDS(server) = sparams.recv_buf.words;
	IPC_RECV_BUF_COUNT(server) = sparams.recv_buf_count;
#if FLASK
	if (server->ipc_state.server_wants_client_sid)
	{
	    server->ipc_state.md.min_msg[0] = SECINITSID_KERNEL;
	    server->ipc_state.md.min_msg[1] = 0xdeadbeef;
	}
	else
	{
	    server->ipc_state.md.min_msg[0] = req_RegionKeeper_fault;
	    server->ipc_state.md.min_msg[1] = region_off;
	}
#else
	server->ipc_state.md.min_msg[0] = req_RegionKeeper_fault;
	server->ipc_state.md.min_msg[1] = region_off;
#endif
	assert(FLUKE_MIN_MSG_WORDS == 2);
	IPC_FINISH_RECEIVE(server, FLUKE_IPC_WAIT_IDEMPOTENT);

	return thread_handoff(client, WAIT_IPC_EXCEPTION, server);
}

static int wait_forever(struct s_thread *cur_th)
{
	static mutex_t dead_mutex;
	static cond_t dead_cond;
	static int dead_inited = 0;

	if (!dead_inited) {
		mutex_init(&dead_mutex);
		cond_init(&dead_cond);
		dead_inited = 1;
	}
	mutex_lock(&dead_mutex);
	return cond_wait(&dead_cond, &dead_mutex, 0);
}

#include "config_gdb_serial.h"
#ifdef CONFIG_GDB_SERIAL
#include <fluke/exception.h>
#include <oskit/c/signal.h>
#include <oskit/gdb_serial.h>
#include <oskit/x86/gdb.h>
#endif

/*
 * Perform an exception RPC to the current thread's keeper.
 */
int
s_thread_md_exception_rpc(struct s_thread *client)
{
	s_task_t *task;
	s_port_t *keeper;
	s_thread_t *server;
	int rc;
	fluke_ipc_params_t sparams;

	assert(client == current_thread());

	/* Break any existing idempotent connection. */
	rc = ipc_client_idemp_disconnect(client);
	if (rc)
		return rc;

	/* XXX if we have no task, we have no keeper,
	   since keepers haven't yet been moved to threads! */
	task = client->task_state.task;
	if (task == 0)
		return wait_forever(client);

	/* Find our keeper port.  If it doesn't exist, block forever. */
	s_task_lock(task);
	keeper = (struct s_port *)link_follow(&task->keeper);
	s_task_unlock(task);
	if (keeper == 0) {
		extern link_t kernel_task_link;
		if (link_compare_ptr(&kernel_task_link, &task->sob)) {
#ifdef CONFIG_GDB_SERIAL
			extern int enable_kdb; /* XXX console.c */
			if (enable_kdb) {
				/*
				 * XXX stolen from test/nest_debug.c
				 * Convert the exception code to a signal
				 * number.
				 */
				int sig;
				struct gdb_state r;
				switch ((client->exc_state.code >> 8) & 0xff) {
				case FLUKE_EXC_BAD_ACCESS:
					sig = SIGSEGV;
					break;

				case FLUKE_EXC_BAD_INSTRUCTION:
					sig = SIGILL;
					break;

				case FLUKE_EXC_ARITHMETIC:
					sig = SIGFPE;
					break;

				case FLUKE_EXC_BREAKPOINT:
					sig = SIGTRAP;
					break;

				case FLUKE_EXC_PAGE_FAULT:
					sig = SIGSEGV;
					break;

				case FLUKE_EXC_INSANITY:
				default:
					sig = SIGSEGV;
					break;
				}

				if (sig != SIGTRAP)
					printf("kserver thread %p va %#x "
					       "exc code %#x subcode %#x\n",
					       client, client->thread_va,
					       client->exc_state.code,
					       client->exc_state.subcode);

				r.edi = client->exc_state.edi;
				r.esi = client->exc_state.esi;
				r.ebp = client->exc_state.ebp;
				r.ebx = client->exc_state.ebx;
				r.edx = client->exc_state.edx;
				r.ecx = client->exc_state.ecx;
				r.eax = client->exc_state.eax;
				r.eip = client->exc_state.eip;
				r.eflags = client->exc_state.eflags;
				r.esp = client->exc_state.esp;

				gdb_serial_signal(&sig, &r);

				client->exc_state.edi = r.edi;
				client->exc_state.esi = r.esi;
				client->exc_state.ebp = r.ebp;
				client->exc_state.ebx = r.ebx;
				client->exc_state.edx = r.edx;
				client->exc_state.ecx = r.ecx;
				client->exc_state.eax = r.eax;
				client->exc_state.eip = r.eip;
				client->exc_state.eflags = r.eflags;
				client->exc_state.esp = r.esp;

				return 0;
			}
#endif
			printf("kserver thread %p hit an exception!\n", client);
#ifdef DEBUG
			s_debug_dump_flob(&client->sob, 1);
#endif
			panic("user thread in kernel task hit exception\n");
		}
		return wait_forever(client);
	}
	assert(keeper->sob.type == FLUKE_PORT);

	/* Try to capture a server thread. */
	rc = s_port_capture_locked(keeper, &server);
	if (rc)
		return rc;
	if (server == 0)
		return wait_forever(client);

	/* Extract the receive parameters from the server thread state. */
#if FLASK
	sparams.recv_buf.words = IPC_RECV_BUF_WORDS(server);
	sparams.recv_buf.buf = (fluke_ipc_word_t*)IPC_RECV_BUF_BUF(server);

	if (!server->ipc_state.server_wants_client_sid)
	{
	    sparams.recv_buf.words -= FLUKE_MIN_MSG_WORDS;
	    sparams.recv_buf.buf += FLUKE_MIN_MSG_SIZE;
	}
#else
	sparams.recv_buf.words = IPC_RECV_BUF_WORDS(server) -
		FLUKE_MIN_MSG_WORDS;
	sparams.recv_buf.buf = (fluke_ipc_word_t*)IPC_RECV_BUF_BUF(server) +
		FLUKE_MIN_MSG_WORDS;
#endif
	sparams.recv_buf_count = IPC_RECV_BUF_COUNT(server);
	sparams.recv_buf_tab = (fluke_ipc_buf_t*)IPC_RECV_BUF_TAB(server);
	sparams.recv_ref_count = IPC_RECV_REF_COUNT(server);
	sparams.recv_ref_tab = (struct fluke_ref**)IPC_RECV_REF_TAB(server);

	if (sparams.recv_buf.words > 0) {
#if FLASK
		fluke_ipc_word_t send_buf[4];
#else
		fluke_ipc_word_t send_buf[2];
#endif
		int nwords;

#if FLASK
		if (server->ipc_state.server_wants_client_sid)
		{
		    send_buf[0] = req_TaskKeeper_thread_exception;
		    send_buf[1] = client->exc_state.code;
		    send_buf[2] = client->exc_state.subcode;
		    send_buf[3] = client->thread_va;
		}
		else
		{
		    send_buf[0] = client->exc_state.subcode;
		    send_buf[1] = client->thread_va;
		}
#else
		send_buf[0] = client->exc_state.subcode;
		send_buf[1] = client->thread_va;
#endif

		/*
		 * XXX only send as many words as the recieve buffer
		 * has space, i.e., potentially truncate the message.
		 * Is there something else we should do instead?
		 */
		nwords = sparams.recv_buf.words;
		if (nwords > sizeof(send_buf) / sizeof(send_buf[0]))
			nwords = sizeof(send_buf) / sizeof(send_buf[0]);

		/*
		 * The first two words go in the min_msg buffer,
		 * but the rest must be copied into user memory.
		 */
		rc = copyout_remote(server, send_buf,
				    (oskit_addr_t)sparams.recv_buf.buf, nwords);
		if (rc)
		{
			server->wait_state.resume_rc = rc;
			thread_ready(server);
			INSTRUMENT_EVENT(client, KR_RESTART, 0);
			return KR_RESTART;
		}
		sparams.recv_buf.words -= nwords;
	}

	/* Update the receiver's register state. */
	client->ipc_state.idemp_server = server;
	server->ipc_state.idemp_client = client;
	IPC_RECV_BUF_WORDS(server) = sparams.recv_buf.words;
	IPC_RECV_BUF_COUNT(server) = sparams.recv_buf_count;
	IPC_RECV_REF_COUNT(server) = sparams.recv_ref_count;
#if FLASK
	if (server->ipc_state.server_wants_client_sid)
	{
	    server->ipc_state.md.min_msg[0] = SECINITSID_KERNEL;
	    server->ipc_state.md.min_msg[1] = 0xdeadbeef;
	}
	else
	{
	    server->ipc_state.md.min_msg[0] = req_TaskKeeper_thread_exception;
	    server->ipc_state.md.min_msg[1] = client->exc_state.code;
	}
#else
	server->ipc_state.md.min_msg[0] = req_TaskKeeper_thread_exception;
	server->ipc_state.md.min_msg[1] = client->exc_state.code;
#endif
	assert(FLUKE_MIN_MSG_WORDS == 2);
	IPC_FINISH_RECEIVE(server, FLUKE_IPC_WAIT_IDEMPOTENT);

	return thread_handoff(client, WAIT_IPC_EXCEPTION, server);
}

#include "anno.h"

/*
 * Perform any necessary cleanups of thread state before
 * this thread's state should be examined.
 */
void
s_thread_md_canceled(struct s_thread *cur_th)
{
#if PCLSR
	struct anno_entry *ae;

	ae = anno_find_lower(&anno_fluke_entry, cur_th->exc_state.eip);
	if (ae && cur_th->exc_state.eip <= ae->val2) {
		/*
		 * The user PC is in a Fluke API entry point implemented
		 * partly or wholly in user mode and requiring some
		 * sort of PC lusring (i.e. rollback or roll-forward).
		 */
		void (*pclsr)(struct s_thread *) = (void *) ae->val3;
		(*pclsr)(cur_th);
	}
#endif
}

#ifdef DEBUG
struct {
	unsigned mask;
	unsigned val;
	char *str;
} ws_tab[] = {
	{ -1,	WAIT_ON_COND,		"for waiting_on condition" },
	{ -1,	WAIT_IPC_SRV_DSCN,	"in ipc_server_disconnect" },
	{ -1,	WAIT_IPC_SRV_RECV,	"in ipc_server_receive" },
	{ -1,	WAIT_IPC_SRV_ASEND,	"in ipc_server_ack_send" },
	{ WAIT_IPC_SRV_RECEIVER,	WAIT_IPC_SRV_RECEIVER,
					"with server as the receiver" },
	{ -1,	WAIT_IPC_SRV_ORECV,	"in ipc_server_over_receive" },
	{ -1,	WAIT_IPC_SRV_SEND,	"in ipc_server_send" },
	{ WAIT_IPC_SRV_SENDER,	WAIT_IPC_SRV_SENDER,
					"with server as the sender" },
	{ -1,	WAIT_IPC_CLI_DSCN,	"in ipc_client_disconnect" },
	{ -1,	WAIT_IPC_CLI_RECV,	"in ipc_client_receive" },
	{ -1,	WAIT_IPC_CLI_ASEND,	"in ipc_client_ack_send" },
	{ WAIT_IPC_CLI_RECEIVER,	WAIT_IPC_CLI_RECEIVER,
					"with client as the receiver" },
	{ -1,	WAIT_IPC_CLI_ORECV,	"in ipc_client_over_receive" },
	{ -1,	WAIT_IPC_CLI_SEND,	"in ipc_client_send" },
	{ WAIT_IPC_CLI_SENDER,	WAIT_IPC_CLI_SENDER,
					"with client as the sender" },
	{ -1,	WAIT_IPC_UNPICKLE,	"for connections to be activated" },
	{ -1,	WAIT_IPC_CALL,		"on explicit idempotent ipc_call" },
	{ -1,	WAIT_IPC_EXCEPTION,	"on idempotent exception IPC" },
#ifdef WAIT_IPC_SECURITY
	{ -1,	WAIT_IPC_SECURITY,	"on security exception IPC" },
#endif
	{ WAIT_IPC_IDEMPOTENT,	WAIT_IPC_IDEMPOTENT,
					"on any idempotent request" },
	{ WAIT_CANCELABLE,	WAIT_CANCELABLE,
					"in a cancelable state" },
	{ -1,	WAIT_IPC_PICKLE,	"for connections to be deactivated" },
	{ -1,	WAIT_STOPPED,		"[stopped]" },
	{ WAIT_ANYTHING,		WAIT_ANYTHING,
					"at any clean point" },
	{ -1,	WAIT_NONE,		"for nothing [ready/running/blocked]" },
	{ 0,	0,			0 }
};

void
s_waitstate_dump(struct wait_state *wsp, int indent)
{
	int i;

	for (i = 0; i < indent; i++) putchar(' ');
	for (i = 0; ws_tab[i].mask; i++)
		if ((wsp->val & ws_tab[i].mask) == ws_tab[i].val)
			break;
	printf("wait_state: wval=%#x (waiting %s), condition=%p\n",
	       wsp->val, ws_tab[i].str, wsp->waiting_on);
	for (i = 0; i < indent; i++) putchar(' ');
	printf("            cancel_pending=%d, resume_rc=%#x\n",
	       wsp->cancel_pending ? 1 : 0, wsp->resume_rc);
}

void
s_thread_dump(s_thread_t *thread, int indent)
{
	int i;

	/* XXX not thread safe, break out mach dep stuff. */

	for (i = 0; i < indent; i++) putchar(' ');
	printf("thread_va=%p gflags=%#x, lflags=%#x\n",
	       (void *) thread->thread_va, thread->gflags, thread->th_lflags);
	for (i = 0; i < indent; i++) putchar(' ');
	printf("stop_cond: queue.head=%p\n", thread->stop_cond.queue.head);

	/* XXX csw */

	for (i = 0; i < indent; i++) putchar(' ');
	printf("sched_state: prio=%d, next=%p\n",
	       thread->sched_state.link.prio, 
	       thread->sched_state.link.next);
	for (i = 0; i < indent; i++) putchar(' ');
	printf("disp_state: "
#if defined CONFIG_SMP || defined CONFIG_USER_SCHED
	       "cpu=%p,  "
#endif
	       "next=%p, last=%p, run_tick=%u\n",
#if defined CONFIG_SMP || defined CONFIG_USER_SCHED
	       thread->disp_state.cpu,
#endif
	       thread->disp_state.next, thread->disp_state.last_thread,
	       thread->disp_state.run_tick);
#ifdef PREEMPTSTATS
	for (i = 0; i < indent; i++) putchar(' ');
	printf("            scheds=%d, lattot=%u, latmin=%u, latmax=%u\n",
	       thread->disp_state.scheds, thread->disp_state.lattotal,
	       thread->disp_state.latmin, thread->disp_state.latmax);
#endif
	s_waitstate_dump(&thread->wait_state, indent);
	for (i = 0; i < indent; i++) putchar(' ');
	printf("exc_state: eax=%#x, ebx=%#x, ecx=%#x, edx=%#x\n",
	       thread->exc_state.eax, thread->exc_state.ebx,
	       thread->exc_state.ecx, thread->exc_state.edx);
	for (i = 0; i < indent; i++) putchar(' ');
	printf("            edi=%#x, esi=%#x, ebp=%#x, esp=%#x, eip=%#x\n",
	       thread->exc_state.edi, thread->exc_state.esi,
	       thread->exc_state.ebp, thread->exc_state.esp,
	       thread->exc_state.eip);
	for (i = 0; i < indent; i++) putchar(' ');
	printf("code=%#x, subcode=%#x\n",
	       thread->exc_state.code, thread->exc_state.subcode);
	for (i = 0; i < indent; i++) putchar(' ');
	printf("task_state: task=%p task_dying=%d\n",
	       thread->task_state.task,
	       thread->task_state.task_dying ? 1 : 0);
	for (i = 0; i < indent; i++) putchar(' ');
	printf("ipc_state: flags=%#x server=%p (link=%p) \n",
	       thread->ipc_state.flags,
	       thread->ipc_state.server, thread->ipc_state.server_link.target);
	for (i = 0; i < indent; i++) putchar(' ');
	printf("                     client=%p (link=%p)\n",
	       thread->ipc_state.client, thread->ipc_state.client_link.target);
	for (i = 0; i < indent; i++) putchar(' ');
	printf("           idemp_server=%p idemp_client=%p\n",
	       thread->ipc_state.idemp_server, thread->ipc_state.idemp_client);
	if (thread->ipc_state.server || thread->ipc_state.server_link.target ||
	    thread->ipc_state.client || thread->ipc_state.client_link.target ||
	    thread->ipc_state.idemp_server || thread->ipc_state.idemp_client) {
		for (i = 0; i < indent; i++) putchar(' ');
		printf("           min_msg: %#x %#x\n",
		       thread->ipc_state.md.min_msg[0], thread->ipc_state.md.min_msg[1]);
	}
	for (i = 0; i < indent; i++) putchar(' ');
	printf("syscalls: calls=%d, uexcept=%d, pfault=%d, secfault=%d\n",
	       thread->kr_stats.calls, thread->kr_stats.uexceptions,
	       thread->kr_stats.pagefaults, thread->kr_stats.secfaults);
	for (i = 0; i < indent; i++) putchar(' ');
	printf("          cancel=%d, restart=%d\n",
	       thread->kr_stats.cancels, thread->kr_stats.restarts);
	for (i = 0; i < indent; i++) putchar(' ');
	printf("md_state: saved: eax=%#x, ecx=%#x, edx=%#x, eip=%#x, eflags=%#x\n",
	       thread->th_saved_eax, thread->th_saved_ecx, thread->th_saved_edx,
	       thread->th_saved_eip, thread->th_saved_eflags);
	for (i = 0; i < indent; i++) putchar(' ');
	printf("          syscall_st:");
	s_thread_print_stacktrace(thread, thread->md.syscall_st,
				  S_THREAD_SYSCALL_ST_SIZE);
}

void
s_thread_print_stacktrace(s_thread_t *thread, unsigned *tbuf, int tbufsize)
{
	int i;

	for (i = 0; i < tbufsize; i++) {
		if (tbuf[i] == 0)
			break;
		printf(" 0x%x", tbuf[i]);
	}
	printf("\n");
}

void
s_thread_store_stacktrace(s_thread_t *thread, unsigned *tbuf, int tbufsize)
{
	unsigned usp = thread->exc_state.esp;
	unsigned ebp = thread->exc_state.ebp;
	unsigned int code, subcode;
	int i;
	s_thread_t *cur_th = current_thread();

	/*
	 * We must save and restore our own pending exception data,
	 * because a page fault caught in copyin_local will clobber them.
	 */
	code = cur_th->exc_state.code;
	subcode = cur_th->exc_state.subcode;

	if (copyin_local(usp, &tbuf[0], 1))
		tbuf[0] = 0;
	else {
		for (i = 1; i < tbufsize; i++) {
			if (ebp == 0 || copyin_local(ebp+4, &tbuf[i], 1)) {
				tbuf[i] = 0;
				break;
			}
			if (copyin_local(ebp, &ebp, 1))
				ebp = 0;
		}
	}

	cur_th->exc_state.code = code;
	cur_th->exc_state.subcode = subcode;
}

#endif
