/*
 * Copyright (c) 1996, 1997, 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * x86-specific thread-related system calls.
 */

#include <fluke/exception.h>
#include <fluke/thread.h>
#include <fluke/x86/syscall.h>

#include "s_thread.h"
#include "exception.h"
#include "debug.h"

#ifndef CONFIG_OPT_THREADS
int sys_thread_get_saved_state()
{
	struct s_thread *cur_th = current_thread();

	if (EXC_SYSPARAM(cur_th, thread_get_saved_state, cur_thread)
	    != cur_th->thread_va)
		EXC_RETURN_INSANITY(NOT_CURRENT_THREAD);

	if (cur_th->lflags & FLUKE_THREAD_EXCEPTION_ENABLE)
		EXC_RETURN_INSANITY(EXCEPTIONS_ENABLED);

	EXC_SYSPARAM(cur_th, thread_get_saved_state, orig_eax) =
		cur_th->md.saved_eax;
	EXC_SYSPARAM(cur_th, thread_get_saved_state, orig_ecx) =
		cur_th->md.saved_ecx;
	EXC_SYSPARAM(cur_th, thread_get_saved_state, orig_edx) =
		cur_th->md.saved_edx;
	EXC_SYSPARAM(cur_th, thread_get_saved_state, orig_eip) =
		cur_th->md.saved_eip;
	EXC_SYSPARAM(cur_th, thread_get_saved_state, orig_eflags) =
		cur_th->md.saved_eflags;

	return 0;
}

int sys_thread_set_saved_state()
{
	struct s_thread *cur_th = current_thread();

	if (EXC_SYSPARAM(cur_th, thread_set_saved_state, cur_thread)
	    != cur_th->thread_va)
		EXC_RETURN_INSANITY(NOT_CURRENT_THREAD);

	if (cur_th->lflags & FLUKE_THREAD_EXCEPTION_ENABLE)
		EXC_RETURN_INSANITY(EXCEPTIONS_ENABLED);

	cur_th->md.saved_eax =
		EXC_SYSPARAM(cur_th, thread_set_saved_state, orig_eax);
	cur_th->md.saved_ecx =
		EXC_SYSPARAM(cur_th, thread_set_saved_state, orig_ecx);
	cur_th->md.saved_edx =
		EXC_SYSPARAM(cur_th, thread_set_saved_state, orig_edx);
	cur_th->md.saved_eip =
		EXC_SYSPARAM(cur_th, thread_set_saved_state, orig_eip);
	cur_th->md.saved_eflags =
		EXC_SYSPARAM(cur_th, thread_set_saved_state, orig_eflags);

	return 0;
}
#endif

#if 1 /* XXX ndef CONFIG_OPT_THREADS */
int sys_thread_return_from_exception()
{
	struct s_thread *cur_th = current_thread();

	if (EXC_SYSPARAM(cur_th, thread_return_from_exception, cur_thread)
	    != cur_th->thread_va)
		EXC_RETURN_INSANITY(NOT_CURRENT_THREAD);

	/*
	 * Set the exception enable flag again.
	 * The user must ensure that it is cleared when this routine is called.
	 */
	if (cur_th->th_lflags & FLUKE_THREAD_EXCEPTION_ENABLE)
		EXC_RETURN_INSANITY(EXCEPTIONS_ENABLED);
	cur_th->th_lflags |= FLUKE_THREAD_EXCEPTION_ENABLE;

	/*
	 * Check to see if an interrupt was delivered while handling
	 * this one; if so, jump right back into the handler.
	 */
	s_ob_lock(&cur_th->sob);
	if (cur_th->gflags & FLUKE_THREAD_INTERRUPT_PENDING) {
		thread_cancel(cur_th);
	}
	s_ob_unlock(&cur_th->sob);

	/*
	 * Restore the saved exception state.
	 */
	cur_th->exc_state.eax = cur_th->th_saved_eax;
	cur_th->exc_state.ecx = cur_th->th_saved_ecx;
	cur_th->exc_state.edx = cur_th->th_saved_edx;
	cur_th->exc_state.eip = cur_th->th_saved_eip;
	cur_th->exc_state.eflags = cur_th->th_saved_eflags;

	/*
	 * If we return 0 here,
	 * the syscall entry code will change the saved eip
	 * to point to the fluke_nop entrypoint,
	 * which is definitely not what we want.
	 * Therefore, return KR_RESTART instead,
	 * which will cause the right thing to happen.
	 */
	return KR_RESTART;
}
#endif /* not CONFIG_OPT_THREADS */

#ifndef CONFIG_OPT_THREADS
int sys_thread_get_handlers()
{
	struct s_thread *cur_th = current_thread();

	if (EXC_SYSPARAM(cur_th, thread_get_handlers, cur_thread)
	    != cur_th->thread_va)
		EXC_RETURN_INSANITY(NOT_CURRENT_THREAD);


	EXC_SYSPARAM(cur_th, thread_get_handlers, trap_handler) =
		cur_th->md.trap_handler;
	EXC_SYSPARAM(cur_th, thread_get_handlers, interrupt_handler) =
		cur_th->md.interrupt_handler;
	EXC_SYSPARAM(cur_th, thread_get_handlers, client_alert_handler) =
		cur_th->md.client_alert_handler;
	EXC_SYSPARAM(cur_th, thread_get_handlers, server_alert_handler) =
		cur_th->md.server_alert_handler;

	return 0;
}

int sys_thread_set_handlers()
{
	struct s_thread *cur_th = current_thread();

	if (EXC_SYSPARAM(cur_th, thread_set_handlers, cur_thread)
	    != cur_th->thread_va)
		EXC_RETURN_INSANITY(NOT_CURRENT_THREAD);

	/*
	 * Set the new handler vectors.
	 */
	cur_th->md.trap_handler =
		EXC_SYSPARAM(cur_th, thread_set_handlers, trap_handler);
	cur_th->md.interrupt_handler =
		EXC_SYSPARAM(cur_th, thread_set_handlers, interrupt_handler);
	cur_th->md.client_alert_handler =
		EXC_SYSPARAM(cur_th, thread_set_handlers, client_alert_handler);
	cur_th->md.server_alert_handler =
		EXC_SYSPARAM(cur_th, thread_set_handlers, server_alert_handler);

	return 0;
}
#endif
