/*
 * Copyright (c) 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#ifndef _FLUKE_KERNEL_U_ENTRY_H_
#define _FLUKE_KERNEL_U_ENTRY_H_

/*
 * We use the macros in this file to define Fluke API entry points that can
 * be done entirely in user-mode and can safely be "rolled forward" to
 * preserve the API atomicity guarantees simply by completing what the
 * user-mode instructions would have done themselves.
 *
 * There are two versions of these macros: #ifdef ASSEMBLER, the macros
 * produce assembly code to appear in a user entry page; #ifndef ASSEMBLER,
 * they produce a `pclsr_<ENTRYNAME>' function to roll the thread forward
 * by manipulating its `struct s_thread *' in the kernel.
 *
 * An entry is defined with these macros like this:
 *	BEGIN_UENTRY(entrypoint_name)
 *	...macroized asm code...
 *	END_UENTRY(entrypoint_name, SIZE)
 * where SIZE is the size of the entrypoint (usually 0x40 or 0x20).
 * It doesn't really matter what the entrypoint_name is; it just names
 * the assembler labels and pclsr_* function, the real meaning comes
 * from the placement of the entrypoint code in the page.
 */

#ifdef ASSEMBLER

#include <oskit/machine/anno.h>

#define UENTRY_KTOU(addr)	((addr)-UENTRY_KPAGE+UENTRY_UPAGE)

#define BEGIN_UENTRY(entry)						      \
entry:
#define END_UENTRY(entry, size)						      \
9:	.space	(size)-(9b-entry);					      \
	ANNO_ENTRY(anno_fluke_entry,					      \
		   UENTRY_KTOU(entry+1), UENTRY_KTOU(9b),		      \
		   pclsr_##entry);					      \
	.text
#define ANNO_UENTRY_PCLSR(label, pclsr)					      \
9:	ANNO_ENTRY(anno_fluke_entry,					      \
		   UENTRY_KTOU(label+1), UENTRY_KTOU(9b),		      \
		   pclsr_##pclsr);					      \
	.text
#define ANNO_UENTRY(label)	ANNO_UENTRY_PCLSR(label, label)


#define RET			ret

#define UTH_BASE_REG		eax /* fluke_thread_t * arg always in %eax */
#define	MOV_UTH_REG(reg, REG)	movl UTH_SAVED_##REG(%##UTH_BASE_REG), %##reg
#define	MOV_REG_UTH(reg, REG)	movl %##reg, UTH_SAVED_##REG(%##UTH_BASE_REG)

#define MOV_HDLR_REG(reg, HDLR)	movl UTH_##HDLR##_HANDLER(%##UTH_BASE_REG), %##reg
#define MOV_REG_HDLR(reg, HDLR)	movl %##reg, UTH_##HDLR##_HANDLER(%##UTH_BASE_REG)

#define	THVA_REG(reg)		movl	%gs:GS_CUR_THREAD, %##reg

#define	SET_LFLAGS(cur_th, flags)	orl	$(flags), UTH_LFLAGS(cur_th)
#define	CLEAR_LFLAGS(cur_th, flags)	andl	$~(flags), UTH_LFLAGS(cur_th)

#include "k_entry_asm.h"	/* defines FLUKE_ENTRY for kernel traps */

#else

#include "s_thread.h"
#include <oskit/anno.h>
#include <fluke/machine/syscall.h>

extern struct anno_table anno_fluke_entry; /* defined in s_thread_md.c */

#define BEGIN_UENTRY(name)	void pclsr_##name(struct s_thread *sth) {
#define END_UENTRY(name, size)	}

#define FLUKE_ENTRY(name)	/* no PC lusering required for kernel traps */

#define RET			sth->exc_state.eip = fluke_nop__entry;

#define MOV_UTH_REG(reg, REG)	sth->exc_state.reg = sth->th_saved_##REG;
#define MOV_REG_UTH(reg, REG)	sth->th_saved_##REG = sth->exc_state.reg;
#define th_saved_EAX		th_saved_eax
#define th_saved_ECX		th_saved_ecx
#define th_saved_EDX		th_saved_edx
#define th_saved_EIP		th_saved_eip
#define th_saved_EFLAGS		th_saved_eflags

#define MOV_HDLR_REG(reg, HDLR)	sth->exc_state.reg = sth->th_##HDLR##_handler;
#define MOV_REG_HDLR(reg, HDLR)	sth->th_##HDLR##_handler = sth->exc_state.reg;
#define th_TRAP_handler		th_trap_handler
#define th_INTERRUPT_handler	th_interrupt_handler
#define th_CLIENT_ALERT_handler	th_client_alert_handler
#define th_SERVER_ALERT_handler	th_server_alert_handler

#define	THVA_REG(reg)		sth->exc_state.reg = sth->thread_va;

#define	SET_LFLAGS(cur_th, flags)	sth->th_lflags |= (flags);
#define	CLEAR_LFLAGS(cur_th, flags)	sth->th_lflags &= ~(flags);

#endif


#endif /* _FLUKE_KERNEL_U_ENTRY_H_ */
