/*
 * Copyright (c) 1996 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Machine-independent definitions for transferring data
 * between kernel and user space and between different user spaces.
 *
 * The actual implementation of this module
 * may be either machine-independent or machine-dependent.
 * A semi-generic machine-independent implementation is provided
 * which copies data remotely by looking up physical fragments
 * in the target task and copies directly to/from physical memory.
 * This is not likely to work too well
 * on machines with virtually-indexed caches because of aliasing problems,
 * so at least on those machines the default implementation should be replaced.
 */
#ifndef _FLUKE_KERNEL_XFER_H_
#define _FLUKE_KERNEL_XFER_H_

#ifndef XFER_STATE_MD
struct xfer_state
{
	/* The default implementation doesn't need any per-thread state.  */
};
#endif /* XFER_STATE_MD */

struct s_thread;

/*
 * Transfer data words from the current thread's user space into kernel space.
 * Returns nonzero if a page fault or other exception occurs in user space;
 * in this case the current thread's exception_state is set appropriately
 * to indicate the type of exception.
 *
 * This routine does not return an indication of how much data was transferred
 * (other than indirectly through the faulting address,
 * and this indication may be unreliable in some situations);
 * thus, this routine should only be used to copy small memory chunks
 * for which it is always reasonable to restart from the beginning on a fault.
 *
 * Note that whereas the copyin/out routines in most kernels deal with bytes,
 * these routines deal only with register-sized words,
 * because no Fluke API calls deal with anything less than word granularity.
 * On most architectures, user code is required to ensure
 * that user-mode addresses and sizes are aligned properly;
 * however, the default implementation of these routines doesn't require this,
 * because they are built on the generic memcpy() function in the kernel.
 */
int copyin_local(
	oskit_addr_t		src_uva,
	void			*dest_kva,
	oskit_size_t		nwords);

/*
 * Transfer data from _another_ thread's user space into kernel space.
 * The other thread must be locked and "passivated" in some way
 * so that we know that we can access its user space safely
 * without it switching tasks on it or somesuch.
 * Currently, this is only called by IPC code;
 * the above requirement is assured because we hold the target's IPC lock
 * and we know it is waiting for us to muck with its address space.
 *
 * If a fault occurs while accessing the remote thread's user space,
 * this routine return's a KR_ exception code to be used by the _remote_ thread;
 * the remote thread's exception state has been set up appropriately.
 * This routine never causes exceptional returns in the local thread;
 * one implication of this is that it is uncancelable.
 * Because of this reason, and for the same reasons as for copyin_local(),
 * this routine should only be used to transfer small, fixed-sized chunks.
 */
int copyin_remote(
	struct s_thread		*src_thread,
	oskit_addr_t		src_uva,
	void			*dest_kva,
	oskit_size_t		nwords);

/*
 * These functions work just like the above two functions,
 * but they copy data from the kernel's address space to the user's.
 */
int copyout_local(
	void			*src_kva,
	oskit_addr_t		dest_uva,
	oskit_size_t		nwords);
int copyout_remote(
	struct s_thread		*dest_thread,
	void			*src_kva,
	oskit_addr_t		dest_uva,
	oskit_size_t		nwords);

/*
 * Transfer data from the current thread's user address space
 * directly into another thread's user address space.
 * Destination thread must be passivated as described above.
 *
 * If an exception occurs in the current thread or we get canceled,
 * this routine returns the appropriate error code,
 * and the current thread's exception_state is set up appropriately.
 * If an exception occurs in the _remote_ thread or it gets canceled,
 * then this routine returns 0
 * but sets *out_remote_rc to the return code for the remote thread;
 * in this case the remote thread's exception_state has been set.
 * In any case, *inout_nwords is set
 * indicating the number of words successfully copied.
 */
int copyto_remote(
	struct s_thread		*dest_thread,
	oskit_addr_t		src_uva,
	oskit_addr_t		dest_uva,
	oskit_size_t		*inout_nwords,
	int			*out_remote_rc);

#endif /* _FLUKE_KERNEL_XFER_H_ */
